//========================================================================
//
// OpenDCC
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      xp_client_if.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-07-08 V0.01 kw copied from DCCsniffer/rs232.c,
//                                modified for Xpressnet
//            2008-07-24 V0.02 kw changed rx to int
//            2008-09-13 V0.03 kw compile flags for:  uart / uart0 / uart1
//            2008-09-18 V0.04 kw complete client low level interface
//            2008-09-20 V0.05 kw first tx running
//            2009-02-12 V0.06 kw adapted for MFT
//            2009-02-20 V0.07 kw bugfix in timer setup
//
//------------------------------------------------------------------------
//
// purpose:   command station / throttle for dcc - client for xpressnet
//
// content:   routines for RS485 Interface of Xpressnet Clients
//            see also following references:
//            -> xpressnet definition at www.lenz.com
//            -> www.opendcc.de
// here:      RX fifo and TX buffer
//            Control of line direction
//            Handling of calls and slot_id
//
// to do:     ACK Messages
//
//========================================================================
// upstream interface (howto):
//
// to send something on xpressnet:
//   a) create message to be sent
//      (this must include header and data, the xor checksum
//      is calculated here)
//   b) copy this message to xp_tx_msg   or
//   c) then set xp_tx_msg_ready to 1.
//   d) if xp_tx_msg_ready is 0 again - the message is sent
//      (at least will be sent soon)
//
// or:
//   b) put the message to xp_buf_out fifo
//
//
// to receive a message:
//   a) all incoming messages are forwarded to Rx_Fifo
//   b) xpressnet_client collects the incomming message and handles
//      slot usage and message tracking
//   c) see xp_client_parser for processing
//
// init:
//   a) be sure to set up timer correctly
//   b) define the usart and direction port
//   c) define your desired slot_mode and slot_id.
//   d) call init_xpressnet_client();
//
// run_time:
//   a) periodically call run_xpressnet_client();
//      this updates slot_id and states
//
//========================================================================
// used hardware resources (see xp_client_hardware.h):
//
//   a) UART:   may be the only UART, UART0 or UART1, see below and header
//   b) TIMER2: this must run at something between 0.2us to 2us.
//              The timer must run in wrap around mode, we use OCR2
//              and the corresponding compare interrupt
//              Note: in PWM Modes OCR is double buffered - we cant use these modes
//   c) PORT:   a port, which controls the direction setting.
//
//========================================================================
// How it works:
//
//  ISR:       | content
//-------------|-----------------------------------------
//  RS485_RX   | get data and put to fifo
//             | check for call to our slot_id
//             | if called: enable timer int (20us)
//             |
//  TIMER      | disable timer int
//             | check tx_request
//             | if yes: prepare message
//             |         enable tx
//             |         enable rs485_tx
//             |
//  RS485_TX   | transmit data
//             | if last data, disable tx
//             |
//  TX_NO_DATA | switch rs485 to rx
//
//========================================================================

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <util/parity.h>
#include <string.h>

#include "config.h"               // general structures and definitions
#include "hardware.h"             // general structures and definitions
#include "cortos.h"               // make consumer ready
#include "xp_client_hardware.h"
#include "xp_client_if.h"

// #include "lcd_text.h"


// #define TRIG  { PORTB |= (1<<PB4); _mydelay_us(5); PORTB &= ~(1<<PB4); }
//
// #define TRIGA    { LED_A_GREEN; _mydelay_us(5); LED_A_OFF; }
// #define TRIGB    { LED_B_GREEN; _mydelay_us(5); LED_B_OFF; }
// #define TRIGC    { LED_C_GREEN; _mydelay_us(5); LED_C_OFF; }
// #define TRIGD    { LED_D_GREEN; _mydelay_us(5); LED_D_OFF; }


#define XP_TX_FIFO  1             // 0: single messages
                                  // 1: use a fifo for output


//========================================================================
// check and expand hardware definitions
//========================================================================

#ifndef XPRESSNET_UART
   #warning SEVERE no UART defined, see xp_client_hardware.h
#endif

#if (XPRESSNET_UART == UART)
    #define my_UDR      UDR
    #define my_UCSRA    UCSRA
        #define my_RXC      RXC
        #define my_TXC      TXC
        #define my_U2X      U2X
        #define my_FE       FE
        #define my_DOR      DOR
        #define my_UDRE     UDRE
    #define my_UCSRB    UCSRB
        #define my_RXCIE    RXCIE
        #define my_TXCIE    TXCIE
        #define my_UDRIE    UDRIE
        #define my_RXEN     RXEN
        #define my_TXEN     TXEN
        #define my_UCSZ2    UCSZ2
        #define my_RXB8     RXB8
        #define my_TXB8     TXB8
    #define my_UCSRC    UCSRC
        #ifdef URSEL
        #define my_URSEL    URSEL
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL
        #define my_UPM1     UPM1
        #define my_UPM0     UPM0
        #define my_USBS     USBS
        #define my_UCSZ1    UCSZ1
        #define my_UCSZ0    UCSZ0
        #define my_UCPOL    UCPOL
    #define my_UBRRH    UBRRH
    #define my_UBRRL    UBRRL

#elif (XPRESSNET_UART == UART0)
    #define my_UDR      UDR0
    #define my_UCSRA    UCSR0A
        #define my_RXC      RXC0
        #define my_TXC      TXC0
        #define my_U2X      U2X0
        #define my_FE       FE0
        #define my_DOR      DOR0
        #define my_UDRE     UDRE0
    #define my_UCSRB    UCSR0B
        #define my_RXCIE    RXCIE0
        #define my_TXCIE    TXCIE0
        #define my_UDRIE    UDRIE0
        #define my_RXEN     RXEN0
        #define my_TXEN     TXEN0
        #define my_UCSZ2    UCSZ02
        #define my_RXB8     RXB80
        #define my_TXB8     TXB80
    #define my_UCSRC    UCSR0C
        #ifdef URSEL0
        #define my_URSEL    URSEL0
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL0
        #define my_UPM1     UPM01
        #define my_UPM0     UPM00
        #define my_USBS     USBS0
        #define my_UCSZ1    UCSZ01
        #define my_UCSZ0    UCSZ00
        #define my_UCPOL    UCPOL0
    #define my_UBRRH    UBRR0H
    #define my_UBRRL    UBRR0L

#elif (XPRESSNET_UART == UART1)
    #define my_UDR      UDR1
    #define my_UCSRA    UCSR1A
        #define my_RXC      RXC1
        #define my_TXC      TXC1
        #define my_U2X      U2X1
        #define my_FE       FE1
        #define my_DOR      DOR1
        #define my_UDRE     UDRE1
    #define my_UCSRB    UCSR1B
        #define my_RXCIE    RXCIE1
        #define my_TXCIE    TXCIE1
        #define my_UDRIE    UDRIE1
        #define my_RXEN     RXEN1
        #define my_TXEN     TXEN1
        #define my_UCSZ2    UCSZ12
        #define my_RXB8     RXB81
        #define my_TXB8     TXB81
    #define my_UCSRC    UCSR1C
        #ifdef URSEL1
        #define my_URSEL    URSEL1
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL1
        #define my_UPM1     UPM11
        #define my_UPM0     UPM10
        #define my_USBS     USBS1
        #define my_UCSZ1    UCSZ11
        #define my_UCSZ0    UCSZ10
        #define my_UCPOL    UCPOL1
    #define my_UBRRH    UBRR1H
    #define my_UBRRL    UBRR1L
#else
  #warning wrong uart!
#endif // XPRESSNET_UART


#ifndef XPRESSNET_TIMER
   #warning SEVERE no TIMER defined, see xp_client_hardware.h
#endif

#if (XPRESSNET_TIMER == TIMER2)
  #ifndef T2_PRESCALER
    //#warning T2_PRESCALER ist missing. Please setup timer with prescaler
    #define T2_PRESCALER 8
  #endif
#elif (XPRESSNET_TIMER == TIMER0)
  #ifndef T0_PRESCALER
    //#warning T0_PRESCALER ist missing. Please setup timer with prescaler
    #define T0_PRESCALER 8
  #endif
#else
  #warning wrong timer!
#endif
//=====================================================================
// Data Section
//=====================================================================
// defines:

#define XP_SWITCH_PERIOD    20L         // after call byte: switch to TX after 20us

// global:

unsigned char slot_id;                  // our slot id, where we are allowed to send

unsigned char xp_active;                // true, if xpressnet traffic is detected

volatile unsigned char xp_tx_msg_ready; // flag: something to send (either fifo or direct)


#if (XP_TX_FIFO == 1)
    #define XP_OUT_BUF_SIZE       16                // fifo for outgoing messages
    t_xp_message  xp_out_buf[XP_OUT_BUF_SIZE];

    volatile unsigned char xp_out_buf_read;         // points to next read
    volatile unsigned char xp_out_buf_write;        // points to next write
                                                    // rd = wr: queue empty
                                                    // rd = wr + 1: queue full
#else
    t_xp_message xp_tx_msg;                         // next message to be sent
#endif

// static:
// a note on buffer size: xpressnet runs at 62k5, thus one byte is 160us;
// 60 bytes = 10ms;
// 
#define X_RxBuffer_Size  64              // mind. 16 / max 255
unsigned int X_RxBuffer[X_RxBuffer_Size];

unsigned char X_rx_read_ptr = 0;        // point to next read
unsigned char X_rx_write_ptr = 0;       // point to next write
unsigned char X_rx_fill = 0;

unsigned char out_msg[17];              // this is the currently send (Header, 15 data, XOR)
unsigned char tx_index;                 // current pointer
unsigned char tx_size;                  // total size
unsigned char tx_xor;                   // checksum


//=====================================================================
// Code Section
//=====================================================================


void XP_flush_rx(void)                 // Flush Receive-Buffer
  {
    uint8_t sreg = SREG;
    cli();
    do
      {
        my_UDR;
      }
    while (my_UCSRA & (1 << my_RXC));

    // R�cksetzen von Receive Complete-Flags
    my_UCSRA |= (1 << my_RXC);
    X_rx_read_ptr = 0;
    X_rx_write_ptr = 0;
    X_rx_fill = 0;
    SREG = sreg;
  }

inline void set_XP_to_receive(void) __attribute__((always_inline));
void set_XP_to_receive(void)
  {
    // for Xpressnet Adapter: Set MOSI to LOW
    HARDWARE_SET_XP_RECEIVE;        // see xp_client_hardware.h
    // X_state = XP_RX;
  }

inline void set_XP_to_transmit(void) __attribute__((always_inline));
void set_XP_to_transmit(void)
  {
    // for Xpressnet Adapter: Set MOSI to HIGH
    HARDWARE_SET_XP_TRANSMIT;        // see xp_client_hardware.h
    // X_state = XP_TX;
  }

void init_xp_client_if(void)
  {
    uint16_t ubrr;
    uint8_t sreg = SREG;

    set_XP_to_receive();

    cli();

    my_UCSRB = 0;                  // stop everything

    // we calculate 100* and add 50 to get correct integer cast;
    // normal speed - pre divider 16 (high speed mode - pre divider 8)

    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*62500L) - 1);
    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*625L) - 100L + 50L) / 100);
    my_UBRRH = (uint8_t) (ubrr>>8);
    my_UBRRL = (uint8_t) (ubrr);
    my_UCSRA = (1 << my_RXC) | (1 << my_TXC) | (0 << my_U2X);  // U2X: Speed Mode

    // FIFOs fuer Ein- und Ausgabe initialisieren

    #if (XP_TX_FIFO == 1)
        xp_out_buf_read = 0;
        xp_out_buf_write = 0;
    #endif

    // UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 9N1, asynchron

    my_UCSRB = (1 << my_RXCIE)      // enable rec_interrupt
           | (0 << my_TXCIE)
           | (0 << my_UDRIE)
           | (1 << my_RXEN)
           | (1 << my_TXEN)
           | (1 << my_UCSZ2)        // set frame length to 9bit;
           | (0 << my_RXB8)
           | (0 << my_TXB8);        // set 9th bit to 0

    my_UCSRC = (1 << my_URSEL)      // must be 1 (selector)
           | (0 << my_UMSEL)        // 0 = asyn mode
           | (0 << my_UPM1)         // 00 = parity disabled
           | (0 << my_UPM0)         //
           | (0 << my_USBS)         // 1 = tx with 2 stop bits
           | (1 << my_UCSZ1)        // 11 = 8 or 9 bits
           | (1 << my_UCSZ0)
           | (0 << my_UCPOL);

    XP_flush_rx();

    my_UCSRA |= (1 << my_TXC);         // clear tx

    SREG = sreg;
}

void stop_xp_client_if(void)
  {
    uint8_t sreg = SREG;

    set_XP_to_receive();

    cli();

    my_UCSRB = 0;                  // stop everything

    my_UCSRA |= (1 << my_TXC);      // clear any pending tx complete flag

    #if (XPRESSNET_TIMER == TIMER2)
        TIMSK &= ~(1<<OCIE2);                   // disabled further timer interrupts
    #elif (XPRESSNET_TIMER == TIMER0)
        TIMSK0 &= ~(1<<OCIE0A);                 // disabled further timer interrupts
    #endif 

    SREG = sreg;
  }


//---------------------------------------------------------------------------
// XP receive ISR
//   - received chars are feed to input fifo, 
//     consumer is the task run_xpressnet_client
//   - enable this task
//   - 16 bit interface
//   - no overrun check!
//   - If callbyte to our address:
//             - start timer
//             - after timer: check message and switch to tx
//
#if (XPRESSNET_UART == UART)
ISR(USART_RXC_vect)
#elif (XPRESSNET_UART == UART0)
 #ifndef USART0_RXC_vect
 #define USART0_RXC_vect USART0_RX_vect
 #endif
ISR(USART0_RXC_vect)
#elif (XPRESSNET_UART == UART1)
 #ifndef USART1_RXC_vect
 #define USART1_RXC_vect USART1_RX_vect
 #endif
ISR(USART1_RXC_vect)
#else
   #warning ISR undefined
#endif
  {
    unsigned char resh, resl;

    if (my_UCSRA & (1<< my_FE))
	  { // Frame Error
        resh = my_UCSRB;
        resl = my_UDR;                // at least: grep data
      }
    else
      {
        if (my_UCSRA & (1<< my_DOR))
          { // DATA Overrun -> Fatal
            // !!!
          }

        resh = my_UCSRB;
        resh = (resh >> my_RXB8) & 0x01;
        resl = my_UDR;

        if (resh)                           // it's a callbyte
          {
            if ((resl & 0x60) == 0x40)      // Inquiry (note: here we do no parity checking)
              {
                if ((resl & 0x1F) == slot_id)
                  {
                    // open timer to check for tx
                    // Note: in normal PWM Modes OCR is double buffered - we cannot use these modes
                    #if (XPRESSNET_TIMER == TIMER2)
                        OCR2 = TCNT2 + (XP_SWITCH_PERIOD * F_CPU / T2_PRESCALER / 1000000L);
                        TIFR |= (1<<OCF2);      // clear any old int
                        TIMSK |= (1<<OCIE2);    // enable compare int
                    #elif (XPRESSNET_TIMER == TIMER0)
                        OCR0A = TCNT0 + (XP_SWITCH_PERIOD * F_CPU / T0_PRESCALER / 1000000L);
                        TIFR0 |= (1<<OCF0A);      // clear any old int
                        TIMSK0 |= (1<<OCIE0A);    // enable compare int
                    #else
                        wrong timer
                    #endif
                  }
              }
          }

        // push to fifo for analysis

       	X_RxBuffer[X_rx_write_ptr] = (resh << 8) | resl;
        X_rx_write_ptr++;
        if (X_rx_write_ptr == X_RxBuffer_Size) X_rx_write_ptr=0;
        X_rx_fill++;          // no check for full, we just do an overrun!

        // make consumer ready

        #ifdef _CORTOS_H_
            set_task_ready(TASK_XPRESSNET);
        #else
            #warning cortos call disabled, use in polling environment only
        #endif
      }
  }

//---------------------------------------------------------------------------
//
// TX Messages are put in a fifo
//
//---------------------------------------------------------------------------


#if (XP_TX_FIFO == 1)
bool fifo_put_xp_out(t_xp_message *new_message)
  {
    unsigned char i;

    memcpy(&xp_out_buf[xp_out_buf_write], new_message, sizeof(xp_out_buf[0]));

    cli();
    xp_out_buf_write++;
    if (xp_out_buf_write == XP_OUT_BUF_SIZE) xp_out_buf_write = 0;
    xp_tx_msg_ready = 1;                        // set flag for TX
    sei();
    
    // check for full
    i = xp_out_buf_write+1;
    if (i == XP_OUT_BUF_SIZE) i = 0;
    if (i == xp_out_buf_read) return(1);        // totally full
    return(0);
  }


bool xp_out_buf_is_full(void)
  {
    unsigned char i;
    
    // check for full
    i = xp_out_buf_write+1;
    if (i == XP_OUT_BUF_SIZE) i = 0;
    if (i == xp_out_buf_read) return(1);        // totally full
    return(0);
  }

/*
bool xp_out_buf_not_empty(void)
  {
    if (xp_out_buf_write == xp_out_buf_read) return(0);
    return(1);
  }

bool get_from_xp_out_buf(t_xp_message *new_message)
  {
    // first do a check before calling this function!
    if (xp_out_buf_write == xp_out_buf_read) return(0);

    memcpy(new_message, &xp_out_buf[xp_out_buf_read], sizeof(t_xp_message));

    xp_out_buf_read++;
    if (xp_out_buf_read == XP_OUT_BUF_SIZE) xp_out_buf_read = 0;   // advance pointer

    return(1);
  } 
*/



#else
bool fifo_put_xp_out(t_xp_message *new_message)
  {
    while( !XP_tx_ready());
    cli();
    xp_tx_msg_ready = 1;
    sei();
    return(0);
  }

#endif

//---------------------------------------------------------------------------
// Timer Interrupt
// if switch time is over - we can send
//

#if (XPRESSNET_TIMER == TIMER2)
    ISR(TIMER2_COMP_vect)
#elif (XPRESSNET_TIMER == TIMER0)
    ISR(TIMER0_COMP_vect)
#else
  #warning wrong timer!
#endif
  {
    #if (XPRESSNET_TIMER == TIMER2)
        TIMSK &= ~(1<<OCIE2);                   // disabled further timer interrupts
    #elif (XPRESSNET_TIMER == TIMER0)
        TIMSK0 &= ~(1<<OCIE0A);                 // disabled further timer interrupts
    #endif 
    if (xp_tx_msg_ready)                        // check tx
      {
        set_XP_to_transmit();
        #if (XP_TX_FIFO == 1)
            memcpy(out_msg, &xp_out_buf[xp_out_buf_read], sizeof(xp_out_buf[0]));
            tx_size = xp_out_buf[xp_out_buf_read].size+1;        // data + header

            xp_out_buf_read++;   // advance pointer
            if (xp_out_buf_read == XP_OUT_BUF_SIZE) xp_out_buf_read = 0;

            if (xp_out_buf_write == xp_out_buf_read) xp_tx_msg_ready = 0;  // clear semaphor

        #else
            memcpy(out_msg, &xp_tx_msg.header, sizeof(xp_tx_msg));
            tx_size = xp_tx_msg.size+1;        // data + header
            xp_tx_msg_ready = 0;               // clear semaphor
        #endif
        tx_index = 0;
        tx_xor = 0;

        my_UCSRA |= (1 << my_TXC);      // clear any pending tx complete flag
        my_UCSRB |= (1 << my_TXEN);     // enable TX
        my_UCSRB |= (1 << my_UDRIE);    // enable TxINT
        my_UCSRB |= (1 << my_TXCIE);    // enable TX complete
      }
  }




//----------------------------------------------------------------------------
// TX: transmit one byte from out_msg to xp
//     if transmitted: a new UART_DATA-IRQ will be created
//     if all data is transmitted, we send the xor
// We transmit 9 bits, but the 8th bit is allways zero.

#if (XPRESSNET_UART == UART)
ISR(USART_UDRE_vect)
#elif (XPRESSNET_UART == UART0)
ISR(USART0_UDRE_vect)
#elif (XPRESSNET_UART == UART1)
ISR(USART1_UDRE_vect)
#else
   #warning ISR undefined
#endif
  {
    if (tx_size > 0)
      {
        // my_UCSRA |= (1 << my_TXC);              // writing a one clears any existing tx complete flag
        // my_UCSRB &= ~(1<<my_TXB8);               // clear bit 8

        my_UDR = out_msg[tx_index];
        tx_xor ^= out_msg[tx_index++];
        tx_size--;
      }
    else
      {
        my_UDR = tx_xor;
        my_UCSRB &= ~(1 << my_UDRIE);           // disable further TxINT
      }
  }


//---------------------------------------------------------------------------
// Halbduplex: auf TXC Complete schalten wir sofort die Richtung um
// Das als NAKED um es wirklich schnell zu bekommen.
// Wenn Optimize wegfaellt, dann darf NAKED nicht mehr benutzt werden - Flags!
// Vermutlich wird alles andere als ISR_NOBLOCK laufen muessen!
//
// ISR(USART1_TX_vect, ISR_NAKED)
//  {
//    set_XP_to_receive();
//    reti();
//  }

#if (XPRESSNET_UART == UART)
ISR(USART_TXC_vect)
#elif (XPRESSNET_UART == UART0)
 #ifndef USART0_TXC_vect
 #define USART0_TXC_vect USART0_TX_vect
 #endif
ISR(USART0_TXC_vect)
#elif (XPRESSNET_UART == UART1)
 #ifndef USART1_TXC_vect
 #define USART1_TXC_vect USART1_TX_vect
 #endif
ISR(USART1_TXC_vect)
#else
   #warning ISR undefined
#endif
  {
    set_XP_to_receive();
  }


//=============================================================================
// Upstream Interface
//-----------------------------------------------------------------------------
// TX:
#if (XP_TX_FIFO == 1)
#else
bool XP_tx_ready (void)
  {
    if (xp_tx_msg_ready == 0)
      {
        return(1);                        // yes, we can take a message
      }
    else
      {
        return(0);
      }
  }
#endif



//------------------------------------------------------------------------------
// RX:
bool XP_rx_ready (void)
{
    if (X_rx_read_ptr != X_rx_write_ptr)
      {
        return(1);     // there is something
      }
    else
      {
        return(0);
      }
}

//-------------------------------------------------------------------
// rx_fifo_read gets one char from the input fifo
//
// there is no check whether a char is ready, this must be
// done before calling with a call to rx_fifo_ready();

unsigned int XP_rx_read (void)
  {
    unsigned int retval;

    retval = X_RxBuffer[X_rx_read_ptr];
    X_rx_read_ptr++;
    if (X_rx_read_ptr == X_RxBuffer_Size) X_rx_read_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    X_rx_fill--;
    SREG = mysreg;

    return(retval);
  }

//-----------------------------------------------------------------------------------
// testroutinen zu Parsertest
// Ersatzroutine, um ein Zeichen in den Empfangsbuffer abzulegen
// Sendepuffer kann im Simulator mit einen Databreakpoint auf UDR �berwacht werden.
void X_push_to_rx(unsigned int d)
  {
	X_RxBuffer[X_rx_write_ptr] = d;

    X_rx_write_ptr++;
    if (X_rx_write_ptr == X_RxBuffer_Size) X_rx_write_ptr=0;
    X_rx_fill++;
  }


//===================================================================================
// security checks for misconfiguration
//===================================================================================

#if (XPRESSNET_TIMER == TIMER2)
  #define XP_TIMER_PRESCALER  T2_PRESCALER
#elif (XPRESSNET_TIMER == TIMER0)
  #define XP_TIMER_PRESCALER  T0_PRESCALER
#else
  #warning wrong timer!
#endif

#if ((XP_SWITCH_PERIOD * (F_CPU / XP_TIMER_PRESCALER / 1000000L)) < 2L)
  #warning either XP_SWITCH_PERIOD too short or PRESCALER too large
  #warning code will not work
#endif

#if ((XP_SWITCH_PERIOD * (F_CPU / XP_TIMER_PRESCALER / 1000000L)) < 10L)
  #warning either XP_SWITCH_PERIOD too short or PRESCALER too large
#endif

#if ((XP_SWITCH_PERIOD * (F_CPU / XP_TIMER_PRESCALER / 1000000L)) > 250L)
  #warning either XP_SWITCH_PERIOD too long or PRESCALER too small
  #warning code will not work
#endif
