//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
// This file is based on appnote from atmel.
//
//-----------------------------------------------------------------
//
// file:      analog.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-02-05 V0.01 started
//
//-----------------------------------------------------------------

#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h" 
#include "keycodes.h"               
#include "keyboard.h"               // hook to fifo_put (system monitor)              




#include "analog.h"

//------------------------------------------------------------------------------
//  @fn adc_init
//!
//! ADC channel initialization. Initialize the ADC with the selected
//! ADC_CHANNEL.
//!
//! @warning See ADC section in datasheet
//!
//! @param voltage_ref   Voltage reference source
//! @param left_adjust   Left adjust result
//! @param adc_channel   Analog channel selection
//!
//! @return range error.
//!          FALSE: Voltage reference source >3,
//!                 Left adjust result >2,
//!          TRUE:  In range
//!
//------------------------------------------------------------------------------
uint8_t  adc_init(uint8_t voltage_ref, unsigned char left_adjust, uint8_t adc_channel)
  {
    uint8_t range_voltage_ref;
    uint8_t range_left_adjust;

    range_voltage_ref = TRUE;
    range_left_adjust = TRUE;

    // --- Enter in ADC
    Disable_adc();
    Disable_adc_it();
    Clear_adc_it_flag();

    // --- AVcc Voltage Reference
    switch (voltage_ref)
    {
        case EXTERNAL_AREF:            // Enable external AREF
            Enable_external_aref();
            break;
        case AVCC_AS_VREF:             // Enable AVCC as Vref
            Enable_avcc_as_vref();
            break;
        case INTERNAL_VREF:            // Enable internal Vref
            Enable_internal_vref();
            break;
        default:                       // Range error
            range_voltage_ref = FALSE;
           break;
    }

    // --- Left Adjust Result
    switch (left_adjust)
    {
        case LEFT_ADJUST:              // Enable left adjust result
            Set_left_adjust();
            break;
        case NO_LEFT_ADJUST:           // Disable left adjust result
            Clear_left_adjust();
            break;
        default:                       // Range error
            range_left_adjust = FALSE;
           break;
    }

    // --- Set Channel number
    Select_adc_channel(adc_channel);

    // --- Set ADC prescaler close to 125kHz
    Set_adc_prescaler(ADC_PRESCALER);

    return (range_voltage_ref | range_left_adjust);
  }

//------------------------------------------------------------------------------
//  @fn adc_single_conversion
//!
//! ADC channel single conversion. Perform an analog to digital single
//! conversion on the selected ADC_CHANNEL.
//!
//! @warning Perform a "adc_init()" before to choose the adc_channel
//!
//! @param adc_input_pin   ADC I/O pin selection
//!
//! @return Conversion value (uint16_t).
//!
//------------------------------------------------------------------------------
uint16_t adc_single_conversion(uint8_t adc_input_pin)
{
    uint8_t  i;
    uint16_t adc_result;
    uint16_t accu_adc_result = 0;

    // --- To save power, the voltage over the ADC is turned off when not used.
    // --- This is done by controlling the voltage from the ADC I/O-pin.

    // Enable ADC I/O-pin
    ADC_PORT_OUT &= ~(1<<adc_input_pin);
    ADC_PORT_DIR &= ~(1<<adc_input_pin);
    ADC_DID_REG  |=  (1<<adc_input_pin);

    // --- Enable the ADC
    Enable_adc();
    // --- This stabilizes Aref rising level after Enable
    for (i=0; i<(1<<(ADC_PRESCALER))>>2; i++) asm("nop");

    // --- Perform a dummy single conversion first
    Start_conv();                   // Do single conversion
    while(!Conv_complete());        // Wait for conversion done, ADIF flag active
    Clear_adc_it_flag();            // Clear ADIF flag

    // --- Do the ADC conversion 16 times for better accuracy
    for(i=0; i<16; i++)
    {
        // --- Perform a single conversion
        Start_conv();
        // --- Wait for conversion done, ADIF flag active
        while(!Conv_complete());
        // --- Read out ADCH/ADCL registers
        adc_result = ADC;
        // --- Accumulate result (32 samples) for later averaging
        accu_adc_result += adc_result;
        // --- Clear ADIF flag
        Clear_adc_it_flag();
    }

    // --- Average the 16 samples
    adc_result = accu_adc_result >> 4;

    // --- Disable the ADC
    Disable_adc();

    return adc_result;
}


//------------------------------------------------------------------------------
//
// System Monitor
//
// we check:  - battery voltage
//              (measured on channel 2, external divider 2, so full range
//              is 5,12V (=1024 bins); we need to multiply x 5 to get mV.)
//            - power supply control
//
//------------------------------------------------------------------------------
//
// batt voltage depends heavily on temperature and load current
// source:
// http://www.panasonic.com/industrial/battery/oem/images/pdf/Panasonic_LiIon_CGR18650AF.pdf

// globals:

unsigned char batt_state;
unsigned char g_vccio_state;
uint16_t batt_voltage;


unsigned char get_vccio_state(void)
  {
    unsigned char retval;
    DIDR0 &= ~(1<<VCCIO_SENSE);    // enable readout
    _mydelay_us(1);                // wait one us
    retval = VCCIO_ACTIV();
    DIDR0 |= (1<<VCCIO_SENSE);   
    return(retval);
  }

unsigned char get_batt_state(void)
  {
    unsigned char bat_key;

    batt_voltage = 5 * adc_single_conversion(2);
    
    if (batt_voltage < 3000) bat_key = KEY_BATT_EMPTY ;
    else if (batt_voltage < 3400) bat_key = KEY_BATT_20_100 ;
    else if (batt_voltage < 3600) bat_key = KEY_BATT_40_100 ;
    else if (batt_voltage < 3900) bat_key = KEY_BATT_60_100 ;
    else bat_key = KEY_BATT_FULL ;

    return(bat_key);
  }


void init_system_monitor(void)
  {
    adc_init(INTERNAL_VREF, NO_LEFT_ADJUST, 2);
    batt_state = get_batt_state();
    g_vccio_state = get_vccio_state();
  }


//--------------------------------------------------------------------------------
//
// Task
//
//--------------------------------------------------------------------------------
// This is a CR_Task for CORTOS


#define MONITOR_TICK_PERIOD  500000L 
#if ((MONITOR_TICK_PERIOD / SYSTICK_PERIOD) == 0)
  #error: MONITOR_TICK_PERIOD too small!
#endif


t_cr_task system_monitor(void)
  {
    if (g_vccio_state != get_vccio_state())
      {
        g_vccio_state = get_vccio_state();
        cr_fifo_put(&key_fifo, KEY_VCCIO);
      }
   
    
    unsigned char new_batt_state;
    
    new_batt_state = get_batt_state();


    if (new_batt_state != batt_state)
      {
        batt_state = new_batt_state;
        cr_fifo_put(&key_fifo, batt_state);
        if (batt_state == KEY_BATT_EMPTY)
          {
            TURN_POWER_OFF;
          }
      }
      
    return(MONITOR_TICK_PERIOD / SYSTICK_PERIOD);
  }
