//------------------------------------------------------------------------
//
// OpenDCC - Throttle MFT
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      analog.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-02-09 V0.01 kw  started
//            2009-03-02 V0.02 kw  added globals for our states
//
//------------------------------------------------------------------------
//
// purpose:   Supply macros and define for analog measurements
//
//------------------------------------------------------------------------
// based on an appnote from atmel for the ADC for AT90CAN128/64/32.


#ifndef _ANALOG_H_
#define _ANALOG_H_

//------------------------------------------------------------------------
// Includes

//------------------------------------------------------------------------
// Defines

#define EXTERNAL_AREF    0
#define AVCC_AS_VREF     1
#define INTERNAL_VREF    3

#define LEFT_ADJUST      1
#define NO_LEFT_ADJUST   0

#define ADC_FULL_RANGE       ((uint16_t)(0x03FF))   // 10-bit ADC
#define INTERNAL_VREF_VALUE  (()(   256))   // 2.56 Volts

#ifndef F_CPU
        #error You must define F_CPU in config.h or makefile
    #elif F_CPU < 500000L
        #define ADC_PRESCALER 0x01
    #elif F_CPU < 1000000L
        #define ADC_PRESCALER 0x02
    #elif F_CPU < 2000000L
        #define ADC_PRESCALER 0x03
    #elif F_CPU < 4000000L
        #define ADC_PRESCALER 0x04
    #elif F_CPU < 8000000L
        #define ADC_PRESCALER 0x05
    #elif F_CPU < 16000000L
        #define ADC_PRESCALER 0x06
    #elif F_CPU < 20000000L
        #define ADC_PRESCALER 0x07
    #else
        #error Highest F_CPU definition
#endif /* F_CPU */


//------------------------------------------------------------------------------
//! Configure ADC Vref as external Aref (AREF pin).
//!
//! @param (none)
//!
//------------------------------------------------------------------------------
#define Enable_external_aref() \
        ( ADMUX  &= ~((1<<REFS1)|(1<<REFS0)) )

//------------------------------------------------------------------------------
//! Configure ADC Vref as AVCC.
//!
//! @param (none)
//!
//------------------------------------------------------------------------------
#define Enable_avcc_as_vref() \
         ( ADMUX  &= ~(1<<REFS1), ADMUX  |=  (1<<REFS0) )

//------------------------------------------------------------------------------
//! Configure ADC Vref as internal vref (2560 mV).
//!
//! @param (none)
//!
//------------------------------------------------------------------------------
#define Enable_internal_vref()   ( ADMUX  |=  ((1<<REFS1)|(1<<REFS0)) )

//------------------------------------------------------------------------------
//! ADC result = ADCH[1,0] / ACDL[7..0].
//!
//! @param (none)
//!
//------------------------------------------------------------------------------
#define Clear_left_adjust()                  (ADMUX  &= ~(1<<ADLAR))

//------------------------------------------------------------------------------
//! ADC result = ADCH[7..0] / ACDL[7,6].
//!
//! @param (none)
//!
//------------------------------------------------------------------------------
#define Set_left_adjust()                    (ADMUX  |=  (1<<ADLAR))

//------------------------------------------------------------------------------
//! Select ADC Channel.
//!
//! @param adc_channel   0 <= adc_channel <= 0x1F
//!
//------------------------------------------------------------------------------
#define Select_adc_channel(adc_channel)      (ADMUX  &= ~(0x1F<<MUX0), ADMUX |= (adc_channel<<MUX0))

//------------------------------------------------------------------------------
//! Configure Prescaler.
//!
//! @param adc_prescaler   0 <= adc_prescaler <= 7
//!
//------------------------------------------------------------------------------
#define Set_adc_prescaler(adc_prescaler)     (ADCSRA &= ~(0x07<<ADPS0), ADCSRA |= (adc_prescaler<<ADPS0))

//! Turn "ON" the ADC.
#define Enable_adc()                         (ADCSRA |=  (1<<ADEN))

//! Turn "OFF" the ADC.
#define Disable_adc()                        (ADCSRA &= ~(1<<ADEN))

//! Configure the High Speed Mode.
#define Set_adc_hsm()                        (ADCSRB |=  (1<<HSM))
#define Clear_adc_hsm()                      (ADCSRB &= ~(1<<HSM))

//! Start ADC conversion in single convertion mode.
#define Start_conv()                         (ADCSRA &= ~(1<<ADATE), ADCSRA |=  (1<<ADSC))

//! ADC conversion complete. If=0 then conversion running.
#define Conv_complete()                      (ADCSRA  &  (1<<ADIF))

//! Configure ADC interrupt.
#define Enable_adc_it()                      (ADCSRA |=  (1<<ADIE))
#define Disable_adc_it()                     (ADCSRA &= ~(1<<ADIE))
#define Clear_adc_it_flag()                  (ADCSRA |=  (1<<ADIF))

//_____ D E C L A R A T I O N S ________________________________________________

//_____ P R O T O T Y P E S ____________________________________________________


//------------------------------------------------------------------------------
//  @fn adc_init
//!
//! ADC channel initialization. Initialize the ADC with the selected
//! ADC_CHANNEL.
//!
//! @warning See ADC section in datasheet
//!
//! @param voltage_ref   Voltage reference source
//! @param left_adjust   Left adjust result
//! @param adc_channel   Analog channel selection
//!
//! @return range error.
//!          FALSE: Voltage reference source >3,
//!                 Left adjust result >2,
//!          TRUE:  In range
//!
uint8_t  adc_init(uint8_t voltage_ref, unsigned char left_adjust, uint8_t adc_channel);

//------------------------------------------------------------------------------
//  @fn adc_single_conversion
//!
//! ADC channel single conversion. Perform an analog to digital single
//! conversion on the selected ADC_CHANNEL.
//!
//! @warning Perform a "adc_init()" before to choose the adc_channel
//!
//! @param adc_input_pin   ADC I/O pin selection
//!
//! @return Conversion value ().
//!
uint16_t adc_single_conversion(uint8_t adc_input_pin);

//______________________________________________________________________________

uint16_t batt_voltage;          // the real voltage at the battery (in mV)
unsigned char batt_state;       // this is our fill state (coded like keys)
unsigned char g_vccio_state;      // true if VCCIO is on (line mode)

void init_system_monitor(void);


#endif  // _ANALOG_H_

