//------------------------------------------------------------------------
//
// OpenDCC - Throttle MFT (My First Throttle)
//
// Copyright (c) 2008, 2009 
//
// Main contributors: Jean Michel Fischer
//                    Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      config.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-18 V0.01 kw  start
//            2009-03-18 V0.04 kw  new knob_mode
//            2010-05-12 V0.05 kw  xbee base station mac
//            2010-06-15 V1.00 kw  changed to 1.00 due to eeproom change
//            2011-01-22 V1.01 kw  added switch_routes and LCD_CONTRAST_DEFAULT
//
//------------------------------------------------------------------------
//
// purpose:   This is the central definition file for the project
//            (data content like eeprom)
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>                 // using sscanf and sprintf increases prog. by 4k!
#include <inttypes.h>
#include <avr/pgmspace.h>          // put var to program memory
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include <string.h>

#include "config.h"                // general structures and definitions


t_power_mode g_power_mode = LINE_POWERED;

//======================================================================

// Die folgende (assemblerhnliche) Defintion der EEPROM-Variablen ist
// erforderlich, um diese compilerunabhngig auf feste Adressen zu legen.
// Auf diese Adressen kann von auen mit den Befehl zum Lesen und Schreiben
// von Sonderoptionen zugegriffen werden.

// The following definitions (quite similar to assembler) are neccesary
// to assign fixed addresses to variables based in EEPROM.
// This is required to allow external access to eeprom with so called
// "special option" commands.

//======================================================================
// this should be linked to beginning of EEPROM!
// Offsets are defined in config.h

unsigned char mft_cfg[] EEMEM =
  {
    [eadr_MFT_version]              = MFT_VERSION,          // 00: main compile version
    [eadr_MFT_subversion]           = MFT_SUBVERSION,       //     sub compile version
    [eadr_MFT_serial_l]             = MFT_SERIAL % 256,     // 02: This is a serial number used to identify different boxes
    [eadr_MFT_serial_h]             = MFT_SERIAL / 256,     // 03:
    [eadr_MFT_year]                 = BUILD_YEAR, 
    [eadr_MFT_month]                = BUILD_MONTH,
    [eadr_MFT_day]                  = BUILD_DAY,
    [eadr_show_splash]              = 10,                   // 07: time in units of 100ms
    [eadr_res8]                     = 0,                    // 08:
    [eadr_knob_push]                = KNOB_PUSH_ESTOP,      // 09: KNOB_PUSH_FS0, KNOB_PUSH_ESTOP
    [eadr_STOP_action]              = 0,                    // 10: MAKES_OFF, MAKES_STOP, MAKES_STOP_MY_LOCOS, 
    [eadr_XBEE_mode]                = 1,                    // 11: 1=XBEE enabled
    [eadr_knob_mode]                = KNOB_MODE_DUALSIDED,  // 12: 0:dual, 1:KNOB_MODE_SINGLESIDED
    [eadr_global_bright_pow]        = 50,                   // 13: brightness in power mode, default = 50%
    [eadr_global_bright_batt]       = 20,                   // 14: brightness in battery mode, default = 20%
    [eadr_global_bright_led]        = 100,    
    [eadr_global_bright_disp]       = 100,    
    [eadr_global_bright_night]      = 50,                   // 17: 50: night = half
    [eadr_res18]                    = 0, 
    [eadr_lcd_contrast]             = LCD_CONTRAST_DEFAULT, // 19: contrast voltage for colour display 
    [eadr_default_dcc_format]       = DCC128,               // 20:
    [eadr_turnout_delay]            = 2,                    // 21: in steps of 50ms
    [eadr_turnout_state_query]      = 1,                    // 22: 1: ask command station
    [eadr_lok_search_entry]         = 1,                    // 23  save loc: to retrieve same lok after power down
    [eadr_connect]                  = 0,                    // 24: preferred connect: 0=none, 1=XP, 2=XBEE

    [eadr_xp_slot_mode]             = 1,                    // 30: 1 = automatic
    [eadr_xp_slot_addr]             = 3,                    // 31: our addr, range 1..31

    [eadr_xbee_base]                = 0,                    // 32 0 =   void, 1..n base station
    [eadr_xbee_base_mac0]           = 0,                    // 33 mac0  
    [eadr_xbee_base_mac1]           = 0,                    // 34 mac0  
    [eadr_xbee_base_mac2]           = 0,                    // 35 mac0  
    [eadr_xbee_base_mac3]           = 0,                    // 36 mac0  
    [eadr_xbee_base_mac4]           = 0,                    // 37 mac0  
    [eadr_xbee_base_mac5]           = 0,                    // 38 mac0  
    [eadr_xbee_base_mac6]           = 0,                    // 39 mac0  
    [eadr_xbee_base_mac7]           = 0,                    // 40 mac0  
                                                        
    [eadr_waddr0_l]                 = (   1 % 256),         // to retrieve turnout addr
    [eadr_waddr0_h]                 = (   1 / 256),
    [eadr_waddr1_l]                 = (   2 % 256),
    [eadr_waddr1_h]                 = (   2 / 256),
    [eadr_waddr2_l]                 = (   3 % 256),
    [eadr_waddr2_h]                 = (   3 / 256),
    [eadr_waddr3_l]                 = (   4 % 256),
    [eadr_waddr3_h]                 = (   4 / 256),

    [eadr_saddr0_l]                 = (  11 % 256),         // to retrieve signal addr
    [eadr_saddr0_h]                 = (  11 / 256),
    [eadr_saddr1_l]                 = (  12 % 256),
    [eadr_saddr1_h]                 = (  12 / 256),
    [eadr_saddr2_l]                 = (  13 % 256),
    [eadr_saddr2_h]                 = (  13 / 256),
    [eadr_saddr3_l]                 = (  14 % 256),
    [eadr_saddr3_h]                 = (  14 / 256),

    [eadr_faddr0_l]                 = (  64 % 256),         // to retrieve feedbacked addr
    [eadr_faddr0_h]                 = (  64 / 256),
    [eadr_faddr1_l]                 = (  65 % 256),
    [eadr_faddr1_h]                 = (  65 / 256),

    [eadr_pc_tid]                   = 0x04,                 // 61: ID when used as PC Throttle
    [eadr_pc_v1]                    = 7,                    // 62: main version
    [eadr_pc_v2]                    = 0,                    // 63: sub version
    [eadr_pc_v3]                    = 5,                    // 64: char
    [eadr_pc_v4]                    = 1,                    // 65: index
    [eadr_switch_search_entry]      = 0,                    // 66 
    [eadr_switch_sca]               = 0,                    // 67 
    [eadr_switch_scb]               = 1,                    // 68 
    [eadr_switch_scc]               = 2,                    // 69 
    [eadr_switch_scd]               = 3,                    // 70	

    [eadr_dreh_addr_l     ]         = 225,                  // 71
    [eadr_dreh_addr_h     ]         = 0,                    // 72
    [eadr_dreh_posis_bit0 ]         = 0b01010101,           // 73
    [eadr_dreh_posis_bit1 ]         = 0,                    // 74
    [eadr_dreh_posis_bit2 ]         = 0b00110100,           // 75
    [eadr_dreh_posis_bit3 ]         = 0b01010101,           // 76
    [eadr_dreh_posis_bit4 ]         = 0,                    // 77
    [eadr_dreh_posis_bit5 ]         = 0b00110100,           // 78
    [eadr_dreh_sho_acc_cmd]         = 1,                    // 79
  };

// we init the clock to false values - this will suppress display if
// don't get a valid fast clock from the command station

t_fast_clock  fast_clock =
  {
    99,
    99,
    99,
    99,
  };
