
//------------------------------------------------------------------------
//
// OpenDCC - Throttle MFT (My First Throttle)
//
// Copyright (c) 2008, 2009, 2010
//
// Main contributors: Jean Michel Fischer
//                    Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      config.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-18 V0.01 kw  start
//            2009-03-18 V0.04 kw  new knob_mode
//            2009-04-07 V0.05 kw  knob_push
//            2010-06-16 V0.06 kw  added signal addr
//            2011-01-22 V0.07 kw  added changes from Volkmar Dierkes:
//                                 Nokia Display, Weichenmenu, Help 
//            2013-10-11 V0.08 kw  added support Drehscheibe (#define DREHSCHEIBE)
//
//------------------------------------------------------------------------
//
// purpose:   This is the central definition file for the project
//
//------------------------------------------------------------------------
//
// content:   1. Project settings
//               1.a) General
//               1.b) Configuration of Software/Hardware Modules
//
//            2. Global variables
//            3. EEPROM definitions (CV's)
//            4. Useful inline routines (delay, Semaphor-operations)
//
//========================================================================
//
#ifndef _CONFIG_H_
#define _CONFIG_H_


//========================================================================
//
// 1. Project Definitions
//
//========================================================================

#define MFT_VERSION       0x01
#define MFT_SUBVERSION    0x04

#define MFT_SERIAL        0x0001   // This is a 16-bit serial number used to identify different boxes.

#define BUILD_YEAR         15      // 10=2010
#define BUILD_MONTH         3      // 1=jan
#define BUILD_DAY          22      // day

#define FALSE  0
#define TRUE   (!FALSE)

// Debugswitches:  0: no debugging, interrupts running -> alive!
//                 1: no ints, some lowlevel test

#define DEBUG  0

#define LED_STOP_DEBUG      0       // 0: regular function, 1: red dimming (heart beat)



//-----------------------------------------------------------------
//
// 1.b) Configuration of installed hardware / software
//
//-----------------------------------------------------------------

#define LCD_EA204           1       // fully supported
#define LCD_REICHELT162     2       // not supported
#define LCD_EA128x64        3       // fully supported
#define LCD_NOKIA6100       4       // in development (by Volkmar Dierkes)

// #define DISPLAY             LCD_EA204
#define DISPLAY             LCD_EA128x64
//#define DISPLAY             LCD_NOKIA6100

#if (DISPLAY == LCD_NOKIA6100)
    #include "lcd_driver_nokia6100.h"
	#define LCD_CONTRAST_DEFAULT NOKIA6100_CONTRAST_DEFAULT
	#define LCD_CONTRAST_MIN     NOKIA6100_CONTRAST_MIN
	#define LCD_CONTRAST_MAX     NOKIA6100_CONTRAST_MAX
#else
	#define LCD_CONTRAST_DEFAULT 22
	#define LCD_CONTRAST_MIN     0
	#define LCD_CONTRAST_MAX     255
#endif

#define LCD_IF_SPI          1       // connect on SPI bus
#define LCD_IF_4BIT         2

#define DISPLAY_INTERFACE   LCD_IF_SPI

// !!!! EA_DOGM_TOPVIEW in lcd_driver_st7565.h


#define DEFAULT_BAUD        BAUD_115200   // on USB

// Board

#define  BOARD_MFT          1       // Version 1.0 or 1.1 of MyFirstThrottle
#define  BOARD_JMF          2       // Protptype by JEAN_MICHEL_FISCHER

#define  BOARD  BOARD_MFT

// see also EA_DOGM_TOPVIEW

// Language

#define GERMAN              1
#define ENGLISH             2
#define LANGUAGE            ENGLISH


#define DCC_F13_F28         1        // 1: add code for functions F13 up to F28
#define LOK_NAME_LENGTH     10       // no of char; multimaus uses 5
                                     // we have a range of 4 to 10
                                     // these are the characters without any trailing 0

#define SWITCH_NAME_LENGTH  16       // no of char
                                     // these are the characters without any trailing 0
#define SWITCH_ADR_NUMBER   7        // no of switches per switch way


#define MENU_SWITCH_TYPE	0		 // 0: Standard 'simple' menu for switches
									 // 1: Using list with 'street of switches'

#define MENU_HELP           0		 // 0: No Help screens
									 // 1: Use Help screens

#define DREHSCHEIBE         0        // 0: standard
                                     // 1: add Menue for turntable


//-- Images

#define PICS_INCLUDED_UK    1        // 1: include english locos


//========================================================================
// 2. Global Variables and types
//========================================================================

// This union allows to access 16 bits as word or as two bytes.
// This approach is (probably) more efficient than shifting.
typedef union
{
    uint16_t as_uint16;
    int16_t  as_int16;
    uint8_t  as_uint8[2];       // avr uses little endian - LSB is uint8[0]
    int8_t   as_int8[2];
    struct                      // if code runs on other endianess - change here
      {
        uint8_t  low8;
        uint8_t  high8;
      };
} t_data16;


typedef enum {DCC14 = 0, DCC27 = 1, DCC28 = 2, DCC128 = 3} t_format;

typedef struct
  {
    unsigned int addr;                  // address (either 7 or 14 bits)
    unsigned char speed;                // 0..28 or 0..126 (real speed)
    t_format format: 2;                 // 00 = 14, 01=27, 10=28, 11=128 speed steps.
                                        // DCC27 is not supported
    unsigned char dir: 1;               // 0 = forward, 1=revers
    unsigned char light: 1;             // function light
    unsigned char other: 1;             // 0: owned by me,   1: owned by other
    unsigned char valid: 1;             // 0: not confirmed  1: record validated by CS,
    unsigned char res: 2;               // reserved
    unsigned char f4_f1: 4;             // function 4 downto 1
    unsigned char f8_f5: 4;             // function 8 downto 5
    unsigned char f12_f9: 4;            // function 12 downto 9
    #if (DCC_F13_F28)
    unsigned char f16_f13: 4;           // function 16 downto 13
    unsigned char f20_f17: 4;           // function 20 downto 17
    unsigned char f24_f21: 4;           // function 24 downto 21
    unsigned char f28_f25: 4;           // function 28 downto 24
    #endif
  }  locomem_t;


typedef enum {STS_OFF           = 0,
              STS_ON            = 1,
              STS_STOP          = 2,
              STS_PROG          = 3,
              STS_PBUSY         = 4,
              STS_BUSY          = 5,
              STS_DISCONNECTED  = 6,
              STS_CMD_ERROR     = 7} t_status;

typedef enum {LINE_POWERED      = 0,
              BATT_POWERED      = 1,
             } t_power_mode;

t_power_mode g_power_mode;


#define SIZE_LOCO_DATABASE     128         // no of locos in database (max. 250)

#define SIZE_SWITCH_DATABASE    16         // no of switch ways in database (max. 250)

enum {OPM_PCMODE,                          // run as throttle for rocrail, tunnelled thr. xpressnet
      OPM_THROTTLE,                        // (XpressNet) throttle
      } operation_mode;

//------------------------------------------------------------------------
//
// Interfaces and Connections
//
//------------------------------------------------------------------------

typedef enum {
      USE_NONE,
      USE_USB,                    // transport to USB
      USE_XPRESSNET,              // XpressNet throttle
      USE_XBEE,                   // Wireless throttle
      USE_CAN,                    // CAN throttle
      } t_used_interface;

t_used_interface used_interface;    // This is the currently used interface

typedef enum {
      DISCONNECTED,
      CONNECT_XPRESSNET,
      CONNECT_XBEE,
      } t_connect_state;



t_format default_dcc_format;            // gets loaded from eeprom

typedef enum {  MAKES_OFF,
                MAKES_STOP,
                MAKES_STOP_MY_LOCOS,
                sizeof_t_STOP_action,
            } t_STOP_action;

t_STOP_action STOP_action;


typedef enum {  KNOB_MODE_DUALSIDED,              // act on both sides
                KNOB_MODE_SINGLESIDED,
                sizeof_t_knob_mode,
            } t_knob_mode;

t_knob_mode knob_mode;

typedef enum {  KNOB_PUSH_FS0,                    // push -> FS0
                KNOB_PUSH_ESTOP,                  // push -> Emergency Stop
                sizeof_t_knob_push,
            } t_knob_push;

t_knob_push knob_push;

typedef struct
  {
    unsigned char minute;
    unsigned char hour;
    unsigned char day_of_week;
    unsigned char ratio;
  } t_fast_clock;

t_fast_clock  fast_clock;

extern unsigned int serial_number;

extern unsigned char throttle_id;          // logische Handregler-ID (0..31)

//========================================================================
// 3. EEPROM Definitions (CV's)
//========================================================================
//
// content is defined in config.c

//------------------------------------------------------------------------
// Config - EEPROM
//------------------------------------------------------------------------
//
// MFT uses EEPROM for Configuration
//
// To achieve a fixed address schema, new sections are introduced - these
// must be allocated during linker run.
// -> see config file of AVR Studio, Memory Sections
//
// ATTENTION: There is a bug in AVR Studio - the memory options are not exported
//            to avr-objcopy, therefore the new sections appear in opendcc.hex,
//            not in opendcc.eep.
// SOLUTION:  manually edit makefile:
//
//             ## Intel Hex file production flags
//             HEX_FLASH_FLAGS = -R .eeprom
//             HEX_FLASH_FLAGS = -R .ee_my_new_sect    <-- add this line
//
// Sections:    Size:   Location:   Content:
// EEMEM                810000      General Config
// EE_***        128    810080

// see config.c for valid ranges and explanation

#define   eadr_MFT_version                0  // major compile version (change when eeprom is modified)
#define   eadr_MFT_subversion             1  // sub compile version
#define   eadr_MFT_serial_l               2  //
#define   eadr_MFT_serial_h               3  //
#define   eadr_MFT_year                   4  //
#define   eadr_MFT_month                  5  //
#define   eadr_MFT_day                    6  //
#define   eadr_show_splash                7  // time in unit of 100ms
#define   eadr_res8                       8  //
#define   eadr_knob_push                  9  // 0: fs0, 1: EStop
#define   eadr_STOP_action               10  // 0: Stop-Key generates OFF, 1: Stop-Key generates STOP
#define   eadr_XBEE_mode                 11  // 0: XBEE off, 2=run XBEE
#define   eadr_knob_mode                 12  // 0: dual sided (DC mode), 1:single sided (AC mode)
#define   eadr_global_bright_pow         13  // brightness, when in power mode
#define   eadr_global_bright_batt        14  // brightness, when in battery mode
#define   eadr_global_bright_led         15  // brightness for leds (range 0..100, default 100)
#define   eadr_global_bright_disp        16  // brightness for display (range 0..100, default 100)
#define   eadr_global_bright_night       17  // brightness reduction when in night mode (range 0..100, default 100)
#define   eadr_res18                     18  //
#define   eadr_lcd_contrast              19  // contrast voltage for graphic display
#define   eadr_default_dcc_format        20  //
#define   eadr_turnout_delay             21  // in steps of 50ms
#define   eadr_turnout_state_query       22  // 0: do not ..; 1: ask command station about turnout position
#define   eadr_lok_search_entry          23  // to retrieve same lok after power down
#define   eadr_connect                   24  // // 0: select on power on, 1= run XP, 2=run XBEE

#define   eadr_xp_slot_mode              30  // manual or automatic
#define   eadr_xp_slot_addr              31  // our addr, range 1..31
#define   eadr_xbee_base                 32  // 0 = void, 1..n base station
#define   eadr_xbee_base_mac0            33  // mac0
#define   eadr_xbee_base_mac1            34  // mac0
#define   eadr_xbee_base_mac2            35  // mac0
#define   eadr_xbee_base_mac3            36  // mac0
#define   eadr_xbee_base_mac4            37  // mac0
#define   eadr_xbee_base_mac5            38  // mac0
#define   eadr_xbee_base_mac6            39  // mac0
#define   eadr_xbee_base_mac7            40  // mac0

#define   eadr_waddr0_l                  41  // weichen - turnout
#define   eadr_waddr0_h                  42
#define   eadr_waddr1_l                  43
#define   eadr_waddr1_h                  44
#define   eadr_waddr2_l                  45
#define   eadr_waddr2_h                  46
#define   eadr_waddr3_l                  47
#define   eadr_waddr3_h                  48

#define   eadr_saddr0_l                  49  // signal - extended accessory
#define   eadr_saddr0_h                  50
#define   eadr_saddr1_l                  51
#define   eadr_saddr1_h                  52
#define   eadr_saddr2_l                  53
#define   eadr_saddr2_h                  54
#define   eadr_saddr3_l                  55
#define   eadr_saddr3_h                  56

#define   eadr_faddr0_l                  57  // last used feedback addr
#define   eadr_faddr0_h                  58
#define   eadr_faddr1_l                  59
#define   eadr_faddr1_h                  60

#define   eadr_pc_tid                    61
#define   eadr_pc_v1                     62
#define   eadr_pc_v2                     63
#define   eadr_pc_v3                     64
#define   eadr_pc_v4                     65
#define   eadr_switch_search_entry       66  // to retrieve same switch after power down
#define   eadr_switch_sca                67
#define   eadr_switch_scb                68
#define   eadr_switch_scc                69
#define   eadr_switch_scd                70

#define   eadr_dreh_addr_l               71  // turntable addr
#define   eadr_dreh_addr_h               72  // turntable addr
#define   eadr_dreh_posis_bit0           73
#define   eadr_dreh_posis_bit1           74
#define   eadr_dreh_posis_bit2           75
#define   eadr_dreh_posis_bit3           76
#define   eadr_dreh_posis_bit4           77
#define   eadr_dreh_posis_bit5           78
#define   eadr_dreh_sho_acc_cmd          79

// try to keep eeprom addresses fixed -> thus does not change loco database

// Error-Codes

#define   ERR_TURNOUT_SPACE               1     // there are too many turnouts operated simultaneously
#define   ERR_POWER_OFF                   2     // can't turn power off (hardware error)
#define   ERR_SLOT_OCCUPIED               3     // xpressnet see's the slot already in use
#define   ERR_NO_SLOT                     4     // xpressnet can't find a empty slot
#define   ERR_DATABASE                   10     // chaining in database korrupt (fatal)
#define   ERR_DATABASE_FULL              11     // data base full, last entry was not stored
#define   ERR_DATABASE_ADDR0             12     // lok with addr 0 in database

#define   ERR_XBEE_TX_OVERRIDE           20     // too many xbee messages, tx buffer overide
#define   ERR_ATCMD_FAILED               21     // At Command to XBee reported error
#define   ERR_XBEE_UNKNOWN_COMMAND       22     // Unexpected AT command answer received
#define   ERR_XBEE_FAILURE               23     // XBee internal problem
#define   ERR_XBEE_UNEXPECTED_FRAME      24     // Frame not expected

//
//#define   ERR_XBEE_UNKNOWN_FRAME         14     // Unexpected Frame received
//#define   ERR_XBEE_CHECKSUM_ERROR        15     // Invalid checksum on Frame
//
//#define   ERR_XBEE_UNEXPECTED_STATE      17     // Modemstatus not expected
//#define   ERR_XBEE_FAULTY_TX_STATUS      18     // TX Status reported error
//#define   ERR_XEEE_NOT_READY             19
//#define   ERR_XBEE_COMMAND_STATUS        24



//========================================================================
// 4.b) Useful inline code
//========================================================================

extern volatile unsigned char Communicate;


#define C_Received        0    // a new DCC was received - issued by ISR(Timer1)
                               //                          cleared by main
#define C_DoSave          1    // a new PORT state should be saved
                               //                        - issued by action
                               //                          cleared by main


static inline unsigned char semaphor_query(unsigned char flag)
       __attribute__((always_inline));

unsigned char semaphor_query(unsigned char flag)
  {
    return (Communicate & (1<<flag));
  }

static inline void semaphor_set(unsigned char flag)
       __attribute__((always_inline));

void semaphor_set(unsigned char flag)
  {
    cli();
    Communicate |= (1<<flag);
    sei();
  }

static inline unsigned char semaphor_get(unsigned char flag)
       __attribute__((always_inline));

unsigned char semaphor_get(unsigned char flag)
  {
    unsigned char value;
    cli();
    value = Communicate & (1<<flag);
    Communicate &= (~(1<<flag));
    sei();
    return(value);
  }


//------------------------------------------------------------------------
// Delay-Macro (all values in us) -> this is busy waiting
// see timer.h for a non busy implementation
//------------------------------------------------------------------------
//
// same macro as in util/delay.h, but use up to some ms.
// The maximal possible delay is 262.14 ms / F_CPU in MHz.
// This is 16ms for 16MHz; longest used delay: 1000us

#ifndef _UTIL_DELAY_H_
  #include <util/delay.h>
#endif

static inline void _mydelay_us(double __us) __attribute__((always_inline));
void
_mydelay_us(double __us)
{
    uint16_t __ticks;
    double __tmp = ((F_CPU) / 4e6) * __us;
    if (__tmp < 1.0)
        __ticks = 1;
    else if (__tmp > 65535)
        __ticks = 0;    /* i.e. 65536 */
    else
        __ticks = (uint16_t)__tmp;
    _delay_loop_2(__ticks);
}

static inline void _restart(void) __attribute__((always_inline));
void
_restart(void)
{
    cli();

    // laut diversen Internetseiten sollte folgender Code laufen -
    // tuts aber nicht, wenn man das Assemblerlistung ansieht.
    // void (*funcptr)( void ) = 0x0000;    // Set up function pointer
    // funcptr();                        // Jump to Reset vector 0x0000

    __asm__ __volatile
    (
       "ldi r30,0"  "\n\t"
       "ldi r31,0"  "\n\t"
       "icall" "\n\t"
     );
}

#endif   // config.h
