//------------------------------------------------------------------------------
//
// OpenDCC - Simple Cooperative Task Scheduler
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//-----------------------------------------------------------------------------
//
// file:      cortos.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-01-12 V0.01 kw  started
//            2009-03-02 V0.02 kw  LED control with compile-switch
//
//-----------------------------------------------------------------------------
//
// purpose:   A small and tiny cooperative scheduler
//
//------------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>

#include "cortos.h"
#include "cortos_tasklist.h"
#include "timer.h"          				// access to systick


// Tasks (here defined again for includes)
//                                          wakeup:
t_cr_task task_idle (void);                 // always
t_cr_task init_xbee (void);                 // by program action
t_cr_task run_navigator (void);             // by keystroke fifo
t_cr_task run_led (void);                   // timing
t_cr_task system_monitor (void);            // timing
t_cr_task programmer (void);                // gui
t_cr_task check_objects (void);             // timing
t_cr_task run_xpressnet_client (void);      // xpressnet fifo
t_cr_task update_display(void);             // gui
t_cr_task cursor(void);                     // timing
t_cr_task run_parser(void);                 // serial fifo
t_cr_task run_xbee_client(void);            // serial fifo



#if (DEBUG != 0)                    // generell debug allowed
    #define DEBUG_CORTOS   0        // 0: no debug messages
                                    // 1: LED turns on if not sleeping 
#endif
//------------------------------------------------------------------------------
//
// Fifo
//
//------------------------------------------------------------------------------

void cr_fifo_init (cr_fifo_t *f, uint8_t *buffer, const uint8_t size)
{
    f->count = 0;
    f->task_to_wakeup = 0;
    f->pread = f->pwrite = buffer;
    f->read2end = f->write2end = f->size = size;
}

uint8_t cr_fifo_put (cr_fifo_t *f, const uint8_t data)
{
    return _inline_fifo_put (f, data);
}

uint8_t cr_fifo_get_wait (cr_fifo_t *f)
{
    while (!f->count);
    
    return _inline_fifo_get (f);    
}

int cr_fifo_get_nowait (cr_fifo_t *f)
{
    if (!f->count)      return -1;
        
    return (int) _inline_fifo_get (f);  
}

uint8_t cr_fifo_filled (cr_fifo_t *f)
{
    return (f->count);  
}

void cr_fifo_set_comsumer (cr_fifo_t *f, t_task_id consumer)
{
    f->task_to_wakeup = consumer;
}

//------------------------------------------------------------------------------
//
// Task
//
//------------------------------------------------------------------------------
// this is the list of all tasks

t_task tasklist[] =
   // id                 ready, wakeup, call
  {  [TASK_IDLE]      = {    1, 0,      task_idle},            // must be the first entry, lowest prio
     [TASK_INIT_XBEE] = {    0, 0,      init_xbee},
     [TASK_SYSMON]    = {    1, 0,      system_monitor},
     [TASK_PROGRAMMER]= {    0, 0,      programmer},
     [TASK_PARSER]    = {    0, 0,      run_parser},
     [TASK_XBEE]      = {    0, 0,      run_xbee_client},
     [TASK_OBJECTS]   = {    1, 0,      check_objects},
     [TASK_NAVIGATOR] = {    1, 0,      run_navigator},
     [TASK_CURSOR]    = {    0, 0,      cursor},
     [TASK_LED]       = {    1, 0,      run_led},
     [TASK_DISPLAY]   = {    0, 0,      update_display},
     [TASK_XPRESSNET] = {    1, 0,      run_xpressnet_client}, // highest prio
  };


void set_task_ready(t_task_id ready_to_run)
  {
    tasklist[ready_to_run].wakeup = get_systick_int();  // ready now!
    tasklist[ready_to_run].ready = 1;
  }

void set_task_blocked(t_task_id ready_to_run)
  {
    tasklist[ready_to_run].ready = 0;
  }

uint8_t task_is_ready(t_task_id test_task)
  {
    unsigned char ready;
    cli();
    ready = tasklist[test_task].ready;
    sei();
    return(ready);
  }

volatile signed int my_systick;

void scheduler(void)
  {
    signed char t_id;
    // unsigned char any_task_ready = 0;
    signed int new_delay;

    // signed int my_systick;

    my_systick = get_systick_int();  // atomic!

    for (t_id = (sizeof(tasklist) / sizeof(tasklist[0]) -1); t_id >= 0; t_id--)
      {
        if (tasklist[t_id].ready)
          {
            // any_task_ready = 1;
            if ((my_systick - tasklist[t_id].wakeup) >= 0)
              {
                new_delay = tasklist[t_id].my_call();
                if (new_delay < 0)                      // if a task returns negativ - it get blocked
                  {
                    tasklist[t_id].ready = 0;
                  }
                else if (new_delay == 0)
                  {
                    tasklist[t_id].wakeup = my_systick;
                  }
                else
                  {
                    tasklist[t_id].wakeup = tasklist[t_id].wakeup + new_delay;
                  }
              }
          }
      }
  }

#include "hardware.h"

t_cr_task task_idle (void)
  {
    set_sleep_mode(SLEEP_MODE_IDLE);
    #if (DEBUG_CORTOS == 1) 
        LED_D_OFF;
    #endif
    sleep_mode(); 
    #if (DEBUG_CORTOS == 1) 
        LED_D_RED;
    #endif
    return(0);     // ready forever
  }
