//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006, 2007, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      database.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-08-25 V0.01 kw  started
//            2008-08-27 V0.02 kw  Transfertask for Database entries
//            2009-01-28 V0.03 kw  LOK_NAME_LENGTH
//            2009-02-28 V0.04 kw  added count_loco_in_database
//            2009-03-15 V0.05 kw  access to waddr move to this file
//            2009-03-31 V0.06 kw  added some values for graphics
//            2010-06-15 V0.07 kw  added extended accessory
//            2013-03-19 V0.08 kw  find_empty_loco_data see 0xff 0xff also as empty
//
//--------------------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   handling of lok names and speed steps
//            (This is not locobuffer, this is the data base,
//            has nothing to do with the actual dcc content)
//
//
//--------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <string.h>

#include "config.h"                // general structures and definitions
#include "timer.h"                 // timerval
#include "cortos.h"                // real time kernel
#include "messages_objects.h"      // actual*
#include "database.h"

#include "display.h"               // display_error



// mega32:   2kByte SRAM, 1kByte EEPROM
// mega644:  4kByte SRAM, 2kByte EEPROM


t_format default_dcc_format;            // dcc default: 0=DCC14, 2=DCC28, 3=DCC128
                                        // this value is read from eeprom

unsigned char actual_database_index;    // the index for actual_lok


//!warning! next_search_index void!
unsigned char next_search_index;

unsigned char cur_database_entry;
unsigned char total_database_entry;    // actual number of locos (not included the position 0)

unsigned char db_message[17];
unsigned char db_message_ready;

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Database for ADDR, FORMAT and NAME
//------------------------------------------------------------------------
//------------------------------------------------------------------------
//
// The default format of Throttle is defined in config.h/.c
// and stored in EEPROM (CV??)
//
//
// We allocate this storage to a fixed address to have access from the
// parser with a constant offset to eeprom
// (parser commands: read and write Special Option)
//
#if (EEPROM_FIXED_ADDR == 1)
 #define EEMEM_LOCO __attribute__((section(".ee_loco")))
#else
 #define EEMEM_LOCO EEMEM
#endif
//
// please add following command to linker:
//
//   -Wl,--section-start=.ee_loco=0x810080

t_locoentry locoentry[SIZE_LOCO_DATABASE] EEMEM_LOCO =
  {
    //  next,   prev,       addr   format      picture      name
    {    1,     10,     {{     3,    DCC28 }}, 0,           {  "Neu?"     } },   // void entry
    {    2,      0,     {{     4,   DCC128 }}, 0x3005,      {  "BR218"    } },
    {    3,      1,     {{     5,   DCC128 }}, 0x3004,      {  "BR290"    } },
    {    4,      2,     {{     6,   DCC128 }}, 0x6002,      {  "LINT"     } },
    {    5,      3,     {{     7,   DCC128 }}, 0x5003,      {  "Et423"    } },
    {    6,      4,     {{    14,    DCC14 }}, 0x2001,      {  "Lok14"    } },
    {    7,      5,     {{    28,    DCC28 }}, 0x4001,      {  "Lok28"    } },
    {    8,      6,     {{   128,   DCC128 }}, 0x5001,      {  "Lok128"   } },
    {    9,      7,     {{   888,   DCC128 }}, 0x5002,      {  "ICE 888"  } },
    {   10,      8,     {{   843,   DCC128 }}, 0x3002,      {  "Trudy"    } },
    {    0,      9,     {{   621,   DCC128 }}, 6,           {  "621 "     } },
  };


t_locoentry lok_neu;   // this hold the NEW? -Lok

//---------------------------------------------------------------

static void _write_loco_name(unsigned char index, char * name)
  {
    unsigned char i;
    for (i=0; i<LOK_NAME_LENGTH; i++)
      {
        eeprom_write_byte(&locoentry[index].name[i], name[i]);
      }
  }


// This searches the full database, independant weather in chain or not!
// return index if successful
// return SIZE_LOCO_DATABASE if not found
/*
unsigned char search_loco(unsigned int addr)
  {
    unsigned char i;
    unsigned char stored_h, stored_l;
    unsigned char addr_low, addr_high;

    addr_low = (unsigned char) addr;
    addr_high = (unsigned char) (addr>>8);

    // search loco, if found: replace name
    for (i=0; i<SIZE_LOCO_DATABASE; i++)
      {
        stored_l = eeprom_read_byte(&locoentry[i].b[0]);
        if (stored_l == addr_low)
          {
            stored_h = eeprom_read_byte(&locoentry[i].b[1]);
            if ((stored_h & 0x3f) == addr_high)
              {
                return(i);
              }
          }
      }
    return(SIZE_LOCO_DATABASE);       // failed
  }
*/

unsigned char search_loco(unsigned int addr)
  {
    unsigned char j;
    unsigned char index = 0;
    unsigned char stored_h, stored_l;
    unsigned char addr_low, addr_high;

    addr_low = (unsigned char) addr;
    addr_high = (unsigned char) (addr>>8);

    // search loco, if found: replace name
    for (j=0; j<total_database_entry; j++)
      {
        index = scan_loco_database_forward(index);
    
        stored_l = eeprom_read_byte(&locoentry[index].b[0]);
        if (stored_l == addr_low)
          {
            stored_h = eeprom_read_byte(&locoentry[index].b[1]);
            if ((stored_h & 0x3f) == addr_high)
              {
                return(index);
              }
          }
      }
    return(SIZE_LOCO_DATABASE);       // failed
  }


// ret true if successful
unsigned char assign_picture_to_loco(unsigned int addr, unsigned int pic_id)
  {
    unsigned char i;
    if ((i = search_loco(addr)) < SIZE_LOCO_DATABASE)
      {
        eeprom_write_word(&locoentry[i].picture_id, pic_id);
        return(1);
      }
    return(0);       // failed
  }


// ret true if successful
unsigned char store_loco_format(unsigned int addr, t_format format)
  {
    unsigned char i;
    t_locoentry test;
    if ((i = search_loco(addr)) < SIZE_LOCO_DATABASE)
      {
        test.b[1] = eeprom_read_byte(&locoentry[i].b[1]);
        test.w.format = format;
        eeprom_write_byte(&locoentry[i].b[1], test.b[1] );
        return(1);
      }
    return(0);       // failed
  }


unsigned int query_picture_to_loco(unsigned int addr, unsigned int *pic_id)
  {
    unsigned char i;
    if ((i = search_loco(addr)) < SIZE_LOCO_DATABASE)
      {
        *pic_id = eeprom_read_word(&locoentry[i].picture_id);
        return(1);
      }
    return(0);       // failed
  }

unsigned char store_loco_name(unsigned int addr, char * name)
  {
    unsigned char i;
    unsigned char last;
    unsigned char addr_low, addr_high;

    addr_low = (unsigned char) addr;
    addr_high = (unsigned char) (addr>>8);

    // search loco, if found: replace name
    if ((i = search_loco(addr)) < SIZE_LOCO_DATABASE)
      {
        _write_loco_name(i, name);
        return(1);
      }

    last = eeprom_read_byte(&locoentry[0].prev);

    i = insert_new_loco_data(last);
    if (i)
      {
        addr_high = addr_high | (default_dcc_format << 6);
        eeprom_write_byte(&locoentry[i].b[1], addr_high);
        eeprom_write_byte(&locoentry[i].b[0], addr_low);
        _write_loco_name(i, name);
        return(1);
      }

     //!warning! ohters: error!!!
     // too many locos with extra format
     return(0);
  }

void clear_loco_database(void)
  {
    unsigned char i;
    unsigned char stored_h, stored_l;

    for (i=1; i<SIZE_LOCO_DATABASE; i++)  // keep first entry
      {
        stored_h = eeprom_read_byte(&locoentry[i].b[1]);
        if (stored_h != 0x00)
          {
            eeprom_write_byte(&locoentry[i].b[1], 0);
          }
        stored_l = eeprom_read_byte(&locoentry[i].b[0]);
        if (stored_l != 0x00)
          {
            eeprom_write_byte(&locoentry[i].b[0], 0);
          }
      }

    // close chain at index 0
    eeprom_write_byte(&locoentry[0].next, 0);
    eeprom_write_byte(&locoentry[0].prev, 0);
    total_database_entry = 1;
    actual_database_index = 0;
  }


// fills structure actual

void get_loco_data(unsigned char index, t_locoentry* actual)
  {
    unsigned char j, data;

    if (index > 0)
      {
        actual->b[0] = eeprom_read_byte(&locoentry[index].b[0]);
        actual->b[1] = eeprom_read_byte(&locoentry[index].b[1]);

        actual->picture_id = eeprom_read_word(&locoentry[index].picture_id);

        for (j=0; j<sizeof(locoentry[0].name); j++)
          {
            if ((data = eeprom_read_byte(&locoentry[index].name[j])))   // yes we have a name
               actual->name[j] = data;
            else actual->name[j] = ' ';
          }
        return;
      }
    else
      { // do no read from eeprom, but use our local values
        memcpy(actual, &lok_neu, sizeof(lok_neu));
        actual->picture_id = 0;
      }
  }


// copy from structure actual to the given database entry
// no new position
void save_loco_data(unsigned char index, t_locoentry* actual)
  {
    unsigned char i;

    if (index > 0)
      {
        eeprom_write_byte(&locoentry[index].b[0], actual->b[0]);
        eeprom_write_byte(&locoentry[index].b[1], actual->b[1]);

        eeprom_write_word(&locoentry[index].picture_id, actual->picture_id);

        for (i=0; i<LOK_NAME_LENGTH; i++)
          {
            eeprom_write_byte(&locoentry[index].name[i], actual->name[i]);
          }
      }
    else
      { // do no write to eeprom, but use our local values
        memcpy(&lok_neu, actual, sizeof(lok_neu));
      }
  }

// 'search forward', returns the index of an entry found;

unsigned char scan_loco_database_forward(unsigned char index)
  {
    unsigned char i;

    i = eeprom_read_byte(&locoentry[index].next);
    if (i >= SIZE_LOCO_DATABASE)
      {
        // panic! chain must be broken
        i = 0;
        display_error(ERR_DATABASE);
      }
    // if (i==0) i = eeprom_read_byte(&locoentry[i].next);
    return(i);
  }


// 'search revers', returns the index of an entry found;

unsigned char scan_loco_database_revers(unsigned char index)
  {
    unsigned char i;

    i = eeprom_read_byte(&locoentry[index].prev);
    if (i >= SIZE_LOCO_DATABASE)
      {
        // panic! chain must be broken
        i = 0;
      }
    // if (i==0) i = eeprom_read_byte(&locoentry[i].prev);
    return(i);
  }



unsigned char format_to_uint8(t_format myformat)
  {  unsigned char i;
    switch(myformat)
      {
        default:   // no break;
        case DCC14:
                i = 14; break;
        case DCC27:
                i = 27; break;
        case DCC28:
                i = 28;  break;
        case DCC128:
                i = 126; break;
      }
    return(i);
  }

// 19.03.2013 leere emprom-Plaetze (also 0xFFFF) werden auch akzeptiert.
unsigned char find_empty_loco_data(void)
  {
    unsigned char i;
    unsigned char stored_h, stored_l;

    for (i=1; i<SIZE_LOCO_DATABASE; i++)         // start from 0
      {
        stored_h = eeprom_read_byte(&locoentry[i].b[1]);
        if (stored_h == 0x00)
          {
            stored_l = eeprom_read_byte(&locoentry[i].b[0]);
            if (stored_l == 0x00)
              {
                // empty found
                return(i);
              }
          }
        if (stored_h == 0xFF)
          {
            stored_l = eeprom_read_byte(&locoentry[i].b[0]);
            if (stored_l == 0xFF)
              {
                // empty found
                return(i);
              }
          }
      }
    return(0);
  }


// this creates a new entry and inserts it after the given index.
// the new index is returned
// index = 0: this is void

unsigned char insert_new_loco_data(unsigned char index)
  {
    unsigned char new_index;
    unsigned char temp_next;

    // 1. search for an empty place
    // 2. rearrange pointers

    if ((new_index = find_empty_loco_data()) != 0)
      {
        temp_next = eeprom_read_byte(&locoentry[index].next);
        eeprom_write_byte(&locoentry[new_index].next, temp_next);
        eeprom_write_byte(&locoentry[index].next, new_index);
        eeprom_write_byte(&locoentry[new_index].prev, index);
        eeprom_write_byte(&locoentry[temp_next].prev, new_index);

        total_database_entry++;
      }
    return(new_index);
  }


void delete_loco_data(unsigned char index)
  {
    unsigned char temp_next, temp_prev;

    if (index > 0)            // we do not allow index 0 to be deleted
      {
        // 1. delete content

        eeprom_write_byte(&locoentry[index].b[0], 0);
        eeprom_write_byte(&locoentry[index].b[1], 0);

        // 2. rearrange pointers

        temp_next = eeprom_read_byte(&locoentry[index].next);
        temp_prev = eeprom_read_byte(&locoentry[index].prev);

        eeprom_write_byte(&locoentry[temp_next].prev, temp_prev);
        eeprom_write_byte(&locoentry[temp_prev].next, temp_next);

        total_database_entry--;

        // if (index == actual_database_index)        // if our ref is stolen
        //  {
        //    actual_database_index = temp_next;
        //  }
      }
  }

// the chain at old is broken and the entry goes after new_index
void move_loco_data(unsigned char new_index, unsigned char old_index)
  {
    unsigned char temp_next, temp_prev;

    // 1. rearrange pointers old place

    temp_next = eeprom_read_byte(&locoentry[old_index].next);
    temp_prev = eeprom_read_byte(&locoentry[old_index].prev);

    eeprom_write_byte(&locoentry[temp_next].prev, temp_prev);
    eeprom_write_byte(&locoentry[temp_prev].next, temp_next);

    // 2. rearrange pointers new place

    temp_next = eeprom_read_byte(&locoentry[new_index].next);
    eeprom_write_byte(&locoentry[old_index].next, temp_next);
    eeprom_write_byte(&locoentry[new_index].next, old_index);
    eeprom_write_byte(&locoentry[old_index].prev, new_index);
    eeprom_write_byte(&locoentry[temp_next].prev, old_index);
  }


unsigned char get_loco_in_database(void)
  {
    return(total_database_entry);
  } 

unsigned char count_loco_in_database(void)
  {
    unsigned char i;
    unsigned char database_entry = 0;
    unsigned char index = 0;
    unsigned char stored_h, stored_l;

    index = 0;
    index = scan_loco_database_forward(index);

    for (i=0; i<SIZE_LOCO_DATABASE; i++)
      {
        if (index != 0)
          {
            database_entry++;
            // check this entry
            stored_l = eeprom_read_byte(&locoentry[i].b[0]);
         
            if (stored_l == 0)
              {
                stored_h = eeprom_read_byte(&locoentry[i].b[1]);
                if ((stored_h & 0x3f) == 0)
                  {
                    display_error(ERR_DATABASE_ADDR0);
                  }
              }
            index = scan_loco_database_forward(index);
          }
      }
    return(database_entry);
  } 








//!warning!

// #if (XPRESSNET_ENABLED == 1)
//=================================================================================
//
// run_database: multitask replacement, must be called in loop
//
// This task transports the loco data base to connected throttles
//
//=================================================================================
enum db_run_states
  {                                 // actual state
     IDLE,
     DB_XMIT,
     DB_WAIT,
     DB_WAIT2
  } db_run_state;

signed char next_database_run;   // timer variable to create a update grid;
#define DB_UPDATE_PERIOD  100000L

// Note: this message must be transmitted as CALL - like if it is transmitted from another client after a call!
void xmit_locoentry(void)
  {
    t_locoentry report;
    unsigned char i; // t_format myformat;

    next_search_index = scan_loco_database_forward(next_search_index);

    if (next_search_index < SIZE_LOCO_DATABASE)
      {
        get_loco_data(next_search_index, &report);
        // db_message[0] = 0xE0;
		db_message[1] = 0xF1;
        db_message[2] = report.b[1] & 0x3F;     // addr high
        db_message[3] = report.b[0];            // addr low
        db_message[4] = cur_database_entry;
        db_message[5] = total_database_entry;
        for (i=0; i<sizeof(report.name); i++)
          {
            if (report.name[i] == 0)
              {
                break;
              }
            db_message[6+i] = report.name[i];
          }
        db_message[0] = 0xE0 + 5 + i;
        db_message_ready = 1;                   // tell xpressnet that we have a message
      }
  }



void run_database(void)
  {
    switch (db_run_state)
      {
        case IDLE:
            break;

        case DB_XMIT:
            if (cur_database_entry >= total_database_entry)
              {
                db_run_state = IDLE; return;        // all done
              }

            xmit_locoentry();
            cur_database_entry++;
            next_database_run = get_systick() + (DB_UPDATE_PERIOD / SYSTICK_PERIOD);    // next wake up
            db_run_state = DB_WAIT;
            break;

        case DB_WAIT:
            if ((signed char)(next_database_run - get_systick()) >= 0)  return;

            db_message_ready = 1;                   // send again
            next_database_run = get_systick() + (DB_UPDATE_PERIOD / SYSTICK_PERIOD);    // next wake up
            db_run_state = DB_WAIT2;
            break;

        case DB_WAIT2:
            if ((signed char)(next_database_run - get_systick()) >= 0)  return;
            db_run_state = DB_XMIT;
            break;
     }
  }


void rewind_database(void)
  {
    next_search_index = 0;
  }

void do_xmit_database(void)
  {
    if (db_run_state != IDLE) return;
    db_run_state = DB_XMIT;
    rewind_database();
    cur_database_entry = 0;     // restart
  }



//---------------------------------------------------------------------------
//
// Init the loco database
//
//---------------------------------------------------------------------------

unsigned char read_curr_index(void)
  {
    return(eeprom_read_byte((void *)eadr_lok_search_entry));
  }

void store_curr_index(unsigned char index)
  {
    eeprom_write_byte((void *)eadr_lok_search_entry, index );
  }


void read_actual_waddr(void)
  {
    t_data16 temp;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_waddr0_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_waddr0_h);
    actual_weiche[0].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_waddr1_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_waddr1_h);
    actual_weiche[1].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_waddr2_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_waddr2_h);
    actual_weiche[2].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_waddr3_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_waddr3_h);
    actual_weiche[3].addr = temp.as_int16;
  }


void write_actual_waddr(void)
  {
    t_data16 temp;

    temp.as_int16 = actual_weiche[0].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_waddr0_l))
        eeprom_write_byte((void *)eadr_waddr0_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_waddr0_h))
        eeprom_write_byte((void *)eadr_waddr0_h, temp.as_uint8[1]);

    temp.as_int16 = actual_weiche[1].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_waddr1_l))
        eeprom_write_byte((void *)eadr_waddr1_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_waddr1_h))
        eeprom_write_byte((void *)eadr_waddr1_h, temp.as_uint8[1]);

    temp.as_int16 = actual_weiche[2].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_waddr2_l))
        eeprom_write_byte((void *)eadr_waddr2_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_waddr2_h))
        eeprom_write_byte((void *)eadr_waddr2_h, temp.as_uint8[1]);

    temp.as_int16 = actual_weiche[3].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_waddr3_l))
        eeprom_write_byte((void *)eadr_waddr3_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_waddr3_h))
        eeprom_write_byte((void *)eadr_waddr3_h, temp.as_uint8[1]);
  }


void read_actual_saddr(void)
  {
    t_data16 temp;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_saddr0_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_saddr0_h);
    actual_signal[0].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_saddr1_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_saddr1_h);
    actual_signal[1].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_saddr2_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_saddr2_h);
    actual_signal[2].addr = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_saddr3_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_saddr3_h);
    actual_signal[3].addr = temp.as_int16;
  }

void write_actual_saddr(void)
  {
    t_data16 temp;

    temp.as_int16 = actual_signal[0].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_saddr0_l))
        eeprom_write_byte((void *)eadr_saddr0_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_saddr0_h))
        eeprom_write_byte((void *)eadr_saddr0_h, temp.as_uint8[1]);

    temp.as_int16 = actual_signal[1].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_saddr1_l))
        eeprom_write_byte((void *)eadr_saddr1_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_saddr1_h))
        eeprom_write_byte((void *)eadr_saddr1_h, temp.as_uint8[1]);

    temp.as_int16 = actual_signal[2].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_saddr2_l))
        eeprom_write_byte((void *)eadr_saddr2_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_saddr2_h))
        eeprom_write_byte((void *)eadr_saddr2_h, temp.as_uint8[1]);

    temp.as_int16 = actual_signal[3].addr;
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_saddr3_l))
        eeprom_write_byte((void *)eadr_saddr3_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_saddr3_h))
        eeprom_write_byte((void *)eadr_saddr3_h, temp.as_uint8[1]);
  }



void init_database(void)
  {
    default_dcc_format = eeprom_read_byte((void *)eadr_default_dcc_format);
    // if misconfigured: use DCC28
    switch(default_dcc_format)
      {
        default:     default_dcc_format = DCC28; break;
        case DCC14:  break;
        case DCC27:  default_dcc_format = DCC28; break;
        case DCC28:  break;
        case DCC128: break;
      }

    actual_database_index = read_curr_index();

    eeprom_read_block(&lok_neu, &locoentry[0],sizeof(lok_neu));        // load preset for custom lok
    lok_neu.w.format = default_dcc_format;

    

    total_database_entry = count_loco_in_database();

    /*
    unsigned char i, j0, j1;
    for (i=0; i<SIZE_LOCO_DATABASE; i++)
      {
        j0 = eeprom_read_byte(&locoentry[i].b[0]);
        j1 = eeprom_read_byte(&locoentry[i].b[1]) & 0x3F;
        j0 = j0 | j1;
        if (j0)  // any address
          {
            if (eeprom_read_byte(&locoentry[i].name[0]) )  // any name
               total_database_entry++;
          }
      }   // we should also check chains

    */
    
    read_actual_waddr();
    read_actual_saddr();
  }




