//------------------------------------------------------------------------
//
// Emulate a general purpose graphics frontend
// 
//------------------------------------------------------------------------
//
// file:      debug_graphics.c
// history:   2010-07-03 V0.01 nn started
//
//------------------------------------------------------------------------
//
// purpose:   gprintf routines
//            
// copyright: THIS IS CONFIDENTIAL CODE
//
//            used hardware: MFT from www.opendcc.de
//
//------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include <string.h>
#include <stdbool.h>

#include "lcd_driver_st7565.h"     // basic calls to display
#include "lcd_graphics.h"

#include "debug_fonts.h"              // get pointer to pixels for a char


//-----------------------------------------------------------------------
// Hier werden die Graphikbefehle des PC auf Treibercalls umgesetzt.
// Fr Zeichenausgaben werden die Fonts in debug_fonts verwendet.
// Dort wird je nach gesetzter Codepage ein Pointer auf 
// width, height, pixels[] errechnet.


// Globals

// diese sind vor dem Aufruf von Fontroutinen von auen zu setzen

unsigned char display_codepage = 0x1F;

void sh_gprint(unsigned char z)
  {
    PGM_P z_ptr;
    z_ptr = get_cp_pointer(z);
    if (z_ptr == NULL)
      {
        // void, no pixel map available
        return;
      }
    
    LCD_drawBMP(z_ptr);
  }


// parameter: pointer to transferred message
//            size: size of message (sum of all display commands)
void parse_display(unsigned char *message, int size)
  {
    unsigned char command;
    unsigned char x1,y1,x2,y2;
    unsigned char px_on, px_off;
    int remaining = size;

    while (remaining > 0)
      {  
        command = *message++;
        switch(command)
          {
            default:    // PANIC! Command not known
                        // LED_A_RED; LED_B_RED; LED_C_RED; LED_D_RED;
                        return;
            case 0x00:  LCD_clr();
                        remaining -= 1;
                        break;

            case 0x28:  // [byte]
                        //  Setze Schriftart
                        //  Bit 7: 1: Schrift invers
                        //         0: Schrift normal
                        //  Bit 6: 1: Overwrite
                        //         0: Clear and Write
                        //  Bit 4-0:  Font
                        //
                        //  folgende Codes sind bisher aufgetreten:
                        //   D4  1101 0100  danach Lok oben links
                        //   DC  1101 1100  danach 'Fahrbetrieb'; dieser Font offenbar 10pixel
                        //   5B  0101 1010  danach Zugname (klein)
                        //   1B  0001 1010  danach Balken => bit 6 ist offenbar write, oder xor
                        //   52  0101 0010  danach Funktionsbilder; offenbar 11 oder 12 pixel gro.
                        //   5B  0101 1011  danach km
                        //   DB  1101 1011  danach Softkeys
                        if (*message & 0x80) display_invers = 0xff;
                        else                 display_invers = 0;
                        if (*message & 0x40) display_keep = 0;      // overwrite
                        else                 display_keep = 0xff;   // or-write
                        display_codepage = *message++ & 0x1F;
                        remaining -= 2;
                        break;

            case 0x69:  //[X],[Y],[char]
                        //  Gibt ein Zeichen an der Stelle X, Y aus   
                        x1 = *message++;
                        y1 = *message++;
                        x2 = *message++;            // this is the char to print
                        LCD_setCursorXY(x1,y1);
                        sh_gprint(x2);
                        remaining -= 4;
                  
            case 0x86:  // [X1], [Y1], [X2], [Y2] 
                        // RECT, X1, Y1, X2, Y2 malt ein nicht geflltes Rechteck auf dem Schirm.
                        // oder doch ein geflltes?
                        x1 = *message++;
                        y1 = *message++;
                        x2 = *message++;
                        y2 = *message++;
                        LCD_clearRect(x1+1,y1+1,x2-1,y2-1);  // testhalber
                        LCD_drawRect(x1,y1,x2,y2);
                        remaining -= 5;
                        break;
       
            case 0x87:  //[X1], [Y1], [X2], [Y2] 
                        //     FILL-AREA, X1, Y1, X2, Y2 und malt ein geflltes Rechteck auf dem Schirm.
                        //     z.B. 00 00 7F 09 also die obersten 10 Pixelreihen, volle Breite.
                        //     weiteres Beispiel: 87 60 37 7F 3F, ist der Fill unten bei MEN
                        x1 = *message++;
                        y1 = *message++;
                        x2 = *message++;
                        y2 = *message++;
                        if (display_invers) LCD_fillRect(x1,y1,x2,y2);
                        else                LCD_clearRect(x1,y1,x2,y2);
                        remaining -= 5;
                        break;

            case 0x89: //[X], [Y], [2 chars]
                        //  Gibt zwei Zeichen an der Stelle X, Y aus.
                        x1 = *message++;
                        y1 = *message++;
                        LCD_setCursorXY(x1,y1);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        remaining -= 5;
                        break;
        
            case 0x95:  //[X1], [Y1], [X2], [Y2] 
                        //         INVERT-AREA, z.B. beim Menuverschieben kommt:
                        //         95 03 0A 7C 12 : obere Zeile invertieren
                        //         95 03 13 7C 1B : unter Zeile invertieren, dadurch wandert der Balken runter.
                        x1 = *message++;
                        y1 = *message++;
                        x2 = *message++;
                        y2 = *message++;
                        LCD_invertRect(x1,y1,x2,y2);
                        remaining -= 5;
                        break;

            case 0xA9:  //[X], [Y], [3 chars]
                        //     Gibt drei Zeichen an der Stelle X, Y aus.
                        //     Beispiel: A9 09 38 41 6C 74 = Alt @ 9,0x38
                        //
                        x1 = *message++;
                        y1 = *message++;
                        LCD_setCursorXY(x1,y1);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        remaining -= 6;
                        break;
        
            case 0xC6:  //[X1], [Y1], [X2], [Y2], [L1], [L2]
                        //     Malt ein strichliertes Rechteck.
                        //     Beispiel: C6 0A 17 74 2D 01 01 
                        //     von x1,y1 10,23 bis 116,45: strichlierter Rahmen (1px on/1px off)
                        x1 = *message++;
                        y1 = *message++;
                        x2 = *message++;
                        y2 = *message++;
                        px_on = *message++;
                        px_off = *message++;
                        LCD_lineRect(x1,y1,x2,y2,px_on,px_off);
                        remaining -= 7;
                        break;
        
            case 0xC9:  //[X], [Y], [4 chars]
                        //     Gibt vier Zeichen an der Stelle X, Y aus.
                        x1 = *message++;
                        y1 = *message++;
                        LCD_setCursorXY(x1,y1);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        sh_gprint(*message++);
                        remaining -= 7;
                        break;
                
            case 0xE9:  //[mode], [Size], [X], [Y], [characters]
                        //     mode war in allen bisherigen calls immer 00, Size ist die Lnge des gesamten
                        //     Commands (also inkl. E9). Bei Charakters wird eine erweiterte ASCII-Tabelle
                        //     verwendet, siehe unten.
                        // E9 00 0D 04 26 4B 65 79 62 6F 61 72 64  = Text "Keyboard" 
                        //
                        message++;        // skip mode
                        x2 = *message++;  // size
                        x1 = *message++;
                        y1 = *message++;
                        LCD_setCursorXY(x1,y1);
                        for (y2 = 5; y2 < x2; y2++)
                          {
                            sh_gprint(*message++);
                          }
                        remaining -= x2;
                                          
                        break;
          }
      }
  }


#if (1==0)


unsigned char intro[] = {0x00,0x28,0x54,0x69,0x09,0x00,0x0E,0x28,0x5F,0xE9,
                         0x00,0x10,0x1E,0x15,0x4D,0x6F,0x64,0x65,0x6C,0x6C,
                         0x62,0x61,0x68,0x6E,0x2D,0xE9,0x00,0x10,0x1E,0x24,
                         0x48,0x61,0x6E,0x64,0x72,0x65,0x67,0x6C,0x65,0x72,
                         0x2D,0xE9,0x00,0x15,0x0D,0x33,0x53,0x74,0x65,0x75,
                         0x65,0x72,0x75,0x6E,0x67,0x73,0x73,0x79,0x73,0x74,
                         0x65,0x6D};

void display_intro(void)
  {
    parse_display(intro, sizeof(intro));
  }

#endif

#if (1==0)
//-------------------------------------------------------------------------------------------
//
//  TEST
//
//-------------------------------------------------------------------------------------------
// Lokschirm
unsigned char first[] = {0x00,
                         0x28,0xD4,
                            0x87,0x00,0x00,0x7F,0x09,
                            0x69,0x05,0x01,0x01,
                         0x28,0xDC,
                            0xE9,0x00,0x10,0x21,0x00,0x46,0x61,0x68,0x72,0x62,0x65,0x74,0x72,0x69,0x65,0x62,
                            0x89,0x74,0x00,0x30,0x31,
                         0x28,0x5B,
                            0xE9,0x00,0x1A,0x0C,0x0D,0x44,0x5F,0x30,0x33,0x38,0x20,0x57,0x61,0x6E,0x6E,0x65,0x6E,0x74,0x65,0x6E,0x64,0x65,0x72,0x5F,0x33,0x38,
                            0x86,0x14,0x17,0x6B,0x21,
                         0x28,0x1B,
                            0xE9,0x00,0x11,0x28,0x19,0x81,0x20,0x30,0x20,0x2F,0x20,0x30,0x20,0x6B,0x6D,0x2F,0x68,
                         0x28,0x52,
                            0xE9,0x00,0x0F,0x09,0x23,0x02,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                         0x28,0x5B,
                            0xE9,0x00,0x0A,0x00,0x2F,0x30,0x30,0x35,0x34,0x32,
                            0xA9,0x19,0x2F,0x20,0x6B,0x6D,
                            0xE9,0x00,0x0A,0x2E,0x2F,0x81,0x20,0x30,0x30,0x30,
                            0xE9,0x00,0x0A,0x44,0x2F,0x20,0x6B,0x6D,0x2F,0x68,
                            0xE9,0x00,0x0A,0x6A,0x2F,0x31,0x39,0x3A,0x34,0x34,0x28,0xDB,
                            0x87,0x00,0x37,0x1E,0x3F,
                            0xA9,0x09,0x38,0x41,0x6C,0x74,         // Alt
                            0x87,0x20,0x37,0x3E,0x3F,
                            0xE9,0x00,0x0C,0x22,0x38,0x4C,0x69,0x73,0x74,0x20,0x20,0x82,
                            0x87,0x40,0x37,0x5E,0x3F,
                            0xE9,0x00,0x0C,0x42,0x38,0x4C,0x69,0x73,0x74,0x20,0x20,0x83,
                            0x87,0x60,0x37,0x7F,0x3F,
                            0xC9,0x64,0x38,0x4D,0x65,0x6E,0xFC    // Menu
                            };

#define  SH_PRINT_1C  0x69   //  [X],[Y],[char]  Gibt ein Zeichen an der Stelle X, Y aus   
#define  SH_PRINT_2C  0x89   //  [X],[Y],[char][char]  Gibt zwei Zeichen an der Stelle X, Y aus   
#define  SH_PRINT_3C  0xA9   //  [X],[Y],[char][char][char]  Gibt drei Zeichen an der Stelle X, Y aus   
#define  SH_PRINT_4C  0xC9   //  [X],[Y],[char][char][char][char]  Gibt vier Zeichen an der Stelle X, Y aus   
#define  SH_PRINT_STR 0xE9   //  [mode], [Size], [X], [Y], [characters]
 
#define  SH_DRAWRECT  0x86   // [X1], [Y1], [X2], [Y2] // malt ein nicht geflltes Rechteck
#define  SH_FILLRECT  0x87   // [X1], [Y1], [X2], [Y2] // malt ein geflltes Rechteck
#define  SH_INVTRECT  0x95   // [X1], [Y1], [X2], [Y2] // invertiert ein Rechteck (Menucursor)
#define  SH_LINERECT  0xC6   // [X1], [Y1], [X2], [Y2], [L1], [L2] //  Malt ein strichliertes Rechteck.

unsigned char xxxxx[] = {0x00,
                         0x28,0xD4,
                            SH_FILLRECT,0x00,0x00,0x7F,0x09,
                            SH_PRINT_1C,   5,  1,0x01,
                         0x28,0xDC,
                            SH_PRINT_STR,0x00,0x10,0x21,0x00,'F','a','h','r','b','e','t','r','i','e','b',
                            SH_PRINT_2C, 116,  0,'0','1',
                         0x28,0x5B,
                            SH_PRINT_STR,0x00,0x1A,0x0C,0x0D,0x44,0x5F,'0','3','8',' ','W','a','n','n','e','n','t','e','n','d','e','r',0x5F,'3','8',
                            0x86,0x14,0x17,0x6B,0x21,
                         0x28,0x1B,
                            SH_PRINT_STR,0x00,0x11,0x28,0x19,0x81,0x20,0x30,0x20,0x2F,0x20,0x30,0x20,0x6B,0x6D,0x2F,0x68,
                         0x28,0x52,
                            SH_PRINT_STR,0x00,0x0F,0x09,0x23,0x02,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                         0x28,0x5B,
                            SH_PRINT_STR,0x00,0x0A,0x00,0x2F,0x30,0x30,0x35,0x34,0x32,
                            SH_PRINT_3C,0x19,0x2F,0x20,0x6B,0x6D,
                            SH_PRINT_STR,0x00,0x0A,0x2E,0x2F,0x81,0x20,0x30,0x30,0x30,
                            SH_PRINT_STR,0x00,0x0A,0x44,0x2F,0x20,0x6B,0x6D,0x2F,0x68,
                            SH_PRINT_STR,0x00,0x0A,0x6A,0x2F,0x31,0x39,0x3A,0x34,0x34,0x28,0xDB,
                            SH_FILLRECT,0x00,0x37,0x1E,0x3F,
                            SH_PRINT_3C,0x09,0x38,0x41,0x6C,0x74,         // Alt
                            SH_FILLRECT,0x20,0x37,0x3E,0x3F,
                            SH_PRINT_STR,0x00,0x0C,0x22,0x38,'L','i','s','t',' ',' ',0x82,
                            SH_FILLRECT,0x40,0x37,0x5E,0x3F,
                            SH_PRINT_STR,0x00,0x0C,0x42,0x38,'L','i','s','t',' ',' ',0x83,
                            SH_FILLRECT,0x60,0x37,0x7F,0x3F,
                            SH_PRINT_4C,0x64,0x38,0x4D,0x65,0x6E,0xFC     // Menu
                            };


unsigned char testdat1[] = {0x00,
                         0x28,0xD4,
                            SH_FILLRECT,0x00,0x00,0x7F,0x09,
                            SH_PRINT_1C,   5,  1,0x01,
                         };
unsigned char testdat2[] = { 0x28,0xDC,
                            SH_PRINT_STR,0x00,0x10,0x21,0x00,'F','a','h','r','b','e','t','r','i','e','b',
                            SH_PRINT_2C, 116,  0,'0','1',
                         };
unsigned char testdat3[] = {0x28,0x5B,
                                SH_PRINT_STR,0x00,0x1A,0x0C,0x0D,0x44,0x5F,'0','3','8',' ','W','a','n','n','e','n','t','e','n','d','e','r',0x5F,'3','8',
                                0x86,0x14,0x17,0x6B,0x21,
                             0x28,0x1B,
                                SH_PRINT_STR,0x00,0x11,0x28,0x19,0x81,0x20,0x30,0x20,0x2F,0x20,0x30,0x20,0x6B,0x6D,0x2F,0x68,
                             0x28,0x52,
                                SH_PRINT_STR,0x00,0x0F,0x09,0x23,0x02,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
                         };
unsigned char testdat4[] = {0x28,0x5B,
                            SH_PRINT_STR,0x00,0x0A,0x00,0x2F,0x30,0x30,0x35,0x34,0x32,
                            SH_PRINT_3C,0x19,0x2F,0x20,0x6B,0x6D,                       // km
                         };
unsigned char testdat5[] = {SH_PRINT_STR,0x00,0x0A,0x2E,0x2F,0x81,0x20,0x30,0x30,0x30,  // 2f = 47
                         };
unsigned char testdat6[] = { SH_PRINT_STR,0x00,0x0A,0x44,0x2F,0x20,0x6B,0x6D,0x2F,0x68,
                            SH_PRINT_STR,0x00,0x0A,0x6A,0x2F,0x31,0x39,0x3A,0x34,0x34,0x28,0xDB,
                            SH_FILLRECT,0x00,0x37,0x1E,0x3F,
                            SH_PRINT_3C,0x09,0x38,0x41,0x6C,0x74,         // Alt
                            SH_FILLRECT,0x20,0x37,0x3E,0x3F,
                            SH_PRINT_STR,0x00,0x0C,0x22,0x38,'L','i','s','t',' ',' ',0x82,
                            SH_FILLRECT,0x40,0x37,0x5E,0x3F,
                            SH_PRINT_STR,0x00,0x0C,0x42,0x38,'L','i','s','t',' ',' ',0x83,
                            SH_FILLRECT,0x60,0x37,0x7F,0x3F,
                            SH_PRINT_4C,0x64,0x38,0x4D,0x65,0x6E,0xFC     // Menu
                          };

unsigned char line_rect[] = {0xC6,   5,   2, 122,  61,   2,   1,};

void display_dotted(void)
  {
    parse_display(line_rect, sizeof(line_rect));
  }



#include "config.h"
#include "hardware.h"
void test_keyb(void)  // return only if STOP pressed
  {
    while ((PINC & (1 << KEY_7)) == 0) {};  // warten auf auslassen

    _mydelay_us(20000L);

    while(1)
      {
         if ((PINC & (1 << KEY_7)) == 0)
          {
            return;
          }
      }
  }

void test(void)
  {
    parse_display(testdat1, sizeof(testdat1));
    update_display_complete();
    test_keyb();

    parse_display(testdat2, sizeof(testdat2));
    update_display_complete();
    test_keyb();

    parse_display(testdat3, sizeof(testdat3));
    update_display_complete();
    test_keyb();

    parse_display(testdat4, sizeof(testdat4));
    update_display_complete();
    test_keyb();

    parse_display(testdat5, sizeof(testdat5));
    update_display_complete();
    test_keyb();

    parse_display(testdat6, sizeof(testdat6));
    update_display_complete();
    test_keyb();
  }
#endif
