//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Jean Michel Fischer, Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//
//-----------------------------------------------------------------
//
// file:      display.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-18 V0.01 jmf started
//            2009-01-13 V0.02 kw  comments, tests
//            2009-01-26 V0.03 kw  splash + set_cursor jetzt ab 0,0
//            2009-03-02 V0.04 kw  display_batt_state added
//            2009-03-14 V0.05 kw  added void display_error
//            2009-03-27 V0.06 kw  added display_softkey
//            2009-06-30 V0.07 kw  added defines for display position
//                             kw  added wireless status display
//                             kw  added display_time
//            2010-05-11 V0.08 kw  added display_connect with heartbeat
//
//-----------------------------------------------------------------
//
// purpose:   supply mid-level routines,
//            showing the content
//
// Interface: downstream: lcd_text.h
//            upstream:
//
//-----------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <util/delay.h>
#include <stdio.h>

#include <inttypes.h>
#include <string.h>
#include <stdbool.h>


#include "config.h"
#include "timer.h"
#include "hardware.h"           // CS for SPI
#include "analog.h"
#include "keycodes.h"
#include "cortos.h"                 // fr messages_objects.h
#include "led.h"

#include "lcd_text.h"
#include "lcd_ks0073_charset.h"

#include "lcd_driver_st7565.h"

#include "messages_objects.h"



#if (DISPLAY == LCD_EA204)
    #define DP_POSX_TIME        8
    #define DP_POSX_CONNECT    13
    #define DP_POSX_BATT       14
    #define DP_POSX_SYSTEM     15
#elif  (DISPLAY == LCD_EA128x64)
    #define DP_POSX_TIME       48
    #define DP_POSX_CONNECT    (48+28)
    #define DP_POSX_BATT       (48+28+8)
    #define DP_POSX_SYSTEM     (48+28+8+14)    // =98
#endif


unsigned char display_time_allowed;       // Flag to suppress unwanted displays




void display_text(unsigned char col, unsigned char row, char * text)
  {
    set_cursor(col, row);
    lcd_string(text);
  }


void display_loc_name(char * name)
  {
    unsigned char i;
    unsigned char fill = 0;
    for(i=0; i< LOK_NAME_LENGTH; i++)
      {
        if (fill) lcd_putc(' ');                     // fill rest with blank
        else
          {
            if (*name) lcd_putc(*name++);
            else
              {
                fill = 1;
                lcd_putc(' ');
              }
          }
      }
  }


//-----------------------------------------------------------------------
//
// display_status and wheelie
//
//-----------------------------------------------------------------------
#if (DISPLAY == LCD_EA204)
// char text_wheelie[]          PROGMEM = {C_VERT_LINE, '/', '-' , C_BACKSLASH};
char text_wheelie[]          PROGMEM = {C_ARROW_LEFT_UP,C_ARROW_LEFT_UP,C_ARROW_RIGHT_UP,C_ARROW_RIGHT_UP,
                                   C_ARROW_RIGHT_DOWN,C_ARROW_RIGHT_DOWN,C_ARROW_LEFT_DOWN,C_ARROW_LEFT_DOWN};
static unsigned char wheelie;


//   01234567890123456789
//  |--------------------|
// 0|               STOP |
// 1|                    |
// 2|                    |
// 3|                    |
//  |--------------------|

char text_sts_on[]           PROGMEM = {" ON  "};
char text_sts_off[]          PROGMEM = {" OFF "};
char text_sts_stop[]         PROGMEM = {"STOP "};
char text_sts_prog[]         PROGMEM = {"PROG "};
char text_sts_busy[]         PROGMEM = {"BUSY "};
char text_sts_pbusy[]        PROGMEM = {"PGM:"};   // + wheelie
char text_sts_disconnected[] PROGMEM = {"<=#=>"};


void display_status(t_status _status)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    get_cursor(&saved_cursor_x, &saved_cursor_y);
    set_cursor(DP_POSX_SYSTEM,0);
    switch (_status)
      {
        case STS_ON:
            lcd_string_P(text_sts_on);
            break;
        case STS_OFF:
            lcd_string_P(text_sts_off);
            break;
        case STS_STOP:
            lcd_string_P(text_sts_stop);
            break;
        case STS_PROG:
            lcd_string_P(text_sts_prog);
            break;
        case STS_BUSY:
            lcd_string_P(text_sts_busy);
            break;
        case STS_PBUSY:
            lcd_string_P(text_sts_pbusy);
            // make a wheelie();
            wheelie++;
            wheelie %= sizeof(text_wheelie);
            lcd_putc(pgm_read_byte(&text_wheelie[wheelie]));
            break;
        case STS_DISCONNECTED:
            lcd_string_P(text_sts_disconnected);
            break;
        case STS_CMD_ERROR:
            set_cursor(19,0);
            lcd_putc('?');
            break;
      }
    set_cursor(saved_cursor_x, saved_cursor_y);
  }
#elif  (DISPLAY == LCD_EA128x64)

#include "bmp/disconnected_glcd.h"
#include "bmp/wheelie_glcd.h"


static unsigned char wheelie;


//   01234567890123456789
//  |--------------------|
// 0|               STOP |
// 1|                    |
// 2|                    |
// 3|                    |
//  |--------------------|



const char text_sts_on[]           PROGMEM = {" ON  "};
const char text_sts_off[]          PROGMEM = {" OFF "};
const char text_sts_stop[]         PROGMEM = {"STOP "};
const char text_sts_prog[]         PROGMEM = {"PROG "};
const char text_sts_busy[]         PROGMEM = {"BUSY "};
const char text_sts_pbusy[]        PROGMEM = {"PGM:"};
const char text_sts_disconnected[] PROGMEM = {"<=#=>"};


void display_status(t_status _status)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    LCD_getCursorXY(&saved_cursor_x, &saved_cursor_y);
    LCD_setCursorXY(DP_POSX_SYSTEM,0);
    switch (_status)
      {
        case STS_ON:
            lcd_string_P(text_sts_on);
            break;
        case STS_OFF:
            lcd_string_P(text_sts_off);
            break;
        case STS_STOP:
            lcd_string_P(text_sts_stop);
            break;
        case STS_PROG:
            lcd_string_P(text_sts_prog);
            break;
        case STS_BUSY:
            lcd_string_P(text_sts_busy);
            break;
        case STS_PBUSY:
            lcd_string_P(text_sts_pbusy);
            // make a wheelie();
            LCD_setCursorXY(DP_POSX_SYSTEM + 23,0);
            wheelie++;
            wheelie %= WHEELIE_GLCD_HEIGHT / 8;
            LCD_fillPage_with_BMP(WHEELIE_GLCD_WIDTH, &wheelie_glcd_bmp[wheelie*WHEELIE_GLCD_WIDTH]);
            break;
        case STS_DISCONNECTED:
            // lcd_string_P(text_sts_disconnected);
            LCD_fillPage_with_BMP(DISCONNECTED_GLCD_WIDTH, &disconnected_glcd_bmp[0]);
            break;
        case STS_CMD_ERROR:
            LCD_setCursorXY(DP_POSX_SYSTEM+24,0);
            lcd_putc('?');
            break;
      }
    LCD_setCursorXY(saved_cursor_x, saved_cursor_y);
  }

#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY
//-----------------------------------------------------------------------



//-----------------------------------------------------------------------
//
// display_batt_state
//
//-----------------------------------------------------------------------
#if (DISPLAY == LCD_EA204)
void display_batt_state(void)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    get_cursor(&saved_cursor_x, &saved_cursor_y);
    set_cursor(DP_POSX_BATT,0);

    switch(batt_state)
      {
        case KEY_BATT_EMPTY:    lcd_putc(C_BLOCK_M4); break;
        case KEY_BATT_20_100:   lcd_putc(C_BLOCK_M3); break;
        case KEY_BATT_40_100:   lcd_putc(C_BLOCK_M2); break;
        case KEY_BATT_60_100:   lcd_putc(C_BLOCK_M1); break;
        case KEY_BATT_FULL:
            if (g_vccio_state)    lcd_putc(C_BULLET);
            else                lcd_putc(C_BLOCK_M0);
            break;
        case KEY_VCCIO:
            if (g_vccio_state)
              {
                g_power_mode = LINE_POWERED;
              }
            else
              {
                g_power_mode = BATT_POWERED;

              }
            restore_brightness(g_power_mode);
            break;
        default: break;
      }
    set_cursor(saved_cursor_x, saved_cursor_y);
  }

#elif  (DISPLAY == LCD_EA128x64)

#include "bmp/batt_glcd.h"

void display_batt_state(void)
  {
    // Batt is 14 pixels wide
    #if (BATT_GLCD_WIDTH != 14)
     #warning check position of BATT BMP
    #endif

    unsigned char saved_cursor_x, saved_cursor_y;
    LCD_getCursorXY(&saved_cursor_x, &saved_cursor_y);
    LCD_setCursorXY(DP_POSX_BATT,0);

    switch(batt_state)
      {
        case KEY_BATT_EMPTY:    LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[0*BATT_GLCD_WIDTH]); break;
        case KEY_BATT_20_100:   LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[1*BATT_GLCD_WIDTH]); break;
        case KEY_BATT_40_100:   LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[2*BATT_GLCD_WIDTH]); break;
        case KEY_BATT_60_100:   LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[3*BATT_GLCD_WIDTH]); break;
        case KEY_BATT_FULL:
            if (g_vccio_state)  LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[5*BATT_GLCD_WIDTH]);
            else                LCD_fillPage_with_BMP(BATT_GLCD_WIDTH, &batt_glcd_bmp[4*BATT_GLCD_WIDTH]);
            break;
        case KEY_VCCIO:
            if (g_vccio_state)                       // das hier zu verarbeiten ist nicht sauber
              {
                g_power_mode = LINE_POWERED;
              }
            else
              {
                g_power_mode = BATT_POWERED;

              }
            restore_brightness(g_power_mode);
            break;
        default: break;
      }
    LCD_setCursorXY(saved_cursor_x, saved_cursor_y);
  }
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY

//-----------------------------------------------------------------------
//
// display_connect_state
//
//-----------------------------------------------------------------------
#if (DISPLAY == LCD_EA204)
void display_connect_state(t_connect_state connect_state)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    get_cursor(&saved_cursor_x, &saved_cursor_y);
    set_cursor(DP_POSX_CONNECT,0);

    switch(connect_state)
      {
        default:
        case DISCONNECTED:        lcd_putc(C_XI);  break;  // greek Xi
        case CONNECT_XPRESSNET:   lcd_putc('X');   break;
        case CONNECT_XBEE:        lcd_putc(C_PSI); break;  // greek Psi
      }
    set_cursor(saved_cursor_x, saved_cursor_y);
  }

#elif  (DISPLAY == LCD_EA128x64)

#include "bmp/connect_glcd.h"

static unsigned char dconnect_phase;                      // for displaying heart beat

void display_connect_state(t_connect_state connect_state)
  {
    // Wireless ICON is 7 pixels wide
    #if (CONNECT_GLCD_WIDTH != 7)
     #warning check position of CONNECT BMP
    #endif

    unsigned char saved_cursor_x, saved_cursor_y;
    LCD_getCursorXY(&saved_cursor_x, &saved_cursor_y);
    LCD_setCursorXY(DP_POSX_CONNECT,0);

    switch(connect_state)
      {
        default:
        case DISCONNECTED:
            LCD_fillPage_with_BMP(CONNECT_GLCD_WIDTH, &connect_glcd_bmp[0*CONNECT_GLCD_WIDTH]); // broken icon
            break;
        case CONNECT_XPRESSNET:
            LCD_fillPage_with_BMP(CONNECT_GLCD_WIDTH, &connect_glcd_bmp[1*CONNECT_GLCD_WIDTH]); // wire icon
            break;
        case CONNECT_XBEE:
            if (dconnect_phase)
              {
                dconnect_phase = 0;
                LCD_fillPage_with_BMP(CONNECT_GLCD_WIDTH, &connect_glcd_bmp[2*CONNECT_GLCD_WIDTH]);
              }
            else
              {
                dconnect_phase = 1;
                LCD_fillPage_with_BMP(CONNECT_GLCD_WIDTH, &connect_glcd_bmp[3*CONNECT_GLCD_WIDTH]);
              }
            break;
      }

    LCD_setCursorXY(saved_cursor_x, saved_cursor_y);
  }
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// display_time
//
//-----------------------------------------------------------------------
#if (DISPLAY == LCD_EA204)
void display_time(void)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    if (display_time_allowed)
      {
        get_cursor(&saved_cursor_x, &saved_cursor_y);
        set_cursor(DP_POSX_TIME,0);
        if (fast_clock.hour < 24 && fast_clock.minute < 60)
          {
            lcd_int02(fast_clock.hour);
            lcd_putc(':');
            lcd_int02(fast_clock.minute);
          }
        set_cursor(saved_cursor_x, saved_cursor_y);
      }
  }

#elif  (DISPLAY == LCD_EA128x64)

void display_time(void)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    if (display_time_allowed)
      {
        LCD_getCursorXY(&saved_cursor_x, &saved_cursor_y);
        LCD_setCursorXY(DP_POSX_TIME,0);
        if (fast_clock.hour < 24 && fast_clock.minute < 60)
          {
            lcd_int02(fast_clock.hour);
            LCD_PF_putchar(':');   // use proportional font insteac lcd_putc(':');
            lcd_int02(fast_clock.minute);
          }
        LCD_setCursorXY(saved_cursor_x, saved_cursor_y);
      }
  }
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY



//-----------------------------------------------------------------------
//
// Softkeys
//
//-----------------------------------------------------------------------
#if (DISPLAY == LCD_EA204)

// no border checking is performed
//  Softkey ranges
//  XX 0 XX | XX 1 XX | XX 2 XX | XX 3 XX
//  0.... 4   5... 9    10...14   15..19

void display_softkey_P(uint8_t key, const char *progmem_data)
  {
    uint8_t sizeX;

    sizeX = strlen_P(progmem_data);
    sizeX /= 2;
    if (sizeX > 2) sizeX = 2;  // limit offset to half width

    switch(key)
      {
        default:
        case 0:
            set_cursor( 0, 3);
            lcd_blank(5);
            set_cursor( 0 + 2 - sizeX, 3);
            lcd_string_P(progmem_data);
            break;
        case 1:
            set_cursor( 5, 3);
            lcd_blank(5);
            set_cursor( 5 + 2 - sizeX, 3);
            lcd_string_P(progmem_data);
            break;
        case 2:
            set_cursor(10, 3);
            lcd_blank(5);
            set_cursor(10 + 2 - sizeX, 3);
            lcd_string_P(progmem_data);
            break;
        case 3:
            set_cursor(15, 3);
            lcd_blank(5);
            set_cursor(15 + 2 - sizeX, 3);
            lcd_string_P(progmem_data);
            break;
      }
  }

void clear_softkey(uint8_t key)
  {
    switch(key)
      {
        default:
        case 0:
            set_cursor( 0, 3);
            lcd_blank(5);
            break;
        case 1:
            set_cursor( 5, 3);
            lcd_blank(5);
            break;
        case 2:
            set_cursor(10, 3);
            lcd_blank(5);
            break;
        case 3:
            set_cursor(15, 3);
            lcd_blank(5);
            break;
      }
  }

#elif  (DISPLAY == LCD_EA128x64)

unsigned char LCD_PF_get_stringsizeX_P(const char *progmem_data)
  {
    unsigned char c;
    uint8_t i = 0;

    while ( (c = pgm_read_byte(progmem_data++)) )
      {
        i += LCD_PF_get_sizeX(c);
      }
    return(i);
  }

unsigned char LCD_PF_get_stringsizeX(const char *data)
  {
    unsigned char c;
    uint8_t i = 0;

    while ( (c = *data++) )
      {
        i += LCD_PF_get_sizeX(c);
      }
    return(i);
  }

void  LCD_PF_string_P(const char *progmem_data)
  {
    unsigned char c;

    while ( (c = pgm_read_byte(progmem_data++)) )
      {
        LCD_PF_putchar(c);
      }
  }

void  LCD_PF_string(const char *data)
  {
    unsigned char c;

    while ( (c = *data++) )
      {
        LCD_PF_putchar(c);
      }
  }
// no border checking is performed
//  Softkey ranges
//  XX 0 XX | XX 1 XX | XX 2 XX | XX 3 XX
//  0....30   32...62   64...94   96..127
void display_softkey(uint8_t key, const char *data)
  {
    uint8_t sizeX;

    sizeX = LCD_PF_get_stringsizeX(data);
    sizeX /= 2;
    if (sizeX > 15) sizeX = 15;  // limit offset to half width

    switch(key)
      {
        default:
        case 0:
            LCD_offPage(7, 0, 30);  // page = 7 => y= 56
            LCD_setCursorXY( 0 + 15 - sizeX, 7 * LCD_PAGE_HEIGHT);
            LCD_PF_string(data);
            LCD_invertPage(7, 0, 30);
            break;
        case 1:
            LCD_offPage(7, 32, 62);  // page = 7 => y= 56
            LCD_setCursorXY(32 + 15 - sizeX, 7 * LCD_PAGE_HEIGHT);
            LCD_PF_string(data);
            LCD_invertPage(7, 32, 62);
            break;
        case 2:
            LCD_offPage(7, 64, 94);  // page = 7 => y= 56
            LCD_setCursorXY(64 + 15 - sizeX, 7 * LCD_PAGE_HEIGHT);
            LCD_PF_string(data);
            LCD_invertPage(7, 64, 94);
            break;
        case 3:
            LCD_offPage(7, 96, 127);  // page = 7 => y= 56
            LCD_setCursorXY(96 + 15 - sizeX, 7 * LCD_PAGE_HEIGHT);
            LCD_PF_string(data);
            LCD_invertPage(7, 96, 127);
            break;
      }
  }

void display_softkey_P(uint8_t key, const char *progmem_data)
  {
    char ram_data[10];
    strncpy_P (ram_data, progmem_data, 10-1);
    ram_data[9] = 0; // terminate anyway
    display_softkey(key, ram_data);
  }

void clear_softkey(uint8_t key)
  {
    switch(key)
      {
        default:
        case 0:
            LCD_offPage(7, 0, 30);  // page = 7 => y= 56
            break;
        case 1:
            LCD_offPage(7, 32, 62);  // page = 7 => y= 56
            break;
        case 2:
            LCD_offPage(7, 64, 94);  // page = 7 => y= 56
            break;
        case 3:
            LCD_offPage(7, 96, 127);  // page = 7 => y= 56
            break;
      }
  }

#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------

void check_status_display(unsigned char key)
  {
    if (key == 0)
      {
        display_status(command_station_status);
        display_batt_state();
        display_time();
        display_connect_state(get_connection_status());
      }
    else
      {
        if ((key >= KEY_BATT_EMPTY) && (key <= KEY_VCCIO))
          {
            display_batt_state();
          }
        else if ((key == KEY_CONNECT) || (key == KEY_DISCONNECT) || (key == KEY_HEARTBEAT))
          {
            display_connect_state(get_connection_status());
          }
        else if (key == KEY_PARSER_STATE)                  // update system info
          {
            // new state info from command station
            // if changed, update display
            display_status(command_station_status);
          }
        else if (key == KEY_PARSER_TIME)
          {
            display_time();
          }
      }
  }

const static char text_error[]           PROGMEM = {"E:"};

void display_error(unsigned char errnum)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    get_cursor(&saved_cursor_x, &saved_cursor_y);
    set_cursor(0,0);
    lcd_string_P(text_error);
    lcd_hex2(errnum);
    lcd_putc('!');
    set_cursor(saved_cursor_x, saved_cursor_y);
  }

void display_error2(unsigned char errnum,unsigned char code)
  {
    unsigned char saved_cursor_x, saved_cursor_y;
    get_cursor(&saved_cursor_x, &saved_cursor_y);
    set_cursor(0,0);
    lcd_string_P(text_error);
    lcd_hex2(errnum);
    lcd_putc(':');
    lcd_hex2(code);
    lcd_putc('!');
    set_cursor(saved_cursor_x, saved_cursor_y);
  }

//-----------------------------------------------------------------------
//
// SPLASH - DISPLAY - CHAR
//
//-----------------------------------------------------------------------

