//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
// Here we do the access to a spi flash
//
//-----------------------------------------------------------------
//
// file:      flash_driver_m25p80.c
// author:    Wolfgang Kufer
// contact:   kufer_at_gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-07-04 V0.07 kw  started
//
//-----------------------------------------------------------------
//
// purpose:   control of a spi driven flash
//
// howto:     3 Layers:
//            Layer 0: doing the SPI_transfer
//            Layer 1: running the commands for m25p80
//            Layer 2: performing read, write, erase
//
//            handling of SPI_CS:
//            we assert and deassert CS in layer 1 to gruup commands
//            we dont use HOLD or PROTECTION
//
//-----------------------------------------------------------------

#include <inttypes.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>          // put var to program memory
#include <util/delay.h>

#include "hardware.h"
#include "keycodes.h"           // table (scan -> keystroke)
#include "timer.h"              // systick
#include "cortos.h"             // cooperative scheduler and fifo

#include "keyboard.h"

//
//
//-----------------------------------------------------------------
//
// Layer 0
//
//-----------------------------------------------------------------

static inline void SPI_init(void)
  {
    // Enable SPI, Master, set clock rate fck/2
    // M25P80 has fast SPI (up to 25MHz), no need to slow down SPI
    // our F_CPU is 10 MHz, so we use fcpu/2; a transfer lasts 1,6us 

    #define SPI_PRESCALER  2

    #if   (SPI_PRESCALER==2)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==4)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==8)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==16)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==32)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==64)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==128)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #else
      #error SPI_PRESCALER is not defined or is not a supported value
    #endif

    SPCR = (0 << SPIE)    // 0 = no interrupt
         | (1 << SPE)     // 1 = enabled
         | (0 << DORD)    // 1 = data order LSB first; 0 = msb first
         | (1 << MSTR)    // 1 = master mode
         | (0 << CPOL)    // 0 = clock low when idle
         | (0 << CPHA)    // 0 = sample a leading edge
         | SPI_PRESCALER_BITS;

    SPSR = (SPI2XMUL);
  }

uint8_t spi_tx_byte(uint8_t data)
  {
    SPDR = data;										
    while(!(SPSR & (1<<SPIF)));
    return (SPDR);					
  }


void spi_send(uint8_t *buffer, uint16_t count)
  {
    uint16_t i;
    for (i=0; i<count; i++)  spi_tx_byte(*buffer++);
  }


//-----------------------------------------------------------------------------------------
//
// Layer 1
//
//-----------------------------------------------------------------------------------------

// Command Bytes fo the M25P80
#define FLASH_WREN  0x06    // write enable
#define FLASH_WRDI  0x04
#define FLASH_RDSR  0x05
#define FLASH_WRSR  0x01
#define FLASH_READ  0x03
#define FLASH_FREAD 0x0B    // fast read
#define FLASH_PP    0x02
#define FLASH_SE    0xD8
#define FLASH_BE    0xC7
#define FLASH_DP    0xB9
#define FLASH_RES   0xAB    // release from deep power down
#define DUMMY       0xAA

// Status Register Masks for the M25P80
#define FLASH_WIP  0x01     // write in progress
#define FLASH_WEL  0x02     // write ena latch
#define FLASH_BP0  0x04     // block protect
#define FLASH_BP1  0x08
#define FLASH_BP2  0x10
#define FLASH_SRWD 0x80     // status write prot

//-----------------------------------------------------------------------------------------
//
// Commands for flash access

// block until write in progress is no longer set in the status register

static inline void flash_block_until_write_done()
  {
    SPI_init(); SPI_ENABLE_FLASH;

    unsigned char status;

    spi_tx_byte(FLASH_RDSR);
    status = spi_tx_byte(0);            // read status
    while(status & FLASH_WIP)
        status = spi_tx_byte(0);

    SPI_DISABLE_FLASH;
}

// enable writing to the flash
static inline void flash_set_write_enable()
{
   SPI_init(); SPI_ENABLE_FLASH;

   spi_tx_byte(FLASH_WREN);

   SPI_DISABLE_FLASH;
}

/** @brief disable writing to the st flash */
static inline void st_flash_write_disable()
{
   SPI_init(); SPI_ENABLE_FLASH;

   spi_tx_byte(FLASH_WRDI);

   SPI_DISABLE_FLASH;
}


// send a 24bit address over the SPI - set the address inside the flash
static inline void set_addr_in_flash(uint32_t addr)
  {
    spi_tx_byte((addr >> 16) & 0xFF);
    spi_tx_byte((addr >>  8) & 0xFF);
    spi_tx_byte((addr >>  0) & 0xFF);
  }

// read the m25p80 flash status register
static inline uint8_t flash_read_status_reg()
  {
    uint8_t status_reg;
    SPI_init(); SPI_ENABLE_FLASH;

    spi_tx_byte(FLASH_RDSR);
    status_reg = spi_tx_byte(0);

    SPI_DISABLE_FLASH;

    return status_reg;
  }

// write to the st flash status register
static inline void flash_write_status_reg(uint8_t new_sr)
{
   flash_block_until_write_done();
   flash_set_write_enable();

   SPI_init(); SPI_ENABLE_FLASH;

   spi_tx_byte(FLASH_WRSR);
   spi_tx_byte(new_sr);

   SPI_DISABLE_FLASH;
}

// erase the entire flash
void flash_erase_complete()
  {
    flash_block_until_write_done();
    flash_set_write_enable();

    SPI_init(); SPI_ENABLE_FLASH;

    spi_tx_byte(FLASH_BE);

    SPI_DISABLE_FLASH;
  }

// enter low power mode
static inline void flash_set_2_deep_powerdown()
  {
    flash_block_until_write_done();
    
    SPI_init(); SPI_ENABLE_FLASH;

    spi_tx_byte(FLASH_DP);

    SPI_DISABLE_FLASH;
  }


// initialize the m25p80 flash and enter deep power down mode

void flash_init(void)
  {
    // Hardware INIT
    
    // see init_port in hardware.c
    
    // we need: CS as output
    // SPI dedicated registers
    
    // initialize into low power mode
    flash_set_2_deep_powerdown();
  }


// exit deep power down mode 
static inline void flash_wakeup()
  {
    SPI_init(); SPI_ENABLE_FLASH;
    spi_tx_byte(FLASH_RES);
    SPI_DISABLE_FLASH;
  }


// exit deep power down mode and/or read the signature
static inline uint8_t flash_read_signature()
{
   SPI_init(); SPI_ENABLE_FLASH;
   spi_tx_byte(FLASH_RES);
   spi_tx_byte(DUMMY);
   spi_tx_byte(DUMMY);
   uint8_t sig = spi_tx_byte(DUMMY);

   SPI_DISABLE_FLASH;

   return sig;
}

//-----------------------------------------------------------------------------------------
//
// Layer 2
//
//-----------------------------------------------------------------------------------------

// read a block of data from addr in flash
// here we use normal read (this is up to 20MHz)
void flash_read(uint32_t addr, uint8_t* buffer, uint16_t count)
  {
    flash_block_until_write_done();

    SPI_init(); SPI_ENABLE_FLASH;

    spi_tx_byte(FLASH_READ);
    set_addr_in_flash(addr);

    uint16_t i;
    for (i=0; i < count; i++)
      {
        buffer[i] = spi_tx_byte(0);
      }

    SPI_DISABLE_FLASH;
  }

// write to a page in flash at addr 
// Note: this is only changing from 1 to 0 - so erase is need before!

void flash_write(uint32_t addr, const uint8_t* buffer, uint16_t count)
  {
    flash_block_until_write_done();
    flash_set_write_enable();

    SPI_init(); SPI_ENABLE_FLASH;

    spi_tx_byte(FLASH_PP);
    set_addr_in_flash(addr);
    spi_send((uint8_t*)buffer, count);

    SPI_DISABLE_FLASH;
  }

// erase the sector which contains address addr /512kBit are erased!
void flash_erase_sector(uint32_t addr)
{
   flash_block_until_write_done();
   flash_set_write_enable();

   SPI_init(); SPI_ENABLE_FLASH;

   spi_tx_byte(FLASH_SE);
   set_addr_in_flash(addr);

   SPI_DISABLE_FLASH;
}



