//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      hardware.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-08 V0.01 kw  started
//            2009-02-29 V0.02 kw  DIDR added
//            2009-03-25 V0.03 kw  bugfix keyboard: disable out if no scan
//            2010-05-05 V0.04 kw  added a58_check();
//
//-----------------------------------------------------------------
#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "config.h"
#include "hardware.h"

static void init_port(void)
  {
    // Port A
    PORTA = (0<< LCD162_B0 )    // out (low)
          | (0<< LCD162_B1 )    // out (low)
          | (0<< LCD162_B2 )    // out
          | (0<< LCD162_B3 )    // out
          | (0<< LCD162_RS )    // out
          | (0<< LCD162_E  )    // out
          | (1<< KEY_SCAN1 )    // high = no scan
          | (1<< KEY_SCAN0 );   // high = no scan

    DDRA  = (1<< LCD162_B0 )    // out
          | (1<< LCD162_B1 )    // out
          | (1<< LCD162_B2 )    // out
          | (1<< LCD162_B3 )    // out
          | (1<< LCD162_RS )    // out
          | (1<< LCD162_E  )    // out
          | (0<< KEY_SCAN1 )    // in
          | (0<< KEY_SCAN0 );   // in

    // Port B
    PORTB = (1<< SPI_CS_EE )    // high = disable
          | (1<< SPI_SCK   )    // 
          | (1<< SPI_MOSI  )    // 
          | (1<< SPI_MISO  )    // pullup
          | (0<< LED_X0    )    // low = off
          | (0<< LED_X1    )    // low = off
          | (0<< LED_X2    )    // low = off
          | (0<< LED_X3    );   // low = off

    DDRB  = (1<< SPI_CS_EE )    // out
          | (1<< SPI_SCK   )    // out  / runs dedicated
          | (1<< SPI_MOSI  )    // out  / runs dedicated
          | (0<< SPI_MISO  )    // in   / runs dedicated
          | (1<< LED_X0    )    // out (red anode)
          | (1<< LED_X1    )    // out (red anode)
          | (1<< LED_X2    )    // out (red anode)
          | (1<< LED_X3    );   // out (red anode)

    // Port C
    PORTC = (1<< KEY_0     )    // pullup
          | (1<< KEY_1     )    // pullup
          | (1<< KEY_2     )    // pullup
          | (1<< KEY_3     )    // pullup
          | (1<< KEY_4     )    // pullup
          | (1<< KEY_5     )    // pullup / shift
          | (1<< KEY_6     )    // pullup / knob
          | (1<< KEY_7     );   // pullup / emergency stop

    DDRC  = (0<< KEY_0     )    // in
          | (0<< KEY_1     )    // in
          | (0<< KEY_2     )    // in
          | (0<< KEY_3     )    // in
          | (0<< KEY_4     )    // in
          | (0<< KEY_5     )    // in / shift
          | (0<< KEY_6     )    // in / knob
          | (0<< KEY_7     );   // in / emergency stop

    // Port D
    PORTD = (0<< MY_CTS    )    // out, enabled
          | (1<< MY_RTS    )    // in, pullup
          | (1<< MY_RXD    )    // in, pullup
          | (1<< MY_TXD    )    // out
          | (1<< ROT_A     )    // in, pullup
          | (0<< TXCAN     )    // out
          | (0<< RXCAN     )    // in
          | (1<< ROT_B     );   // in, pullup

    DDRD  = (1<< MY_CTS    )    // out
          | (0<< MY_RTS    )    // in
          | (0<< MY_RXD    )    // in
          | (1<< MY_TXD    )    // out
          | (0<< ROT_A     )    // in
          | (1<< TXCAN     )    // out
          | (0<< RXCAN     )    // in
          | (0<< ROT_B     );   // in

    // Port E
    PORTE = (1<< XP_RXD      )    // in
          | (0<< XP_TXD      )    // out
          | (1<< XP_RE       )    // out, high = disable
          | (0<< LED_XSTOP   )    // out (red anode), 0 = off
          | (0<< LED_YSTOP   )    // out (red cathode), 0 = off
          | (0<< LCD_PWM     )    // out, low = off
          | (1<< CAN_DISABL  )    // out
          | (0<< XP_DE       );   // out, low = disable

    DDRE  = (0<< XP_RXD      )    // in
          | (1<< XP_TXD      )    // out
          | (1<< XP_RE       )    // out
          | (1<< LED_XSTOP   )    // out (red anode)
          | (1<< LED_YSTOP   )    // out (red cathode)
          | (1<< LCD_PWM     )    // out
          | (1<< CAN_DISABL  )    // out
          | (1<< XP_DE       );   // out


    // Port F
    PORTF = (0<< VCCIO_SENSE )     // in analog (VCCIO), no pullup
          | (1<< POWER_ON    )     // out, high = turn on
          | (0<< VINT        )     // in analog (VINT/2)
          | (1<< A58         )     // in (pullup, for board detection)
          | (0<< LED_Y0      )     // out (red cathode), low = off
          | (0<< LED_Y1      )     // out (red cathode), low = off
          | (0<< LED_Y2      )     // out (red cathode), low = off
          | (0<< LED_Y3      );    // out (red cathode), low = off

    DDRF  = (0<< VCCIO_SENSE )     // in analog (VCCIO)
          | (1<< POWER_ON    )     // out
          | (0<< VINT        )     // in analog (VINT/2)
          | (0<< A58         )     // in (reserved)
          | (1<< LED_Y0      )     // out (red cathode)
          | (1<< LED_Y1      )     // out (red cathode)
          | (1<< LED_Y2      )     // out (red cathode)
          | (1<< LED_Y3      );    // out (red cathode)


// Port G
    PORTG = (1<< XBEE_ENABL   )     // out (low = enabled)
          | (0<< XBEE_SLEEP   )     // out (low=wake-up)
          | (0<< SPI_ADR0     )     // out
          | (0<< FTDI_RESET   )     // out (low = disabled)
          | (1<< SPI_CS_LCD   );    // out

    DDRG  = (1<< XBEE_ENABL   )     // out (low = enabled)
          | (1<< XBEE_SLEEP   )     // out
          | (1<< SPI_ADR0     )     // out
          | (1<< FTDI_RESET   )     // out
          | (1<< SPI_CS_LCD   );    // out

    DIDR0 = 0xFF;                   // disable all digital input on analog pins
    DIDR1 = 0xFF;
  }


void init_interrupt(void)
  {
  // External Interrupt(s) initialization
  // INT0: Off
  // INT1: Off
  // INT2: Off
  // INT3: Off
  // INT4: Off
  // INT5: Off
  // INT6: Off
  // INT7: Off
  // INT7 Mode: Low level

  // Timer(s)/Counter(s) Interrupt(s) initialization
 
    TIMSK0 =   (0<<OCIE0A)
             | (0<<TOIE0); 


    TIMSK1 =   (0<<ICIE1)     // input capture
             | (0<<OCIE1A)
             | (0<<OCIE1B)
             | (0<<OCIE1C)
             | (0<<TOIE1); 

    TIMSK2 =   (0<<OCIE2A)
             | (0<<TOIE2); 

    TIMSK3 =   (0<<ICIE3)
             | (0<<OCIE3A)
             | (0<<OCIE3B)
             | (0<<OCIE3C)
             | (0<<TOIE3); 

    sei();              // Global enable interrupts

  }

void init_hardware(void)
  {
    init_port();
    // init_interrupt();
  }


// true if MFT, false if Falko
unsigned char a58_check(void)
 {
    unsigned char retval;
    DIDR0 &= ~(1<<A58);         // enable readout
    _mydelay_us(1);             // wait one us
    retval = PINF & (1<<A58);   // if true: MFT; if false: DualThrottle (wired)
    DIDR0 |= (1<<A58);   
    return(retval);
 } 


