//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      hardware.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-08 V0.01 kw  started
//            2009-03-01 V0.02 kw  bugfix for VCCIO_SENSE
//            2009-03-25 V0.03 kw  bugfix keyboard: disable out if no scan
//
//-----------------------------------------------------------------
//
// purpose:   throtte for dcc
// content:   central definitions used in the project
//            all hardware project settings are done here!!
//            
//            1.   Prozessor, Timing
//                 (change here, if OpenDCC-Throttle is run on a different uP)
//            2.   IOs
//                 (change here, if OpenDCC-Throttle is run on a different board)
//
//-----------------------------------------------------------------
#ifndef __HARDWARE_H__
#define __HARDWARE_H__


//========================================================================
// 1. Processor Definitions
//========================================================================
//

// PROCESSOR: one of: __AVR_AT90can128__

// if changed: check timings, timer settings, baudrate

#if (__AVR_AT90CAN128__)
  // atmega32:   2kByte SRAM, 1kByte EEPROM
  #define SRAM_SIZE    4096
  #define EEPROM_SIZE  4096
  #define EEPROM_BASE  0x810000L
#else
  #warning: severe: no supported processor  
#endif
//
#ifndef F_CPU
// prevent compiler error by supplying a default 
# warning "F_CPU not defined for <config.h>", set default to 10MHz 
# define F_CPU 10000000UL
// if changed: check every place where it is used (possible range overflow in preprocessor!)
#endif

//------------------------------------------------------------------------
// Fuses of AT90can128 (just for documentation)
//------------------------------------------------------------------------
// Lock Bit Byte
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  1       1       1       1       1       1       1       1       default, not locked
//
// Extended Fuse Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//                                  BOD2    BOD1    BOD0    TA0SEL  111=no BOD (*), 101=2,7V, 100=4,3V
//  1       1       1       1       0       1       0       1
//
// Fuse High Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  OCDEN   JTAGEN  SPIEN   WDTON   EESAVE  BOOTSZ1 BOOTSZ0 BOOTRST
//  1       1       0       1       1       1       0       0
//  disabl  disabl  enabl   default no save =0xFC00         enter boot
//
// Fuse Low Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  CKDIV8  CKOUT   SUT1    SUT0    CKSEL3  CKSEL2  CKSEL1  CKSEL0
//  1       1       0       0       1       1       1       0
//  no div  no clk  crystal, BOD ena, 10 MHz    




//========================================================================
// 2. Port Definitions
//========================================================================

// Port A
#define LCD162_B0   0     // out
#define LCD162_B1   1     // out
#define LCD162_B2   2     // out
#define LCD162_B3   3     // out
#define LCD162_RS   4     // out
#define LCD162_E    5     // out
#define KEY_SCAN1   6     // out
#define KEY_SCAN0   7     // out

// Port B
#define SPI_CS_EE   0     // out
#define SPI_SCK     1     // out  / runs dedicated
#define SPI_MOSI    2     // out  / runs dedicated
#define SPI_MISO   	3     // in   / runs dedicated
#define LED_X0    	4     // out (red anode)
#define LED_X1    	5     // out (red anode)
#define LED_X2    	6     // out (red anode)
#define LED_X3    	7     // out (red anode)

// Port C
#define KEY_0       0     // in
#define KEY_1       1     // in
#define KEY_2       2     // in
#define KEY_3       3     // in
#define KEY_4       4     // in
#define KEY_5       5     // in / shift  -> this is used to enter bootloader
#define KEY_6       6     // in / knob
#define KEY_7       7     // in / emergency stop

// Port D
#define MY_CTS      0     // out
#define MY_RTS      1     // in
#define MY_RXD      2     // in
#define MY_TXD      3     // out
#define ROT_B       4     // in
#define TXCAN       5     // out
#define RXCAN       6     // in
#define ROT_A       7     // in

// Port E
#define XP_RXD      0     // in
#define XP_TXD      1     // out
#define XP_RE       2     // out
#define LED_XSTOP   3     // out (red anode)
#define LED_YSTOP   4     // out (red cathode)
#define LCD_PWM     5     // out
#define CAN_DISABL  6     // out
#define XP_DE       7     // out

// Port F
#define VCCIO_SENSE 0     // in analog (VCCIO, pulldown)
#define POWER_ON    1     // out (keep high for power on)
#define VINT        2     // in analog (VINT/2)
#define A58         3     // in (pullup; on DUAL MFT: extern 1k to gnd)
#define LED_Y0    	4     // out (red cathode)
#define LED_Y1    	5     // out (red cathode)
#define LED_Y2    	6     // out (red cathode)
#define LED_Y3    	7     // out (red cathode)

// Port G
#define XBEE_ENABL  0     // out (low = enabled)
#define XBEE_SLEEP  1     // out (high=set XBEE to sleep, low=wake up)
#define SPI_ADR0    2     // out
#define FTDI_RESET  3     // out
#define SPI_CS_LCD  4     // out


//------------------------------------------------------------------------
// Definitions for the keyboard (see also keyboard.c)

#define PIN_PHASE_A   PIND      // this is the input port for 'A'
#define BIT_PHASE_A   ROT_A     // may be 0..7
#define PHASE_A       (PIN_PHASE_A & (1<<BIT_PHASE_A))

#define PIN_PHASE_B   PIND      // this is the input port for 'B'
#define BIT_PHASE_B   ROT_B     // may be 0..7
#define PHASE_B       (PIN_PHASE_B & (1<<BIT_PHASE_B))

#define PIN_PUSH_KNOB PINC      // this is the input port for 'KNOB'
#define BIT_PUSH_KNOB KEY_6     // may be 0..7

#define SET_SCANLINE0_to_0   { DDRA |= (1<<KEY_SCAN0);  PORTA &= ~(1<<KEY_SCAN0);}
#define SET_SCANLINE0_to_1   { PORTA |= (1<<KEY_SCAN0); DDRA &= ~(1<<KEY_SCAN0);} 
#define SET_SCANLINE1_to_0   { DDRA |= (1<<KEY_SCAN1);  PORTA &= ~(1<<KEY_SCAN1); }
#define SET_SCANLINE1_to_1   { PORTA |= (1<<KEY_SCAN1); DDRA &= ~(1<<KEY_SCAN1);} 

#define KEY_PIN       PINC

//------------------------------------------------------------------------
// Definitions for the ADC (see also analog.c)


#define ADC_PORT_IN        PINF
#define ADC_PORT_DIR       DDRF
#define ADC_PORT_OUT      PORTF
#define ADC_DID_REG       DIDR0

//------------------------------------------------------------------------
// LED-Control and IO-Macros
// LEDs are connected between two IOs, we can turn it red, green or off.

#define LED_STOP_RED     {PORTE |= (1<<LED_XSTOP); PORTE &= ~(1<<LED_YSTOP);}
#define LED_STOP_GREEN   {PORTE &= ~(1<<LED_XSTOP); PORTE |= (1<<LED_YSTOP);}
#define LED_STOP_OFF     {PORTE &= ~(1<<LED_XSTOP); PORTE &= ~(1<<LED_YSTOP);}

#define LED_A_RED        {PORTB |= (1<<LED_X0); PORTF &= ~(1<<LED_Y0);}
#define LED_A_GREEN   	 {PORTB &= ~(1<<LED_X0); PORTF |= (1<<LED_Y0);}
#define LED_A_OFF        {PORTB &= ~(1<<LED_X0); PORTF &= ~(1<<LED_Y0);}

#define LED_B_RED        {PORTB |= (1<<LED_X1); PORTF &= ~(1<<LED_Y1);}
#define LED_B_GREEN   	 {PORTB &= ~(1<<LED_X1); PORTF |= (1<<LED_Y1);}
#define LED_B_OFF        {PORTB &= ~(1<<LED_X1); PORTF &= ~(1<<LED_Y1);}

#define LED_C_RED        {PORTB |= (1<<LED_X2); PORTF &= ~(1<<LED_Y2);}
#define LED_C_GREEN   	 {PORTB &= ~(1<<LED_X2); PORTF |= (1<<LED_Y2);}
#define LED_C_OFF        {PORTB &= ~(1<<LED_X2); PORTF &= ~(1<<LED_Y2);}

#define LED_D_RED        {PORTB |= (1<<LED_X3); PORTF &= ~(1<<LED_Y3);}
#define LED_D_GREEN   	 {PORTB &= ~(1<<LED_X3); PORTF |= (1<<LED_Y3);}
#define LED_D_OFF        {PORTB &= ~(1<<LED_X3); PORTF &= ~(1<<LED_Y3);}


#define MY_CTS_HIGH      PORTD |= (1<<MY_CTS)
#define MY_CTS_LOW       PORTD &= ~(1<<MY_CTS)
#define MY_CTS_STATE     (PORTD & (1<<MY_CTS))

// RTS is aktive low
#define MY_RTS_DETECT    (!(PIND & (1<<MY_RTS)))

#define SET_XBEE_TO_SLEEP   PORTG |= (1<<XBEE_SLEEP)
#define SET_XBEE_TO_WAKEUP  PORTG &= ~(1<<XBEE_SLEEP)

#define ENABLE_USB          {PORTG |= (1<<XBEE_ENABL); PORTG |= (1<<FTDI_RESET);}
#define ENABLE_XBEE         {PORTG &= ~(1<<FTDI_RESET); PORTG &= ~(1<<XBEE_ENABL);}
#define DISABLE_USB_XBEE    {PORTG |= (1<<XBEE_ENABL); PORTG &= ~(1<<FTDI_RESET);}
#define ENABLE_CAN          PORTE &= ~(1<<CAN_DISABL)
#define DISABLE_CAN         PORTE |= (1<<CAN_DISABL)
#define ENABLE_XP_RX        PORTE &= ~(1<<XP_RE)
#define DISABLE_XP_RX       PORTE |= (1<<XP_RE)
#define ENABLE_XP_TX        PORTE |= (1<<XP_DE)
#define DISABLE_XP_TX       PORTE &= ~(1<<XP_DE)
#define DISABLE_XP          PORTE |= (1<<XP_RE); PORTE |= (1<<XP_DE)

#define TURN_POWER_ON       PORTF |= (1<<POWER_ON)
#define TURN_POWER_OFF      PORTF &= ~(1<<POWER_ON)

#define VCCIO_ACTIV()       (PINF & (1<<VCCIO_SENSE))  // note: first enable digital readout (DIDR0)

#define SPI_ENABLE_LCD      PORTG &= ~(1 << SPI_CS_LCD)
#define SPI_DISABLE_LCD     PORTG |= (1 << SPI_CS_LCD)

#define SPI_ENABLE_FLASH    PORTB &= ~(1 << SPI_CS_EE)
#define SPI_DISABLE_FLASH   PORTB |= (1 << SPI_CS_EE)


#define SPI_ADR0_LOW        PORTG &= ~(1 << SPI_ADR0)
#define SPI_ADR0_HIGH       PORTG |= (1 << SPI_ADR0)

// Calls

extern void init_hardware(void);

extern unsigned char a58_check(void);    // true if MFT, false if Falko

#endif   // hardware.h
