//========================================================================
//
// OpenDCC - Rotary Encoder - Keyboard
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      keyboard.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-23 V0.01 kw  started
//            2009-01-14 V0.02 kw  added different speedup
//            2009-02-01 V0.03 kw  added release code for shift key
//            2009-04-03 V0.04 kw  added is_shift_pressed()
//
//------------------------------------------------------------------------
//
// purpose:   create keystrokes from rotary encoder
//            here: hardware interface, accelleration functions
//
// content:   1. rotary encoder (bullet proof and with accelerator)
//            2. keyboard encoder (n-key rollover)
//            3. interface upstream
//
//========================================================================

#include <inttypes.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>          // put var to program memory
#include <util/delay.h>

#include "hardware.h"
#include "keycodes.h"           // table (scan -> keystroke)
#include "timer.h"              // systick
#include "cortos.h"             // cooperative scheduler and fifo

#include "keyboard.h"

//========================================================================
// CONFIG

#define KEYBUF_SIZE       32       // type ahead buffer - this also carries
                                   // rotary keystrokes

//========================================================================
// DATA

uint8_t key_buffer[KEYBUF_SIZE];

cr_fifo_t key_fifo;


//========================================================================
// Section 1: rotary encoder
//========================================================================
//
// This file handles the rotary knob by timer based polling.
// A readout with pin interrupts would bear the risk of heavy CPU load if
// encoder stops at a toggle point.
// Also debouncing would not be possible.
//
//------------------------------------------------------------------------
// interface: 1. call init_knob();
//            2. whenever the knob is rotated, a keystroke is issued
//            3. the calling program must ask get_knob_rotate() to
//               receive the actual rotation - this value may be zero,
//               even if a knob_keystroke was issued!
//               get_knob_rotate() returns the sum of rotation steps.
//------------------------------------------------------------------------
//
// Signal from rotary encoder
// using pull-up, switch against GND
//
// clockwise:
//       |      |       |      |      |<-detent
// A: _______XXXXXXX________XXXXXXX________
//       |      |       |      |      |
// B: XXX________XXXXXXX_______XXXXXXX_____
//       |      |       |      |      |
// counterclockwise:
//       |      |       |      |      |
// A: _______XXXXXXX________XXXXXXX________
//       |      |       |      |      |
// B: ___XXXXXXX________XXXXXXX_______XXXXX
//       |      |       |      |      |
//
// Note: there may be: bouncing up to 3ms
//                     sliding noise up to 2ms
// Source: Alps EC11*
//=====================================================================
// Connection:
//
// Connect Phase A to ROT_A
//         (this is the stable phase when in detent position)         
// Connect Phase B to ROT_B
// Connect Knob to any Pin
//=====================================================================
// Howto
// detection: debouce and evaluate any change on ROT_A
// states:
//    A goes high, B high: a step counterclockwise
//    A goes high, B low: a step clockwise
//    A goes low,  B high: a step clockwise
//    A goes low,  B low: a step counterclockwise
//
// acceleration (speed_up):
// if the last step is younger then i.e.
//     30ms: advance 10*
//     60ms: advance  5*
//    100ms: advance  2*
//=====================================================================


#ifndef PIN_PHASE_A
	#warning local defines of rot inputs - this should be done in hardware.h
	#define PIN_PHASE_A   PIND      // this is the input port for 'A'
	#define BIT_PHASE_A   ROT_A     // may be 0..7
	#define PHASE_A       (PIN_PHASE_A & (1<<BIT_PHASE_A))
	
	#define PIN_PHASE_B   PIND      // this is the input port for 'B'
	#define BIT_PHASE_B   ROT_B     // may be 0..7
	#define PHASE_B       (PIN_PHASE_B & (1<<BIT_PHASE_B))
	
	#define PIN_PUSH_KNOB PINC   		// this is the input port for 'KNOB'
	#define BIT_PUSH_KNOB KEY6      // may be 0..7
#endif // PIN_PHASE_A


#define SPEEP_UP_TABLE_LENGTH  5

struct s_speedup
  {
    signed int threshold;       // given in systick_10ms
    signed char mult;           // multiplier
  };

struct s_speedup speedup[SPEEP_UP_TABLE_LENGTH];

const struct s_speedup speedup_none[SPEEP_UP_TABLE_LENGTH] PROGMEM =  
  {{ ( 15000L/SYSTICK_10MS_PERIOD), 1 },
   { ( 30000L/SYSTICK_10MS_PERIOD), 1 },
   { ( 50000L/SYSTICK_10MS_PERIOD), 1 },
   { (140000L/SYSTICK_10MS_PERIOD), 1 },
   { (300000L/SYSTICK_10MS_PERIOD), 1 },
  };

const struct s_speedup speedup_normal[SPEEP_UP_TABLE_LENGTH] PROGMEM = 
  {{ ( 15000L/SYSTICK_10MS_PERIOD), 10},
   { ( 30000L/SYSTICK_10MS_PERIOD), 7 },
   { ( 50000L/SYSTICK_10MS_PERIOD), 5 },
   { (140000L/SYSTICK_10MS_PERIOD), 3 },
   { (300000L/SYSTICK_10MS_PERIOD), 2 },
  };

const struct s_speedup speedup_strong[SPEEP_UP_TABLE_LENGTH] PROGMEM = 
  {{ ( 15000L/SYSTICK_10MS_PERIOD), 20},
   { ( 30000L/SYSTICK_10MS_PERIOD), 14 },
   { ( 50000L/SYSTICK_10MS_PERIOD), 7 },
   { (140000L/SYSTICK_10MS_PERIOD), 4 },
   { (300000L/SYSTICK_10MS_PERIOD), 2 },
  };

void set_knob_speedup_none(void)
  {
    memcpy_P(speedup, speedup_none, sizeof(speedup));
  }

void set_knob_speedup_normal(void)
  {
    memcpy_P(speedup, speedup_normal, sizeof(speedup));
  }

void set_knob_speedup_strong(void)
  {
    memcpy_P(speedup, speedup_strong, sizeof(speedup));
  }



// signed counters are used to debounce the slider
#define ROT_DEBOUNCE  2   // this is the limit of the counter
                          // (3 = we need 3 polls to detect a change)
                          // valid: 1: very fast, but no debouncing
                          //        2: very fast, small debounce
                          //        3: fast, very good debounce, but
                          //           may fail to see extreme rotations
signed char debounce_A;   
signed char debounce_B;
unsigned char debounce_KNOB;
#define ROT_DEBOUNCE_SHIFT 0x03

// accelerate function - if the knob is turned fast, we increase the steps
volatile signed char rotate;     // accumulated rot-steps (atomic access)
signed int last_knobturn;        // systick of last knob turn
signed char last_knobturn_dir;   // direction of last knob turn

// this routine must be called every tick
void knob_poll(void)
  {
    // Interrupt part - to be called every 1-2ms!
    // could be copied to timer ISR or called in multitask environment every ms
    // note: this code also handles slider drop outs on any line; debounce filtering
    //       is done in both ways!

    signed char step = 0;
    if (PHASE_B)
      {
        debounce_B += 1; if (debounce_B > ROT_DEBOUNCE-1) debounce_B = ROT_DEBOUNCE-1;
      }
    else
      {
        debounce_B -= 1; if (debounce_B < -ROT_DEBOUNCE) debounce_B = -ROT_DEBOUNCE;
      }

    if (PHASE_A)
      {
        if (debounce_A >= 0) // no change;
          {
            debounce_A += 1;
            if (debounce_A > ROT_DEBOUNCE-1) debounce_A = ROT_DEBOUNCE-1;
          }
        else
          {
            debounce_A += 1;
            if (debounce_A >= 0)
              {
                // rising edge, evaluate B
                if (debounce_B >= 0) { step = -1; }
                else  { step = 1; }
              }
          }
      }
    else
      {
        if (debounce_A < 0) // no change;
          {
            debounce_A -= 1;
            if (debounce_A < -ROT_DEBOUNCE) debounce_A = -ROT_DEBOUNCE;
          }
        else
          {
            debounce_A -= 1;
            if (debounce_A < 0)
              {
                // falling edge, evaluate B
                if (debounce_B >= 0)  { step = 1; }
                else                  { step = -1; }
              }
          }
      }

    // now do some acceleration and accumulation
   
    if (step == 0)
      { // prevent accelerator from wrap around
        signed int s_tick;
        s_tick = get_systick_int();
        if (s_tick - last_knobturn > 1000000L / SYSTICK_PERIOD)
          {
            last_knobturn = s_tick;
            last_knobturn_dir = 0;
          }
      }
    else
      {
        signed int s_tick, compare;
        s_tick = get_systick_int();
        compare = s_tick - last_knobturn;
        last_knobturn = s_tick;
        
        if (last_knobturn_dir == step)
          { 
            unsigned char j;
            for (j=0; j<sizeof(speedup)/sizeof(speedup[0]); j++)
              {
                if (compare <= speedup[j].threshold) 
                  {
                    step *= speedup[j].mult;
                    break;
                  }
              }
          }
        else
          {
            last_knobturn_dir = step;                // no accelerator if dir was changed
          }
        rotate = rotate + step;
        if (rotate > 100) rotate = 100;
        if (rotate < -100) rotate = -100;
        cr_fifo_put(&key_fifo, KEY_KNOB_TURNED);        // dump only the message, not the number.
     }

    #define DECODE_KNOB_PUSH 0     // is done with normal keyboard (so here commented)
    #if (DECODE_KNOB_PUSH == 1)
        // now handle knob (connected against GND)
        if (PIN_PUSH_KNOB & (1<<BIT_PUSH_KNOB))
          {
            debounce_KNOB <<= 1;
            debounce_KNOB |= 1;
          }
        else
          {
            if ((debounce_KNOB & ROT_DEBOUNCE_SHIFT) == 0) ; // schon gedrueckt gewesen
            else 
              {
                debounce_KNOB <<= 1;
                if ((debounce_KNOB & ROT_DEBOUNCE_SHIFT) == 0) ; // wurde jetzt gedrueckt
                  {
                    fifo_put(&key_fifo, KEY_KNOB_PUSHED);
                    set_task_ready(TASK_NAVIGATOR);
                  }
              }
          }
    #endif
  }



//========================================================================
// Section 2: keyboard encoder
//======================================================================== 
//
// Keys for throttle
//
// Bit:      7       6        5       4       3       2       1       0
// SCAN1:    Stop    Knob     Shift   <leer>  D       C       B       A 
// SCAN0:    Stop    Knob     Shift   Up      Right   Center  Left    Down 

#ifndef nop
   #define nop() __asm volatile ("nop")
#endif


#define KEY_DEBOUNCE    4                //  4 ticks => 6ms

#define SHIFT_MASK (1 << 5)              // Position of shift key

static unsigned char old_scan0;
static unsigned char old_scan1;
static unsigned char change_scan0;
static unsigned char change_scan1;
static unsigned char debounce0;
static unsigned char debounce1;


unsigned char key_code_table[] = {
    // Scan line 0
    KEY_DOWN,       // Bit 0
    KEY_LEFT,       // Bit 1
    KEY_CENTER,     // Bit 2
    KEY_RIGHT,      // Bit 3
    KEY_UP,         // Bit 4
    KEY_SHIFT,      // if only shift is pressed
    KEY_KNOB,
    KEY_STOP,
    // Scan line 0 with shift pressed
    KEY_SH_DOWN,
    KEY_SH_LEFT,
    KEY_SH_CENTER,
    KEY_SH_RIGHT,
    KEY_SH_UP,
    KEY_SHIFT,       // if only shift is pressed
    KEY_SH_KNOB,
    KEY_SH_STOP,
    // Scan line 1
    KEY_A,
    KEY_B,
    KEY_C,
    KEY_D,
    // Scan line 1 with shift pressed
    KEY_SH_A,
    KEY_SH_B,
    KEY_SH_C,
    KEY_SH_D,
    };
    
void keyboard_poll(void)
  {
    unsigned char new_scan0, new_scan1, new_pressed, new_released;
    unsigned char i, mask;
    unsigned char key_code;

    SET_SCANLINE0_to_0; nop(); nop(); nop(); nop();     // wait 200ns to settle lines

    new_scan0 = ~KEY_PIN;                 // read line, invert - we switch against gnd
    if (new_scan0 ^ old_scan0)
      {
        if (debounce0 == 0)
          {
            change_scan0 = new_scan0;
            debounce0++;
          }
        else
          {
            if (change_scan0 == new_scan0)
              {
                debounce0++;
                if (debounce0 >= KEY_DEBOUNCE)
                  {
                    new_pressed = new_scan0 & ~old_scan0;
                    if (new_pressed)
                      {  // new key pressed
                        mask = 1;
                        for (i=0; i<8; i++)            // evaluate complete row
                          {
                            if (new_pressed & mask)
                              {
                                if (new_scan0 & SHIFT_MASK)
                                  {
                                    key_code = key_code_table[i + 8];
                                  }
                                else
                                  {
                                    key_code = key_code_table[i];
                                  }
                                cr_fifo_put(&key_fifo, key_code);
                              }
                            mask = mask << 1;
                          }
                      }
                    new_released = ~new_scan0 & old_scan0;
                    if (new_released)
                      {  // new key release, but we only are interested in SHIFT
                        if (new_released & SHIFT_MASK)
                          {
                            key_code = KEY_SHIFT_RELEASE;
                            cr_fifo_put(&key_fifo, key_code);
                          }
                      }
                    
                    old_scan0 = new_scan0;
                    debounce0 = 0;
                  }
              }
            else
                debounce0 = 0;
          }       
      }
    SET_SCANLINE0_to_1;
    nop(); nop(); nop(); nop();
    SET_SCANLINE1_to_0;
    nop(); nop(); nop(); nop();
    new_scan1 = ~KEY_PIN;                 // read line, invert - we switch against gnd
    if (new_scan1 ^ old_scan1)
      {
        if (debounce1 == 0)
          {
            change_scan1 = new_scan1;
            debounce1++;
          }
        else
          {
            if (change_scan1 == new_scan1)
              {
                debounce1++;
                if (debounce1 >= KEY_DEBOUNCE)
                  {
                    new_pressed = new_scan1 & ~old_scan1;
                    if (new_pressed)
                      {  // new key pressed
                        mask = 1;
                        for (i=0; i<4; i++)     // !! only keys A, B, C, D
                          {
                            if (new_pressed & mask)
                              {
                                if (new_scan1 & SHIFT_MASK)
                                  {
                                    key_code = key_code_table[i + 20];
                                  }
                                else
                                  {
                                    key_code = key_code_table[i + 16];
                                  }
                                cr_fifo_put(&key_fifo, key_code);
                              }
                            mask = mask << 1;
                          }
                      }
                    old_scan1 = new_scan1;
                    debounce1 = 0;
                  }
              }
            else
                debounce1 = 0;
          }
      }
    SET_SCANLINE1_to_1;
  }



//=====================================================================
// Upstream Interface
//=====================================================================

signed int get_knob_rotate(void)
  {
    signed char retval;
    cli();
    retval = rotate;
    rotate = 0;
    sei();
    return(retval);
  }

unsigned char is_shift_pressed(void)      // shift key
  {
    return (old_scan1 & SHIFT_MASK);
  }

unsigned char keystroke_present(void)
  {
    return (cr_fifo_filled(&key_fifo));
  }

unsigned char get_wait_keystroke(void)
  {
    while (!keystroke_present()) ;     // busy waiting! -> multitask
    return (cr_fifo_get_wait(&key_fifo));
  }


void init_knob(void)
  {
    rotate = 0;                        // accumulated steps
    last_knobturn = get_systick_int();
    last_knobturn_dir = 0;             // no direction
    debounce_A = 0;
    debounce_B = 0;
    debounce_KNOB = 0;

    set_knob_speedup_normal();
  }


void init_keyboard(void)
  {
    cr_fifo_init (&key_fifo, key_buffer, KEYBUF_SIZE);

    SET_SCANLINE0_to_0;
    old_scan0 = ~KEY_PIN;                 // read line, invert - we switch against gnd
    SET_SCANLINE0_to_1;

    SET_SCANLINE1_to_0;
    old_scan1 = ~KEY_PIN;                 // read line, invert - we switch against gnd
    SET_SCANLINE1_to_1;
    
    debounce0 = 0; 
    debounce1 = 0;       
    init_knob();
  }
