//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Jean Michel Fischer, Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
// This code is initially based on:
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial 
//
// Here we do the access to a lcd display, no special functions
// for MFT contained (so this is a general purpose file)
//
//-----------------------------------------------------------------
//
// file:      lcd_driver_ks0073.c
// author:    Wolfgang Kufer
// contact:   kufer_at_gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-18 V0.01 jmf started
//            2009-01-13 V0.02 kw  comments, tests
//            2009-01-26 V0.03 kw  set_cursor jetzt ab 0,0
//            2009-03-07 V0.04 kw  get_cursor neu dazu
//            2009-03-20 V0.05 kw  nach init cursor off
//            2009-03-27 V0.06 kw  #if DISPLAY added
//            2009-04-04 V0.07 kw  cursor_on_at added
//
//-----------------------------------------------------------------
//
// purpose:   control of a spi driven display
//            Controller: KS0073 / HD44780
//
// Interface: downstream: we use hardware SPI,
//                        so we have to observe the the mutex on SPI
//            upstream: we keep trace of controller state,
//                      so we offer an own mutex for the display access
//
// 2 do: better code, mutex, attributes
//
//-----------------------------------------------------------------
// Die Pinbelegung ist ueber defines in lcd-routines.h einstellbar

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <util/delay.h>
#include <stdio.h>

#include "config.h"
#include "cortos.h"
#include "hardware.h"           // CS for SPI

//#include "ks0073_charset.h"

#include "lcd_driver_ks0073.h"

//#include <avr/interrupt.h>

#if (DISPLAY == LCD_EA204)

//-----------------------------------------------------------------
// Low level interface for text display
//-----------------------------------------------------------------

#if (DISPLAY_INTERFACE == LCD_IF_4BIT)
// erzeugt den Enable-Puls
void lcd_enable(void)
{
   // Bei Problemen ggf. Pause gemaess Datenblatt des LCD Controllers einfuegen
   // http://www.mikrocontroller.net/topic/81974#685882
   LCD_PORT |= (1<<LCD_EN);
    _delay_us(1);                   // kurze Pause
   // Bei Problemen ggf. Pause gemaess Datenblatt des LCD Controllers verlaengern
   // http://www.mikrocontroller.net/topic/80900
   LCD_PORT &= ~(1<<LCD_EN);
}

// sendet ein Datenbyte an das LCD

void lcd_write_data(unsigned char temp1)
{
   unsigned char temp2 = temp1;

   LCD_PORT |= (1<<LCD_RS);         // RS auf 1 setzen

   temp1 = temp1 >> 4;
   temp1 = temp1 & 0x0F;
   LCD_PORT &= 0xF0;
   LCD_PORT |= temp1;               // setzen
   lcd_enable();

   temp2 = temp2 & 0x0F;
   LCD_PORT &= 0xF0;
   LCD_PORT |= temp2;               // setzen
   lcd_enable();

   _delay_us(42);
}

// sendet einen Befehl an das LCD

void lcd_write_command(unsigned char cmd)
{
   unsigned char temp;

   LCD_PORT &= ~(1<<LCD_RS);        // RS auf 0 setzen

   temp = cmd >> 4;              		// oberes Nibble holen
   temp = temp & 0x0F;            	// maskieren
   LCD_PORT &= 0xF0;
   LCD_PORT |= temp;               	// setzen
   lcd_enable();

   temp = cmd & 0x0F;            		// unteres Nibble holen und maskieren
   LCD_PORT &= 0xF0;
   LCD_PORT |= temp;               	// setzen
   lcd_enable();

   _delay_us(42);
}



#elif (DISPLAY_INTERFACE == LCD_IF_SPI)

static unsigned char 		rs;	        // used to keep track of rs state
static unsigned char 		rw;	        // used to keep track of rw state

static inline void SPI_init(void)
  {
    // Set MOSI, SCK and SS output, all others input
    // DDR_SPI = (1 << P_MOSI) | (1 << P_SCK) | (1 << P_SS);
    // make SS output high
    // PORT_SPI |= (1 << P_SS);
    // all the above is done in hardware.h 

    // Enable SPI, Master, set clock rate fck/16
    // ks0073 needs 0.5us ... 20us
    // our F_CPU is 10 MHz, so we use fcpu/8; a transfer lasts 6,4us 

    #define SPI_PRESCALER  16  // was 8

    #if   (SPI_PRESCALER==2)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==4)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==8)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==16)
      #define SPI_PRESCALER_BITS   ((0<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==32)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (1 << SPI2X)
    #elif (SPI_PRESCALER==64)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(0<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #elif (SPI_PRESCALER==128)
      #define SPI_PRESCALER_BITS   ((1<<SPR1)|(1<<SPR0))
      #define SPI2XMUL             (0 << SPI2X)
    #else
      #error SPI_PRESCALER is not defined or is not a supported value
    #endif

    #if (F_CPU / SPI_PRESCALER > 2000000L)
      #warning SPI clock too fast for KS0073
    #elif (F_CPU / SPI_PRESCALER < 50000L)
      #warning SPI clock too slow for KS0073
    #endif
  
    SPCR = (0 << SPIE)    // 0 = no interrupt
         | (1 << SPE)     // 1 = enabled
         | (1 << DORD)    // 1 = data order LSB first; 0 = msb first
         | (1 << MSTR)    // 1 = master mode
         | (0 << CPOL)    // 0 = clock low when idle
         | (0 << CPHA)    // 0 = sample a leading edge
         | SPI_PRESCALER_BITS;

    SPSR = (SPI2XMUL);
  }




unsigned char SPI_xmit (unsigned char aData) 
  {
    SPI_ENABLE_LCD;                // CS low

    _delay_us(42);

    SPDR = aData;										
    while(!(SPSR & (1<<SPIF)));

    return (SPDR);					
  }

#define LCD_SPI_SYNC  0x1F
#define LCD_SPI_RD    0x20
#define LCD_SPI_WR    0x00
#define LCD_SPI_RS    0x40

static unsigned char spi_command;
static unsigned char spi_data;

void lcd_write_command (unsigned char cmd)  // writecmd
{
	spi_command = cmd;

    SPI_ENABLE_LCD;                // CS low
    
    unsigned char lower;
	unsigned char upper;
	lower = cmd % 16;
	upper = cmd / 16;
	if ((rs != 0) || (rw != 0))
	{
		SPI_xmit (LCD_SPI_SYNC + LCD_SPI_WR);
        rs = 0;
        rw = 0;	
	}
	SPI_xmit (lower);
	SPI_xmit (upper);
	
	_delay_us(42);
    // SPI_DISABLE_LCD;	  !! Achtung: Timing fr Disable kontrollieren	
}

unsigned char lcd_read (unsigned char reg)  // readcmd
{
	unsigned char result;

    SPI_ENABLE_LCD;                // CS low
	if (reg != 0)
	  {
		result = SPI_xmit (LCD_SPI_SYNC + LCD_SPI_RD + LCD_SPI_RS);	
        rw = 1; rs = 1;
	  }
    else
      {
		result = SPI_xmit (LCD_SPI_SYNC + LCD_SPI_RD);	
        rw = 1; rs = 0;
	  }

	_delay_us(42);
    // SPI_DISABLE_LCD;	  !! Achtung: Timing fr Disable kontrollieren
    return(result);
}


void lcd_write_data (unsigned char dat)
{
    SPI_ENABLE_LCD;                // CS low
	spi_data = dat;
    
    unsigned char lower;
	unsigned char upper;
	lower = dat % 16;
	upper = dat / 16;
	if ((rs !=1) || (rw != 0))
	{
		SPI_xmit (LCD_SPI_SYNC + LCD_SPI_WR + LCD_SPI_RS);
        rs = 1;
        rw = 0;
	}
	SPI_xmit (lower);
	SPI_xmit (upper);
	
	_delay_us(50);
    _delay_us(50);
    _delay_us(50);
    // SPI_DISABLE_LCD;		  !! Achtung: Timing fr Disable kontrollieren
}


#endif // (DISPLAY_INTERFACE == LCD_IF_4BIT)





// Initialisierung:
// Muss ganz am Anfang des Programms aufgerufen werden.

void lcd_init(void)
{
#if (DISPLAY_INTERFACE == LCD_IF_4BIT)
   LCD_DDR = LCD_DDR | 0x0F | (1<<LCD_RS) | (1<<LCD_EN);   // Port auf Ausgang schalten

   // muss 3mal hintereinander gesendet werden zur Initialisierung

   _delay_ms(15);
   LCD_PORT &= 0xF0;
   LCD_PORT |= 0x03;
   LCD_PORT &= ~(1<<LCD_RS);      // RS auf 0
   lcd_enable();

   _delay_ms(5);
   lcd_enable();

   _delay_ms(1);
   lcd_enable();
   _delay_ms(1);

   // 4 Bit Modus aktivieren
   LCD_PORT &= 0xF0;
   LCD_PORT |= 0x02;
   lcd_enable();
   _delay_ms(1);

#elif (DISPLAY_INTERFACE == LCD_IF_SPI)

    // we need some hardware

    SPI_init();

	_delay_ms(30);
	rs = 1;
	lcd_write_command(0x20          // function set 
                     |(1 << 4)      // DL: 0= 4 Bit, 1= 8 Bit
                     |(0 << 3)      // N:  0= 1-line Display, 1=2-line (see also NW)
                     |(1 << 2)      // RE: 1= extension register
                     |(0 << 1)      // DH: 0= dot scroll
                     |(0 << 0));    // REV:0= no revers

	lcd_write_command(0x08          // extended function set
                     |(0 << 2)      // FW: 0= 5-dot font,  1= 6-dot font
                     |(0 << 1)      // BW: 0= normal cursor, 1=invert cursor
                     |(1 << 0));    // NW: 0= 1 or 2 lines, 1= 4-lines
	
    lcd_write_command(0x20          // function set 
                     |(1 << 4)      // DL: DL: 0= 4 Bit, 1= 8 Bit
                     |(0 << 3)      // N:  0= 1-line Display, 1=2-line (see also NW)
                     |(0 << 2)      // RE: 0= no extension register
                     |(0 << 1)      // DH: 0= dot scroll
                     |(0 << 0));    // REV:0= no revers

	lcd_write_command(0x08          // display on/off
                     |(1 << 2)      // D:  0= display off,  1= display on
                     |(0 << 1)      // C:  0= cursor off, 1= cursor on
                     |(0 << 0));    // B:  0= blink off, 1= blink on
	
    _delay_ms(2);
	lcd_write_command(LCD_CLEAR_DISPLAY);
	_delay_ms(2);
	lcd_write_command(0x04          // entry mode
                     |(1 << 1)      // ID: 0= cursor decrement, 1= cursor increment
                     |(0 << 0));    // S:  0= shift disable, 1= shift enable
#endif // (DISPLAY_INTERFACE)

   // for LCD_REICHELT162
   // 4Bit / 2 Zeilen / 5x7
   // lcd_write_command(0x28);
}


unsigned char pixel_haken[] PROGMEM =
  { 0b00000001,
    0b00000010,
    0b00000010,
    0b00010010,
    0b00001010,
    0b00000100,
    0b00000100,
    0b00000000,
  };

void lcd_set_userchar(unsigned char my_addr, unsigned char *pixel)
  {
    unsigned char cg_set_addr;
    unsigned char i;
    
    cg_set_addr = my_addr * 8;    // Bit 3,4,5
    cg_set_addr |= 0b01000000;
    for (i=0; i<8; i++)
      {
        // Mit dem Kommando "CG RAM Address Set" wird der
        // ASCII Code (Bit 3,4,5) und die entsprechende
        // Pixelzeile (Bit 0,1,2) des Zeichens angew?t.
        lcd_write_command(cg_set_addr + i);
        lcd_write_data(pgm_read_byte(pixel++));
      }
  } 


// Sendet den Befehl zur Loeschung des Displays

void lcd_clear(void)
{
   lcd_write_command(LCD_CLEAR_DISPLAY);
   _delay_ms(5);
}

// Sendet den Befehl: Cursor Home

void lcd_home(void)
{
   // xSemaphoreTake( mySem, portMAX_DELAY ); 
   lcd_write_command(LCD_CURSOR_HOME);
   _delay_ms(5);
   // xSemaphoreGive( mySem ); 

}

void lcd_reset(void)
  {
    lcd_clear();
  }

// setzt den Cursor in Zeile y (0..3) Spalte x (0..20)
// Origin: upper top left corner
void set_cursor_addr(uint8_t x, uint8_t y)
{
  uint8_t tmp;

  switch (y) {
      case 0: tmp=0x80+0x00+x; break;    // 1. Zeile
      case 1: tmp=0x80+0x20+x; break;    // 2. Zeile
      case 2: tmp=0x80+0x40+x; break;    // 3. Zeile
      case 3: tmp=0x80+0x60+x; break;    // 4. Zeile
      default: return;                   // fuer den Fall einer falschen Zeile
    }

  lcd_write_command(tmp);
}

void cursor_off(void)
  {
	lcd_write_command(0x08          // display on/off
                 |(1 << 2)      // D:  0= display off,  1= display on
                 |(0 << 1)      // C:  0= cursor off, 1= cursor on
                 |(0 << 0));    // B:  0= blink off, 1= blink on
  }

void cursor_on(void)
  {
	lcd_write_command(0x08          // display on/off
                 |(1 << 2)      // D:  0= display off,  1= display on
                 |(1 << 1)      // C:  0= cursor off, 1= cursor on
                 |(0 << 0));    // B:  0= blink off, 1= blink on
  }

void cursor_blink(void)
  {
	lcd_write_command(0x08          // display on/off
                 |(1 << 2)      // D:  0= display off,  1= display on
                 |(0 << 1)      // C:  0= cursor off, 1= cursor on
                 |(1 << 0));    // B:  0= blink off, 1= blink on
  }

void cursor_on_at(uint8_t x, uint8_t y, uint8_t size)
  {
    set_cursor_addr(x + size - 1, y);  // set text cursor to last char of input field
    cursor_on();
  }
//---------------------------------------------------------------------------------
//
// dummy to fullfill link requirements from MFT
//
#ifdef _CORTOS_H_
t_cr_task update_display(void)
  {
    return(-1);                         // fall asleep
  }

t_cr_task cursor(void)
  {
    return(-1);                         // fall asleep
  }

#endif

void enable_display_update(void)
  {  // we _are_ up to date
  }



#elif (DISPLAY == LCD_REICHELT162)

// setzt den Cursor in Zeile y (0..3) Spalte x (0..20)
// Origin: upper top left corner
void set_cursor(uint8_t x, uint8_t y)
{
  uint8_t tmp;

  cursor_x = x;
  cursor_y = y;

  switch (y) {
    case 0: tmp=0x80+0x00+x; break;    // 1. Zeile
    case 1: tmp=0x80+0x40+x; break;    // 2. Zeile
    case 2: tmp=0x80+0x10+x; break;    // 3. Zeile
    case 3: tmp=0x80+0x50+x; break;    // 4. Zeile
    default: return;                   // fuer den Fall einer falschen Zeile
  }

  lcd_write_command(tmp);
}

#endif // (DISPLAY == LCD_EA204)
