//----------------------------------------------------------------
//
// OpenDCC
// Throttle MFT 1.0
//
// Copyright (c) 2008-2009 Jean Michel Fischer, Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
// This code is initially based on:
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial 
//
//-----------------------------------------------------------------
//
// file:      lcd_routines.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-18 V0.01 jmf started
//            2009-01-13 V0.01 kw  comments, tests
//            2009-03-07 V0.04 kw  get_cursor neu dazu
//
//-----------------------------------------------------------------
//
// purpose:   control of a char lcd display
//            supported controllers: KS0073
//            either spi oder 4 bit parallel
//
//-----------------------------------------------------------------

#ifndef _LCD_DRIVER_KS0073_H_
#define _LCD_DRIVER_KS0073_H_

#include <stdlib.h>

//-----------------------------------------------------------------
//
// Configuration
//
//-----------------------------------------------------------------
#define LCD_EA204           1
#define LCD_REICHELT162     2


#ifndef DISPLAY
 #warning DISPLAY define is missing, we define the DISPLAY locally
 #define DISPLAY   LCD_EA204
#endif

#define LCD_IF_SPI      1
#define LCD_IF_4BIT     2

#ifndef DISPLAY_INTERFACE
 #warning we define the DISPLAY_INTERFACE locally
 #define DISPLAY_INTERFACE    LCD_IF_SPI
#endif

#ifndef F_CPU
   #warning Severe: F_CPU is not defined! using 10MHz instead
   #define F_CPU 10000000L
#endif

//-----------------------------------------------------------------------------

#if (DISPLAY_INTERFACE == LCD_IF_4BIT)
	// Pinbelegung für das LCD, an verwendete Pins anpassen
	
	#define LCD_PORT      PORTA
	#define LCD_DDR       DDRA
	#define LCD_RS        PA4
	#define LCD_EN        PA5
	// DB4 bis DB7 des LCD sind mit PA0 bis PA3 des AVR verbunden
    // see also: hardware.h
#elif (DISPLAY_INTERFACE == LCD_IF_SPI)
    // note: IM of Controller must be set low
    // we assume to be connected to the SPI-Controller of the ATMEL
    // AVR		          LCD
    // MOSI ------------> SID
    // SCLK ------------> SCLK
	// SPI_CS_LCD ------> /CS
  
    #define SPI_SELECT_LCD    PORTG &= ~(1<<SPI_CS_LCD);
    #define SPI_DESELECT_LCD  PORTG |= (1<<SPI_CS_LCD);
#else
    #warning no or wrong DISPLAY_INTERFACE
#endif // (DISPLAY_INTERFACE)

//-----------------------------------------------------------------------------
// Lowlevel routines:

void lcd_enable(void);                      // enables the display
void lcd_write_data(unsigned char temp1);   // send one data byte to the LCD
void lcd_command(unsigned char temp1);

//-----------------------------------------------------------------------------
// Highlevel routines:

typedef struct
  {
    unsigned char x;         // col, range 0..19
    unsigned char y;         // row, range 0..3
    unsigned char string[];
  }
  t_text;


void lcd_init(void);
void lcd_home(void);
void lcd_clear(void);
void lcd_reset(void);        // for ks0073 this is equal to clear

void set_cursor_addr(uint8_t x, uint8_t y);

void cursor_on_at(uint8_t x, uint8_t y, uint8_t size);
void cursor_on(void);
void cursor_blink(void);
void cursor_off(void);

void lcd_putc(char data);                           // puts data at cursor position
void lcd_putc_cnt(unsigned char count, char data);  // puts count data at cursor position
void lcd_blank(unsigned char toblank);              // puts blanks at cursor position

void lcd_string_P(const char *progmem_data);        // print string from Progmem at current pos.
void lcd_string(char *data);                        // print string  at current pos.

void lcd_int(int data);
void lcd_int2(int data);
void lcd_int02(int data);
void lcd_int3(int data);                            // print int, 3 chars wide, leading blank
void lcd_int4(int data);
void lcd_int04(int data);
void lcd_int5(int data);
void lcd_hex(int data);
void lcd_bool(unsigned char data);                  // display a single logic bin at cursor position.

void enable_display_update(void);

// LCD Befehle (only local usage)

#define LCD_OFF             0x00        // 39us
#define LCD_CLEAR_DISPLAY 	0x01        // 1.53ms
#define LCD_CURSOR_HOME     0x02        // 1.53ms
#define LCD_ON              0x08        // 39us

#define LCD_DISP_ON              0x0C   // display on, cursor off 
#define LCD_DISP_ON_BLINK        0x0D   // display on, cursor off, blink char
#define LCD_DISP_ON_CURSOR       0x0E   // display on, cursor on   
#define LCD_DISP_ON_CURSOR_BLINK 0x0F   // display on, cursor on, blink char 



#endif /* _LCD_DRIVER_KS0073_H_ */
