//----------------------------------------------------------------
//
// OpenDCC
// Gateway / MFT
//
// Copyright (c) 2008-2009 Jean Michel Fischer, Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
// This code is initially based on:
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial 
//
// Here we do the access to a lcd display, no special functions
// for MFT contained (so this is a general purpose file)
//
//-----------------------------------------------------------------
//
// file:      lcd_text.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-18 V0.01 jmf started
//            2009-01-13 V0.02 kw  comments, tests
//            2009-01-26 V0.03 kw  set_cursor jetzt ab 0,0
//            2009-03-07 V0.04 kw  get_cursor neu dazu
//            2009-03-20 V0.05 kw  nach init cursor off
//            2009-03-27 V0.06 kw  #if DISPLAY added
//
//-----------------------------------------------------------------
//
// purpose:   supply midlevel routines for a graphic or char lcd
//
// Interface: downstream: we use either:
//            lcd_driver_ks0073
//            lcd_driver_st7565
//
//-----------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <util/delay.h>
#include <stdio.h>

#include "config.h"
#include "cortos.h"
#include "hardware.h"           // CS for SPI

#include "lcd_ks0073_charset.h"

#include "lcd_driver_ks0073.h"
#include "lcd_driver_st7565.h"

//#include <avr/interrupt.h>

uint8_t cursor_x;
uint8_t cursor_y;

//------------------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)

// Schreibt ein Zeichen auf das LCD

void lcd_putc(char data)
  {
    lcd_write_data(data);
  }

void set_cursor(uint8_t x, uint8_t y)
  {
    cursor_x = x;
    cursor_y = y;
    set_cursor_addr(x, y);
  }


#elif  (DISPLAY == LCD_EA128x64)


// count from 0
void set_cursor(uint8_t x, uint8_t y)
{
  cursor_x = x;
  cursor_y = y;
  LCD_setCursorXY(x * 6, y * 8);    // remap to graphic
}


void lcd_putc(char data)
  {
    LCD_putchar(data);              // graphic print function
  }

#else
  #error: DISPLAY is not defined!
#endif // (DISPLAY == LCD_EA204)
//------------------------------------------------------------------------------------

void get_cursor(uint8_t* x, uint8_t* y)
  {
    *x = cursor_x;
    *y = cursor_y;
  }

void lcd_putc_cnt(unsigned char count, char data)
  {
    while(count > 0)
      {
        lcd_putc(data);
        count--;
      }
  }

void lcd_blank(unsigned char toblank)
  {
    lcd_putc_cnt(toblank, ' ');
  }


// Schreibt einen String auf das LCD

void lcd_string(char *data)
{
    while(*data) {
        lcd_putc(*data);
        data++;
    }
}

/* print string from program memory on lcd (no auto linefeed) */
void lcd_string_P(const char *progmem_data)
  {
    unsigned char c;

    while ( (c = pgm_read_byte(progmem_data++)) )
      {
        lcd_putc(c);
      }
  }

void lcd_pstring_P(t_text *data)
  {
    unsigned char x,y;
    x = pgm_read_byte(data++);
    y = pgm_read_byte(data++);
    set_cursor(x, y);
    lcd_string_P((char *)data);
  }


void lcd_int(int data) {
  char buffer[20];
  itoa(data, buffer, 10);
  lcd_string(buffer);
}

void lcd_int2(int data) {           // leading blank, right adjusted
  char r0[5];
  sprintf( r0, "%2.0d", data );
  lcd_string(r0);
}

void lcd_int02(int data) {          // with leading 0
  char r0[5];
  sprintf( r0, "%02d", data );
  lcd_string(r0);
}

void lcd_int3(int data) {
  char r0[5];
  sprintf( r0, "%3.0d", data );
  lcd_string(r0);
}


void lcd_int4(int data) {
  char r0[6];
  sprintf( r0, "%4.0d", data );
  lcd_string(r0);
}

void lcd_int04(int data) {         // with leading 0
  char r0[6];
  sprintf( r0, "%04d", data );
  lcd_string(r0);
}

void lcd_int5(int data) {
  char r0[7];
  sprintf( r0, "%5.0d", data );
  lcd_string(r0);
}


void lcd_hex2(uint8_t data) 
  {
    char r0[6];
    sprintf( r0, "%02X", data );
    lcd_string(r0);
  }

void lcd_hex4(uint16_t data)
  {
    lcd_hex2((uint8_t)(data >> 8));
    lcd_hex2((uint8_t)data);
  }

void lcd_hex8(uint32_t data)
  {
    lcd_hex4((uint16_t)(data >> 16));
    lcd_hex4((uint16_t)data);
  }




void lcd_bool(unsigned char digit)
  {
    if(digit) lcd_putc(C_BLOCK_M2);
    else      lcd_putc(C_BLOCK_SMALL);
  }


