#warning: CHECK Ledbrightness 255..100
//------------------------------------------------------------------------
//
// OpenDCC - LED Controller
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      led.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-23 V0.01 kw  started
//            2009-02-26 V0.02 kw  added routines for led access
//            2009-03-02 V0.03 kw  bugfix for turn_soft (underflow of DELTA)
//            2009-03-29 V0.04 kw  added led_bright .. additional gain
//            2009-04-09 V0.05 kw  added more blink modes
//            2010-07-30 V0.06 kw  added led_orange
//------------------------------------------------------------------------
//
// purpose:   LED Control
//
//------------------------------------------------------------------------
//
// interface: 1. call led_set_...();
//            2. 
// Interface:
// Howto:  every application calls led_set_....
//               this sets the corresponding values in our database and changes
//               the hardware setting.
// A cortos task is used to run these settings. (cooperative)

//------------------------------------------------------------------------
//

#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include <util/delay.h>

#include "hardware.h"
#include "config.h"
#include "cortos.h"
#include "timer.h"


#include "led.h"

//===========================================================================
// A note on brightness
// Every LED and the backlight runs with a timer based PWM, 256 steps.
// We use internally brightness values from 0..100.
//
// A mapping curve (built of 4 linear segments) maps this range to 0..255.
// The curve is linear for low values and gets more gain for brighter values. 
//
// Brightness values from the application are multiplied with global_bright
//
//===========================================================================
//
// how calculation is done:
//
// 1. select start value:   brightness_pow  or brightness_batt
// 2. check time: use either 100% (day) or reduced value
// 3. the result is global_brightness
// 4. for backlight: global_brightness * backlight_brightness
// 5. for LED:  global_brightness * led_brightness

// Global Memory

// static unsigned int global_bright;  // 0 = dark, 256 = fully on
static unsigned char backlight_brightness;   // normalized brightness 0..100
static unsigned char global_brightness;      // local copy of brightness 0..100;
static unsigned int global_brightness0_256;  // 0 = dark, 256 = fully on
static unsigned int led_brightness0_256;     // 0 = dark, 256 = fully on
                                             // used for calculations

// 5 LEDs: 0 = Softkey A, 1=B, ... 4 = Stop

#define LED_STATE_A 0
#define LED_STATE_B 1

typedef struct
  {
    unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
    unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
    unsigned char volume;       // aktueller Istwert
    unsigned char target;       // aktueller Zielwert
    unsigned char delta;        // aktuelle Schrittweite
    t_led_color   color;        // aktuelle Farbe

    unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
    unsigned char act_flash;    // Counter dazu
    unsigned char pause;        // wenn != 0: pulse repetierend
    
    unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
    t_led_color   color_A;
    unsigned char light_A_val;  // aktueller Zielwert in Richtung A
    unsigned char delta_A;      // aktuelles Increment in Richtung A

    unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
    t_led_color   color_B;
    unsigned char light_B_val;  // aktueller Zielwert in Richtung B
    unsigned char delta_B;      // aktuelles Decrement in Richtung B
  } t_led;

t_led led[NUM_OF_LED];  // 5

// define some handy delta values

#define DELTA_FAST    100
#define DELTA_200ms   (100 / (200000L / LED_TICK_PERIOD))  // ramp up in 200ms
#define DELTA_400ms   (100 / (400000L / LED_TICK_PERIOD))  // ramp up in 400ms
#define DELTA_1s      (100 / (1000000L / LED_TICK_PERIOD))  // ramp up in 1s

#if (DELTA_1s == 0)
  #warning severe DELTA_1s is too small - use a larger LED_TICK_PERIOD
#endif


//===========================================================================
// tools
// input:  volume in the range of 0 ... 100; (note: > 100 is alway max.)
// output: pwmval
//  GAIN    VAL in %    VAL in uint8
//  1         0           0
//  1        25          25
//  2        26          27
//  2        47          69
//  3        48          72
//  3        73         147
//  4        74         151
//  4       100         255

unsigned char convert_volume_2_pwm(unsigned char volume)
  {
    unsigned char pwm;
    if (volume <= 25) pwm = volume;
    else if (volume <=47)
      {
        volume -= 25;
        pwm = 2 * volume;
        pwm += 25;
      }
    else if (volume <=73)
      {
        volume -= 48;
        pwm = 3 * volume;
        pwm += 72;
      }
    else if (volume <=100)
      {
        volume -= 74;
        pwm = 4 * volume;
        pwm += 151;
      }
    else pwm = 255;
    return(pwm);
  }

// multiply two values range 0..100; result is range 0..100;
// we do:
// result = ( (((a * b) * 5) / 256) * 132 / 256 )
// this is fast and keeps the error low; 
unsigned char mult_norm100(unsigned char a, unsigned char b)
  {
    t_data16 result;

    result.as_uint16 = a * b;
    result.as_uint16 *= 5;
    result.as_uint16 = result.low8 * 132;
    return(result.low8);
  }

//===========================================================================



void set_backlight()
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    //          0..100                 0..256
    my_volume = backlight_brightness * global_brightness0_256;     // now we have a range of 0...25600

    if (my_volume == 0)
      {
        TCCR3A &=  ~(1 << COM3C1);
        TCCR3A &=  ~(1 << COM3C0);
      }
    else
      {
        pwm = convert_volume_2_pwm(my_volume / 256);
        OCR3CH = 0;
        OCR3CL = pwm;
        TCCR3A |=   (1 << COM3C1);     // 10=clear on compare match, set at top
        TCCR3A &=  ~(1 << COM3C0);
      }
  }


// Softkey A is connected between OCR2A and Pin LED_Y0
// X = Bit 3, OCR2A
// Y = LED_Y0
// Red: X=1, Y=0; Green: X=0, Y=1

void set_LED_A(t_led_color color, unsigned char volume)
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    my_volume = volume * global_brightness0_256;   // now we have a range of 0...25600
    my_volume /= 256;
    my_volume *= led_brightness0_256;              // now we have a range of 0...25600

    if (volume == 0)
      {
        LED_A_OFF; // get both X and Y = 0;
        TCCR2A &=  ~(1 << COM2A1);      // X = normal port
        TCCR2A &=  ~(1 << COM2A0);  
      }
    else
      {
        switch(color)
          {
            default:  // no break;
            case LED_COLOR_OFF:
                LED_A_OFF; // get both X and Y = 0;
                TCCR2A &=  ~(1 << COM2A1);      // X = normal port
                TCCR2A &=  ~(1 << COM2A0);
                break;
            case LED_COLOR_RED:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR2A = pwm;
                TCCR2A |=   (1 << COM2A1);     // X = OCR, 10=clear on compare match, set at top
                TCCR2A &=  ~(1 << COM2A0);
                LED_A_RED;
                break;
            case LED_COLOR_GREEN:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR2A = pwm;
                TCCR2A |=   (1 << COM2A1);     // X = OCR, 11=set compare match, clear at top
                TCCR2A |=   (1 << COM2A0);
                LED_A_GREEN;            // Y = 1
                break;
          }
      }
  }

// Softkey B is connected between OCR1A and Pin LED_Y1
// X = OCR1A
// Y = LED_Y1
// Red: X=1, Y=0; Green: X=0, Y=1

void set_LED_B(t_led_color color, unsigned char volume)
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    my_volume = volume * global_brightness0_256;   // now we have a range of 0...25600
    my_volume /= 256;
    my_volume *= led_brightness0_256;              // now we have a range of 0...25600


    if (volume == 0)
      {
        LED_B_OFF; // get both X and Y = 0;
        TCCR1A &=  ~(1 << COM1A1);      // X = normal port
        TCCR1A &=  ~(1 << COM1A0);
      }
    else
      {
        switch(color)
          {
            default:  // no break;
            case LED_COLOR_OFF:
                LED_B_OFF; // get both X and Y = 0;
                TCCR1A &=  ~(1 << COM1A1);      // X = normal port
                TCCR1A &=  ~(1 << COM1A0);
                break;
            case LED_COLOR_RED:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1A = pwm;
                TCCR1A |=   (1 << COM1A1);     // X = OCR, 10=clear on compare match, set at top
                TCCR1A &=  ~(1 << COM1A0);
                LED_B_RED;
                break;
            case LED_COLOR_GREEN:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1A = pwm;
                TCCR1A |=   (1 << COM1A1);     // X = OCR, 11=set compare match, clear at top
                TCCR1A |=   (1 << COM1A0);
                LED_B_GREEN;            // Y = 1
                break;
          }
      }
  }

// Softkey C is connected between OCR1B and Pin LED_Y2
// X = OCR1B
// Y = LED_Y2
// Red: X=1, Y=0; Green: X=0, Y=1

void set_LED_C(t_led_color color, unsigned char volume)
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    my_volume = volume * global_brightness0_256;   // now we have a range of 0...25600
    my_volume /= 256;
    my_volume *= led_brightness0_256;              // now we have a range of 0...25600

    if (volume == 0)
      {
        LED_C_OFF; // get both X and Y = 0;
        TCCR1A &=  ~(1 << COM1B1);      // X = normal port
        TCCR1A &=  ~(1 << COM1B0);
      }
    else
      {
        switch(color)
          {
            default:  // no break;
            case LED_COLOR_OFF:
                LED_C_OFF; // get both X and Y = 0;
                TCCR1A &=  ~(1 << COM1B1);      // X = normal port
                TCCR1A &=  ~(1 << COM1B0);
                break;
            case LED_COLOR_RED:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1B = pwm;
                TCCR1A |=   (1 << COM1B1);     // X = OCR, 10=clear on compare match, set at top
                TCCR1A &=  ~(1 << COM1B0);
                LED_C_RED;
                break;
            case LED_COLOR_GREEN:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1B = pwm;
                TCCR1A |=   (1 << COM1B1);     // X = OCR, 11=set compare match, clear at top
                TCCR1A |=   (1 << COM1B0);
                LED_C_GREEN;            // Y = 1
                break;
          }
      }
  }

// Softkey D is connected between OCR1C and Pin LED_Y3
// X = OCR1C
// Y = LED_Y3
// Red: X=1, Y=0; Green: X=0, Y=1

void set_LED_D(t_led_color color, unsigned char volume)
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    my_volume = volume * global_brightness0_256;   // now we have a range of 0...25600
    my_volume /= 256;
    my_volume *= led_brightness0_256;              // now we have a range of 0...25600

    if (volume == 0)
      {
        LED_D_OFF; // get both X and Y = 0;
        TCCR1A &=  ~(1 << COM1C1);      // X = normal port
        TCCR1A &=  ~(1 << COM1C0);
      }
    else
      {
        switch(color)
          {
            default:  // no break;
            case LED_COLOR_OFF:
                LED_D_OFF; // get both X and Y = 0;
                TCCR1A &=  ~(1 << COM1C1);      // X = normal port
                TCCR1A &=  ~(1 << COM1C0);
                break;
            case LED_COLOR_RED:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1C = pwm;
                TCCR1A |=   (1 << COM1C1);     // X = OCR, 10=clear on compare match, set at top
                TCCR1A &=  ~(1 << COM1C0);
                LED_D_RED;
                break;
            case LED_COLOR_GREEN:
                pwm = convert_volume_2_pwm(my_volume / 256);
                OCR1C = pwm;
                TCCR1A |=   (1 << COM1C1);     // X = OCR, 11=set compare match, clear at top
                TCCR1A |=   (1 << COM1C0);
                LED_D_GREEN;            // Y = 1
                break;
          }
      }
  }



// volume: 0..100
void set_LED_STOP(t_led_color color, unsigned char volume)
  {
    unsigned int my_volume;
    unsigned char pwm;
    
    my_volume = volume * global_brightness0_256;   // now we have a range of 0...25600
    my_volume /= 256;
    my_volume *= led_brightness0_256;              // now we have a range of 0...25600

    if (volume == 0)
      {
        LED_STOP_OFF; // get both X and Y = 0;
        TCCR3A &=  ~(1 << COM3B1);      // Y = normal port
        TCCR3A &=  ~(1 << COM3B0);
        TCCR3A &=  ~(1 << COM3A1);      // X = normal port
        TCCR3A &=  ~(1 << COM3A0);
      }
    else
      {
        switch(color)
          {
            default:  // no break;
            case LED_COLOR_OFF:
                LED_STOP_OFF; // get both X and Y = 0;
                TCCR3A &=  ~(1 << COM3B1);      // Y = normal port
                TCCR3A &=  ~(1 << COM3B0);
                TCCR3A &=  ~(1 << COM3A1);      // X = normal port
                TCCR3A &=  ~(1 << COM3A0);
                break;
            case LED_COLOR_RED:
                pwm = convert_volume_2_pwm(my_volume / 256);
                TCCR3A &=  ~(1 << COM3B1);      // Y = normal port
                TCCR3A &=  ~(1 << COM3B0);
                OCR3AH = 0;
                OCR3AL = pwm;
                TCCR3A |=   (1 << COM3A1);     // X = OCR, 10=clear on compare match, set at top
                TCCR3A &=  ~(1 << COM3A0);
                LED_STOP_RED;
                break;
            case LED_COLOR_GREEN:
                pwm = convert_volume_2_pwm(my_volume / 256);
                TCCR3A &=  ~(1 << COM3B1);      // Y = normal port
                TCCR3A &=  ~(1 << COM3B0);
                OCR3AH = 0;
                OCR3AL = pwm;
                TCCR3A |=   (1 << COM3A1);     // X = OCR, 11=set on compare match, clear at top
                TCCR3A |=   (1 << COM3A0);
                LED_STOP_GREEN;            // Y = 1
                break;
          }
      }
  }


// parameter:
// i: the LED
// color: LED_COLOR_RED, LED_COLOR_GREEN, ...
// brightness: 0..100
void set_led_to(unsigned char i, t_led_color color, unsigned char brightness)
  {
    switch(i)
      {
        case 0: set_LED_A(color, brightness);     break;
        case 1: set_LED_B(color, brightness);     break;
        case 2: set_LED_C(color, brightness);     break;
        case 3: set_LED_D(color, brightness);     break;
        case 4: set_LED_STOP(color, brightness);  break;
        default: break;
      }    
  }


// parameter bright: 0..100
void set_global_bright(unsigned char bright)
  {
    unsigned char i;
    unsigned int temp;

    global_brightness = bright;
    
    // scale this value up 100 -> 256; 2,56 =~ 41/16

    temp = bright * 41;
    temp /= 16;

    // global_bright = temp;        // save to static var
    global_brightness0_256 = temp;        // save to static var
    
    set_backlight();                      // redo backlight
    
    for (i=0; i<NUM_OF_LED; i++)          // redo all leds
      {
        set_led_to(i,led[i].color,led[i].volume);
      }
   }

unsigned char get_global_bright(void)
  {
    return(global_brightness);
  }

//----------------------------------------------------------------------------------

// run_led implements the state engine for leds
//  1: check out whether we should approch a new lightval
//  2: check blinking and state changes
//  return code: desired sleep time in systicks

t_cr_task run_led(void)
  {
    unsigned char i;

    for (i=0; i<NUM_OF_LED; i++)
      {
        // part 1: check state changes

        if (led[i].rest != 0)
          {
            if (--led[i].rest == 0)
              {
                 if (led[i].state == LED_STATE_A)
                  {
                    if (led[i].act_flash == led[i].flashes)
                      {
                        led[i].rest = led[i].pause;
                        led[i].act_flash = 0;
                      }
                    else
                      {
                        led[i].rest = led[i].time_B;
                      }
                    if (led[i].color != led[i].color_B)
                      {
                        led[i].color = led[i].color_B;
                        led[i].volume = led[i].light_B_val +1; // force update 
                      }
                    led[i].state = LED_STATE_B;
                    led[i].target = led[i].light_B_val;
                    led[i].delta = led[i].delta_B;
                  }
                else
                  {
                    led[i].act_flash++;
                    led[i].rest = led[i].time_A;
                    
                    if (led[i].color != led[i].color_A)
                      {
                        led[i].color = led[i].color_A;
                        led[i].volume = led[i].light_A_val +1;
                      }
                    led[i].state = LED_STATE_A;
                    led[i].target = led[i].light_A_val;
                    led[i].delta = led[i].delta_A;
                  }
              }
          }

        // part 2: check dimming
        
        unsigned char my_bright;
        
        my_bright = led[i].volume;
   
        if (my_bright == led[i].target);
        else if (my_bright > led[i].target)
          {
            if (led[i].target + led[i].delta > my_bright)
              {
                led[i].volume = led[i].target;
                set_led_to(i,led[i].color,led[i].volume);
              }
            else
              {
                led[i].volume = my_bright - led[i].delta;
                set_led_to(i,led[i].color,led[i].volume);
              }
          }
        else
          {
            if (my_bright + led[i].delta >= led[i].target)
              {
                led[i].volume = led[i].target;
                set_led_to(i,led[i].color,led[i].volume);
              }
            else
              {
                led[i].volume = my_bright + led[i].delta;
                set_led_to(i,led[i].color,led[i].volume);
              }
          }
      }
    return(LED_TICK_PERIOD / SYSTICK_PERIOD);
  }


void start_led(unsigned char i)
  {
    if (led[i].state == LED_STATE_A)
      {
        led[i].volume = led[i].light_A_val +1;  // force update
        led[i].target = led[i].light_A_val;
        led[i].color = led[i].color_A;
        led[i].delta = led[i].delta_A;
      }
    else 
      {
        led[i].volume = led[i].light_B_val +1;
        led[i].target = led[i].light_B_val;
        led[i].color = led[i].color_B;
        led[i].delta = led[i].delta_B;
      }
  }

#define FULL_BRIGHT   100

// the following single color led controls only use the dimm engine,
// not the led_state engine

void turn_led_red_soft(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].target = FULL_BRIGHT;
    if (led[led_enum].volume == FULL_BRIGHT) led[led_enum].volume++;
    led[led_enum].color = LED_COLOR_RED;
    led[led_enum].delta = DELTA_1s;
  }

void turn_led_green_soft(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].target = FULL_BRIGHT;
    if (led[led_enum].volume == FULL_BRIGHT) led[led_enum].volume++;
    led[led_enum].color = LED_COLOR_GREEN;
    led[led_enum].delta = DELTA_1s;
  }

void turn_led_red_immediate(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].target = FULL_BRIGHT;
    if (led[led_enum].volume == FULL_BRIGHT) led[led_enum].volume++;
    led[led_enum].color = LED_COLOR_RED;
    led[led_enum].delta = DELTA_FAST;
  }

void turn_led_green_immediate(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].target = FULL_BRIGHT;
    if (led[led_enum].volume == FULL_BRIGHT) led[led_enum].volume++;
    led[led_enum].color = LED_COLOR_GREEN;
    led[led_enum].delta = DELTA_FAST;
  }

void turn_led_off_soft(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].target = 0;
    if (led[led_enum].volume == 0) led[led_enum].volume++;
    led[led_enum].delta = DELTA_1s;
  }

void turn_led_off(unsigned char led_enum)
  {
    led[led_enum].rest = 0;
    led[led_enum].volume = 1;
    led[led_enum].target = 0;
    led[led_enum].delta = DELTA_FAST;
  }


/*
// make a series of flashes, then a longer pause
void flash_errorcode(unsigned char count)
  {
    led[4].act_flash = 1;
    led[4].flashes = count;
    led[4].pause   = 700000L / LED_TICK_PERIOD;
    led[4].time_B  = 240000L / LED_TICK_PERIOD;
    led[4].time_A  = 120000L / LED_TICK_PERIOD;
    led[4].rest    = 120000L / LED_TICK_PERIOD;
    led[4].state = LED_STATE_A;
    led[4].color_A = LED_COLOR_RED;
    led[4].color_B = LED_COLOR_OFF;
    start_led(4);
    // check
    #if ((700000L / LED_TICK_PERIOD) > 255L)
      #warning: LED_TICK_PERIOD too small
    #endif
  }
*/


//!warning! disconnected mit memcopy machen!

#define NR_ON  (240000L / LED_TICK_PERIOD)

t_led led_disconnect[4] = 
  {
    {
        LED_STATE_B,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        NR_ON,          // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        1,              // unsigned char volume;       // aktueller Istwert
        0,              // unsigned char target;       // aktueller Zielwert
        20,             // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_GREEN,// t_led_color   color;        // aktuelle Farbe

        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        0,              // unsigned char pause;        // wenn != 0: pulse repetierend
    
        NR_ON,          // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_RED,  // t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        1,              // unsigned char delta_A;      // aktuelles Increment in Richtung A

        5 * NR_ON,      // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_GREEN,// t_led_color   color_B;
        100,            // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        1,              // unsigned char delta_B;      // aktuelles Decrement in Richtung B
    },
    {
    },
    {
    },
    {
    }
  };



void show_disconnected(void)
  {
    // we make a red 'night rider' on the softkeys
    // A  _X_____X_____X___
    // B  X_X___X_X___X_X__
    // C  ___X_X___X_X___X_
    // D  ____X_____X_____X

    
    led[0].state = LED_STATE_B;
    led[0].rest = NR_ON;
    led[0].time_A = NR_ON;
    led[0].time_B = 5 * NR_ON;
    led[0].act_flash = 0;
    led[0].flashes = 0;
    led[0].color_A = LED_COLOR_RED;
    led[0].color_B = LED_COLOR_OFF;

    led[1].state = LED_STATE_A;
    led[1].rest = NR_ON;
    led[1].time_A = NR_ON;
    led[1].time_B = 1 * NR_ON;
    led[1].act_flash = 1;
    led[1].flashes = 2;
    led[1].pause = 3 * NR_ON;
    led[1].color_A = LED_COLOR_RED;
    led[1].color_B = LED_COLOR_OFF;

    led[2].state = LED_STATE_B;
    led[2].rest = 3 * NR_ON;
    led[2].time_A = NR_ON;
    led[2].time_B = 1 * NR_ON;
    led[2].act_flash = 0;
    led[2].flashes = 2;
    led[2].pause = 3 * NR_ON;
    led[2].color_A = LED_COLOR_RED;
    led[2].color_B = LED_COLOR_OFF;

    led[3].state = LED_STATE_B;
    led[3].rest = 4 * NR_ON;
    led[3].time_A = NR_ON;
    led[3].time_B = 5 * NR_ON;
    led[3].act_flash = 0;
    led[3].flashes = 0;
    led[3].color_A = LED_COLOR_RED;
    led[3].color_B = LED_COLOR_OFF;
    start_led(0);
    start_led(1);
    start_led(2);
    start_led(3);
  }




#define BLINK_SOFT  (1000000L / LED_TICK_PERIOD)
#define BLINK_FAST  (300000L / LED_TICK_PERIOD)
#define BLINK       (500000L / LED_TICK_PERIOD)


const t_led DAT_led_blink_red_soft PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_SOFT,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_RED,  // t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_SOFT,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_SOFT,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_RED,  // t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        10,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_SOFT,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_RED,  // t_led_color   color_B;
        0,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        5,              // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };

void blink_led_red_soft(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_blink_red_soft, sizeof(DAT_led_blink_red_soft));
  }

const t_led DAT_led_blink_red_fast PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_FAST,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_RED,  // t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_FAST,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_FAST,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_RED,  // t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        34,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_FAST,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_RED,  // t_led_color   color_B;
        0,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        34,             // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };

void blink_led_red_fast(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_blink_red_fast, sizeof(DAT_led_blink_red_fast));
  }



const t_led DAT_led_blink_green_soft PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_SOFT,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_GREEN,// t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_SOFT,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_SOFT,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_GREEN,// t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        10,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_SOFT,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_GREEN,// t_led_color   color_B;
        0,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        10,             // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };


void blink_led_green_soft(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_blink_green_soft, sizeof(DAT_led_blink_green_soft));
  }


const t_led DAT_led_blink_green_fast PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_FAST,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_GREEN,// t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_FAST,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_FAST,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_GREEN,// t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        25,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_FAST,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_GREEN,// t_led_color   color_B;
        0,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        25,             // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };


void blink_led_green_fast(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_blink_green_fast, sizeof(DAT_led_blink_green_fast));
  }


const t_led DAT_led_blink_red_green PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_FAST,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_GREEN,// t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_FAST,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_FAST,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_RED,  // t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        25,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_FAST,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_GREEN,// t_led_color   color_B;
        100,            // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        25,             // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };


void blink_led_red_green(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_blink_red_green, sizeof(DAT_led_blink_red_green));
  }

const t_led DAT_led_orange PROGMEM = 
  {
        LED_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        1,              // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        LED_COLOR_GREEN,// t_led_color color;          // aktuelle Farbe
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        1,              // unsigned char pause;        // wenn != 0: pulse repetierend
    
        1,               // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        LED_COLOR_RED,  // t_led_color   color_A;
        100,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        100,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        1,              // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        LED_COLOR_GREEN,// t_led_color   color_B;
        100,            // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        100,            // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };


void led_orange(unsigned char led_enum)
  {
    memcpy_P (&led[led_enum], &DAT_led_orange, sizeof(DAT_led_orange));
  }


void restore_brightness(t_power_mode power_mode)
  {
    unsigned char temp;

    if (power_mode == BATT_POWERED)
      {
        temp = eeprom_read_byte((void *)eadr_global_bright_batt);
      }
    else
      {
        temp = eeprom_read_byte((void *)eadr_global_bright_pow);
      }
    set_global_bright(temp);
  }

void save_brightness(void)
  {
    unsigned char temp;

    temp = get_global_bright();
    eeprom_write_byte((void *)eadr_global_bright_pow, temp);
  }

void set_led_brightness(unsigned char value)
  {
    unsigned char i;
    unsigned int temp;

    // scale this value up 100 -> 256; 2,56 =~ 41/16
    temp = value * 41;
    temp /= 16;
    led_brightness0_256 = temp;           // save to static var

    for (i=0; i<NUM_OF_LED; i++)          // redo all leds
      {
        set_led_to(i,led[i].color,led[i].volume);
      }
  }

void set_disp_brightness(unsigned char value)
  {
    backlight_brightness = value;
    set_backlight();
  }


void init_led(void)
  {
    unsigned char temp;

    set_led_brightness(eeprom_read_byte((void *)eadr_global_bright_led));
    
    backlight_brightness = eeprom_read_byte((void *)eadr_global_bright_disp);

    restore_brightness(LINE_POWERED);                   // get global brightness from eeprom
    #warning: brightness fixed to power_mode LINE_POWERED

    for(temp = 0; temp < NUM_OF_LED; temp++) turn_led_off(temp);    

    #if (LED_STOP_DEBUG == 1)
      blink_led_red_soft(LED_STOP);
    #endif
  }





// debug only

static unsigned char bootloader_led;

void advance_bootloader_led(void)
  {
    bootloader_led--;
    bootloader_led &= 0x07;  // 8 states
    switch (bootloader_led)
      {
        case 0: LED_C_OFF; LED_D_GREEN; break;
        case 1: LED_D_OFF; LED_A_RED; break;
        case 2: LED_A_OFF; LED_B_RED; break;
        case 3: LED_B_OFF; LED_C_RED; break;
        case 4: LED_C_OFF; LED_D_RED; break;
        case 5: LED_D_OFF; LED_A_GREEN; break;
        case 6: LED_A_OFF; LED_B_GREEN; break;
        case 7: LED_B_OFF; LED_C_GREEN; break;
      }
  }


