//------------------------------------------------------------------------
//
// OpenDCC - Throttle MFT
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-12 V0.01 kw started
//            2013-07-25 V0.02 kw XBEE - parserstart wird erzwungen
//
//------------------------------------------------------------------------
//
// purpose:   A program for a versatile model railroad throttle
//
//            see config.h for system configuration
//
//------------------------------------------------------------------------
//
// style guide (mandatory!):
//  1. every change must be mentioned in the history with date and name
//  2. debug messages must be inside #ifdef DEBUG
//  3. Es wird im Allman-Stil eingerckt, standard 4 chars
//  4. Windows cr+lf, blanks statt tabs.
//  5. deutsche Sonderzeichen soweit als moeglich vermeiden.
//  6. (temporre) Codierungslcken sind mit //!warning! zu kennzeichnen.
//  7. Soweit moeglich: alle Konstanten (Strings) ins Flash (progmem)
//  8. Version, subversion und build muessen nachgezogen werden


#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // cooperative scheduler and fifo
#include "timer.h"
#include "led.h"
#include "lcd_text.h"
#include "display.h"                // midlevel display routines
#include "keyboard.h"

#include "xp_client_parser.h"       // init xp
#include "analog.h"
#include "messages_objects.h"
#include "serial.h"                 // serial io
#include "database.h"
#include "menu_navigator.h"

#include "xbee_if.h"
// #include "main.h"

//----------------------------------------------------------------------------------
//
// System wide global data
// (status of throttle itself)
//
unsigned char timeout_to_shutdown;  // this is the time until the 
                                    // 
unsigned int serial_number;

unsigned char throttle_id;          // logische Handregler-ID (0..31)

    

// parameter time_to_wait [unit: 10ms]
void wait_time(int time_to_wait)
  {
    signed int my_systick_10ms;
    my_systick_10ms = get_systick_10ms();  // atomic!
 //   while ((get_systick_10ms() - my_systick_10ms) <= (1000000L / SYSTICK_10MS_PERIOD));
    while ((get_systick_10ms() - my_systick_10ms) <= time_to_wait);
  }



//----------------------------------------------------------------------------------
//
// MAIN
//
//----------------------------------------------------------------------------------

 
static signed char main_systick;
static signed char led_systick;

int main(void)
  {
    unsigned char splash_time;
    t_used_interface prefered_interface;                                // could either be xpressnet or xbee

    init_hardware();                                                    // ports

    if (eeprom_read_byte((void *)eadr_MFT_version) != MFT_VERSION)      // check if eeprom loaded
      {
        // oops, no data loaded or wrong version! 
        LED_A_RED;LED_B_RED;LED_C_RED;LED_D_RED;
        SET_SCANLINE1_to_0;
        while(1)
          {
            if ((PINC & (1<<KEY_3)) == 0)
    	      {
    		    TURN_POWER_OFF;
              }
          }
      }

    serial_number = eeprom_read_byte((void *)eadr_MFT_serial_h) * 256 +
                    eeprom_read_byte((void *)eadr_MFT_serial_l);

    throttle_id   = eeprom_read_byte((void *)eadr_pc_tid); // logische Handregler-ID (0..31)

    init_database();                    // load system settings and locos
    init_timer();                       // create systick, led and xp timer
    init_led();                         // LED needs timers
    init_keyboard();                    // keys and knob
    init_system_monitor();              // power and battery control
    init_objects();                     // all state infos about power, loks, turnouts

    LED_A_GREEN;

    operation_mode = OPM_THROTTLE;      // this is currently fix, later we will load this from eeprom
    prefered_interface = USE_NONE;      

    lcd_init();                         // LCD (either char 20x4 or graphic 128 x 64)

    init_xpressnet_client();            // Xpressnet Interface (Parser)

    LED_B_GREEN;

    //!warning! determine powermode (if vcciosence == high)

    main_systick = get_systick();
    led_systick = get_systick();

    sei();

    splash_time = eeprom_read_byte((void *)eadr_show_splash);
    if (splash_time) show_splash_mft();         // from menu_splash.c
    if (splash_time < 3) splash_time = 3;       // mind. 300ms (need time for xbee!)
    wait_time(splash_time * 10);

    detect_xbee_module();                // this will update xbee_module_detected (start 200ms after power up)

    if(xbee_module_detected)  { LED_A_GREEN; }
    else                      { LED_A_RED; }

    while (keystroke_present())  get_wait_keystroke();  // clean up buffer

    init_navigator();                                   // this sets gui to SCREEN MODE


    if (a58_check())
      {
        switch(eeprom_read_byte((void *)eadr_connect))
          {
            default:
            case 0: 
                if(xbee_module_detected)
                  {
                    ENABLE_XBEE;
                    init_rs232(DEFAULT_BAUD, TASK_XBEE); 
                    set_task_ready(TASK_INIT_XBEE);
                    prefered_interface = USE_XBEE;
                  }
                else
                  {
                    ENABLE_USB;
                    init_rs232(DEFAULT_BAUD, TASK_PARSER);
                    prefered_interface = USE_XPRESSNET;             // we override any setting if no module
                  }
                break;
            case 1:
                // ENABLE_USB;
                init_rs232(DEFAULT_BAUD, TASK_PARSER);
                prefered_interface = USE_XPRESSNET; 
                break;
            case 2:
                ENABLE_XBEE;
                init_rs232(DEFAULT_BAUD, TASK_XBEE); 
                set_task_ready(TASK_INIT_XBEE);
                prefered_interface = USE_XBEE;
                break;
          }
      }
    else
      { // Falko spezial-schnitz
        init_rs232(DEFAULT_BAUD, TASK_PARSER);
        prefered_interface = USE_XPRESSNET; 
      }
    
    #warning SET_XBEE_TO_SLEEP
    //!!  SET_XBEE_TO_SLEEP;

    // now we must put a keystroke into the fifo
    // whenever we have a preferred interface (which is up and running)
    // we simple push the appropriate key into the fifo (same as being pressed)
    switch(prefered_interface)
      {
        default:
        case USE_NONE:
            cr_fifo_put(&key_fifo, KEY_SH_STOP);
            break;
        case USE_USB:
            ENABLE_USB;
            cr_fifo_put(&key_fifo, KEY_SH_STOP);
            break;
        case USE_XPRESSNET:
            ENABLE_USB;                                    // no module, so open USB
            cr_fifo_put(&key_fifo, KEY_SH_A);              // simulate keystroke for Xpressnet - enter SCREEN_LOK
            break;
        case USE_XBEE:
            cr_fifo_put(&key_fifo, KEY_SH_STOP);
            cr_fifo_put(&key_fifo, KEY_SH_B);             // hk: keystroke needed to get parser to work - if XBEE gets connected, it should
                                                         // send the stroke
            break;
        case USE_CAN:
            cr_fifo_put(&key_fifo, KEY_SH_STOP);
            break;
      }                                                  // may be we will have other keys in future
    
    // now enter MAIN LOOP!
    while(1)
      {
        if ((signed char)(get_systick() - main_systick) > 0)
          {
            main_systick = get_systick();
            knob_poll();                    // here we go with hook functions
            keyboard_poll();
          }

        scheduler();                        // run task engine (from cortos.c)
      }
  }


#if (1==0)

// some old code fragment - may be we could reuse it for a test bench

    //!warning! this switch should move to scheduler!
    switch(operation_mode)
      {
        case OPM_THROTTLE:

            
           break;
        case OPM_THROTTLE:
           LED_A_GREEN;LED_B_GREEN;LED_C_GREEN;LED_D_RED;
           SET_SCANLINE1_to_0;
            while(1)
              {
                if ((PINC & (1<<KEY_3)) == 0)
        	      {
        		    TURN_POWER_OFF;
                  }
              };
            break;   
        default:   
           while(1)
             {

                        unsigned char backlight_volume = 50;
                        unsigned char key;
                        signed int j;

                        if (keystroke_present())
                          {
                            key = get_wait_keystroke();

                            sendchar( key );
                            if (key == KEY_KNOB_TURNED)
                              {
                                signed int rot;
                                rot = get_knob_rotate();
                                if (rot == 0) ;
                                else if (rot > 0)
                                   for (j = 0; j < rot; j++)  sendchar( '+' );
                                else 
                                   for (j = 0; j < -rot; j++)  sendchar( '-' );

                                if (rot == 0) ;
                                else if (rot > 0)
                                  { backlight_volume += rot;
                                    if (backlight_volume > 100) backlight_volume = 100;
                                    set_global_bright(backlight_volume);
                                  }
                                else
                                  { if (backlight_volume > -rot) backlight_volume = backlight_volume - (-rot);
                                    else backlight_volume = 0;
                                    set_global_bright(backlight_volume);
                                  } 
                                set_global_bright(backlight_volume);
                              }


                			switch(key)
                			  {
                			    case KEY_A: blink_led_red(LED_A); break;
                			    case KEY_B: blink_led_red(LED_B); break;
                			    case KEY_C: blink_led_red(LED_C); break;
                			    case KEY_D: blink_led_red(LED_D); break;
                			    case KEY_STOP: blink_led_red(LED_STOP); break;
                			    case KEY_SH_A: blink_led_green(LED_A); break;
                			    case KEY_SH_B: blink_led_green(LED_B);  break;
                			    case KEY_SH_C: blink_led_green(LED_C); break;
                			    case KEY_SH_D: blink_led_green(LED_D);  break;
                			    case KEY_SH_STOP: blink_led_green(LED_STOP); break;
                				case KEY_CENTER: turn_led_off(LED_A);
                                                 turn_led_off(LED_B); 
                                                 turn_led_off(LED_C); 
                                                 turn_led_off(LED_D);
                                                 turn_led_off(LED_STOP);  
                                 
                                    display_status(STS_DISCONNECTED);      // show status at upper right corner
                                    
                                                  break;
                                case KEY_KNOB:
                                    lcd_init(); 
                                    break;
                			    default: break;
                			  }       
              }
          }
      }
  }
#endif



