//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_clock.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2010-06-16 V0.01 kw  startet
//            2010-06-22 V0.07 kw  msg_transmit
//            2014-08-17 V0.08 kw  added english menu text
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for clock setup
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"

//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------
//
// we use the array actual_extacc from messages_objects.c
//
      
static unsigned char edit_focus;   // da steht der Cursor gerade
static t_fast_clock  my_clock;

//------------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   012345678901234567890
//  |--------------------|
// 0|Clock          ON   |
// 1| Uhrzeit:  hh:mm    |
// 2| Faktor:      12    |
// 3| 8:00           set |
//  |--------------------|
//

#if (LANGUAGE == GERMAN)
    const static char text_uhrzeit[]     PROGMEM = "Uhrzeit:";
    const static char text_faktor[]      PROGMEM = "Faktor:";


    const static char text_clock[]      PROGMEM = "Clock";
    const static char text_sk_A[]       PROGMEM = "7:58";
    const static char text_sk_B[]       PROGMEM = "11:58";
    const static char text_sk_C[]       PROGMEM = "18:30";
    const static char text_sk_D[]       PROGMEM = "set";

#elif  (LANGUAGE == ENGLISH)
    const static char text_uhrzeit[]     PROGMEM = "Time:";
    const static char text_faktor[]      PROGMEM = "Factor:";


    const static char text_clock[]      PROGMEM = "Clock";
    const static char text_sk_A[]       PROGMEM = "7:58";
    const static char text_sk_B[]       PROGMEM = "11:58";
    const static char text_sk_C[]       PROGMEM = "18:30";
    const static char text_sk_D[]       PROGMEM = "set";
#else
    #error LANGUAGE is void
#endif


#define UHR_ZEILE          1
#define FAKTOR_ZEILE       2
#define UHR_SPALTE        11

static void display_uhr()
  {
    cursor_off();
    set_cursor(UHR_SPALTE,UHR_ZEILE);
    lcd_int2(my_clock.hour);
    lcd_putc(':');
    lcd_int02(my_clock.minute);
    set_cursor(UHR_SPALTE+3,UHR_ZEILE+2);
    lcd_int2(my_clock.ratio);    
  }

static void set_cursor_editfocus()
  {
    switch (edit_focus)
      {
        default:
        case 0: cursor_on_at(UHR_SPALTE,UHR_ZEILE,2); break;
        case 1: cursor_on_at(UHR_SPALTE+3,UHR_ZEILE,2); break;
        case 2: cursor_on_at(UHR_SPALTE+3,FAKTOR_ZEILE,2); break;
      }
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_clock);
    check_status_display(0);

    set_cursor(3,UHR_ZEILE-1);
    lcd_string_P(text_uhrzeit);
    set_cursor(UHR_SPALTE,UHR_ZEILE-1);
    lcd_putc('h');
    lcd_putc('h');
    lcd_putc(':');
    lcd_putc('m');
    lcd_putc('m');
    set_cursor(3,FAKTOR_ZEILE);
    lcd_string_P(text_faktor);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)


//   0123456789012345678901234
//  |------------------------|
// 0|Clock              ON   |
// 1|                        |
// 2| Uhrzeit:   HH:MM       |
// 3|            12:34       |
// 4| Faktor:        8       |
// 5|  7:58 11:58 18:30  set |
//  |------------------------|
//

#if (LANGUAGE == GERMAN)
    const static char text_uhrzeit[]     PROGMEM = "Uhrzeit:";
    const static char text_faktor[]      PROGMEM = "Faktor:";


    const static char text_clock[]      PROGMEM = "Clock";
    const static char text_sk_A[]       PROGMEM = "7:58";
    const static char text_sk_B[]       PROGMEM = "11:58";
    const static char text_sk_C[]       PROGMEM = "18:30";
    const static char text_sk_D[]       PROGMEM = "set";

#elif  (LANGUAGE == ENGLISH)
    const static char text_uhrzeit[]     PROGMEM = "Time:";
    const static char text_faktor[]      PROGMEM = "Factor:";


    const static char text_clock[]      PROGMEM = "Clock";
    const static char text_sk_A[]       PROGMEM = "7:58";
    const static char text_sk_B[]       PROGMEM = "11:58";
    const static char text_sk_C[]       PROGMEM = "18:30";
    const static char text_sk_D[]       PROGMEM = "set";
#else
    #error LANGUAGE is void
#endif

#define UHR_ZEILE    3
#define UHR_SPALTE  12

static void display_uhr()
  {
    cursor_off();
    set_cursor(UHR_SPALTE,UHR_ZEILE);
    lcd_int2(my_clock.hour);
    lcd_putc(':');
    lcd_int02(my_clock.minute);
    set_cursor(UHR_SPALTE+3,UHR_ZEILE+2);
    lcd_int2(my_clock.ratio);    
  }

static void set_cursor_editfocus()
  {
    switch (edit_focus)
      {
        default:
        case 0: cursor_on_at(UHR_SPALTE,UHR_ZEILE,2); break;
        case 1: cursor_on_at(UHR_SPALTE+3,UHR_ZEILE,2); break;
        case 2: cursor_on_at(UHR_SPALTE+3,UHR_ZEILE+2,2); break;
      }
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_clock);
    check_status_display(0);

    set_cursor(3,UHR_ZEILE-1);
    lcd_string_P(text_uhrzeit);
    set_cursor(UHR_SPALTE,UHR_ZEILE-1);
    lcd_putc('h');
    lcd_putc('h');
    lcd_putc(':');
    lcd_putc('m');
    lcd_putc('m');
    set_cursor(3,UHR_ZEILE+2);
    lcd_string_P(text_faktor);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------


void action_clock(t_navicode par)
  {
    // unsigned char i;
    signed int stemp;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            display_time_allowed = 1;
            show_header();
            display_softkey_P(0, text_sk_A); 
            display_softkey_P(1, text_sk_B); 
            display_softkey_P(2, text_sk_C); 
            display_softkey_P(3, text_sk_D);
            turn_led_green_soft(3); 

            memcpy(&my_clock, &fast_clock, sizeof(fast_clock));

            display_uhr();

            set_cursor_editfocus();
            break;

        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:
                    edit_focus++;
                    if (edit_focus == 3) edit_focus = 0;
                    break;
                case KEY_SH_KNOB:
                    if (edit_focus == 0) edit_focus = 3;
                    edit_focus--;
                    break;
                case KEY_A:
                    my_clock.minute = 58;
                    my_clock.hour =  7;
                    turn_led_red_soft(3);
                    display_uhr();
                    break;
                case KEY_B:
                    my_clock.minute = 58;
                    my_clock.hour =  11;
                    turn_led_red_soft(3);
                    display_uhr();
                    break;
                case KEY_C:
                    my_clock.minute = 30;
                    my_clock.hour =  18;
                    turn_led_red_soft(3);
                    display_uhr();
                    break;
                case KEY_D:
                    turn_led_green_soft(3); 
                    msg_send_clock_set(&my_clock);
                    msg_transmit();
                    break;
                case KEY_SH_A:  // with shift: send again
                case KEY_SH_B:
                case KEY_SH_C:
                case KEY_SH_D:
                    break;
                case KEY_KNOB_TURNED:
                    switch (edit_focus)
                      {
                        case 0:
                            stemp = my_clock.hour;
                            stemp += get_knob_rotate();
                            while (stemp < 0) stemp += 24;
                            while (stemp >= 24) stemp -= 24;
                            my_clock.hour = stemp;
                            break;
                        case 1:
                            stemp = my_clock.minute;
                            stemp += get_knob_rotate();
                            while (stemp < 0)
                              {
                                stemp += 60;
                                if (my_clock.hour == 0) my_clock.hour = 24;
                                my_clock.hour--;
                              }
                            while (stemp >= 60) 
                              {
                                stemp -= 60;
                                my_clock.hour++;
                                if (my_clock.hour >= 24) my_clock.hour = 0;
                              }
                            my_clock.minute = stemp;
                            break;
                        case 2:
                            stemp = my_clock.ratio;
                            stemp += get_knob_rotate();
                            if (stemp < 1) stemp = 1;
                            if (stemp > 24) stemp = 24;
                            my_clock.ratio = stemp;
                            break;
                      }
                    display_uhr();
                    turn_led_red_soft(3);
                    break;
              }
            set_cursor_editfocus();
                    
            break;
        case MENU_CLOSE:
            break;
      }
  }
