//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_cs_cfg.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-02-22 V0.02 kw  added list for cvnames
//            2009-02-28 V0.03 kw  cursor after init, bugfix cvnames
//            2009-03-09 V0.04 kw  white blocks fixed
//            2009-03-31 V0.05 kw  wrong addr for softkey 1 fixed
//            2009-04-04 V0.06 kw  cursor_off before writing to display
//            2009-07-03 V0.07 kw  CV2 and CV26 added
//            2010-06-22 V0.08 kw  msg_transmit
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screen for configuration of the command station
//            (this is done with an extension to xpressnet)
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>


#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "lcd_ks0073_charset.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"


//   01234567890123456789
//  |--------------------|
// 0|Prog ( OK )    STOP |
// 1|CV      addr val=$$ |
// 2|cv-name 1024 255=FF |
// 3|Read Writ           |
//  |--------------------|

static const char text_cscfg[]  PROGMEM = "ConfOpenDCC";

static const char text_header[]    PROGMEM = {' ','V','a','r',' ',' ',' ',' ','a','d','d','r',' ','v','a','l','=',C_DOLLAR,C_DOLLAR,0};

static const char text_sk_read[]  PROGMEM = "Read";
static const char text_sk_write[]  PROGMEM = "Writ";


//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------


#if (DISPLAY == LCD_EA204)
  #define PROG_CUR_Y    1
#elif  (DISPLAY == LCD_EA128x64)
  #define PROG_CUR_Y    2
#else
  #warning DISPLAY not defined

#endif



//------------------------------------------------------------------------
// routines for menu_cs_cfg
//------------------------------------------------------------------------

extern void display_cv_value(void);  // from prog menu

/*
void display_cv_value(void)
  {
	if (current_CV.result >= PRG_OKAY)
      {
        set_cursor(13,2);
    	lcd_int3( current_CV.value);
        lcd_putc('=');
        lcd_hex( current_CV.value);
      }
    else
      {
        set_cursor(13,2);
    	lcd_putc_cnt(3,'?');
        lcd_putc('=');
        lcd_putc_cnt(2,'?');
      }  
  }
*/

#define ODCC_VAR_NAME_SIZE 7
#define ODCC_VAR_HELP_SIZE 22


struct s_ODCCName
  {
    unsigned int value;
    char name[ODCC_VAR_NAME_SIZE+1];
    char help[ODCC_VAR_HELP_SIZE+1];
  };

const struct s_ODCCName ODCCNames[] PROGMEM =
  {
    {   0,"Version","read only             "},
    {   1,"Baud   ","0=9600,1=19200,2=38400"},
    {   2,"Mode   ","rddf rrrx; f=fast clk "},
    {   3,"DMXlow ","acc.addr for dbase xmt"},
    {   4,"DMXhigh","acc.addr for dbase xmt"},
    {   7,"S88mode","0=off, 1=s88, 2=feedbk"},
    {   8,"S88sum ","total sum s88 bits / 8"},
    {   9,"S88cnt1","string 1 / left side  "},
    {  10,"S88cnt2","string 2 / middle     "},
    {  11,"S88cnt3","string 3 / right /@xp "},
    {  12,"InvAcc ","0=none, 1=lenz, 2=p50x"},
    {  13,"ReptAcc","repeat of acc commands"},
    {  15,"p50mode","0=mixed, 1=p50x only  "},
    {  16,"FBoffst","offset for feedback   "},
    {  17,"FBmode ","0=ack, 1=nack, 2=posit"},
    {  18,"Reptres","repeat of resets@prog "},
    {  19,"Reptprg","repeat of prog cmds   "},
    {  20,"Reptpom","repeat of pom commands"},
    {  21,"Reptspd","repeat of spd commands"},
    {  22,"Reptfct","repeat of fct commands"},
    {  24,"DCCdflt","0=DCC14,2=DCC28,3=126 "},
    {  25,"BiDi   ","0=off,1=cutout,3=id   "},
    {  26,"FastClk","0=off,1-31=multiply   "},
    {  29,"XPFBmap","0=512,1=all fb, 2=trnt"},
    {  30,"S88time","given in units of 10us"},
    {  31,"S88ctFB","virtual s88 bits      "},
    {  34,"Moff-T ","time to main off, *5ms"},
    {  35,"Poff-T ","time to porg off, *5ms"},
    {  36,"ExtOff ","0=off, 1=ext off enab."},
    {  37,"ExtStpT","ext off ignore time   "},
    {  39,"Ser.No.","serial number         "},
  };



void display_var_number( unsigned int varnumber)
  {
    unsigned char i, found=0;
	
    for (i=0; i < sizeof(ODCCNames) / sizeof(ODCCNames[0]); i++)
      {
        if ( varnumber == pgm_read_word(&ODCCNames[i].value) )
          {
            found = 1;
            set_cursor(1, PROG_CUR_Y+1);
            lcd_string_P(ODCCNames[i].name);
            set_cursor(1, PROG_CUR_Y+2);
            lcd_string_P(ODCCNames[i].help);
          }
      }
    if (!found)
      {
        set_cursor(1, PROG_CUR_Y+1);
        lcd_blank(ODCC_VAR_NAME_SIZE);   // strlen_P(cvNames[0].name)
        set_cursor(1, PROG_CUR_Y+2);
        lcd_blank(ODCC_VAR_HELP_SIZE);   // strlen_P(cvNames[0].help)
      }


	set_cursor(8, PROG_CUR_Y+1);
    lcd_int4( varnumber);
  }

//------------------------------------------------------------------------------

typedef enum 
  {
    EDIT_ADDR,
    EDIT_DATA,
    N_of_EDIT
  } t_edit_focus;

static t_edit_focus edit_focus;

static void set_edit_cursor(void)
  {
    switch(edit_focus)
      {
        default:
        case N_of_EDIT:       break;
        case EDIT_ADDR:
            cursor_on_at(8, PROG_CUR_Y+1,4); break;
        case EDIT_DATA:
            cursor_on_at(13,PROG_CUR_Y+1,3); break;            
      }
  }

static void move_edit_pointer_forward(void)
  {
    edit_focus++;
    if (edit_focus == N_of_EDIT) edit_focus = 0;
    set_edit_cursor();
  }

static void move_edit_pointer_revers(void)
  {
    if (edit_focus == 0) edit_focus = N_of_EDIT;
    edit_focus--;
    set_edit_cursor();
  }




void action_cs_cfg(t_navicode par)
  {
    signed int my_var;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            lcd_clear();
            lcd_string_P(text_cscfg);
            check_status_display(0);

            set_cursor(0,1);
            lcd_string_P(text_header);
            display_softkey_P(0,text_sk_read);
            display_softkey_P(1,text_sk_write);
            turn_led_green_soft(LED_A);
            turn_led_red_soft(LED_B);
            turn_led_off_soft(LED_C);
            turn_led_off_soft(LED_D);
            
            edit_focus = EDIT_ADDR;
            current_CV.result = PRG_CV_VOID;           
            display_cv_value();
            display_var_number(current_CV.number);
            msg_send_cs_cfg_read(current_CV.number);  // und gleich anfragen
            msg_transmit();
            set_edit_cursor();
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:                      // change focus forward
                    move_edit_pointer_forward();
                    break;
                case KEY_SH_KNOB:                   // change focus revers
                    move_edit_pointer_revers();
                    break;
                case KEY_KNOB_TURNED:               // cvValue
                    cursor_off();
                    switch(edit_focus)
                      {
                        case EDIT_ADDR:
                            current_CV.result = PRG_CV_VOID;
                            my_var = current_CV.number;
                            my_var += get_knob_rotate();
                            if (my_var < 0) my_var = 0;
                            if (my_var > 1024) my_var = 1024;
                            current_CV.number = (unsigned int) my_var;
                            display_var_number(current_CV.number);
                            display_cv_value();
                            msg_send_cs_cfg_read(current_CV.number);  // und gleich anfragen
                            msg_transmit();
                            break;
                        case EDIT_DATA:
                            current_CV.result = PRG_CV_CHANGED;
                            my_var = current_CV.value;
                            my_var += get_knob_rotate();
                            if (my_var < 0) my_var = 0;
                            if (my_var > 255) my_var = 255;
                            current_CV.value = (unsigned char) my_var;
                            display_cv_value();
                            break;
                        case N_of_EDIT:
                            break;
                      }
                    set_edit_cursor();                    
                    break;
                case KEY_UP:                        // nextCV
                    cursor_off();
                    if (current_CV.number < 1024) 
                      {
                        current_CV.number++;
                        current_CV.result = PRG_CV_VOID;
                      }
                    display_var_number(current_CV.number);
                    display_cv_value();
                    msg_send_cs_cfg_read(current_CV.number);  // und gleich anfragen
                    msg_transmit();
                    set_edit_cursor();
                    break;
                case KEY_DOWN:                      // prevCV
                    cursor_off();
                    if (current_CV.number > 1)
                      {
                        current_CV.number--;
                        current_CV.result = PRG_CV_VOID;
                      }
                    display_var_number(current_CV.number);
                    display_cv_value();
                    msg_send_cs_cfg_read(current_CV.number);  // und gleich anfragen
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_A:                    // CV read
                    msg_send_cs_cfg_read(current_CV.number);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_B:                    // CV write
                    msg_send_cs_cfg_write(current_CV.number, current_CV.value);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_C:
                    break;
                case KEY_D:
                    break;
                case KEY_SH_C:
                    break;
                case KEY_SH_D:
                    break;
                case KEY_PARSER_CFG:          // CFG result received
                    cursor_off();
                    display_cv_value();
                    set_edit_cursor();                    
                    break;
              }
            break;
        case MENU_CLOSE:
            cursor_off();
            turn_led_off_soft(LED_A);
            turn_led_off_soft(LED_B);
            break;
      }
  }
