//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_dreh.c
// author:    Hans Kurscheidt
// contact:   none
// history:   2013-10-11 V0.01 kw  ported to mft
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for turnout operation
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"

#if (DREHSCHEIBE == 1)

//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------
// 
#define DEC_SVEN_BRANDT             1                   // 1 Sven Brandt
#define DEC_RAILWAY_LAUF            2                   // 2  Railway Lauf
#define DREHSCHEIBEN_DECODER		DEC_SVEN_BRANDT		

// Operation Codes
// 0..48 turn
#define DREH_OP_LICHT				80							
#define DREH_OP_L180				81		
#define DREH_OP_R180				82
#define DREH_OP_TURN_SENSE_LEFT		83
#define	DREH_OP_TURN_SENSE_RIGHT	84
#define DREH_OP_TURN75_RIGHT		85
#define DREH_OP_TURN75_LEFT			86


#ifndef DREHSCHEIBEN_DECODER
	#warning: "No Target-Decoder defined"
#endif

#if (DREHSCHEIBEN_DECODER == DEC_SVEN_BRANDT)
    // adr offsets

    #define DEC_LICHT	0			
    #define TURN180		1
    #define TURN_SENSE	3
    #define POS1		4
    #define TURN75		2
#else
    #error DREHSCHEIBEN_DECODER wrong
#endif
   
// 48 Graphik-Positionen fr Gleisabgnge
const uint8_t __attribute__ ((progmem)) dreh_posis[][2] = {
	{ 0,32}, { 1,28}, { 1,23}, { 2,20}, { 4,17}, { 6,15}, { 8,13}, {10,11},
	{13,10}, {16, 9}, {19, 9}, {23, 9}, {27, 9}, {30, 9}, {33, 9}, {36, 9},
	{39,10}, {42,11}, {46,13}, {48,15}, {50,17}, {51,20}, {52,23}, {53,27},
	{54,32}, {54,36}, {53,41}, {51,43}, {49,46}, {47,48}, {45,50}, {43,52},
	{40,53}, {38,55}, {35,55}, {32,55}, {27,55}, {24,55}, {21,55}, {18,55}, 
	{15,54}, {12,53}, { 8,51}, { 7,48}, { 5,46}, { 3,44}, { 2,42}, { 1,38}};


static uint8_t dreh_pos = 0;				// 0 = unknown, otherwise 1..48
#warning die folgenden Definitionen sollten ins PROGMEM

static unsigned char dreh_headline[] = "Drehscheibe";
static unsigned char dreh_softk[] = "POS1 |L180 |R180 | LI";
static unsigned char dreh_softk_shift[] = "L7,5 | R7,5| <  |  > ";
static unsigned char dreh_knob_pos[] = "Position";
static uint8_t dreh_licht; uint8_t dreh_sho_acc_cmd;
static uint16_t dreh_acc_address;			// drehscheiben accessory DECODER address local copy
static uint8_t dreh_new_knob_posi;

/* Hier werden die x,y-Paare der bestckten Gleisabgnge gem EEMEM bitmap angelegt. Diese dienen dem Vektor, der die 
"Drehscheibenbrcke darstellt als Anfangs- und Endpunkte und zur Zeichnung der Gleisabgnge in der DS-Grafik. Der INDEX 
luft von 0 und entspricht der Dekoderposition. Man unterscheide die Postionen der Gleisabgnge auf der DS (1..48) und 
die Positionen der bestckten Gleisabgnge!! z.B. Positionen 1(immer!), 3,5,10,12,13 entsprechend Dekoderpositionen 1,2,3,4,5,6!
Deshalb wird zur Vektorausgabe die Gleisposition ausgegeben und zum Anfahren der Position der Index = Dekoderposition.
*/

static uint8_t dreh_stubs[50];					// 0 = Delimiter nach letzter bestckter Position
static signed int dreh_knob_accu;				// Index in dreh_stubs[50];


//------------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   012345678901234567890
//  |--------------------|
// 0|Clock          ON   |
// 1| Uhrzeit:  hh:mm    |
// 2| Faktor:      12    |
// 3| 8:00           set |
//  |--------------------|
//

#error NO turnable menu on char display; Set define DREHSCHEIBE to 0
//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)


//   0123456789012345678901234
//  |------------------------|
// 0|Clock              ON   |
// 1|                        |
// 2| Uhrzeit:   HH:MM       |
// 3|            12:34       |
// 4| Faktor:        8       |
// 5|  7:58 11:58 18:30  set |
//  |------------------------|
//

#include "bmp/id_dreh_glcd.h"     // Drehscheibe OHNE Gleisabgnge

/*
static void display_uhr()
  {
    cursor_off();
    set_cursor(UHR_SPALTE,UHR_ZEILE);
    lcd_int2(my_clock.hour);
    lcd_putc(':');
    lcd_int02(my_clock.minute);
    set_cursor(UHR_SPALTE+3,UHR_ZEILE+2);
    lcd_int2(my_clock.ratio);    
  }

static void set_cursor_editfocus()
  {
    switch (edit_focus)
      {
        default:
        case 0: cursor_on_at(UHR_SPALTE,UHR_ZEILE,2); break;
        case 1: cursor_on_at(UHR_SPALTE+3,UHR_ZEILE,2); break;
        case 2: cursor_on_at(UHR_SPALTE+3,UHR_ZEILE+2,2); break;
      }
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_clock);
    check_status_display(0);

    set_cursor(3,UHR_ZEILE-1);
    lcd_string_P(text_uhrzeit);
    set_cursor(UHR_SPALTE,UHR_ZEILE-1);
    lcd_putc('h');
    lcd_putc('h');
    lcd_putc(':');
    lcd_putc('m');
    lcd_putc('m');
    set_cursor(3,UHR_ZEILE+2);
    lcd_string_P(text_faktor);
  }
*/
//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------

//-------------------------------------------------------------------------------
/* signum function */
int sgn(int x){
	return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}

void dreh_vect (int xstart, int ystart, int xend, int yend, uint8_t value)
/*--------------------------------------------------------------
 * Bresenham-Algorithmus: Linien auf Rastergerten zeichnen
 *
 * Eingabeparameter:
 *    int xstart, ystart        = Koordinaten des Startpunkts
 *    int xend, yend            = Koordinaten des Endpunkts
 *
 * Ausgabe:
 *    void SetPixel(int x, int y) setze ein Pixel in der Grafik
 *         (wird in dieser oder aehnlicher Form vorausgesetzt)
 *---------------------------------------------------------------
 */
{
    int x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;
 
/* Entfernung in beiden Dimensionen berechnen */
   dx = xend - xstart;
   dy = yend - ystart;
 
/* Vorzeichen des Inkrements bestimmen */
   incx = sgn(dx);
   incy = sgn(dy);
   if(dx<0) dx = -dx;
   if(dy<0) dy = -dy;
 
/* feststellen, welche Entfernung grer ist */
   if (dx>dy)
   {
      /* x ist schnelle Richtung */
      pdx=incx; pdy=0;    /* pd. ist Parallelschritt */
      ddx=incx; ddy=incy; /* dd. ist Diagonalschritt */
      es =dy;   el =dx;   /* Fehlerschritte schnell, langsam */
   } else
   {
      /* y ist schnelle Richtung */
      pdx=0;    pdy=incy; /* pd. ist Parallelschritt */
      ddx=incx; ddy=incy; /* dd. ist Diagonalschritt */
      es =dx;   el =dy;   /* Fehlerschritte schnell, langsam */
   }
 /* Initialisierungen vor Schleifenbeginn */
 x = xstart;
 y = ystart;
 err = el/2;
 if (value) LCD_pixelOn(x,y); else LCD_pixelOff(x,y);
 
 /* Pixel berechnen */
 for(t=0; t<el; ++t) /* t zaehlt die Pixel, el ist auch Anzahl */
 {
	 /* Aktualisierung Fehlerterm */
	 err -= es;
	 if(err<0)
	 {
		 /* Fehlerterm wieder positiv (>=0) machen */
		 err += el;
		 /* Schritt in langsame Richtung, Diagonalschritt */
		 x += ddx;
		 y += ddy;
	 } else
	 {
		 /* Schritt in schnelle Richtung, Parallelschritt */
		 x += pdx;
		 y += pdy;
	 }
	 if (value) LCD_pixelOn(x,y); else LCD_pixelOff(x,y);
 }
} /* gbham() */


void dreh_disp_raw (void)		// Zeichne Drehscheibe OHNE Gleisabgnge
{	
	cursor_off();		
	lcd_clear();
	LCD_setCursorXY(0,0);
	LCD_puts(dreh_headline);

	display_status(command_station_status);
	display_batt_state();
	display_connect_state (get_connection_status());
	
	LCD_setCursorXY(0,56);	
	LCD_puts(dreh_softk);
		
	LCD_setCursorXY(0,8);
    LCD_drawBMP_aligned ( (PGM_VOID_P) id_dreh_glcd_bmp, 64, 6);
}

void dreh_declutter(void)
{
uint8_t i;

	for (i=0; i<14; i++)			// de-clutter inner circle
		{
		dreh_vect (21-i, 18+i, 33+i, 18+i, 0);
		dreh_vect (21-i, 45-i, 33+i, 45-i, 0);
		}
}
void dreh_disp_stubs(void)	// Zeichne Gleisabgnge, wenn POS gesetzt/bekannt
{
	uint8_t i = 0, index;
	
	while (dreh_stubs[i] !=0)
	{
		index = dreh_stubs[i++]-1;
		dreh_vect (27, 32, pgm_read_byte (&dreh_posis[index][0]), pgm_read_byte (&dreh_posis[index][1]),1);
		dreh_declutter();
	}
}

void dreh_send_acc_cmd (uint8_t dreh_acc_cmd)
  {
    char dreh_r0[4];
    uint16_t tmp;

	switch (dreh_acc_cmd)
	  {
		case 0 ... 48:
            tmp = (dreh_acc_address + POS1 + dreh_acc_cmd / 2) * 2 + (dreh_acc_cmd & 1) ;
    		msg_send_accessory(tmp,1);
    		if (dreh_sho_acc_cmd)
    	      {
    			sprintf( dreh_r0, "%3.0d", tmp/2);
    			LCD_setCursorXY(72,40);
    			LCD_number_6px(dreh_r0);
    	      }
            break;
        case DREH_OP_LICHT:
            dreh_licht ^= 1;
            msg_send_accessory((dreh_acc_address + DEC_LICHT) *2 + dreh_licht, 1 );
		    break;
		case DREH_OP_R180:
            msg_send_accessory((dreh_acc_address + TURN180) *2 + 0, 1 );
            break;
		case DREH_OP_L180:
            msg_send_accessory((dreh_acc_address + TURN180) *2 + 1, 1 );
            break;
		case DREH_OP_TURN_SENSE_LEFT:
            msg_send_accessory((dreh_acc_address + TURN_SENSE) *2 + 1, 1 );
            break;
		case DREH_OP_TURN_SENSE_RIGHT:
            msg_send_accessory((dreh_acc_address + TURN_SENSE) *2 + 0, 1 );
            break;
    	case DREH_OP_TURN75_LEFT:
            msg_send_accessory((dreh_acc_address + TURN75) *2 + 1, 1 );
            break;
    	case DREH_OP_TURN75_RIGHT:
            msg_send_accessory((dreh_acc_address + TURN75) *2 + 0, 1 );
            break;			
		default: break;
     }
  }

void dreh_disp_bridge(uint8_t posi)
{
	posi--;			// posi 1 is index [0]
	if (posi <=23)
		dreh_vect (pgm_read_byte (&dreh_posis[posi][0]), pgm_read_byte (&dreh_posis[posi][1]), pgm_read_byte (&dreh_posis[posi+24][0]), pgm_read_byte (&dreh_posis[posi+24][1]),1);
	else
		dreh_vect (pgm_read_byte (&dreh_posis[posi][0]), pgm_read_byte (&dreh_posis[posi][1]), pgm_read_byte (&dreh_posis[posi-24][0]), pgm_read_byte (&dreh_posis[posi-24][1]),1);
	
}

void dreh_copy_stubs (void)
{
uint8_t i,j,k,x, mask;

	k=0;
	for (i=0; i<6; i++)
	{
		x = eeprom_read_byte ((void *)eadr_dreh_posis_bit0 + i);    // ziemlich dreckig, dieser Zugriff
		mask = 1;
		for (j=1; j<=8; j++)
		{
			if  ( x & mask)
			{
				dreh_stubs[k++] = j+i*8;
			}
		mask <<=1;
		}
	}
	dreh_stubs[k] = 0;				// delimiter
}

void dreh_set_pos(uint8_t posi)
{
	char dreh_r0[3];
	
	dreh_disp_raw();
	LCD_setCursorXY(64,24);
	LCD_puts(dreh_knob_pos);
	dreh_disp_stubs();
	dreh_disp_bridge(posi);						// send POSI  sh. oben
	sprintf( dreh_r0, "%2.0d", posi);
	LCD_setCursorXY(72,32);
	LCD_number_6px(dreh_r0);
	
    #if (DREHSCHEIBEN_DECODER == DEC_SVEN_BRANDT)
	    dreh_send_acc_cmd((dreh_knob_accu %24));	// send INDEX	sh. oben
    #else
	    dreh_send_acc_cmd(dreh_knob_accu);
    #endif
}

//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------


void action_dreh(t_navicode par)
  {
    char tmp_dreh_r0[3];

    check_status_display(par.key);

	switch (par.menucode)
	{
		case MENU_OPEN:
        
			dreh_acc_address = eeprom_read_word ( ( void *) eadr_dreh_addr_l) - 1 ; // -1, xp_addressing
			dreh_sho_acc_cmd = eeprom_read_byte ( ( void *) eadr_dreh_sho_acc_cmd) ;
			
			dreh_disp_raw();
			dreh_copy_stubs();
										// hier knnte man die Posi vom Dekoder auslesen, falls untersttzt.
			if (dreh_pos)	
			  {
				dreh_disp_stubs();		// >0 = we know our position
				dreh_set_pos(dreh_pos);
				turn_led_green_soft(LED_A);
			  }			
			else blink_led_red_soft(LED_A);
            turn_led_off(LED_B); turn_led_off(LED_C); turn_led_off(LED_D);  // all other leds of
            
            #warning display_softkey_P
            /* display_softkey_P(0, text_sk_A); 
            display_softkey_P(1, text_sk_B); 
            display_softkey_P(2, text_sk_C); 
            display_softkey_P(3, text_sk_D);
            */
            break;

        case MENU_UPDATE:
			switch (par.key)
			  {		
				case KEY_SHIFT:
					LCD_setCursorXY(0,56);
					LCD_puts(dreh_softk_shift);
					break;
					
				case KEY_SHIFT_RELEASE:
					LCD_setCursorXY(0,56);
					LCD_puts(dreh_softk);
					break;
					
				case KEY_SH_A:
					if (!dreh_pos)	break;
					else dreh_send_acc_cmd(DREH_OP_TURN75_LEFT);
				    break;
										// read Pos from decoder
				case KEY_SH_D:						// set dir LEFT
					if (!dreh_pos)	break;
					else dreh_send_acc_cmd(DREH_OP_TURN_SENSE_LEFT);
				    break;
										
				case KEY_SH_C:						// set dir RIGHT
					if (!dreh_pos)	break;
					else dreh_send_acc_cmd(DREH_OP_TURN_SENSE_RIGHT);
				    break;
										
				case KEY_SH_B:
					if (!dreh_pos)	break;
					else dreh_send_acc_cmd(DREH_OP_TURN75_RIGHT);
					break;
	
				case KEY_A:							// goto POS1
					turn_led_green_soft(LED_A);
					dreh_pos=1; dreh_knob_accu=0;
					dreh_set_pos(1);
					break;
		
				case KEY_B:							// turn left 180
					if (!dreh_pos) break;
					else
				      {
						dreh_send_acc_cmd(DREH_OP_L180);
					  }
				    break;
	
				case KEY_C:							// turn right 180
					if (!dreh_pos)	break;
					else
					  {
					    dreh_send_acc_cmd(DREH_OP_R180);
					  }
				    break;
				case KEY_D:							// house light on		
					dreh_send_acc_cmd(DREH_OP_LICHT);
				    break;
	
				case KEY_KNOB_TURNED:
					if (!dreh_pos) break;
					else
					  {
						dreh_declutter();
						dreh_knob_accu += get_knob_rotate();
						if (dreh_knob_accu <0)	dreh_knob_accu = 0;
						if (dreh_stubs[dreh_knob_accu] == 0) dreh_knob_accu = 0;			// EoL reached
						dreh_new_knob_posi = dreh_stubs[dreh_knob_accu];
						dreh_disp_bridge(dreh_new_knob_posi);
						sprintf( tmp_dreh_r0, "%2.0d",dreh_new_knob_posi );
						#warning das sollte ber eine Displayroutine gehen, nicht knallhart hinten in die Graphik rein.
                        LCD_setCursorXY(72,32);
						LCD_number_6px(tmp_dreh_r0);
						if (dreh_new_knob_posi > dreh_pos) 
                               dreh_send_acc_cmd(DREH_OP_TURN_SENSE_RIGHT);
                        else   dreh_send_acc_cmd(DREH_OP_TURN_SENSE_LEFT);
					  }
				    break;	
				case KEY_KNOB:
					if (!dreh_pos) break;
						else
						{					
							dreh_pos = dreh_new_knob_posi;
							dreh_set_pos(dreh_pos);
						}
				    break;
				
				case KEY_CENTER:
				case KEY_DOWN:    
				case KEY_UP:          
				    break;
				default: break;
			  }
		    break;				
		case MENU_CLOSE:
		    break;
		
		default: break;
	}	
}
	
#endif // #if (DREHSCHEIBE == 1)
