//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_extacc.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2010-06-12 V0.01 kw  copied from weiche
//            2010-06-22 V0.02 kw  msg_transmit
//            2014-08-17 V0.03 kw  added english menu text
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for extended accessory
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"

//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------
//
// we use the array actual_extacc from messages_objects.c
//
      
static unsigned char edit_focus;   // da steht der Cursor gerade
static unsigned char edit_signal;  // dieses Signal wird aktuell bearbeitet

//------------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   012345678901234567890
//  |--------------------|
// 0|Signal         ON   |
// 1| Adresse   Begriff  |
// 2|    1234        31  |
// 3| memA memB memC memD|
//  |--------------------|
//

#if (LANGUAGE == GERMAN)
    const static char text_adresse[]     PROGMEM = "Adresse";
    const static char text_begriff[]     PROGMEM = "Begriff";
#elif  (LANGUAGE == ENGLISH)
    const static char text_adresse[]     PROGMEM = "Address";
    const static char text_begriff[]     PROGMEM = "Aspect";
#else
    #error LANGUAGE is void
#endif

const static char text_extacc[]     PROGMEM = "Signal";
const static char text_sk_A[]       PROGMEM = "memA";
const static char text_sk_B[]       PROGMEM = "memB";
const static char text_sk_C[]       PROGMEM = "memC";
const static char text_sk_D[]       PROGMEM = "memD";

#define SADDR_ZEILE  2

static void display_signal_state(unsigned char signal)
  {
    cursor_off();
    set_cursor(4,SADDR_ZEILE);
    lcd_int4(actual_signal[signal].addr);
    set_cursor(16,SADDR_ZEILE);
    lcd_int2(actual_signal[signal].state);
  }

static void set_cursor_editfocus()
  {
    if (edit_focus) cursor_on_at(16,SADDR_ZEILE,2);
    else cursor_on_at(4,SADDR_ZEILE,4);
  }


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_extacc);
    check_status_display(0);

    set_cursor(1,SADDR_ZEILE-1);
    lcd_string_P(text_adresse);
    set_cursor(11,SADDR_ZEILE-1);
    lcd_string_P(text_begriff);
  }


//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)


//   0123456789012345678901234
//  |------------------------|
// 0|Signal             ON   |
// 1|                        |
// 2| Adresse     Begriff    |
// 3|    1234          31    |
// 4|                        |
// 5| memA  memB  memC  memD |
//  |------------------------|
//

#if (LANGUAGE == GERMAN)
    const static char text_adresse[]     PROGMEM = "Adresse";
    const static char text_begriff[]     PROGMEM = "Begriff";
#elif  (LANGUAGE == ENGLISH)
    const static char text_adresse[]     PROGMEM = "Address";
    const static char text_begriff[]     PROGMEM = "Aspect";
#else
    #error LANGUAGE is void
#endif

const static char text_extacc[]     PROGMEM = "Signal";
const static char text_sk_A[]       PROGMEM = "memA";
const static char text_sk_B[]       PROGMEM = "memB";
const static char text_sk_C[]       PROGMEM = "memC";
const static char text_sk_D[]       PROGMEM = "memD";

#define SADDR_ZEILE  5

static void display_signal_state(unsigned char signal)
  {
    cursor_off();
    set_cursor(4,SADDR_ZEILE);
    lcd_int4(actual_signal[signal].addr);
    set_cursor(18,SADDR_ZEILE);
    lcd_int2(actual_signal[signal].state);
  }

static void set_cursor_editfocus()
  {
    if (edit_focus) cursor_on_at(17,SADDR_ZEILE,3);
    else cursor_on_at(4,SADDR_ZEILE,4);
  }


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_extacc);
    check_status_display(0);

    set_cursor(1,SADDR_ZEILE-2);
    lcd_string_P(text_adresse);
    set_cursor(13,SADDR_ZEILE-2);
    lcd_string_P(text_begriff);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------


void action_extacc(t_navicode par)
  {
    // unsigned char i;
    signed int stemp;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            display_time_allowed = 1;
            show_header();
            display_softkey_P(0, text_sk_A); 
            display_softkey_P(1, text_sk_B); 
            display_softkey_P(2, text_sk_C); 
            display_softkey_P(3, text_sk_D);
            turn_led_green_soft(edit_signal); 

            read_actual_saddr();

            display_signal_state(edit_signal);

            // until now, there is no query
            // for (i=0; i<4; i++)
            //  {
            //    // msg_send_accessory_query(actual_signal[i].addr);  // signal addr
            //  }
            // msg_transmit();

            
            set_cursor_editfocus();
            break;

        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:
                    if (edit_focus)
                      {
                        msg_send_extended_accessory(actual_signal[edit_signal].addr, actual_signal[edit_signal].state);
                        msg_transmit();
                      }
                    else edit_focus = 1;
                    break;
                case KEY_SH_KNOB:
                    edit_focus = 0;
                    break;
                case KEY_A:
                case KEY_B:
                case KEY_C:
                case KEY_D:
                    turn_led_off(edit_signal); 
                    edit_signal = par.key & 0x03;
                    turn_led_green_soft(edit_signal); 
                    display_signal_state(edit_signal);
                    edit_focus = 0;
                    // switch_turnout(actual_weiche[i].addr, actual_weiche[i].state);
                    break;
                case KEY_SH_A:  // with shift: send again
                case KEY_SH_B:
                case KEY_SH_C:
                case KEY_SH_D:
                    turn_led_off(edit_signal); 
                    edit_signal = par.key & 0x03;
                    turn_led_green_soft(edit_signal); 
                    display_signal_state(edit_signal);
                    edit_focus = 0;
                    // switch_turnout(actual_weiche[i].addr, actual_weiche[i].state);
                    break;
                case KEY_KNOB_TURNED:
                    if (edit_focus)        // aspect
                      {
                        stemp = actual_signal[edit_signal].state;
                        stemp += get_knob_rotate();
                        if (stemp < 0) stemp = 0;
                        if (stemp > 31) stemp = 31;
                        actual_signal[edit_signal].state = stemp;
                      }
                    else
                      {
                        stemp = actual_signal[edit_signal].addr;
                        stemp += get_knob_rotate();
                        if (stemp < 1) stemp = 1;
                        if (stemp > 2040) stemp = 2040;
                        actual_signal[edit_signal].addr = stemp;
                        actual_signal[edit_signal].state = 0;   // void
                      }
                    display_signal_state(edit_signal);
                    //!! msg_send_accessory_query(actual_weiche[edit_focus].addr);
                    //!! msg_transmit();
                    break;
                case KEY_PARSER_SIGNAL:
                    display_signal_state(edit_signal);
                    break;
              }
            set_cursor_editfocus();
                    
            break;
        case MENU_CLOSE:
            write_actual_saddr();
            break;
      }
  }
