//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_feedback.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-09-24 V0.02 kw  conditional assembly for displays
//            2010-06-22 V0.03 kw  msg_transmit
//            2010-08-03 V0.04 kw  query at MENU_OPEN
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screen for feedback display
//
//------------------------------------------------------------------------
// see menu_navigator on how to handle menus
//

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_ks0073_charset.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"



//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------

static unsigned char fb_focus;




//-----------------------------------------------------------------
// tool routines for menu feedback
//-----------------------------------------------------------------

void read_actual_faddr(void)
  {
    t_data16 temp;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_faddr0_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_faddr0_h);
    fb_addr[0] = temp.as_int16;
    temp.as_uint8[0]= eeprom_read_byte((void *)eadr_faddr1_l);
    temp.as_uint8[1]= eeprom_read_byte((void *)eadr_faddr1_h);
    fb_addr[1] = temp.as_int16;
  }

void write_actual_faddr(void)
  {
    t_data16 temp;

    temp.as_int16 = fb_addr[0];
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_faddr0_l))
        eeprom_write_byte((void *)eadr_faddr0_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_faddr0_h))
        eeprom_write_byte((void *)eadr_faddr0_h, temp.as_uint8[1]);

    temp.as_int16 = fb_addr[1];
    if (temp.as_uint8[0] != eeprom_read_byte((void *)eadr_faddr1_l))
        eeprom_write_byte((void *)eadr_faddr1_l, temp.as_uint8[0]);
    if (temp.as_uint8[1] != eeprom_read_byte((void *)eadr_faddr1_h))
        eeprom_write_byte((void *)eadr_faddr1_h, temp.as_uint8[1]);

  }

//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   01234567890123456789
//  |--------------------|
// 0|Feedback       STOP |
// 1|Addr: 123  Addr: 345|
// 2|7.......0          .|
// 3|..X. XX..  ..XX ..X.|
//  |--------------------|


char text_feedback[]  PROGMEM = "Feedback";
char text_faddr[]  PROGMEM = "Addr:";
char text_legend[] PROGMEM = {C_NUM_EXP_7, 
                              C_NUM_EXP_6,
                              C_NUM_EXP_5,
                              C_NUM_EXP_4, ' ',
                              C_NUM_EXP_3,
                              C_NUM_EXP_2,
                              C_NUM_EXP_1,
                              C_NUM_EXP_0, 0 };

void display_array(unsigned char data)
  {
    unsigned char i, mask;
    
    mask = 0x80;

    for (i=0; i< 8; i++)
      {
        if (i==4) lcd_putc(' ');
        lcd_bool(data & mask);
        mask >>= 1;
      }
  }

void display_fb_array(unsigned char my_focus)
  {
    if (my_focus == 0)
      {
        set_cursor(5,1);
        lcd_int4(fb_addr[0]);
        set_cursor(0,3);
        display_array(fb_array[fb_addr[0]]);
      }
    else
      {
        set_cursor(16,1);
        lcd_int4(fb_addr[1]);
        set_cursor(11,3);
        display_array(fb_array[fb_addr[1]]);
      }
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_feedback);
    check_status_display(0);

    set_cursor(0,1);
    lcd_string_P(text_faddr);
    
    set_cursor(11,1);
    lcd_string_P(text_faddr);

    set_cursor(0,2);
    lcd_string_P(text_legend);

    set_cursor(11,2);
    lcd_string_P(text_legend);
  }

static void set_cursor_to_focus(void)
  {
    cursor_on_at(5 + 11*fb_focus, 1, 4);
  }
//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

//   01234567890123456789
//  |--------------------|
// 0|Feedback       STOP |
// 1|Addr: 123  Addr: 345|
// 2|7.......0          .|
// 3|..X. XX..  ..XX ..X.|
//  |--------------------|

#define FEEDB_CUR_Y  2

const char text_feedback[]  PROGMEM = "Feedback";
const char text_faddr[]  PROGMEM = "Addr:";
const char text_legend[] PROGMEM = {C_NUM_EXP_7, 
                              C_NUM_EXP_6,
                              C_NUM_EXP_5,
                              C_NUM_EXP_4, ' ',
                              C_NUM_EXP_3,
                              C_NUM_EXP_2,
                              C_NUM_EXP_1,
                              C_NUM_EXP_0, 0 };

void display_array(unsigned char data)
  {
    unsigned char i, mask;
    
    mask = 0x80;

    for (i=0; i< 8; i++)
      {
        if (i==4) lcd_putc(' ');
        lcd_bool(data & mask);
        mask >>= 1;
      }
  }

void display_fb_array(unsigned char my_focus)
  {
    if (my_focus == 0)
      {
        set_cursor(5,FEEDB_CUR_Y);
        lcd_int4(fb_addr[0]);
        set_cursor(0,FEEDB_CUR_Y+2);
        display_array(fb_array[fb_addr[0]]);
      }
    else
      {
        set_cursor(16,FEEDB_CUR_Y);
        lcd_int4(fb_addr[1]);
        set_cursor(11,FEEDB_CUR_Y+2);
        display_array(fb_array[fb_addr[1]]);
      }
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_feedback);
    check_status_display(0);

    set_cursor(0,FEEDB_CUR_Y);
    lcd_string_P(text_faddr);
    
    set_cursor(11,FEEDB_CUR_Y);
    lcd_string_P(text_faddr);

    set_cursor(0,FEEDB_CUR_Y+1);
    lcd_string_P(text_legend);

    set_cursor(11,FEEDB_CUR_Y+1);
    lcd_string_P(text_legend);
  }

static void set_cursor_to_focus(void)
  {
    cursor_on_at(5 + 11*fb_focus, FEEDB_CUR_Y, 4);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------

void action_feedback(t_navicode par)
  {
    check_status_display(par.key);

    switch (par.menucode) {
        case MENU_OPEN:
            show_header();

            read_actual_faddr();

            msg_send_feedback_query(fb_addr[fb_focus]);   // with xp we trace it always, but with xbee we have
            msg_transmit();                               // to send a query (and it won't harm on xpressnet)
            
            display_fb_array(1);
            display_fb_array(0);
            fb_focus = 0;
            set_cursor_to_focus();
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_PARSER_FB:
                    display_fb_array(0);
                    display_fb_array(1);
                    set_cursor_to_focus();
                    break;
                case KEY_KNOB:
                    if (fb_focus != 0) fb_focus = 0;
                    else fb_focus = 1;
                    set_cursor_to_focus();
                    break;
                case KEY_KNOB_TURNED:
                    fb_addr[fb_focus] += get_knob_rotate();
                    if (fb_addr[fb_focus] < 0)    fb_addr[fb_focus] = 0;
                    if (fb_addr[fb_focus] > FB_ARRAY_SIZE - 1)  fb_addr[fb_focus] = FB_ARRAY_SIZE - 1;
                    msg_send_feedback_query(fb_addr[fb_focus]);
                    msg_transmit();

                    display_fb_array(fb_focus);
                    set_cursor_to_focus();
                    break;
              }
            break;
        case MENU_CLOSE:
            break;
      }
  }
