//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_lok.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-02-09 V0.02 kw  make code modular to support
//                                 different displays
//            2009-03-02 V0.03 kw  added sending of functions and light
//            2009-03-07 V0.04 kw  added LED control
//            2009-03-14 V0.05 jmf if dir changed, send message
//                             kw  added lok cache; if KEY_DOWN first walk through cache
//            2009-03-18 V0.04 kw  new knob_mode
//            2009-03-31 V0.05 kw  added display_lok_picture
//            2009-04-04 V0.06 kw  special routines for ADDR display
//            2009-04-07 V0.07 kw  knob push neu dazu
//            2009-05-06 V0.08 kw  dead zone at knob zero gets extended if knob is rotating.
//            2010-06-22 V0.09 kw  msg_transmit();
//            2010-06-30 V0.10 kw  .other cleared at entry and new lok, functions now static
//            2014-08-17 V0.11 kw  added english texts
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for lok
//            see menu_navigator.c and .h on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_ks0073_charset.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"
#include "lok_graphic.h"

t_locoentry actual_lok;                 // a copy of the dataset of
                                        // currently addressed lok
                                        // This is tyically identical to
                                        // content of the eeprom at
                                        // actual_database_index, but could differ 

static signed int speed_0_holdtime;     // for dual sided knob, we need to remember
                                        // the time when we arrived a speed o to
                                        // create a 'gap' in knob oberation
                                        // Note: the compare has a wrap around
                                        // after 10min; wrong reaction for 700ms
                                        // (It's a feature, not a bug)

#define KNOB_AT_ZERO_TIME  1000000L     // ignore timeout in DC mode

//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------
//
#if (DISPLAY == LCD_EA204)
//
//  Routines for a 4 x 20 character text display
//
//
//   01234567890123456789
//  |--------------------|
// 0| Lok           STOP |
// 1|12345:BBII    =#= * |
// 2|<<<<<<<:126:>>>>>>> |
// 3|F01* F02  F03  F04* |
//  |--------------------|

#if (LANGUAGE == GERMAN)
    static char text_lok[]  PROGMEM = "Lok  ";
#elif  (LANGUAGE == ENGLISH)
    static char text_lok[]  PROGMEM = "Loc  ";
#else
    #error LANGUAGE is void
#endif

static unsigned char old_display_val = 0;
static unsigned char old_display_dir = 0;

static unsigned char old_other = 0;

static void display_loc_speed_cleanup(void)
  {
    old_display_val = 0;
    old_display_dir = 0;
    old_other = 0;
  }

void display_loc_speed(t_format format, unsigned char speed, unsigned char dir)
  {
    unsigned char i;

    if (dir)
      {
        set_cursor(7,2);
        lcd_putc('<');
        lcd_int3(speed);
        lcd_putc(C_SQ_BRAQ_CLOSE);
      }
    else
      {
        set_cursor(7,2);
        lcd_putc(C_SQ_BRAQ_OPEN);
        lcd_int3(speed);
        lcd_putc('>');
      }

    unsigned char display_val = 0;
    // convert speed to a value in the range 0..7
    if (speed != 0)
      {
        switch(format)
          {
            case DCC14:
                display_val = speed * 2;
                display_val = display_val + 3;
                display_val = display_val / 4;
                break;
            case DCC28:
                display_val = speed + 3;
                display_val = display_val / 4;
                break;
            case DCC128:
                for (i = 0; i<(7+1); i++)
                  {
                    if (i * 18 >= speed)
                      {
                        display_val = i;
                        break;
                      } 
                  }
                break; 
            default:
                break;
          }
      }

    if (dir != old_display_dir)
      {
        if (old_display_dir) set_cursor(0,2);
        else  set_cursor(12,2);
        lcd_blank(7);                           // wipe out old stuff
        old_display_dir = dir;
        old_display_val = 0;                    // and kill any reminder
      }

    if (display_val != old_display_val)
      {
        if (dir)
          {
            if (display_val > old_display_val)      // add some chars
              {
                set_cursor(0 + 7 - display_val,2);
                lcd_putc_cnt(display_val - old_display_val, C_ARROW_L_LIGHT);
              }
            else                                    // remove
              {
                set_cursor(0 + 7 - old_display_val,2);
                lcd_blank(old_display_val - display_val);
              }
          }
        else 
          {
            if (display_val > old_display_val)      // add some chars
              {
                set_cursor(12 + old_display_val,2);
                lcd_putc_cnt(display_val - old_display_val, C_ARROW_R_LIGHT);
              }
            else                                    // remove
              {
                set_cursor(12 + display_val,2);
                lcd_blank(old_display_val - display_val);
              }
           }
        old_display_val = display_val;
      }
  }

static void display_loc_light(unsigned char light)
  {
    set_cursor(19,1);
    if(light) lcd_putc('*');
    else      lcd_putc(C_BLOCK_SMALL);
  }

void display_loc_addr(unsigned int addr)
  {
    set_cursor(0,1);
    lcd_int5(addr);
  }

void display_loc_addr_name()
  {
    display_loc_addr( actual_lok.w.addr);
    lcd_putc(':');
    display_loc_name( (char *) actual_lok.name);
  }

void display_loc_other(bool other)
  {
    if (other != old_other)
      {
        set_cursor(16,1);
        if (other) lcd_string("=#=");
        else lcd_blank(3);
        old_other = other;
      }
  }

//                           01234567890123456789
char text_f4_1[]   PROGMEM = "F01  F02  F03  F04  ";
char text_f8_5[]   PROGMEM = "F05  F06  F07  F08  ";
char text_f12_9[]  PROGMEM = "F09  F10  F11  F12  ";
char text_f16_13[] PROGMEM = "F13  F14  F15  F16  ";
char text_f20_17[] PROGMEM = "F17  F18  F19  F20  ";
#if (DCC_F13_F28)
char text_f24_21[] PROGMEM = "F21  F22  F23  F24  ";
char text_f28_25[] PROGMEM = "F25  F26  F27  F28  ";
#endif


// fnum=0..27
void display_softkey_function(unsigned char fg)
  {
    unsigned char func_bits;
    set_cursor(0,3);
    switch(fg)
      {
        default: // no break;
        case 0:
            lcd_string_P(text_f4_1);
            func_bits = current_loc.f4_f1;
            break;
        case 1:
            lcd_string_P(text_f8_5);
            func_bits = current_loc.f8_f5;
            break;
        case 2:
            lcd_string_P(text_f12_9);
            func_bits = current_loc.f12_f9;
            break;
        case 3:
            lcd_string_P(text_f16_13);
            func_bits = current_loc.f16_f13;
            break;
        case 4:
            lcd_string_P(text_f20_17);
            func_bits = current_loc.f20_f17;
            break;
        case 5:
            lcd_string_P(text_f24_21);
            func_bits = current_loc.f24_f21;
            break;
        case 6:
            lcd_string_P(text_f28_25);
            func_bits = current_loc.f28_f25;
            break;
      }
    set_cursor(3,3);
    lcd_bool(func_bits & 0x01);
    if (func_bits & 0x01)  turn_led_green_soft(LED_A);
    else turn_led_off_soft(LED_A);
    
    set_cursor(8,3);
    lcd_bool(func_bits & 0x02);
    if (func_bits & 0x02)  turn_led_green_soft(LED_B);
    else turn_led_off_soft(LED_B);
    
    set_cursor(13,3);
    lcd_bool(func_bits & 0x04);
    if (func_bits & 0x04)  turn_led_green_soft(LED_C);
    else turn_led_off_soft(LED_C);
    
    set_cursor(18,3);
    lcd_bool(func_bits & 0x08);
    if (func_bits & 0x08)  turn_led_green_soft(LED_D);
    else turn_led_off_soft(LED_D);
    
  }

static void display_loc_picture()  {}   // no picture in text mode

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_lok);
    check_status_display(0);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

#if (LANGUAGE == GERMAN)
    static const char text_lok[]   PROGMEM = "Lok  ";
    static const char text_addr[]  PROGMEM = "Adr: ";
#elif  (LANGUAGE == ENGLISH)
    static const char text_lok[]   PROGMEM = "Loco ";
    static const char text_addr[]  PROGMEM = "Addr:";
#else
    #error LANGUAGE is void
#endif

static unsigned char old_display_val = 0;
static unsigned char old_display_dir = 0;

static unsigned char old_other = 0;

static const unsigned char bmp_center[]  PROGMEM = {0xFF};
static const unsigned char bmp_forward[]  PROGMEM =  {0x39, 0x11};   // > (filled)
static const unsigned char bmp_revers[]   PROGMEM =  {0x11, 0x39};   // < (filled)
static const unsigned char bmp_gap[]      PROGMEM =  {0x01};  // only line at top 
static const unsigned char bmp_bar[]      PROGMEM =  {0x7D};  // solid scroll var

#include "bmp/Segment8x14_glcd.h"
static const unsigned char bmp_blank[]    PROGMEM =  {0x00, 0x00, 0x00, 0x00};  // solid scroll var

#define COORDS_SPEED_X  64
#define COORDS_SPEED_Y  24

void bmp_printf(unsigned char data)
  {
    char r0[4];
        
    sprintf( r0, "%3.0d", data );
            
    uint8_t i, j;
    for (j=0; j<3; j++)
      {
        if (r0[j] == ' ') 
          {
            for(i=0; i < SEGMENT8X14_GLCD_WIDTH; i++)
              {
                LCD_setCursorXY(COORDS_SPEED_X + j * (SEGMENT8X14_GLCD_WIDTH+2) + i, COORDS_SPEED_Y);
                LCD_drawBMP_aligned(bmp_blank, 1, 2);
              }
          }
        else if ( r0[j] >= '0' && r0[j] <= '9' ) 
          {
            LCD_setCursorXY(COORDS_SPEED_X + j * (SEGMENT8X14_GLCD_WIDTH+2), COORDS_SPEED_Y);
            LCD_drawBMP_aligned(&segment8x14_glcd_bmp[(r0[j]-'0') * (SEGMENT8X14_GLCD_WIDTH * 2)], 
                                SEGMENT8X14_GLCD_WIDTH,
                                2); 
          }
        // LCD_drawBMP_aligned(bmp_blank, 2, 2);  
      }
  }



void display_loc_speed(t_format format, unsigned char speed, unsigned char dir)
  {
    unsigned char display_val = 0;
    unsigned char i, gap;
    unsigned int temp;
    // convert speed to a value in the range 0..60
    // see speedcalc.xls
    if (speed != 0)
      {
        switch(format)
          {
            case DCC14:
                temp = speed * 1096 + 128;  // 128 = rounding
                display_val = temp / 256;
                break;
            case DCC28:
                temp = speed * 548 + 128;
                display_val = temp / 256;
                break;
            case DCC128:
                temp = speed * 121 + 196;   // larger offset to get a result for speed 1
                display_val = temp / 256;
                break;
            default:
                break;
          }
      }

    // speed text display
    bmp_printf(speed);

    // speed graphic display
    // repaint completely

    LCD_setCursorXY(0,5*8);
    LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // first line
    
    if (dir)
      {
        gap = 60 - display_val;
        for (i=0; i < gap; i++)
            LCD_fillPage_with_BMP(sizeof(bmp_gap),bmp_gap);
        
        LCD_fillPage_with_BMP(sizeof(bmp_revers),bmp_revers);  // < (filled)
        
        for (i=0; i < display_val; i++)
            LCD_fillPage_with_BMP(sizeof(bmp_bar),bmp_bar);
        
        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // middle line @63 = (1 + 60 + 2)
        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);

        for (i=0; i < 60+sizeof(bmp_forward); i++)
            LCD_fillPage_with_BMP(sizeof(bmp_gap),bmp_gap);

        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // last line       
      }
    else
      {
        for (i=0; i < 60+sizeof(bmp_revers); i++)
            LCD_fillPage_with_BMP(sizeof(bmp_gap),bmp_gap);

        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // middle line @63 = (1 + 62 + 1)
        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);
        
        for (i=0; i < display_val; i++)
            LCD_fillPage_with_BMP(sizeof(bmp_bar),bmp_bar);
        
        LCD_fillPage_with_BMP(sizeof(bmp_forward),bmp_forward);  // > (filled)

        gap = 60 - display_val;
        for (i=0; i < gap; i++)
            LCD_fillPage_with_BMP(sizeof(bmp_gap),bmp_gap);

        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // last line       
      }        
  }


static void display_loc_speed_cleanup(void)
  {
    old_display_val = 0;
    old_display_dir = 0;
    old_other = 2;              // void value
  }


#include "bmp/bulb15_glcd.h"

#define COORDS_BULB_X  112
#define COORDS_BULB_Y  24

static void display_loc_light(unsigned char light)
  {
    #define BULB_PAGES (BULB15_GLCD_HEIGHT / LCD_PAGE_HEIGHT / 2)
    
    LCD_setCursorXY(COORDS_BULB_X, COORDS_BULB_Y);
    if (light) LCD_drawBMP_aligned(&bulb15_glcd_bmp[BULB15_GLCD_WIDTH * BULB_PAGES], BULB15_GLCD_WIDTH, BULB_PAGES);
    else       LCD_drawBMP_aligned(&bulb15_glcd_bmp[0], BULB15_GLCD_WIDTH, BULB_PAGES);
  }


#include "bmp/text_addr_6px_glcd.h"


static void display_loc_addr(unsigned int addr)
  {
    #define _LOC_ADDR_NEW
    #ifdef _LOC_ADDR_NEW
        char r0[7];
        LCD_setCursorXY(0, 4 * LCD_PAGE_HEIGHT);
        LCD_drawBMP_aligned(&text_addr_6px_glcd_bmp[0], TEXT_ADDR_6PX_GLCD_WIDTH, 1);
        LCD_setCursorXY(0+TEXT_ADDR_6PX_GLCD_WIDTH, 4 * LCD_PAGE_HEIGHT);
        sprintf( r0, "%5.0d", addr );
        LCD_number_6px(r0);
        LCD_setCursorXY(63, 4 * LCD_PAGE_HEIGHT);
        LCD_fillPage_with_BMP(sizeof(bmp_center),bmp_center);  // vertikal line at end
    #else
        LCD_setCursorXY( 3, 4 * LCD_PAGE_HEIGHT);
        LCD_PF_string_P(text_addr);
        set_cursor(5,4);
        lcd_int5(addr);
    #endif
  }

static void display_loc_addr_name()
  {
    display_loc_addr( actual_lok.w.addr);
    set_cursor(11,2);
    display_loc_name( (char *) actual_lok.name);
  }

#include "bmp/knob_glcd.h"

#define COORDS_KNOB_X  96
#define COORDS_KNOB_Y  24

#define KNOB_PAGES (KNOB_GLCD_HEIGHT / LCD_PAGE_HEIGHT / 2)
    
static void display_loc_other(bool other)
  {
    if (other != old_other)
      {
        if (other) 
          {
            LCD_setCursorXY(COORDS_KNOB_X, COORDS_KNOB_Y);
            LCD_drawBMP_aligned(&knob_glcd_bmp[KNOB_GLCD_WIDTH * KNOB_PAGES], KNOB_GLCD_WIDTH, KNOB_PAGES);
          }
        else 
          {
            LCD_setCursorXY(COORDS_KNOB_X, COORDS_KNOB_Y);
            LCD_drawBMP_aligned(&knob_glcd_bmp[0], KNOB_GLCD_WIDTH, KNOB_PAGES);
          }
        old_other = other;
      }
  }




// make a specual line with all functions
// 0123456789012345678901234567890123456789012345678901234567890123
//0****************************************************************
//1*              *               *               *
//2* ***  ** ***  *  **    **     *  **    **     *
//3* *   *    *   *  **    **     *  **    **     *
//4* **  *    *   *  **    **     *  **    **     *
//5* *    **  *   *  ** ** ** **  *  ** ** ** **  *
//6*              *               *               *
//7****************************************************************

//The LSB is the top most pixel, the MSB the bottom.

static const unsigned char bmp_fct[]  PROGMEM = {0xFF, 0x81, 
                                  0xBD, 0x95, 0x85,     // F
                                  0x81,
                                  0x99, 0xA5, 0xA5,     // C
                                  0x81,
                                  0x85, 0xBD, 0x85,     // T
                                  0x81, 0x81, 0xFF};

static const unsigned char bmp_line[]  PROGMEM = {0xFF}; 
static const unsigned char bmp_one[]  PROGMEM =  {0xBD, 0xBD}; 
static const unsigned char bmp_zero[]  PROGMEM = {0xA1, 0xA1}; 
static const unsigned char bmp_fgap[]  PROGMEM =  {0x81}; 


static void _display_function_fg(unsigned char bits)
  {
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    if (bits & 0x01) LCD_fillPage_with_BMP(sizeof(bmp_one),bmp_one);
    else             LCD_fillPage_with_BMP(sizeof(bmp_zero),bmp_zero);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    if (bits & 0x02) LCD_fillPage_with_BMP(sizeof(bmp_one),bmp_one);
    else             LCD_fillPage_with_BMP(sizeof(bmp_zero),bmp_zero);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    if (bits & 0x04) LCD_fillPage_with_BMP(sizeof(bmp_one),bmp_one);
    else             LCD_fillPage_with_BMP(sizeof(bmp_zero),bmp_zero);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    if (bits & 0x08) LCD_fillPage_with_BMP(sizeof(bmp_one),bmp_one);
    else             LCD_fillPage_with_BMP(sizeof(bmp_zero),bmp_zero);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    LCD_fillPage_with_BMP(sizeof(bmp_fgap),bmp_fgap);
    LCD_fillPage_with_BMP(sizeof(bmp_line),bmp_line);
  }

static void display_function(void)
  {
     LCD_setCursorXY(0,6*8);
     LCD_fillPage_with_BMP(sizeof(bmp_fct),bmp_fct);

     _display_function_fg(current_loc.f4_f1);
     _display_function_fg(current_loc.f8_f5);
     _display_function_fg(current_loc.f12_f9);
     _display_function_fg(current_loc.f16_f13);
     _display_function_fg(current_loc.f20_f17);
     _display_function_fg(current_loc.f24_f21);
     _display_function_fg(current_loc.f28_f25);
  }

// fnum=0..27
static void display_softkey_function(unsigned char fg)
  {
    char func_text[5];
    unsigned char func_bits;
    unsigned char func_num;

    display_function();    

    switch(fg)
      {
        default: // no break;
        case 0:     func_num = 1;
                    func_bits = current_loc.f4_f1;
                    break;
        case 1:     func_num = 5;
                    func_bits = current_loc.f8_f5;
                    break;
        case 2:     func_num = 9;
                    func_bits = current_loc.f12_f9;
                    break;
        case 3:     func_num = 13;
                    func_bits = current_loc.f16_f13;
                    break;
        case 4:     func_num = 17;
                    func_bits = current_loc.f20_f17;
                    break;
        case 5:     func_num = 21;
                    func_bits = current_loc.f24_f21;
                    break;
        case 6:     func_num = 25;
                    func_bits = current_loc.f28_f25;
                    break;
      }
    func_text[0] = 'F';
    sprintf(&func_text[1], "%02d", func_num );
    if (func_bits & 0x01) 
      {
        func_text[3] = C_BLOCK_M2;
        turn_led_green_soft(LED_A);
      }
    else 
      {
        turn_led_off_soft(LED_A);
        func_text[3] = C_BLOCK_SMALL;
      }
    func_text[4] = 0;
    display_softkey(0, func_text);

    sprintf(&func_text[1], "%02d", ++func_num );
    if (func_bits & 0x02) 
      {
        func_text[3] = C_BLOCK_M2;
        turn_led_green_soft(LED_B);
      }
    else 
      {
        turn_led_off_soft(LED_B);
        func_text[3] = C_BLOCK_SMALL;
      }
    func_text[4] = 0;
    display_softkey(1, func_text);

    sprintf(&func_text[1], "%02d", ++func_num );
    if (func_bits & 0x04) 
      {
        func_text[3] = C_BLOCK_M2;
        turn_led_green_soft(LED_C);
      }
    else 
      {
        turn_led_off_soft(LED_C);
        func_text[3] = C_BLOCK_SMALL;
      }
    func_text[4] = 0;
    display_softkey(2, func_text);

    sprintf(&func_text[1], "%02d", ++func_num );
    if (func_bits & 0x08) 
      {
        func_text[3] = C_BLOCK_M2;
        turn_led_green_soft(LED_D);
      }
    else 
      {
        turn_led_off_soft(LED_D);
        func_text[3] = C_BLOCK_SMALL;
      }
    func_text[4] = 0;
    display_softkey(3, func_text);    
  }

static void display_loc_picture()
  {
    LCD_setCursorXY(0,8);
    LCD_loc_picture(actual_lok.picture_id);
  }
    
static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_lok);
    check_status_display(0);
  }



//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY




//------------------------------------------------------------------------
//
// Cache for Locos
//
//------------------------------------------------------------------------

static unsigned char my_cache_mode;    // 0: go to normal list
                                       // 1: go to cache list

struct LokCacheEntry
  {
    unsigned char db_index;
    unsigned char position;
  };

struct LokCacheEntry LokCache[] =
  {
    { 0,  0 },
    { 0,  0 },
    { 0,  0 },
    { 0,  0 },
  };

static void fade_cache(void)
  {
    unsigned char j;
    for ( j=0; j < sizeof(LokCache)/sizeof(LokCache[0]); j++)
      {
        if (LokCache[j].position) LokCache[j].position++;    // fade only valid cache items
      }         
  }

// put lok newindex on top of list
static void cache_lok(unsigned int newindex)
  {
    unsigned char i;
    if (newindex == 0) return;      // never cache the 'new?' Lok
    my_cache_mode = 1;              // activate cache mode for next cursor down key

    // search if lok exits
    for ( i=0; i < sizeof(LokCache)/sizeof(LokCache[0]); i++)
      {
        if (LokCache[i].db_index == newindex)
          {
            if (LokCache[i].position != 1)
              {
                // rearrange priority
                fade_cache();
                LokCache[i].position = 1;  // we are top
              }
            return;
          }
      }
    // does not exist
    fade_cache();
    for ( i=0; i < sizeof(LokCache)/sizeof(LokCache[0]); i++)
      {
        if (LokCache[i].position == 0)     // empty cache item, take it
          {
            LokCache[i].db_index = newindex;
            LokCache[i].position = 1;  // we are top
            return;
          }
        if ( LokCache[i].position == (sizeof(LokCache)/sizeof(LokCache[0])+1) )     // oldest cache item, take it
          {
            LokCache[i].db_index = newindex;
            LokCache[i].position = 1;  // we are top
            return;
          }
      }
  }

// get lok below current_index
// returns 0 if end of cache
static unsigned char restore_cache(unsigned int current_index)
  {
    unsigned char i;
    unsigned char position = 0;
    // get position
    for ( i=0; i < sizeof(LokCache)/sizeof(LokCache[0]); i++)
      {
        if (LokCache[i].db_index == current_index)
          {
            position = LokCache[i].position;
          }
      }
    if (position)
      {
        position++;
        for ( i=0; i < sizeof(LokCache)/sizeof(LokCache[0]); i++)
          {
            if (LokCache[i].position == position)
              {
                return(LokCache[i].db_index);
              }
          }
      }
    return(0);  // not found
  }

//------------------------------------------------------------------------
//
// Tools and Config
//
//------------------------------------------------------------------------


static unsigned char my_fg;

static void clear_current_loc(void)
  {
    current_loc.speed = 0;                 // 0..28 or 0..126
    current_loc.format = actual_lok.w.format;
    current_loc.dir = 0;                   // 0 = forward, 1=revers
    current_loc.light = 0;             // function light
    current_loc.f4_f1 = 0;             // function 4 downto 1
    current_loc.f8_f5 = 0;             // function 8 downto 5
    current_loc.f12_f9 = 0;            // function 12 downto 9
    #if (DCC_F13_F28)
    current_loc.f16_f13 = 0;           // function 16 downto 13
    current_loc.f20_f17 = 0;           // function 20 downto 16
    current_loc.f24_f21 = 0;           // function 24 downto 21
    current_loc.f28_f25 = 0;           // function 28 downto 25
    #endif
    current_loc.other = 0;             // 0: owned by me
  }

static void set_new_actual_lok(unsigned char new_index)
  {
    actual_database_index = new_index;
    get_loco_data(actual_database_index, &actual_lok);
    
    msg_send_askforlocinfo(actual_lok.w.addr);
    msg_transmit();
    
    speed_0_holdtime = get_systick_10ms() - KNOB_AT_ZERO_TIME / SYSTICK_10MS_PERIOD;  // make 'old'
    current_loc.format = actual_lok.w.format;  // just copy (tmp)
    current_loc.speed = 0;
    current_loc.dir = 0;
    current_loc.other = 0;


    display_loc_picture();
    display_loc_addr_name();
    display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
    display_loc_light(current_loc.light);
    display_loc_other(current_loc.other);

    store_curr_index(actual_database_index);
  }

static void toggle_function(unsigned char fnum)
  {
    unsigned char temp, bitmask;
    
    temp = fnum & 0x03;
    bitmask = 1 << temp;

    if (fnum < 4)
      {
        current_loc.f4_f1 ^= bitmask;
        display_softkey_function(0);
      }
    else if (fnum < 8)
      {
        current_loc.f8_f5 ^= bitmask;
        display_softkey_function(1);
      }
    else if (fnum < 12)
      {
        current_loc.f12_f9 ^= bitmask;
        display_softkey_function(2);
      }
    else if (fnum < 16)
      {
        current_loc.f16_f13 ^= bitmask;
        display_softkey_function(3);
      }
    else if (fnum < 20)
      {
        current_loc.f20_f17 ^= bitmask;
        display_softkey_function(4);
      }
    else if (fnum < 24)
      {
        current_loc.f24_f21 ^= bitmask;
        display_softkey_function(5);
      }
    else // if (fnum < 28)
      {
        current_loc.f28_f25 ^= bitmask;
        display_softkey_function(6);
      }
  }

static unsigned char accumulate_knob(unsigned char invert)
  {
    signed int accu;
    unsigned char speedsteps;

    accu = current_loc.speed;
    if (invert) accu -= get_knob_rotate();
    else        accu += get_knob_rotate();
    if (accu < 0)  accu = 0;
    switch (current_loc.format)
      {
        default:
        case DCC14: speedsteps = 14;  break;
        case DCC27: speedsteps = 27;  break;
        case DCC28: speedsteps = 28;  break;
        case DCC128:speedsteps = 126; break;
      }
    if (accu > speedsteps)   accu = speedsteps;

    return(accu);
  }

//------------------------------------------------------------------------
//
// Menu Handler
//
//------------------------------------------------------------------------

void action_lok(t_navicode par)
  {
    unsigned char temp;
    unsigned char lok_dreh = 0;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            cursor_off();
            clear_current_loc();
            
            // we assume actual_lok is loaded
            // ask command station and try to get the real data for actual_lok
            // 
            msg_send_askforlocinfo(actual_lok.w.addr);
            msg_transmit();

            // in the mean time, let's do some init code
            display_time_allowed = 1;
            knob_mode = eeprom_read_byte((void *)eadr_knob_mode);
            speed_0_holdtime = get_systick_10ms() - KNOB_AT_ZERO_TIME / SYSTICK_10MS_PERIOD;  // make 'old'
            show_header();
  
            display_loc_picture();
            display_loc_addr_name();
            
            display_loc_speed_cleanup();   // clear cache

            my_fg = 0;
            my_cache_mode = 1;
            
            // now check for the answer of the command station
            // we could: wait some time - or drop a KEY_PARSER_LOK event to get the update
            // instead: we do it here, even if the answer isn't there
            display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
            display_loc_light(current_loc.light);
            display_softkey_function(my_fg);
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                case KEY_SHIFT:
                    my_fg = my_fg + 1;
                    if (my_fg > 6) my_fg = 0;
                    display_softkey_function(my_fg);
                    break;
                case KEY_SHIFT_RELEASE:
                    if (last_key == KEY_SHIFT)   // shift pressed and released
                      {
                        t_data16 shifttimeout;
                        shifttimeout.as_int16 = get_systick_10ms() - last_key_time;
                        if (shifttimeout.as_uint16 > 1200000L / SYSTICK_10MS_PERIOD) 
                          {  // long press on shift
                            if (my_fg > 0) my_fg -= 1;
                            display_softkey_function(my_fg);
                          }
                      }
                    else
                      {
                        if (my_fg > 0) my_fg -= 1;
                        display_softkey_function(my_fg);
                      }
                    break;
                
                case KEY_SH_A:
                case KEY_SH_B:
                case KEY_SH_C:
                case KEY_SH_D:
                    // no break;
                case KEY_A:
                case KEY_B:
                case KEY_C:
                case KEY_D:
                    temp = par.key & 0x03;  // get lower two bits
                    temp += (my_fg * 4);
                    toggle_function(temp);   // temp = 0..31
                    msg_send_loc_function(temp);
                    msg_transmit();
                    
                    current_loc.other = 0;
                    display_loc_other(current_loc.other);
                    cache_lok(actual_database_index);
                    break;

                case KEY_KNOB:
                    if (knob_mode == KNOB_MODE_SINGLESIDED)
                      {
                        if (current_loc.speed != 0)
                          {
                            current_loc.speed = 0;
                            if (knob_push == KNOB_PUSH_FS0)
                              {
                                msg_send_loc_drive(actual_lok.w.addr, actual_lok.w.format, current_loc.speed, current_loc.dir);
                                msg_transmit();
                              }
                            else // KNOB_PUSH_ESTOP      
                              {
                                msg_send_loc_stop(actual_lok.w.addr);
                                msg_transmit();
                              }
                          }
                        else
                          {
                            if (current_loc.dir)  current_loc.dir = 0;
                            else                  current_loc.dir = 1;
                            msg_send_loc_drive(actual_lok.w.addr, actual_lok.w.format, current_loc.speed, current_loc.dir);
                            msg_transmit();
                          }
                      }
                    else // KNOB_MODE_DUALSIDED
                      {
                        if (current_loc.speed != 0)   // here we do only a stop, no reversing
                          {
                            current_loc.speed = 0;
                            if (knob_push == KNOB_PUSH_FS0)
                              {
                                msg_send_loc_drive(actual_lok.w.addr, actual_lok.w.format, current_loc.speed, current_loc.dir);
                                msg_transmit();
                              }
                            else // KNOB_PUSH_ESTOP      
                              {
                                msg_send_loc_stop(actual_lok.w.addr);
                                msg_transmit();
                              }
                            speed_0_holdtime = get_systick_10ms() - KNOB_AT_ZERO_TIME / SYSTICK_10MS_PERIOD;  // make 'old'
                          }
                      }
                    display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
              
                    current_loc.other = 0;
                    display_loc_other(current_loc.other);
                    cache_lok(actual_database_index);
                    break;

                case KEY_KNOB_TURNED:
                    if (knob_mode == KNOB_MODE_SINGLESIDED)
                      {
                        lok_dreh = accumulate_knob(0);   // not inverted
                      }
                    else // KNOB_MODE_DUALSIDED
                      {
                        if (current_loc.speed) 
                          {
                            lok_dreh = accumulate_knob(current_loc.dir);  // invert, if revers speed
                            if (lok_dreh == 0) speed_0_holdtime = get_systick_10ms();  // we reached 0, remember time
                          }
                        else
                          {
                            signed char stemp;
                            t_data16 holdtimeout;
                            stemp = get_knob_rotate();
                            // possible chance to change dir
                            if (current_loc.dir)
                              { // we are revers
                                if (stemp < 0)
                                  {
                                    lok_dreh = 1;  // kill down any accelerate
                                  }
                                else if (stemp > 0)
                                  {
                                    holdtimeout.as_int16 = get_systick_10ms() - speed_0_holdtime;
                                    if (holdtimeout.as_uint16 > KNOB_AT_ZERO_TIME / SYSTICK_10MS_PERIOD) 
                                      {
                                        current_loc.dir = 0;
                                        lok_dreh = 0;
                                        current_loc.speed = 1;   // fool compare below, force an update
                                      }
                                    speed_0_holdtime = get_systick_10ms();  // 2009-05-06 keep hold as long as knob is turned
                                  }
                              }
                            else
                              { // we are forward
                                if (stemp > 0)
                                  {
                                    lok_dreh = 1;  // kill down any accelerate
                                  }
                                else if (stemp < 0)
                                  {
                                    holdtimeout.as_int16 = get_systick_10ms() - speed_0_holdtime;
                                    if (holdtimeout.as_uint16 > KNOB_AT_ZERO_TIME / SYSTICK_10MS_PERIOD) 
                                      {
                                        current_loc.dir = 1;
                                        lok_dreh = 0;
                                        current_loc.speed = 1;   // fool compare, force an update
                                      }
                                    speed_0_holdtime = get_systick_10ms();  // 2009-05-06 keep hold as long as knob is turned
                                  }
                              }
                          }
                      }
                      
                    if (current_loc.speed != lok_dreh)
                      {
                        current_loc.speed = lok_dreh;
                        msg_send_loc_drive(actual_lok.w.addr, actual_lok.w.format, current_loc.speed, current_loc.dir);
                        msg_transmit();
                        display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
                        current_loc.other = 0;
                        display_loc_other(current_loc.other);
                        cache_lok(actual_database_index);
                      }
                    break;

                case KEY_CENTER:
                    current_loc.light = !current_loc.light;
                    msg_send_loc_light();
                    msg_transmit();
                    display_loc_light(current_loc.light);

                    current_loc.other = 0;
                    display_loc_other(current_loc.other);
                    cache_lok(actual_database_index);
                    break;

                case KEY_PARSER_LOK:
                    display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
                    display_loc_light(current_loc.light);
                    display_loc_other(current_loc.other);
                    display_softkey_function(my_fg);
                    break;                       
                
                case KEY_PARSER_STOLEN:
                    if( actual_lok.w.addr != 0)
                      {
                        display_loc_speed(current_loc.format, current_loc.speed, current_loc.dir);
                        display_loc_light(current_loc.light);
                        display_loc_other(current_loc.other);   // if lok has been stolen
                        display_softkey_function(my_fg);
                      }
                    break;

                case KEY_DOWN:                          // call up prev LOC in cache
                    if (my_cache_mode) 
                      {
                        temp = restore_cache(actual_database_index);
                        if (temp)
                          {
                            set_new_actual_lok(temp);
                          }
                        else
                          {
                            my_cache_mode = 0;
                            temp = scan_loco_database_forward(actual_database_index);
                            set_new_actual_lok(temp);
                          }
                      }
                    else
                      {
                        temp = scan_loco_database_forward(actual_database_index);
                        set_new_actual_lok(temp);
                      }
                    break;
                case KEY_SH_DOWN:                      // call up last loc in list
                    my_cache_mode = 0;
                    temp = scan_loco_database_forward(actual_database_index);
                    set_new_actual_lok(temp);
                    break;
                case KEY_SH_UP:
                case KEY_UP:                        // call up next LOC in list
                    my_cache_mode = 0;
                    temp = scan_loco_database_revers(actual_database_index);                    
                    set_new_actual_lok(temp);
                    break;
              }
            break;
        case MENU_CLOSE:
            break;
      }
  }

