//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_lok.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01     start
//            2009-01-30 V0.02 kw  rewritten completely
//            2009-03-31 V0.03 kw  added display_lok_picture
//            2009-04-03 V0.04 kw  added shift scroll
//            2010-06-22 V0.05 kw  msg_transmit();
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screen for edit a lok
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>


#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_ks0073_charset.h"
#include "lcd_text.h"
#include "display.h"
#include "lok_graphic.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "programmer.h"
#include "menu_navigator.h"
#include "menu_lok.h"               // actual lok


//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------

char throttle_charset[] = 
  { '1','2','3','4','5','6','7','8','9','0',
    ':','!','?','.','&','-','+','(',')',
    '#','*','<','>',' ',
    'A','B','C','D','E','F','G','H','I','J',
    'K','L','M','N','O','P','Q','R','S','T',
    'U','V','W','X','Y','Z',
    'a','b','c','d','e','f','g','h','i','j',
    'k','l','m','n','o','p','q','r','s','t',
    'u','v','w','x','y','z'};



static unsigned char changed;           // flag, if edit_lok is modified
static unsigned char my_index;          // my pointer to database

static int my_charset_index;            // my pointer for char edit
static unsigned char my_edit_index;     // actual edit field


t_locoentry edit_lok;   // a copy of the dataset of currently addressed lok


const char text_editlok[]  PROGMEM = "Lok (Edit)";
const char text_addr[]  PROGMEM = "Addr:";
const char text_name[]  PROGMEM = "Name:";
const char text_fs[]  PROGMEM = "FS:";
const char text_picture[]  PROGMEM = "Picture:";
const char text_edit_sk_del[]   PROGMEM = {C_ARROW_RIGHT_DOWN,'D','E','L',0};
const char text_edit_sk_org[]   PROGMEM = {'=','O','R','G',0}; 
const char text_edit_sk_sav[]   PROGMEM = {C_ARROW_RIGHT_UP,'S','A','V',0}; 
const char text_edit_sk_add[]   PROGMEM = {C_ARROW_RIGHT_UP,'A','D','D',0}; 
const char text_edit_sk_name[]  PROGMEM = {'>','N','A','M','E', 0};                                 
                               //  01234567890123456789
const char text_edit_sk_ptrd[]  PROGMEM = "PtRD";
const char text_edit_sk_ptwr[]  PROGMEM = "PtWR";
const char text_edit_sk_blank[] PROGMEM = " ";




void search_charset_index(char data)
  {
    unsigned char i;
    for (i=0; i < sizeof(throttle_charset); i++)
      {
        if (throttle_charset[i] == data)
          {
            my_charset_index = i;
            return;
          }
      }
    my_charset_index = 0;  // default, if not found
  }


//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------
//
#if (DISPLAY == LCD_EA204)
// Menu:
//   01234567890123456789
//  |--------------------|
// 0|Edit           STOP |
// 1| Addr:12345 FS:128  |
// 2| Name:BBII          |
// 3| DEL  -    -    -   |
//  |--------------------|

// and if a change was made:
// 3| ORG  SAV  ADD  NAME|
//  |--------------------|

typedef enum 
  {
    EDIT_ADDR,
    EDIT_FS,
    EDIT_NAME,
    N_of_EDIT
  } t_edit_focus;

static t_edit_focus edit_focus;




static void set_edit_cursor(void)
  {
    switch(edit_focus)
      {
        case N_of_EDIT:           break;
        case EDIT_ADDR:
            cursor_on_at(6,1,5); break;
        case EDIT_FS:
            cursor_on_at(15,1,3);  break;            
        case EDIT_NAME:
            set_cursor( 6+my_edit_index,2);  // also put the write pointer here
            cursor_on_at( 6+my_edit_index,2,1); break;         
      }
    search_charset_index(edit_lok.name[my_edit_index]);
  }


static void display_edit_loc_picture(unsigned int id) {};  // here void

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_editlok);
    check_status_display(0);
    set_cursor(1,1);
    lcd_string_P(text_addr);
    
    set_cursor(12,1);
    lcd_string_P(text_fs);
    set_cursor(1,2);
    lcd_string_P(text_name);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

//   01234567890123456789
//  |--------------------|
// 0|Edit           STOP |
// 1| Addr:12345 FS:128  |
// 2| Name:BBII          |
// 3|           ######## |
// 4|           ######## |
// 5| Picture:  ######## |
// 6|           ######## |
// 7| DEL  -    -    -   |
//  |--------------------|

typedef enum 
  {
    EDIT_ADDR,
    EDIT_FS,
    EDIT_NAME,
    EDIT_PICTURE,
    N_of_EDIT
  } t_edit_focus;

static t_edit_focus edit_focus;

static void set_edit_cursor(void)
  {
    switch(edit_focus)
      {
        case N_of_EDIT:           break;
        case EDIT_ADDR:
            cursor_on_at(6,1,5); break;
        case EDIT_FS:
            cursor_on_at(15,1,3);  break;            
        case EDIT_NAME:
            set_cursor( 6+my_edit_index,2);  // also put the write pointer here
            cursor_on_at( 6+my_edit_index,2,1); break;         
        case EDIT_PICTURE:
            cursor_on_at(10,5,11); break;    // dirty: this is a pixel field, but we use text cursor     
      }
    search_charset_index(edit_lok.name[my_edit_index]);
  }

static void display_edit_loc_picture(unsigned int id)
  {
    LCD_setCursorXY(60,24);
    LCD_loc_picture(id);
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_editlok);
    check_status_display(0);

    set_cursor(1,1);
    lcd_string_P(text_addr);
    set_cursor(12,1);
    lcd_string_P(text_fs);
    set_cursor(1,2);
    lcd_string_P(text_name);
    set_cursor(1,5);
    lcd_string_P(text_picture);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY

void display_edit_loc_addr(unsigned int addr)
  {
    set_cursor(6,1);
    lcd_int5(addr);
  }

void display_edit_format(t_format my_format)
  {
    set_cursor(15,1);
    switch(my_format)
      {
        default:
        case DCC14: lcd_int3(14);break;
        case DCC27: lcd_int3(27);break;
        case DCC28: lcd_int3(28);break;
        case DCC128:lcd_int3(126);break;
      }
  }

static void display_edit_loc_name(char * name)
  {
    unsigned char i;
    set_cursor(6,2);
    for(i=0; i< LOK_NAME_LENGTH; i++)
      {
        if (*name) lcd_putc(*name++);
        else lcd_putc(' ');                     // fill rest with blank
      }
  }


static void move_edit_pointer_forward(void)
  {
    if (edit_focus == EDIT_NAME) 
      {
        my_edit_index++;
        if (my_edit_index == LOK_NAME_LENGTH)
          {
            my_edit_index=0;
            edit_focus++;
          }
      }
    else
        edit_focus++;
    if (edit_focus == N_of_EDIT) edit_focus = 0;
    set_edit_cursor();
  }

static void move_edit_pointer_revers(void)
  {
    if (edit_focus == EDIT_NAME) 
      {
        if (my_edit_index != 0) my_edit_index--;
        else
          {
            if (edit_focus == 0) edit_focus = N_of_EDIT;
            edit_focus--;
          }
      }
    else
      {
        if (edit_focus == 0) edit_focus = N_of_EDIT;
        edit_focus--;
        if (edit_focus == EDIT_NAME) my_edit_index = LOK_NAME_LENGTH - 1;
      }
    set_edit_cursor();
  }


void display_edit_lok(void)
  {
    cursor_off();
    display_edit_format(edit_lok.w.format);
    display_edit_loc_addr(edit_lok.w.addr);
    display_edit_loc_name((char *)edit_lok.name);
    display_edit_loc_picture(edit_lok.picture_id);
    set_edit_cursor();
  }

void load_and_display_lok(unsigned char index)
  {
    get_loco_data(index, &edit_lok);
    display_edit_lok();
  }

static void show_softkeys_noshift(void)
  {
    if (changed)
      {
        display_softkey_P(0, text_edit_sk_org);
        display_softkey_P(1, text_edit_sk_sav);
        display_softkey_P(2, text_edit_sk_add);
        display_softkey_P(3, text_edit_sk_name);    // org sav add name
        turn_led_green_soft(LED_A);
        turn_led_green_soft(LED_B);
        turn_led_green_soft(LED_C);
        turn_led_green_soft(LED_D);
      }
    else
      {
        display_softkey_P(0, text_edit_sk_del);
        display_softkey_P(1, text_edit_sk_blank);
        display_softkey_P(2, text_edit_sk_blank);
        display_softkey_P(3, text_edit_sk_blank);    // del
        turn_led_red_soft(LED_A);
        turn_led_off_soft(LED_B);
        turn_led_off_soft(LED_C);
        turn_led_off_soft(LED_D);
      }
    set_edit_cursor();
  }


//------------------------------------------------------------------------
//
// Menu-Handler
//
//------------------------------------------------------------------------

void action_lokedit(t_navicode par)
  {
    signed char s_temp;
    unsigned char u_temp;
    signed int my_addr;

    check_status_display(par.key);
    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();

            changed = 0;
            show_softkeys_noshift();

            my_index = actual_database_index;

            my_edit_index = 0;
            edit_focus = EDIT_ADDR;
            load_and_display_lok(my_index);
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_SHIFT:
                    display_softkey_P(0, text_edit_sk_ptrd);
                    display_softkey_P(1, text_edit_sk_ptwr);
                    display_softkey_P(2, text_edit_sk_blank);
                    display_softkey_P(3, text_edit_sk_blank);
                    turn_led_green_soft(LED_A);
                    turn_led_red_soft(LED_B);
                    turn_led_off_soft(LED_C);
                    turn_led_off_soft(LED_D);
                    break;
                case KEY_SHIFT_RELEASE:
                    show_softkeys_noshift();
                    break;
                case KEY_UP:
                    changed = 0;
                    show_softkeys_noshift();
                    my_index = scan_loco_database_revers(my_index);
                    load_and_display_lok(my_index);
                    break;
                case KEY_DOWN:          
                    changed = 0;
                    show_softkeys_noshift();
                    my_index = scan_loco_database_forward(my_index);
                    load_and_display_lok(my_index);
                    break;
                case KEY_KNOB:          // change focus forward
                    move_edit_pointer_forward();
                    break;
                case KEY_SH_KNOB:          // change focus revers
                    move_edit_pointer_revers();
                    break;
                case KEY_KNOB_TURNED:
                    cursor_off();
                    if (changed == 0)
                      {
                        changed = 1;
                        show_softkeys_noshift();
                      }
                    switch(edit_focus)
                      {
                        case EDIT_ADDR:
                            my_addr = edit_lok.w.addr;
                            my_addr += get_knob_rotate();
                            if (my_addr < 1) my_addr = 1;
                            if (my_addr > 10239) my_addr = 10239;   // 10239 = 0x27FF
                            edit_lok.w.addr = (unsigned int) my_addr;
                            display_edit_loc_addr(edit_lok.w.addr);
                            break;
                        case EDIT_FS:
                            s_temp = get_knob_rotate();
                            switch(edit_lok.w.format)
                              {
                                default:
                                case DCC14: if (s_temp > 0) edit_lok.w.format = DCC28;
                                            if (s_temp < 0) edit_lok.w.format = DCC128;
                                            break;
                                case DCC27:
                                            break;
                                case DCC28: if (s_temp > 0) edit_lok.w.format = DCC128;
                                            if (s_temp < 0) edit_lok.w.format = DCC14;
                                            break;
                                case DCC128:if (s_temp > 0) edit_lok.w.format = DCC14;
                                            if (s_temp < 0) edit_lok.w.format = DCC28;
                                            break;
                              }
                            display_edit_format(edit_lok.w.format);
                            break;
                        case EDIT_NAME:
                            my_charset_index  += get_knob_rotate();
                            while (my_charset_index < 0) my_charset_index += sizeof(throttle_charset);
                            while (my_charset_index >= sizeof(throttle_charset)) my_charset_index -= sizeof(throttle_charset);
                            lcd_putc(throttle_charset[my_charset_index]);
                            edit_lok.name[my_edit_index] = throttle_charset[my_charset_index];
                            break;
                        #if  (DISPLAY == LCD_EA128x64)
                        case EDIT_PICTURE:
                            my_addr = get_knob_rotate();
                            if (my_addr > 0) 
                              {
                                if (is_shift_pressed())
                                  {
                                    edit_lok.picture_id |= 0x0FFF;
                                    edit_lok.picture_id = scan_picture_forward(edit_lok.picture_id);
                                  }
                                else
                                  {
                                    edit_lok.picture_id = scan_picture_forward(edit_lok.picture_id);
                                  }
                              }
                            if (my_addr < 0) 
                              {
                                if (is_shift_pressed())
                                  {
                                    edit_lok.picture_id &= 0xF000;
                                    edit_lok.picture_id = scan_picture_revers(edit_lok.picture_id);
                                  }
                                else
                                  {
                                    edit_lok.picture_id = scan_picture_revers(edit_lok.picture_id);
                                  }
                              }
                            display_edit_loc_picture(edit_lok.picture_id);
                            break;
                        #endif
                        case N_of_EDIT:
                            break;
                      }
                    set_edit_cursor();
                    break;
                case KEY_CENTER:
                  {
                    memcpy(&actual_lok, &edit_lok, sizeof(edit_lok));   // give back edit copy // but no save
                    break;
                  }
                case KEY_A:      
                    if (changed)    // ORG
                      {
                        changed = 0;
                        show_softkeys_noshift();
                        my_edit_index = 0;
                        memcpy(&edit_lok, &actual_lok, sizeof(edit_lok));   // make new local copy

                        edit_focus = EDIT_ADDR;
                        display_edit_lok();
                      }
                    else  // delete
                      {
                        u_temp = scan_loco_database_forward(my_index);
                        delete_loco_data(my_index);
                        if (actual_database_index == my_index)             // also modify actual index
                          {
                            my_index = u_temp;
                            actual_database_index = my_index;
                          }
                        else
                          {
                            my_index = u_temp;
                          }
                        edit_focus = EDIT_ADDR;
                        load_and_display_lok(my_index);
                      }
                    break;
                case KEY_B:
                    if (changed)    // change entry
                      {
                        changed = 0;
                        show_softkeys_noshift();
                        save_loco_data(my_index, &edit_lok);
                      }
                    else            // sort
                      {
                        
                      }
                    break;
                case KEY_C:
                    if (changed)    // add entry
                      {
                        changed = 0;
                        show_softkeys_noshift();
                        if ((u_temp = insert_new_loco_data(my_index)))
                          {
                            my_index = u_temp;
                            save_loco_data(my_index, &edit_lok);
                            memcpy(&actual_lok, &edit_lok, sizeof(edit_lok));   // also load the actual lok
                          }
                        else
                          {
                            display_error(ERR_DATABASE_FULL);
                          }
                      }
                    else            // nothing
                      {
                        
                      }
                    break;
                case KEY_D:         // name predef
                    cursor_off();
                    edit_lok.name[0]='L';
                    sprintf((char *) &edit_lok.name[1], "%04d", edit_lok.w.addr );
                    display_edit_loc_name((char *)edit_lok.name);
                    set_edit_cursor();
                    break;
                case KEY_SH_A:      // read addr from prog track
                    start_programmer_readaddr();
                    break;     
                case KEY_SH_B:      // write addr to prog track     
                    start_programmer_writeaddr(edit_lok.w.addr);
                    break;
                case KEY_PROGRAMMER:
                    if (prg_state == PRG_DONE)
                      {
                        if (edit_lok.w.addr != progged_lok_addr)
                          {
                            changed = 1;
                            edit_lok.w.addr = progged_lok_addr;
                            display_edit_loc_addr(edit_lok.w.addr);
                            show_softkeys_noshift();
                          }
                      }
                    msg_send_cv_end();
                    msg_transmit();
                    break;
              }
            break;
        case MENU_CLOSE:
            cursor_off();
            turn_led_off_soft(LED_A);
            turn_led_off_soft(LED_B);
            turn_led_off_soft(LED_C);
            turn_led_off_soft(LED_D);
            switch (par.key)
              {
                default:
                    break;
                case KEY_CENTER:
                  {
                    if (actual_lok.w.addr == edit_lok.w.addr)
                      {
                        if (actual_lok.w.format == edit_lok.w.format)   // stop loc and tell command station
                          {
                            msg_send_loc_drive(edit_lok.w.addr, edit_lok.w.format, 0 /* speed */, current_loc.dir);
                            msg_transmit();
                          }
                      }                    
                    memcpy(&actual_lok, &edit_lok, sizeof(edit_lok));   // give back edit copy // but no save
                    break;
                  }
                
                case KEY_A:
                    break;
              }
            break;
      }
  }



