//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_lok.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-01-30 V0.02 kw  rewritten completely
//            2009-02-23 V0.03 kw  added database receive, LED control
//            2009-03-27 V0.04 kw  added call to softkey control
//            2009-03-27 V0.05 kw  minor bug fix im lokmove
//            2009-04-06 V0.06 kw  longer list for graphic display
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screen for edit a lok
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>


#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_ks0073_charset.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"
#include "menu_lok.h"               // actual lok


//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------

static unsigned char ppprev;
static unsigned char pprev;
static unsigned char prev;
static unsigned char curr;
static unsigned char next;
static unsigned char nnext;
static unsigned char nnnext;
static unsigned char scrollmode;

//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
static char text_listlok[]  PROGMEM = "List";
static char text_view[]  PROGMEM = "(view)";
static char text_sort[]  PROGMEM = "(sort)";
static char text_recv[]  PROGMEM = "(recv)";
           



//   01234567890123456789
//  |--------------------|
// 0|List (view)    STOP |
// 1| 12345:BBII-12345128|
// 2|*    3:Dampfer    14|
// 3|    99:any          |
//  |--------------------|


static char text_14[]  PROGMEM = {' ',        C_NUM_EXP_1,C_NUM_EXP_4,0};
static char text_27[]  PROGMEM = {' ',        C_NUM_EXP_2,C_NUM_EXP_7,0};
static char text_28[]  PROGMEM = {' ',        C_NUM_EXP_2,C_NUM_EXP_8,0};
static char text_126[] PROGMEM = {C_NUM_EXP_1,C_NUM_EXP_2,C_NUM_EXP_6,0};
 

void display_format(t_format my_format)
  {
    switch(my_format)
      {
        default:
        case DCC14: // lcd_int3(14);
                    lcd_string_P(text_14);
                    break;
        case DCC27: // lcd_int3(27);
                    lcd_string_P(text_27);
                    break;
        case DCC28: // lcd_int3(28);
                    lcd_string_P(text_28);
                    break;
        case DCC128:// lcd_int3(126);
                    lcd_string_P(text_126);
                    break;
      }
  }


void display_lok_entry(unsigned char index)
  {
    t_locoentry my_lok;
    
    get_loco_data(index, &my_lok); 
    lcd_int5(my_lok.w.addr);
    lcd_putc(':');

    display_loc_name((char *)my_lok.name);
    lcd_blank(10-LOK_NAME_LENGTH);
    #if (LOK_NAME_LENGTH > 10)
        #warning display overflow
    #endif                  
    display_format(my_lok.w.format);
  }


// static void display_scroll_balken(void)
//  {
//    set_cursor(19,1);
//    lcd_putc(C_ARROW_UP_SMALL);
//    set_cursor(19,2);
//    lcd_putc(C_VERT_LINE); 
//    set_cursor(19,3);
//    lcd_putc(C_ARROW_DOWN_SMALL);
//  }


void display_scrollmode(void)
  {
    set_cursor(0,2);
    if (scrollmode)
      {
        lcd_putc(C_ARROW_RIGHT_SMALL);       // show pickup sign
        set_cursor(5,0);
        lcd_string_P(text_view); 
      }
    else
      {
        lcd_putc(C_ARROW_UP_DOWN);
        set_cursor(5,0);
        lcd_string_P(text_sort); 
      }     
  }


static void display_all_lok(void)
  {
    set_cursor(1,1);
    display_lok_entry(prev);

    set_cursor(0,2);
    if (scrollmode) lcd_putc(C_ARROW_RIGHT_SMALL);       // show pickup sign
    else lcd_putc(C_ARROW_UP_DOWN); 
    display_lok_entry(curr);
    set_cursor(1,3);
    display_lok_entry(next);

    // display_scroll_balken(); passt leider nicht mehr
    cursor_off();
  }


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_listlok);
    check_status_display(0);
  }


//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

static const char text_listlok[]  PROGMEM = "List";
static const char text_view[]  PROGMEM = "(view)";
static const char text_sort[]  PROGMEM = "(sort)";
static const char text_recv[]  PROGMEM = "(recv)";
           



//   01234567890123456789
//  |--------------------|
// 0|List (view)    STOP |
// 1| 12345:BBII-12345128|
// 2|   218:BR218     128|
// 3|     3:Dampfer    14|
// 4|*   99:any          |
// 5|   100:any2         |
// 6|   111:2323         |
// 7|    99:any          |
//  |--------------------|


static const char text_14[]  PROGMEM = {' ',        C_NUM_EXP_1,C_NUM_EXP_4,0};
static const char text_27[]  PROGMEM = {' ',        C_NUM_EXP_2,C_NUM_EXP_7,0};
static const char text_28[]  PROGMEM = {' ',        C_NUM_EXP_2,C_NUM_EXP_8,0};
static const char text_126[] PROGMEM = {C_NUM_EXP_1,C_NUM_EXP_2,C_NUM_EXP_6,0};
 

void display_format(t_format my_format)
  {
    switch(my_format)
      {
        default:
        case DCC14: // lcd_int3(14);
                    lcd_string_P(text_14);
                    break;
        case DCC27: // lcd_int3(27);
                    lcd_string_P(text_27);
                    break;
        case DCC28: // lcd_int3(28);
                    lcd_string_P(text_28);
                    break;
        case DCC128:// lcd_int3(126);
                    lcd_string_P(text_126);
                    break;
      }
  }


void display_lok_entry(unsigned char index)
  {
    t_locoentry my_lok;
    
    get_loco_data(index, &my_lok); 
    lcd_int5(my_lok.w.addr);
    lcd_putc(':');

    display_loc_name((char *)my_lok.name);
    
    lcd_blank(10-LOK_NAME_LENGTH);   // fill rest
    #if (LOK_NAME_LENGTH > 10)
        #warning display overflow
    #endif                  
    display_format(my_lok.w.format);
  }


// static void display_scroll_balken(void)
//  {
//    set_cursor(19,1);
//    lcd_putc(C_ARROW_UP_SMALL);
//    set_cursor(19,2);
//    lcd_putc(C_VERT_LINE); 
//    set_cursor(19,3);
//    lcd_putc(C_ARROW_DOWN_SMALL);
//  }


void display_scrollmode(void)
  {
    set_cursor(0,4);
    if (scrollmode)
      {
        lcd_putc(C_ARROW_RIGHT_SMALL);       // show pickup sign
        set_cursor(5,0);
        lcd_string_P(text_view); 
      }
    else
      {
        lcd_putc(C_ARROW_UP_DOWN);
        set_cursor(5,0);
        lcd_string_P(text_sort); 
      }     
  }


static void display_all_lok(void)
  {
    set_cursor(1,1);
    display_lok_entry(ppprev);

    set_cursor(1,2);
    display_lok_entry(pprev);

    set_cursor(1,3);
    display_lok_entry(prev);

    set_cursor(0,4);
    if (scrollmode) lcd_putc(C_ARROW_RIGHT_SMALL);       // show pickup sign
    else lcd_putc(C_ARROW_UP_DOWN); 
    display_lok_entry(curr);

    set_cursor(1,5);
    display_lok_entry(next);

    set_cursor(1,6);
    display_lok_entry(nnext);

    set_cursor(1,7);
    display_lok_entry(nnnext);

    // display_scroll_balken(); passt leider nicht mehr
    cursor_off();
  }


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_listlok);
    check_status_display(0);
  }




//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY

static void get_database_indeces(void)
  {
    curr = actual_database_index;
    next = scan_loco_database_forward(curr);
    nnext = scan_loco_database_forward(next);
    nnnext = scan_loco_database_forward(nnext);
    prev = scan_loco_database_revers(curr);
    pprev = scan_loco_database_revers(prev);
    ppprev = scan_loco_database_revers(pprev);
  }
//-----------------------------------------------------------------------
//
// Menu Handler (list mode)
//
//-----------------------------------------------------------------------

void action_loklist(t_navicode par)
  {
    signed char s_temp;
    unsigned char u_temp;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();
            
            turn_led_green_soft(LED_A);
            
            scrollmode = 1;
            display_scrollmode();

            get_database_indeces();        // ppprev ... curr ... nnnext

            display_all_lok();      
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_CENTER:
                case KEY_KNOB:             // pick up lok and exit
                    break;
                case KEY_SH_KNOB:          // change to move (handled in navigator)
                    break;
                case KEY_KNOB_TURNED:
                    s_temp = get_knob_rotate();
                    if (s_temp > 0)
                      {
                        u_temp = scan_loco_database_forward(nnnext);
                        ppprev = pprev;
                        pprev = prev;
                        prev = curr;
                        curr = next;
                        next = nnext;
                        nnext = nnnext;
                        nnnext = u_temp;
                        display_all_lok();
                      }
          			    if (s_temp < 0)
                      {
              	        u_temp = scan_loco_database_revers(ppprev);
                        nnnext = nnext;
                        nnext = next;
                        next = curr;
                        curr = prev;
                        prev = pprev;
                        pprev = ppprev;
                        ppprev = u_temp;
                        display_all_lok();
                      }
                    // display_scroll_balken(); passt leider nicht mehr
                    break;

                case KEY_A:         // no softkeys  
                    break;
                case KEY_B:
                    break;
                case KEY_C:
                    break;
                case KEY_D:         
                    break;
              }
            break;
        case MENU_CLOSE:
            turn_led_off(LED_A);
            switch (par.key)
              {
                default:
                    break;
                case KEY_CENTER:
                case KEY_KNOB:          // pick lok and exit
                    actual_database_index = curr;
                    get_loco_data(actual_database_index, &actual_lok);
                    
                    //!warning! kw:debug xp_send_askforlocinfo( actual_lok.w.addr);
                    //!warning!
                    
                    current_loc.speed = 0;
                    current_loc.dir = 0;
                    
                    current_loc.format = actual_lok.w.format;  // just copy (tmp)
                    break;
              }
            break;
      }
  }

//----------------------------------------------------------------------------
//
// Menu-Entry (move inside the loklist)
//
//----------------------------------------------------------------------------

void action_lokmove(t_navicode par)
  {
    signed char s_temp;
    unsigned char u_temp;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();
            
            scrollmode = 0;
            display_scrollmode();
           
            get_database_indeces();        // ppprev ... curr ... nnnext

            display_all_lok();
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:          // pick loco and return
                  break;
                case KEY_SH_KNOB:          // change focus revers
                  break;
                case KEY_KNOB_TURNED:
                    s_temp = get_knob_rotate();
                    if (s_temp > 0)
                      {
                        u_temp = scan_loco_database_forward(nnnext);
                        ppprev = pprev;
                        pprev = prev;
                        prev = next;
                        next = nnext;
                        nnext = nnnext;
                        if (u_temp == curr)     // oops, my old position
                          {
                            u_temp = scan_loco_database_forward(u_temp);
                          }
                        nnnext = u_temp;
                        display_all_lok();
                      }
                    if (s_temp < 0)
                      {
                        u_temp = scan_loco_database_revers(ppprev);
                        nnnext = nnext;
                        nnext = next;
                        next = prev;
                        prev = pprev;
                        pprev = ppprev;
                        if (u_temp == curr)
                          {
                            u_temp = scan_loco_database_revers(u_temp);
                          }
                        ppprev = u_temp;
                        display_all_lok();
                      }
                    // display_scroll_balken(); passt leider nicht mehr
                    // ev. bei graphik display
                    break;

                case KEY_CENTER:
                    break;
                case KEY_A:      
                    break;
                case KEY_B:
                    break;
                case KEY_C:
                    break;
                case KEY_D:         // name predef
                    break;
              }
            break;
        case MENU_CLOSE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_CENTER:
                case KEY_KNOB:                      // move lok and exit
                    move_loco_data(prev, curr);     // insert here;
                    actual_database_index = curr;   
                    break;
              }
            break;
      }
  }


//----------------------------------------------------------------------------
//
// Menu-Entry for receive Lok List
//
//----------------------------------------------------------------------------

static char rec_enabled;

//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   01234567890123456789
//  |--------------------|
// 0|List (view)    STOP |
// 1|Receive Loco Names  |
// 2|(12) 1234:Dampf12345|
// 3|    exit            |
//  |--------------------|

static char text_what[]           PROGMEM = "Receive Loco Names";
static char text_query[]          PROGMEM = "Sure?";
static char text_receiving[]      PROGMEM = "Receiving ...";
static char text_sk_yes[]         PROGMEM = "yes";
static char text_sk_no[]          PROGMEM = "no";
static char text_sk_blank[]       PROGMEM = " ";
static char text_sk_exit[]        PROGMEM = "exit";


void display_rec_item(void)
  {
    set_cursor(0,2);
    lcd_putc('(');
    lcd_int2(new_database_entry.idx + 1);
    // lcd_putc('/');
    // lcd_int2(new_database_entry.size);
    lcd_putc(')');
    lcd_int4(new_database_entry.addr);
    lcd_putc(':');
    display_loc_name(new_database_entry.name);   
  }

static void display_rec_mode(void)
  {
    set_cursor(5,0);
    lcd_string_P(text_recv);
  }

void display_rec_query(void)
  {
    set_cursor(0,1);
    lcd_string_P(text_what); 
    set_cursor(0,2);
    lcd_string_P(text_query); 
    display_softkey_P(0, text_sk_yes);
    display_softkey_P(1, text_sk_no);

  }

void display_rec_accepted(void)
  {
    set_cursor(0,2);
    lcd_string_P(text_receiving);
    display_softkey_P(0, text_sk_blank);
    display_softkey_P(1, text_sk_exit);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

//   01234567890123456789
//  |--------------------|
// 0|List (view)    STOP |
// 1|Receive Loco Names  |
// 2|(12) 1234:Dampf12345|
// 3|    exit            |
//  |--------------------|

static const char text_what[]           PROGMEM = "Receive Loco Names";
static const char text_query[]          PROGMEM = "Sure?";
static const char text_receiving[]      PROGMEM = "Receiving ...";
static const char text_sk_yes[]         PROGMEM = "yes";
static const char text_sk_no[]          PROGMEM = "no";
static const char text_sk_blank[]       PROGMEM = " ";
static const char text_sk_exit[]        PROGMEM = "exit";


void display_rec_item(void)
  {
    set_cursor(0,3);
    lcd_putc('(');
    lcd_int2(new_database_entry.idx + 1);
    // lcd_putc('/');
    // lcd_int2(new_database_entry.size);
    lcd_putc(')');
    lcd_int4(new_database_entry.addr);
    lcd_putc(':');
    display_loc_name(new_database_entry.name);   
  }

static void display_rec_mode(void)
  {
    set_cursor(5,0);
    lcd_string_P(text_recv);
  }

void display_rec_query(void)
  {
    set_cursor(0,2);
    lcd_string_P(text_what); 
    set_cursor(0,3);
    lcd_string_P(text_query); 
    display_softkey_P(0, text_sk_yes);
    display_softkey_P(1, text_sk_no);
  }

void display_rec_accepted(void)
  {
    set_cursor(0,3);
    lcd_string_P(text_receiving);
    display_softkey_P(0, text_sk_blank);
    display_softkey_P(1, text_sk_exit);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY

//----------------------------------------------------------------------------
//
// Menu-Entry (receive loklist)
//
//----------------------------------------------------------------------------
void action_lokreceive(t_navicode par)
  {
    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();
            display_rec_mode();
            
            turn_led_green_soft(LED_A);
            turn_led_red_soft(LED_B);
            
            display_rec_query();
            rec_enabled = 0;
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:          // pick lok and retur
                  break;
                case KEY_SH_KNOB:          // change focus revers
                  break;
                case KEY_PARSER_DATABASE:
                    if (rec_enabled)
                      {
                        display_rec_item();
                        if (!store_loco_name(new_database_entry.addr, new_database_entry.name))
                          {
                            display_error(ERR_DATABASE_FULL);    // failed, we are full!
                          }
                      }
                    break;

                case KEY_CENTER:
                    break;
                case KEY_A:
                     if (!rec_enabled)
                      {
                        turn_led_off_soft(LED_A);
                        turn_led_green_soft(LED_B);
                        actual_database_index = 0;      // clear global pointers to database 
                        //!warning! clear_loco_database();
                        rec_enabled = 1;
                        display_rec_accepted(); 
                      }
                    break;
                case KEY_B:         // exit (handled by upper layer)
                    break;
                case KEY_C:
                    break;
                case KEY_D:         // name predef
                    break;
              }
            break;
        case MENU_CLOSE:
            turn_led_off(LED_A);
            turn_led_off(LED_B);
            switch (par.key)
              {
                default:
                    break;
                case KEY_B:         // exit (handled by upper layer)
                    break;
                case KEY_CENTER:
                case KEY_KNOB:
                     
                    break;
              }
            break;
      }
  }



