
//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_mode.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-03-24 V0.02 kw  conditional assembly for displays
//            2009-03-30 V0.03 kw  lcd_reset after power down
//            2009-07-12 V0.04 kw  XBEE Mode; blocking of TASK_PARSER
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for mode select and power off
//
//------------------------------------------------------------------------
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"
#include "serial.h"                 // select consumer

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"

#include "xbee_if.h"

#include "menu_navigator.h"

            


//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   01234567890123456789
//  |--------------------|
// 0|Mode           STOP |
// 1|                    |
// 2| Operation Mode     |
// 3| XP   XBEE  RR  OFF |
//  |--------------------|


char text_mode[]  PROGMEM = "Mode";
char text_mode2[]  PROGMEM = "Operation Mode";
char text_sk_mode_xp[]  PROGMEM = "XP";
char text_sk_mode_wrp[]  PROGMEM = "WRP";
char text_sk_mode_pc[]  PROGMEM = "PC";
char text_sk_mode_off[]  PROGMEM = "OFF";


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_mode);
    check_status_display(0);

    set_cursor(1,2);
    lcd_string_P(text_mode2);
    display_softkey_P(0, text_sk_mode_xp);
    display_softkey_P(1, text_sk_mode_wrp);
    display_softkey_P(2, text_sk_mode_pc);
    display_softkey_P(3, text_sk_mode_off);

    cursor_off();
  }

char text_connect_xp[]  PROGMEM = "XP ...";
char text_connect_wrp[]  PROGMEM = "WRP...";

static void display_connecting_xp(void)
  {
    set_cursor(1,2);
    lcd_string_P(text_connect_xp);
  }

static void display_connecting_wrp(void)
  {
    set_cursor(1,2);
    lcd_string_P(text_connect_wrp);
  }

static char text_xbee_ok[]          PROGMEM = "OK";
static char text_xbee_netscan[]     PROGMEM = "net scan";
static char text_xbee_nopan[]       PROGMEM = "no net";
static char text_xbee_novalidpan[]  PROGMEM = "net void";
static char text_xbee_nojoin[]      PROGMEM = "no join";
static char text_xbee_router_full[] PROGMEM = "net full";


static void display_xbee_ai(void)
  {
    set_cursor(7,2);
    lcd_blank(16);
    set_cursor(7,2);
    switch (xbee_ai_state)
      {
        default:
            lcd_putc('?');
            lcd_hex2(xbee_ai_state);
            break;
        case XBEE_AI_OKAY:
            lcd_string_P(text_xbee_ok);
            break;
        case XBEE_AI_NET_SCAN:
            lcd_string_P(text_xbee_netscan);
            break;
        case XBEE_AI_NO_PANs:
            lcd_string_P(text_xbee_nopan);
            break;
        case XBEE_AI_NO_VALID_PANs:
            lcd_string_P(text_xbee_novalidpan);
            break;
        case XBEE_AI_ROUTER_FULL_R:
        case XBEE_AI_ROUTER_FULL_E:
            lcd_string_P(text_xbee_router_full);
            break;
        case XBEE_AI_NO_JOIN:
        case XBEE_AI_JOIN_FAILED:
        case XBEE_AI_NO_RESPONSE:
            lcd_string_P(text_xbee_nojoin);
            break;
      }
  }


//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

//   01234567890123456789
//  |--------------------|
// 0|Mode           STOP |
// 1|                    |
// 2| Operation Mode     |
// 3| XPress WRP PC  OFF |
//  |--------------------|


const char text_mode[]  PROGMEM = "Mode";
const char text_mode2[]  PROGMEM = "Operation Mode";
const char text_sk_mode_xp[]  PROGMEM = "XP";
const char text_sk_mode_wrp[]  PROGMEM = "WRP";
const char text_sk_mode_pc[]  PROGMEM = "PC";
const char text_sk_mode_pcx[]  PROGMEM = "PCX";
const char text_sk_mode_off[]  PROGMEM = "OFF";


static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_mode);
    check_status_display(0);

    set_cursor(1,2);
    lcd_string_P(text_mode2);
    if (a58_check())
      {
        display_softkey_P(0, text_sk_mode_xp);
        display_softkey_P(1, text_sk_mode_wrp);
        display_softkey_P(2, text_sk_mode_pc);
        display_softkey_P(3, text_sk_mode_off);
      }
    else
      { // Falko spezial-schnitz
        display_softkey_P(0, text_sk_mode_xp);
      }
    cursor_off();
  }


char const text_connect_xp[]  PROGMEM = "connecting XP ...";
char const text_connect_wrp[]  PROGMEM = "connecting WRP...";

static void display_connecting_xp(void)
  {
    set_cursor(1,3);
    lcd_string_P(text_connect_xp);
  }

static void display_connecting_wrp(void)
  {
    set_cursor(1,3);
    lcd_string_P(text_connect_wrp);
  }

static const char text_xbee_ai[]  PROGMEM = "AIstat: ";
static const char text_xbee_ok[]          PROGMEM = "OK";
static const char text_xbee_netscan[]     PROGMEM = "net scan";
static const char text_xbee_nopan[]       PROGMEM = "no net";
static const char text_xbee_novalidpan[]  PROGMEM = "net void";
static const char text_xbee_nojoin[]      PROGMEM = "no join";
static const char text_xbee_router_full[] PROGMEM = "net full";


static void display_xbee_ai(void)
  {
    set_cursor(1,4);
    lcd_blank(16);
    set_cursor(1,4);
    lcd_string_P(text_xbee_ai);
    switch (xbee_ai_state)
      {
        default:
            lcd_putc('?');
            lcd_hex2(xbee_ai_state);
            break;
        case XBEE_AI_OKAY:
            lcd_string_P(text_xbee_ok);  // we are connected to vaild net
            break;
        case XBEE_AI_NET_SCAN:
            lcd_string_P(text_xbee_netscan);
            break;
        case XBEE_AI_NO_PANs:
            lcd_string_P(text_xbee_nopan);
            break;
        case XBEE_AI_NO_VALID_PANs:
            lcd_string_P(text_xbee_novalidpan);
            break;
        case XBEE_AI_ROUTER_FULL_R:
        case XBEE_AI_ROUTER_FULL_E:
            lcd_string_P(text_xbee_router_full);
            break;
        case XBEE_AI_NO_JOIN:
        case XBEE_AI_JOIN_FAILED:
        case XBEE_AI_NO_RESPONSE:
            lcd_string_P(text_xbee_nojoin);
            break;
      }
  }

//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler (Menu Mode)
//
//-----------------------------------------------------------------------

static const char text_bootloader[]  PROGMEM = "Bootloader";

extern enum {
	IDLE,
	WAIT_FOR_MODULE,
	WAIT_FOR_ID,
	WAIT_FOR_SH,
	WAIT_FOR_SL,
	WAIT_FOR_HV,
	WAIT_FOR_VR,              // up to here cold boot
	WAIT_FOR_AI,
	WAIT_FOR_CH,
	WAIT_FOR_MY,
	INIT_DONE,
} init_xbee_state;

void action_mode(t_navicode par)
  {
    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();
            turn_led_green_soft(LED_A);
            if (a58_check())
              {
                turn_led_green_soft(LED_B);
                turn_led_red_soft(LED_D);
              }
            else
              { // Falko spezial-schnitz
              }
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_SHIFT:
                    display_softkey_P(2, text_sk_mode_pcx);
                    break;
                case KEY_SHIFT_RELEASE:
                    display_softkey_P(2, text_sk_mode_pc);
                    break;
                case KEY_KNOB:
                    break;
                case KEY_KNOB_TURNED:
                    break;

                case KEY_A:             // enter Xpressnet Mode
                    operation_mode = OPM_THROTTLE;
                    msg_start_interface(USE_XPRESSNET);         // when connected: KEY_CONNECT will occur, thus leaving the menu
                    ENABLE_USB;
                    serial_set_consumer_task(TASK_PARSER);
                    display_connecting_xp();
                    break;
                case KEY_SH_C:                              // enter Special-Remote-Operation with XBEE
                    // no break;                            // 1. we will get a connect (pushing us to MENU_LOK)
                                                            // 2. we will get a WRP-DEBUG-ANNOUNCE (pushing us to MENU_DEBUG)
                case KEY_B: 
                    operation_mode = OPM_THROTTLE;
                    ENABLE_XBEE;                             // enter XBEE Mode - turn off USB!
                    used_interface = USE_XBEE;
                    serial_set_consumer_task(TASK_XBEE);
                    
                    if (init_xbee_state == IDLE) set_task_ready(TASK_INIT_XBEE);   // hk:  idle
					else msg_start_interface(USE_XBEE);				// anything but idle, should restart XBEE
						
                    display_connecting_wrp();
                    break;

                case KEY_C:                                 // enter Remote with Xpressnet
                    break;
                
                    break;
                
                case KEY_XBEE_AI:
                    if (used_interface == USE_XBEE)        // nur wenn wir versuchen, mit XBEE zu reden
                      {
                        display_xbee_ai();                 // dann den join status anzeigen
                      }
                    break;

                case KEY_CENTER:
                    break;
                case KEY_D:
                    //!warning! kw: say goodbye and smooth fade out lights
                    lcd_clear();
                    set_cursor(0,0);
                    lcd_string_P(text_bootloader);   // prepare display for next power up :-)
                    _delay_ms(20);  //!!! this is wrong - need scheduler to update display
                    // lcd_reset();
                    cli();
                    TURN_POWER_OFF;
                    _delay_ms(20);
                    sei();
                    display_error(ERR_POWER_OFF);
                    break;
              }
            break;
        case MENU_CLOSE:
            switch (par.key)
              {
                default:
                     break;
                case KEY_SH_A:             // enter Xpressnet Mode
                    operation_mode = OPM_THROTTLE;
                    msg_start_interface(USE_XPRESSNET);
                    ENABLE_USB;
                    serial_set_consumer_task(TASK_PARSER);
                    break;
                case KEY_SH_B:             // enter XBEE Mode - turn off USB!
                    operation_mode = OPM_THROTTLE;
                    ENABLE_XBEE;
                    serial_set_consumer_task(TASK_XBEE);
                    msg_start_interface(USE_XBEE);
                    break;
                case KEY_C:             // enter Xpressnet Tunnel Mode
                                        // or enter emulation mode for pc throttle
                    ENABLE_USB;
                    operation_mode = OPM_PCMODE;
                    break;
                case KEY_SH_D:          // enter BL
                    ENABLE_USB;
                    cli();
                    asm volatile ("jmp 0x1F000");
              }
            turn_led_off(LED_D);
            turn_led_off(LED_B);
            turn_led_off(LED_A);
            break;
      }
  }




