//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_navigator.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-03-13 V0.02 kw  added config command station
//            2009-03-20 V0.03 kw  added time stamp on lastkey
//            2009-09-24 V0.04 kw  conditional assembly for displays
//            2010-06-16 V1.00 kw  new arranged, clock added
//            2013-10-11 V1.01 kw  added turntable from kurscheidt
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   a real tiny navigation system
//
//========================================================================
// upstream interface (howto):
//
// to use the menu system:
//   a) run_navigator(); this checks for a new keystroke
//      evalutes hotkeys and navigation keys.
//      other keys are forwarded to the individual screens
//   b) navigation is defined in struct s_Menu
//
//========================================================================
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>
#include <stdio.h>
#include <util/delay.h>

#include "config.h"
#include "cortos.h"                 // real time kernel
#include "keycodes.h"
#include "keyboard.h"
#include "timer.h"
#include "database.h"
#include "messages_objects.h"
#include "display.h"


#include "menu_navigator.h"
#include "menu_lok.h"               // init for lok

#include "lcd_text.h"


//========================================================================
// This is the central menu structure
struct s_Menu 
  {
    Ptr_To_Function action; // a pointer the screen function,
                            // call parameter: t_navicode;
                            //                 this includes a opcode (OPEN, CLOSE, UPDATE)
                            //                 and the keystroke
    t_navigate *navigate;   // a pointer to a list of keystrokes and target screens
  };

//
// each screen must have a function where it handles keystrokes
// we repeat here the 'headers'.
void action_lok(t_navicode par);
void action_weiche(t_navicode par);
void action_extacc(t_navicode par);
void action_cv(t_navicode par);
void action_cs_cfg(t_navicode par);
void action_setup(t_navicode par);
void action_clock(t_navicode par);
void action_lokedit(t_navicode par);
void action_loklist(t_navicode par);
void action_lokmove(t_navicode par);
void action_lokreceive(t_navicode par);
void action_feedback(t_navicode par);
void action_mode(t_navicode par);
void action_sysstat(t_navicode par);
void action_remote(t_navicode par);
void action_about(t_navicode par);
void action_cv29(t_navicode par);
void action_cv_addr(t_navicode par);
#if (DREHSCHEIBE == 1) 
 void action_dreh(t_navicode par);
#endif

t_navigate navigate_mode[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },      // reenter mode - we need this to reinit at power up
    { KEY_SH_A,     SCREEN_LOK },       // Xpressnet (forced, without connect)
    { KEY_SH_B,     SCREEN_LOK },       // XBEE      (forced, without connect)
    { KEY_C,        SCREEN_REMOTE },    // Remote Mode with tunnel on Xpressnet (rocrail)
    //{ KEY_SH_C,     SCREEN_TC },      // Remote Mode with XBEE (tc)
                                        // KEY_D: turn off
    { KEY_CONNECT,  SCREEN_LOK },       // when connected - leave mode
    { KEY_CENTER,   SCREEN_ABOUT },     // show about  
    { KEY_SH_CENTER,SCREEN_LOK },       // simple return
    { 0, 0 }, };// end of list
    
t_navigate navigate_about[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_CENTER,   SCREEN_MODE },      
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_RIGHT,    SCREEN_LOK },
    { 0, 0 }, };// end of list

t_navigate navigate_lok[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_RIGHT,    SCREEN_WEICHE },
    { KEY_LEFT,     SCREEN_CV },
    { KEY_SH_CENTER,SCREEN_LOKLIST },
    { 0, 0 }, };// end of list
    
t_navigate navigate_weiche[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_RIGHT,    SCREEN_EXTACC },     // in prog - first select addr
    { KEY_LEFT,     SCREEN_LOK },
    #if (DREHSCHEIBE == 1) 
    { KEY_CENTER,   SCREEN_DREH },       // turntable menu
    #else
    { KEY_CENTER,   SCREEN_CV },         // prog menu
    #endif
     { 0, 0 }, };// end of list

t_navigate navigate_signal[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_RIGHT,    SCREEN_LOK },
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_CENTER,   SCREEN_CV },         // prog menu
    { 0, 0 }, };// end of list

t_navigate navigate_cv29[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_LEFT,     SCREEN_SETUP },
    { KEY_RIGHT,    SCREEN_LOK }, 
    { KEY_SH_CENTER,SCREEN_CS_CFG },
    { KEY_UP,       SCREEN_CV },
    { KEY_DOWN,     SCREEN_CV_ADDR },
    { 0, 0 }, };// end of list

t_navigate navigate_cv[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_LEFT,     SCREEN_SETUP },
    { KEY_RIGHT,    SCREEN_LOK }, 
    { KEY_SH_CENTER,SCREEN_CS_CFG },
    { KEY_UP,       SCREEN_CV_ADDR },
    { KEY_DOWN,     SCREEN_CV29 },
    { 0, 0 }, };// end of list

t_navigate navigate_cv_addr[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_LEFT,     SCREEN_SETUP },
    { KEY_RIGHT,    SCREEN_LOK }, 
    { KEY_SH_CENTER,SCREEN_CS_CFG },
    { KEY_UP,       SCREEN_CV29 },
    { KEY_DOWN,     SCREEN_CV },
    { 0, 0 }, };// end of list

t_navigate navigate_setup[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_SH_CENTER,SCREEN_SYSSTAT },    // debug screen
    { KEY_CENTER,   SCREEN_CLOCK },      
    { KEY_LEFT,     SCREEN_FEEDBACK },
    { KEY_RIGHT,    SCREEN_LOK },
    { 0, 0 }, };// end of list

t_navigate navigate_clock[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_SH_CENTER,SCREEN_SYSSTAT },    // debug screen
    { KEY_CENTER,   SCREEN_SETUP },
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_RIGHT,    SCREEN_LOK },
    { 0, 0 }, };// end of list

t_navigate navigate_feedback[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_RIGHT,    SCREEN_LOK },
    { KEY_LEFT,     SCREEN_LOK },
    { 0, 0 }, };// end of list

t_navigate navigate_lokedit[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_CENTER,   SCREEN_LOK },
    { KEY_SH_CENTER,SCREEN_LOKLIST },
    { KEY_RIGHT,    SCREEN_LOK },
    { KEY_LEFT,     SCREEN_LOK },
    { 0, 0 }, };// end of list

t_navigate navigate_loklist[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_CENTER,   SCREEN_LOK },       // pick lok
    { KEY_KNOB,     SCREEN_LOK },       // pick lok
    { KEY_SH_KNOB,  SCREEN_LOKMOVE },
    { KEY_SH_CENTER,SCREEN_LOKEDIT },
    { KEY_RIGHT,    SCREEN_LOK },       // break
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_A,        SCREEN_LOKRECV },
    { 0, 0 }, };// end of list

t_navigate navigate_lokmove[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_CENTER,   SCREEN_LOK },       // pick lok
    { KEY_KNOB,     SCREEN_LOKLIST },   // pick lok
    { KEY_SH_CENTER,SCREEN_LOKEDIT },
    { KEY_RIGHT,    SCREEN_LOK },
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_A,        SCREEN_LOKRECV },
    { 0, 0 }, };// end of list

t_navigate navigate_lokreceive[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
    { KEY_RIGHT,    SCREEN_LOK },
    { KEY_LEFT,     SCREEN_LOK },
    { KEY_CENTER,   SCREEN_LOKLIST },   // pick lok
    { KEY_B,        SCREEN_LOKLIST },   // pick lok
    { 0, 0 }, };// end of list


t_navigate navigate_remote[] = 
  { { KEY_SH_STOP,  SCREEN_MODE },
    { 0, 0 }, };// end of list
    
#if (DREHSCHEIBE == 1) 
 t_navigate navigate_dreh[] =
  { { KEY_SH_STOP,  SCREEN_MODE },
       { KEY_RIGHT,    SCREEN_LOK },    // always back to lok
       { KEY_LEFT,     SCREEN_LOK },
    { 0, 0 }, };// end of list 
#endif


struct s_Menu Menu[] =
  { [SCREEN_MODE] =     {action_mode, navigate_mode},
    [SCREEN_LOK] =      {action_lok, navigate_lok},
    [SCREEN_WEICHE] =   {action_weiche, navigate_weiche},
    [SCREEN_EXTACC] =   {action_extacc, navigate_signal},
    [SCREEN_CV] =       {action_cv, navigate_cv},
    [SCREEN_CV29] =     {action_cv29, navigate_cv29},
    [SCREEN_CV_ADDR] =  {action_cv_addr, navigate_cv_addr},
    [SCREEN_CS_CFG] =   {action_cs_cfg, navigate_cv},       // use same navigation
    [SCREEN_SETUP] =    {action_setup, navigate_setup},
    [SCREEN_CLOCK] =    {action_clock, navigate_clock},
    [SCREEN_SYSSTAT] =  {action_sysstat, navigate_setup},   // use same navigation
    [SCREEN_ABOUT] =    {action_about, navigate_about},
    [SCREEN_LOKEDIT] =  {action_lokedit, navigate_lokedit},
    [SCREEN_LOKLIST] =  {action_loklist, navigate_loklist},
    [SCREEN_LOKMOVE] =  {action_lokmove, navigate_lokmove},
    [SCREEN_LOKRECV] =  {action_lokreceive, navigate_lokreceive},
    [SCREEN_FEEDBACK] = {action_feedback, navigate_feedback},
    [SCREEN_REMOTE] =   {action_remote, navigate_remote}, 
#if (DREHSCHEIBE == 1) 
    [SCREEN_DREH] =     {action_dreh, navigate_dreh},
#endif
     };


//------------------------------------------------------------------------------
//
//  Menu Handler
//
//------------------------------------------------------------------------------
//
// t_cr_task
signed int run_navigator(void)
  {
    t_navigate *actual_navi;
    t_navicode temp;

    if (keystroke_present())
      {
        temp.key = get_wait_keystroke();

        //!warning! kw:  // PROG : someone else is programming, lock the interface
        //if( power_status == PROG ) {
        //  display_programming();
        //}
          {
            if (temp.key == KEY_STOP)
              { // Hotkey = Sonderfunktion
                switch (command_station_status)
                  {
                    case STS_ON:
                        switch (STOP_action)
                          {
                            default:  // sizeof_t_STOP_action
                            case MAKES_OFF:
                                msg_send_state(STS_OFF);    // if ON - turn it OFF
                                msg_transmit();
                                break;
                            case MAKES_STOP:
                                msg_send_state(STS_STOP);
                                msg_transmit();
                                break;
                            case MAKES_STOP_MY_LOCOS:
                                msg_send_all_loc_stop();
                                msg_transmit();
                                break;                                              
                          }
                        break;
                    case STS_OFF:
                        msg_send_state(STS_ON);
                        msg_transmit();
                        break;
                    case STS_STOP:
                        msg_send_state(STS_ON);
                        msg_transmit();
                        break;
                    case STS_PROG:
                        msg_send_cs_status_request();
                        msg_transmit();
                        break;
                    case STS_PBUSY:
                    case STS_BUSY:
                        msg_send_cs_status_request();
                        msg_transmit();
                        break;
                    case STS_DISCONNECTED:
                        msg_send_cs_status_request();
                        msg_transmit();
                        break;
                    case STS_CMD_ERROR:                 // there has been a command error
                        msg_send_cs_status_request();
                        msg_transmit();
                        break;
                  }
                return(0);    // ready again
              }

            // now evaluate key and scan for navigate
            actual_navi = &Menu[actual_screen].navigate[0];

            int i = 0;
            while (actual_navi->key != 0)
              {
                if (actual_navi->key == temp.key)
                  {
                    // navigation key found, lets go away
                    temp.menucode = MENU_CLOSE;
                    Menu[actual_screen].action(temp);
                    // this is our new screen
                    actual_screen = actual_navi->index_of_target_screen;
                    temp.menucode = MENU_OPEN;
                    display_time_allowed = 0;
                    Menu[actual_screen].action(temp);
                    enable_display_update();
                    return(0);
                  }
                actual_navi = &Menu[actual_screen].navigate[i++];
              }

            // nothing to go away - forward keystroke to action.
            temp.menucode = MENU_UPDATE;
            Menu[actual_screen].action(temp);
            enable_display_update();

            if (temp.key < KEY_PARSER)            // real keystroke
              {
                last_key = temp.key;                // save for history
                last_key_time = get_systick_10ms();
              }
          }
        return(0);  // we are ready again
      }
    else return(-1);  // lets fall asleep
  }


void init_navigator(void)
  {
    last_key = 0;
    last_key_time = get_systick_10ms();

    STOP_action = eeprom_read_byte((void *)eadr_STOP_action);
    cr_fifo_set_comsumer (&key_fifo, TASK_NAVIGATOR);    // hook up to key_fifo!

    get_loco_data(actual_database_index, &actual_lok);  // preload actual_lok

    actual_screen = SCREEN_MODE;                        // Display content is still either splash or void
  }


//------------------------------------------------------------------------------
//
//  Menu for direct interaction with a PC program (like rocrail)
//
//------------------------------------------------------------------------------
//

void perform_pc_command(void)
  {
    switch (pc_message[2])
      {
        case 0x00:
            // 0x33 0x?? 0x00 0x00 leave tunnel mode
            // 0x33 0x?? 0x00 0x01 enter tunnel mode
            if (pc_message[3] == 0)         // if 0: leave rocrailmode
              {
                cr_fifo_put(&key_fifo, KEY_SH_STOP);     // simulate a key stroke for Mode-Menu
                cr_fifo_put(&key_fifo, KEY_A);           // simulate a key stroke for Mode XP
              }        
            if (pc_message[3] == 1)         // if 1: enter rocrailmode
              {
                cr_fifo_put(&key_fifo, KEY_SH_STOP);     // simulate a key stroke for Mode-Menu
                cr_fifo_put(&key_fifo, KEY_B);           // simulate a key stroke for Mode RR
              }        
            break;
        case 0x01: // productID and serial
            msg_send_PID();
            msg_transmit();
            break;
        case 0x10: // Display String (x,y,string)
              {
                unsigned char i;
                unsigned char len = pc_message[0] & 0x0F;

                if( pc_message[5] == '$')
                  { // clc
                    lcd_clear();
                  }
                else
                  {
                    set_cursor(pc_message[3],pc_message[4]);
                    for( i = 5; i < len+1; i++)
                        lcd_putc(pc_message[i]);
                  }
              }
            break;
        case 0x11: // Display String but answer
            break;
        case 0x12: // Display String blinking
            break;
        case 0x20: // LED command
            break;
      }
  }



static char const text_menuhead_rr[]  PROGMEM = "Connecting to PC ...";
static char const text_menuhead_id[]  PROGMEM = "ID: ";

void action_remote(t_navicode par)
  {
    int temp;
    check_status_display(par.key);
    switch (par.menucode)
      {
        case MENU_OPEN:
            lcd_clear();
            set_cursor(1,2);
            lcd_string_P(text_menuhead_rr);
            set_cursor(1,3);
            lcd_string_P(text_menuhead_id);
            lcd_int02(eeprom_read_byte((void *)eadr_pc_tid));

         
            msg_send_pc_hello();
            msg_transmit();

            // check_status_display(0);
            
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    if (par.key <= KEY_LAST_REAL_KEY)
                      {
                        msg_send_pc_key(par.key, 0);
                      }
                case KEY_KNOB_TURNED:
                    temp = get_knob_rotate();
                    if (temp > 127) temp = 127;
                    if (temp < -127) temp = -127;
                    msg_send_pc_key(par.key, temp);
                    break;
                case KEY_PARSER_REMOTE:
                    perform_pc_command();
                    break;
              }
            break;
        case MENU_CLOSE:
            msg_send_pc_goodbye();
            msg_transmit();
            break;
      }
  }

