//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_prog.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-02-22 V0.02 kw  added list for cvnames
//            2009-02-28 V0.03 kw  cursor after init, bugfix cvnames
//            2009-03-09 V0.04 kw  white blocks fixed
//            2009-08-07 V0.05 kw  different cursor positions for char and grahpic
//            2010-03-31 V0.06 kw  added PoM read
//            2010-11-04 V0.06 kw  added current_cv.addr
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for programming
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>


#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "lcd_ks0073_charset.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "programmer.h"
#include "menu_navigator.h"
#include "menu_lok.h"



#if (DISPLAY == LCD_EA204)
  #define PROG_CUR_Y    1
#elif  (DISPLAY == LCD_EA128x64)
  #define PROG_CUR_Y    2
#else
  #warning DISPLAY not defined

#endif

//   01234567890123456789
//  |--------------------|
// 0|Prog ( OK )    STOP |
// 1|CV      addr val=$$ |
// 2|cv-name 1024 255=FF |
// 3|Read Writ RPoM WPoM |
//  |--------------------|

const char text_prog[]  PROGMEM = "Prog";

static const char text_header[]    PROGMEM = {' ','C','V',' ',' ',' ',' ',' ','a','d','d','r',' ','v','a','l','=',C_DOLLAR,C_DOLLAR,0};

static const char text_sk_read[]  PROGMEM = "Read";
static const char text_sk_write[]  PROGMEM = "Writ";
static const char text_sk_RPoM[]  PROGMEM = "RPoM";
static const char text_sk_WPoM[]  PROGMEM = "WPoM";
//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------




//------------------------------------------------------------------------
// routines for menu_prog
//------------------------------------------------------------------------


#define PRG_TXT_SIZE 6

struct s_PRG_text
  {
    char name[PRG_TXT_SIZE+1];
  };

const struct s_PRG_text text_prg[] PROGMEM =
  {
    [PRG_CV_VOID] =            { "( -- )"},
    [PRG_CV_WAIT_FOR_ANSWER] = { "(wait)"},
    [PRG_CV_SHORT]           = { "(shrt)"},
    [PRG_CV_NOLOK]           = { "(open)"},
    [PRG_CV_OKAY]            = { "( OK )"},
    [PRG_CV_CHANGED]         = { "(new )"},
  };


void display_prog_state (t_prog_result result)
  {
    set_cursor(5,0);
    lcd_string_P(text_prg[result].name);     
  }

void display_cv_value(void)
  {
	if (current_CV.result >= PRG_CV_OKAY)
      {
        set_cursor(13,PROG_CUR_Y+1);
    	lcd_int3( current_CV.value);
        lcd_putc('=');
        lcd_hex2( current_CV.value);
      }
    else
      {
        set_cursor(13,PROG_CUR_Y+1);
    	lcd_putc_cnt(3,'?');
        lcd_putc('=');
        lcd_putc_cnt(2,'?');
      }  
  }


#define CV_NAME_SIZE 7

struct s_CVName
  {
    unsigned int value;
    char name[CV_NAME_SIZE+1];
  };

const struct s_CVName cvNames[] PROGMEM =
  {
    {   1,"Addr   "},
    {   2,"v(min) "},
    {   3,"(+)acc "},
    {   4,"(-)acc "},
    {   5,"v(max) "},
    {   6,"v(mid) "},
    {   7,"Version"},
    {   8,"Vendor "},
    {  17,"LAddrH "},
    {  18,"LAddrL "},
    {  19,"Consist"},
    {  20,"Cst:1-8"},
    {  21,"Cst:9.."},
    {  28,"BiDi   "},
    {  29,"config "},
    {  30,"Errors "},
    {  33,"F0f-map"},
    {  34,"F0r-map"},
    {  35,"F1-map "},
    {  36,"F2-map "},
    {  37,"F3-map "},
    {  38,"F4-map "},
    {  39,"F5-map "},
    {  40,"F6-map "},
    {  41,"F7-map "},
    {  42,"F8-map "},
    {  65,"Kick   "},
    {  67,"Curve01"},
    {  68,"Curve02"},
    {  69,"Curve03"},
    {  70,"Curve04"},
    {  71,"Curve05"},
    {  72,"Curve06"},
    {  76,"Curve10"},
    {  81,"Curve15"},
    {  86,"Curve20"},
    {  91,"Curve25"},
    {  92,"Curve26"},
    {  93,"Curve27"},
    {  94,"Curve28"},
  };



void display_cv_number( unsigned int cvnumber)
  {
    unsigned char i, found=0;

	set_cursor(1,PROG_CUR_Y+1);
    for (i=0; i < sizeof(cvNames) / sizeof(cvNames[0]); i++)
      {
        if ( cvnumber == pgm_read_word(&cvNames[i].value) )
          {
            found = 1;
            lcd_string_P(cvNames[i].name);
            break;
          }
      }
    if (!found) lcd_blank(CV_NAME_SIZE);   // strlen_P(cvNames[0].name)

	set_cursor(8,PROG_CUR_Y+1);
    lcd_int4( cvnumber);
  }

//------------------------------------------------------------------------------

typedef enum 
  {
    EDIT_ADDR,
    EDIT_DATA,
    N_of_EDIT
  } t_edit_focus;

static t_edit_focus edit_focus;

static void set_edit_cursor(void)
  {
    switch(edit_focus)
      {
        default:
        case N_of_EDIT:       break;
        case EDIT_ADDR:
            cursor_on_at( 8,PROG_CUR_Y+1,4); break;
        case EDIT_DATA:
            cursor_on_at(13,PROG_CUR_Y+1,3); break;            
      }
  }

static void move_edit_pointer_forward(void)
  {
    edit_focus++;
    if (edit_focus == N_of_EDIT) edit_focus = 0;
    set_edit_cursor();
  }

static void move_edit_pointer_revers(void)
  {
    if (edit_focus == 0) edit_focus = N_of_EDIT;
    edit_focus--;
    set_edit_cursor();
  }


void action_cv(t_navicode par)
  {
    signed int my_var;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            lcd_clear();
            lcd_string_P(text_prog);
            check_status_display(0);
            
            set_cursor(0,PROG_CUR_Y);
            lcd_string_P(text_header);
            display_softkey_P(0, text_sk_read);
            display_softkey_P(1, text_sk_write);
            display_softkey_P(2, text_sk_RPoM);
            display_softkey_P(3, text_sk_WPoM);
            turn_led_green_soft(LED_A);
            turn_led_red_soft(LED_B);
            turn_led_green_soft(LED_C);
            turn_led_red_soft(LED_D);
            
            edit_focus = EDIT_ADDR;
            current_CV.result = PRG_CV_VOID;  
            current_CV.addr = 0;         
            display_cv_value();
            display_cv_number(current_CV.number);
            set_edit_cursor();
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:                      // change focus forward
                    move_edit_pointer_forward();
                    break;
                case KEY_SH_KNOB:                   // change focus revers
                    move_edit_pointer_revers();
                    break;
                case KEY_KNOB_TURNED:               // cvValue
                    switch(edit_focus)
                      {
                        case EDIT_ADDR:
                            current_CV.result = PRG_CV_VOID;
                            my_var = current_CV.number;
                            my_var += get_knob_rotate();
                            if (my_var < 1) my_var = 1;
                            if (my_var > 1024) my_var = 1024;
                            current_CV.number = (unsigned int) my_var;
                            display_cv_number(current_CV.number);
                            display_cv_value();
                            break;
                        case EDIT_DATA:
                            current_CV.result = PRG_CV_CHANGED;
                            my_var = current_CV.value;
                            my_var += get_knob_rotate();
                            if (my_var < 0) my_var = 0;
                            if (my_var > 255) my_var = 255;
                            current_CV.value = (unsigned char) my_var;
                            display_cv_value();
                            break;
                            break;
                        case N_of_EDIT:
                            break;
                      }
                    set_edit_cursor();                    
                    break;
                //note: up & down are stolen by navigator (can't use them here)
                /*
                case KEY_UP:                        // nextCV
                    if (current_CV.number < 1024) 
                      {
                        current_CV.number++;
                        current_CV.result = PRG_CV_VOID;
                      }
                    display_cv_number(current_CV.number);
                    display_cv_value();
                    set_edit_cursor();
                    break;
                case KEY_DOWN:                      // prevCV
                    if (current_CV.number > 1)
                      {
                        current_CV.number--;
                        current_CV.result = PRG_CV_VOID;
                      }
                    display_cv_number(current_CV.number);
                    display_cv_value();
                    set_edit_cursor();                    
                    break;
                */
                case KEY_A:                    // CV read
                    msg_send_cv_read(current_CV.number);
                    msg_transmit();
                    display_prog_state(current_CV.result);
                    set_edit_cursor();                    
                    break;
                case KEY_B:                    // CV write
                    msg_send_cv_write(current_CV.number, current_CV.value);
                    msg_transmit();
                    display_prog_state(current_CV.result);
                    set_edit_cursor();                    
                    break;
                case KEY_C:                    // PoM read
                    current_CV.addr = actual_lok.w.addr;
                    msg_send_PoM_cv_read(current_CV.addr, current_CV.number);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_D:                    // PoM write
                    current_CV.addr = actual_lok.w.addr;
                    msg_send_PoM_cv_write(current_CV.addr, current_CV.number, current_CV.value);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_SH_A:                    // PoM read accessory
                    // msg_send_cv_read(??? current_CV.number);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_SH_B:                    // PoM write  accessory
                    // msg_send_cv_write(??? current_CV.number, current_CV.value);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_SH_C:                    // PoM read ext accessory
                    // msg_send_cv_read(current_CV.number);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_SH_D:                    // PoM write ext accessory
                    // msg_send_cv_write(current_CV.number, current_CV.value);
                    msg_transmit();
                    set_edit_cursor();                    
                    break;
                case KEY_PARSER_PROG:
                    display_prog_state(current_CV.result);
                    display_cv_value();
                    set_edit_cursor();                    
                    break;
                case KEY_PARSER_STATE:
                    if (current_CV.result == PRG_CV_WAIT_FOR_ANSWER)
                      {
                        display_status(command_station_status);   // make wheelie
                      }
                    else
                      {
                        display_cv_value();
                        display_status(command_station_status);
                      }
                    set_edit_cursor();                    
                    break;
              }
            break;
        case MENU_CLOSE:
            msg_send_cv_end();
            msg_transmit();
            cursor_off();
            turn_led_off_soft(LED_A);
            turn_led_off_soft(LED_B);
            turn_led_off_soft(LED_C);
            turn_led_off_soft(LED_D);
            break;
      }
  }


//====================================================================================
//
// CV29 support
//
//   01234567890123456789
//  |--------------------|
// 0|Prog ( OK )    STOP |
// 1|CV29 (?)   7654 3210|
// 2|0=Dir std. 0011 0011|
// 3|Read Writ RPoM WPoM |
//  |--------------------|

unsigned char my_CV29;
unsigned char my_CV29_valid;

static unsigned char edit_cv29_focus;

const char text_cv29_header[]  PROGMEM = {'C','V','2','9',' ',
                                    '(',' ',')',' ',' ',
                                    ' ',C_NUM_EXP_7,C_NUM_EXP_6,C_NUM_EXP_5,C_NUM_EXP_4,
                                    ' ',C_NUM_EXP_3,C_NUM_EXP_2,C_NUM_EXP_1,C_NUM_EXP_0,0};


#define CV29_HELP_LENGTH  (10+1)
const struct
  {
    char name[CV29_HELP_LENGTH];
  } cv29_helptext[] PROGMEM =
  {
   // 1234567890
    {"0=Dir std."},   // Bit 0
    {"1=Dir rev."},
    {"0=14FS    "},
    {"1=28/128FS"},
    {"0=DCC     "},
    {"1=DC+DCC  "},
    {"0=Bidi off"},
    {"1=BiDi on "},
    {"0=Spd dflt"},
    {"1=Spd curv"},
    {"0=sht.Addr"},
    {"1=lng.Addr"},
    {"0=reserved"},
    {"1=reserved"},
    {"0=LocoDec "},  // Bit 7
    {"1=AccyDec "},
  };

//!warning! CV28
const struct
  {
    char name[CV29_HELP_LENGTH];
  } cv28_helptext[] PROGMEM =
  {
   // 1234567890
    {"0=no send "},   // Bit 0
    {"1=ch1:addr"},
    {"0=no send "},   // Bit 1
    {"1=ch2:data"},
    {"0=no send "},   // Bit 2
    {"1=ch1:ack "},
    {"0=reserved"},   // Bit 3
    {"1=reserved"},
    {"0=reserved"},   // Bit 4
    {"1=reserved"},
    {"0=reserved"},   // Bit 5
    {"1=reserved"},
    {"0=reserved"},   // Bit 6
    {"1=reserved"},
    {"0=reserved"},   // Bit 7
    {"1=reserved"},
  };


void display_cv_valid(void)
  {
    set_cursor(6,PROG_CUR_Y);
    if (my_CV29_valid) lcd_putc('+');
    else lcd_putc('?');
  }


void display_bitwise(unsigned char data)
  {
    unsigned char i, mask;
    
    mask = 0x80;

    set_cursor(11,PROG_CUR_Y+1);
    for (i=0; i< 8; i++)
      {
        if (i==4) lcd_putc(' ');
        if (data & mask) lcd_putc('1');
        else lcd_putc('0');
        mask >>= 1;
      }
  }


void show_cv29_helptext(unsigned char bitpos)
  {
    set_cursor(0,PROG_CUR_Y+1);
    unsigned char index;
    index = bitpos * 2;
    index += (my_CV29 >> bitpos) & 0x01;
    lcd_string_P(cv29_helptext[index].name);
  }

void set_cv29_edit_focus(unsigned char bitpos)
  {
    if (bitpos > 3) bitpos++;
    cursor_on_at(19-bitpos,PROG_CUR_Y+1,1);
  }



void action_cv29(t_navicode par)
  {
    // signed int my_var;
    unsigned char my_bit;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            lcd_clear();
            lcd_string_P(text_prog);
            check_status_display(0);
            
            set_cursor(0,PROG_CUR_Y);
            lcd_string_P(text_cv29_header);
            display_softkey_P(0, text_sk_read);
            display_softkey_P(1, text_sk_write);
            display_softkey_P(2, text_sk_RPoM);
            display_softkey_P(3, text_sk_WPoM);
            turn_led_green_soft(LED_A);
            turn_led_red_soft(LED_B);
            turn_led_green_soft(LED_C);
            turn_led_red_soft(LED_D);
            
            edit_cv29_focus = 7;
            current_CV.result = PRG_CV_VOID;           
            current_CV.addr = 0;
            my_CV29_valid = 0;
            display_cv_valid();
            display_bitwise(my_CV29);
            show_cv29_helptext(edit_cv29_focus);
            set_cv29_edit_focus(edit_cv29_focus);
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:                      // change focus forward
                    edit_cv29_focus--;
                    edit_cv29_focus %= 8;
                    show_cv29_helptext(edit_cv29_focus);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_SH_KNOB:                   // change focus revers
                    edit_cv29_focus++;
                    edit_cv29_focus %= 8;
                    show_cv29_helptext(edit_cv29_focus);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_KNOB_TURNED:               // cvValue
                    my_bit = 1 << edit_cv29_focus;
                    my_CV29 ^= my_bit;
                    my_CV29_valid = 0;
                    display_cv_valid();
                    display_bitwise(my_CV29);
                    show_cv29_helptext(edit_cv29_focus);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_UP:                        // nextCV
                    break;
                case KEY_DOWN:                      // prevCV
                    break;
                case KEY_A:                    // CV read
                    msg_send_cv_read(29);
                    msg_transmit();
                    display_prog_state(current_CV.result);
                    my_CV29_valid = 0;
                    display_cv_valid();
                    show_cv29_helptext(edit_cv29_focus);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_B:                    // CV write
                    msg_send_cv_write(29, my_CV29);
                    msg_transmit();
                    display_prog_state(current_CV.result);
                    show_cv29_helptext(edit_cv29_focus);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_C:                    // PoM read
                    current_CV.addr = actual_lok.w.addr;
                    msg_send_PoM_cv_read(current_CV.addr, 29);
                    msg_transmit();
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_D:                    // PoM write
                    current_CV.addr = actual_lok.w.addr;
                    msg_send_PoM_cv_write(current_CV.addr, 29, my_CV29);
                    msg_transmit();
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_SH_C:                    // PoM read accessory
                    // msg_send_cv_read(?? current_CV.number);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_SH_D:                    // PoM write  accessory
                    // msg_send_cv_write(current_CV.number, current_CV.value);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_PARSER_PROG:
                    display_prog_state(current_CV.result);
                    if (current_CV.result == PRG_CV_OKAY)
                      {
                        my_CV29 = current_CV.value;
                        my_CV29_valid = 1;
                      }
                    else my_CV29_valid = 0;
                    display_cv_valid();
                    display_bitwise(my_CV29);
                    set_cv29_edit_focus(edit_cv29_focus);
                    break;
                case KEY_PARSER_STATE:
                    if (current_CV.result == PRG_CV_WAIT_FOR_ANSWER)
                      {
                        // display_status(command_station_status);   // make wheelie
                      }
                    else
                      {
                      }
                    break;
              }
            break;
        case MENU_CLOSE:
            msg_send_cv_end();
            cursor_off();
            turn_led_off_soft(LED_A);
            turn_led_off_soft(LED_B);
            turn_led_off_soft(LED_C);
            turn_led_off_soft(LED_D);
            break;
      }
  }

//====================================================================================
//
// Lok Adresse Programming
//
//   01234567890123456789
//  |--------------------|
// 0|Prog ( OK )    STOP |
// 1| Lok Adresse: 12345 |
// 2|                    |
// 3|Read Writ RPoM WPoM |
//  |--------------------|


const char text_lok_header[]  PROGMEM = " Lok Adresse:";


static void display_prog_loc_addr(unsigned int addr)
  {
    set_cursor(14,PROG_CUR_Y);
    lcd_int5(addr);
    cursor_on_at(14,PROG_CUR_Y,5);
  }

static int my_addr;

void action_cv_addr(t_navicode par)
  {

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            lcd_clear();
            lcd_string_P(text_prog);
            check_status_display(0);
            
            set_cursor(0,PROG_CUR_Y);
            lcd_string_P(text_lok_header);
            display_softkey_P(0, text_sk_read);
            display_softkey_P(1, text_sk_write);
            turn_led_green_soft(LED_A);
            turn_led_red_soft(LED_B);
            turn_led_off_soft(LED_C);
            turn_led_off_soft(LED_D);

            my_addr = actual_lok.w.addr;

            current_CV.result = PRG_CV_VOID;
            current_CV.addr = 0;
            display_prog_loc_addr(my_addr);          // incl. set_cursor       
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_KNOB:                      // change focus forward
                    break;
                case KEY_KNOB_TURNED:               // cvValue
                    my_addr += get_knob_rotate();
                    if (my_addr < 1) my_addr = 1;
                    if (my_addr > 10239) my_addr = 10239;   // 10239 = 0x27FF
                    display_prog_loc_addr(my_addr);
                    break;
                case KEY_UP: 
                    break;
                case KEY_DOWN: 
                    break;
                case KEY_A:                    // CV read
                    start_programmer_readaddr();
                    break;     
                case KEY_B:                    // CV write
                    start_programmer_writeaddr(my_addr);
                    break;
                case KEY_C: 
                    break;
                case KEY_D: 
                    break;
                case KEY_SH_C: 
                    break;
                case KEY_SH_D: 
                    break;
                case KEY_PARSER_PROG:
                    display_prog_state(current_CV.result);
                    display_prog_loc_addr(my_addr);
                    break;
                case KEY_PROGRAMMER:
                    if (prg_state == PRG_DONE)
                      {
                        if (my_addr != progged_lok_addr)
                          {
                            my_addr = progged_lok_addr;
                            display_prog_loc_addr(my_addr);
                          }
                      }
                    msg_send_cv_end();
                    msg_transmit();
                    break;
              }
            break;
        case MENU_CLOSE:
            msg_send_cv_end();
            msg_transmit();
            cursor_off();
            turn_led_off_soft(LED_A);
            turn_led_off_soft(LED_B);

            break;
      }
  }




