//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_setup.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-01-29 V0.02 kw  added code for brightness
//            2009-03-09 V0.03 kw  color change if saved
//            2009-03-18 V0.04 kw  new knob_mode
//            2009-03-27 V0.05 kw  contrast added
//            2009-03-29 V0.06 kw  LED backlight balance added
//            2009-04-07 V0.07 kw  Knob Push configurable
//            2009-07-22 V0.08 kw  XBEE Mode, Item-Line depends on display
//            2010-07-24 V0.09 kw  added pc_tid
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for setup
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"
#include "timer.h"
#include "led.h"
#include "serial.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "database.h"
#include "display.h"
#include "messages_objects.h"
#include "menu_navigator.h"
#include "xbee_if.h"                // xbee_module_detected

const char text_setup[]  PROGMEM = "Setup";
static const char text_sk_def[]    PROGMEM = "DEF";
static const char text_sk_sav[]    PROGMEM = "SAVE";
static const char text_sk_blank[]    PROGMEM = " ";


// setup items (to be implemented):
// 
//   overall brightness power mode
//   overall brightness batt mode
//   display / led brightness
//   shutdown timeout
//   knob behaviour: normal, fast, turbo
//   power warning at: voltage
//  *DCC default format
//   Menu restrictions: drive only, 
//                      drive + switch, drive + switch + prog + feed

static unsigned char setup_index = 0;

typedef void (*HandleFnct)( unsigned char key );
 
struct SetupEntry {
  char     Text[12];
  HandleFnct Function;
};


#if (DISPLAY == LCD_EA204)

    #define ITEM_CY 1


#elif  (DISPLAY == LCD_EA128x64)

    #define ITEM_CY 2

#endif

//-----------------------------------------------------------------------------------

static int my_splash_time;

static void display_splash_time(void)
  {
    set_cursor(14, ITEM_CY);
    lcd_int4(my_splash_time * 100);
    lcd_putc('m');
    lcd_putc('s');
    cursor_on_at(14, ITEM_CY, 4);  // x, y, size
  }

void HandleSplashTime( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_splash_time += get_knob_rotate();
            if (my_splash_time < 0) my_splash_time = 0;
            if (my_splash_time > 99) my_splash_time = 99;   // 9,9s = 99*100ms
            display_splash_time();
            break;
        case 0:      // init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_splash_time = eeprom_read_byte((void *)eadr_show_splash);
            display_splash_time();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            eeprom_write_byte((void *)eadr_show_splash, my_splash_time);
            break;
      }
  }

//-----------------------------------------------------------------------------------

static int my_bright;  // also used in balance menu

static void display_my_bright(void)
  {
    set_cursor(15,ITEM_CY);
    lcd_int3(my_bright);
    lcd_putc('\%');
    cursor_on_at(15, ITEM_CY, 3);     // x, y, size
  }

void HandleBrightness( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_bright += get_knob_rotate();
            if (my_bright < 0) my_bright = 0;
            if (my_bright > 100) my_bright = 100;
            set_global_bright(my_bright);
            display_my_bright();
            break;
        case 0:         					// init
            turn_led_green_immediate(LED_D);
            my_bright = get_global_bright();
            display_my_bright();
            break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            restore_brightness(LINE_POWERED);
            #warning we should check power_mode for brightness setting
            my_bright = get_global_bright();
            set_global_bright(my_bright);
            display_my_bright();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            save_brightness();
            break;
      }
  }

//-----------------------------------------------------------------------------------

#if (DISPLAY == LCD_EA128x64)
static int my_contrast;

static void display_my_contrast(void)
  {
    set_cursor(15,ITEM_CY);
    lcd_int2(my_contrast);
    cursor_on_at(15, ITEM_CY, 2);
  }

void HandleContrast( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_contrast += get_knob_rotate();
            if (my_contrast < 0) my_contrast = 0;
            if (my_contrast > 0x3F) my_contrast = 0x3F;
            display_my_contrast();
            LCD_contrast(my_contrast);
            break;
        case 0:         					// init
            turn_led_green_immediate(LED_D);
            my_contrast = eeprom_read_byte((void *)eadr_lcd_contrast);
            display_my_contrast();
            LCD_contrast(my_contrast);
            break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_contrast = eeprom_read_byte((void *)eadr_lcd_contrast);
            display_my_contrast();
            LCD_contrast(my_contrast);
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            eeprom_write_byte((void *)eadr_lcd_contrast, my_contrast);
            break;
      }
  }

#endif

//-----------------------------------------------------------------------------------

static void _display_balance()
  {
    set_cursor(13,ITEM_CY);
    lcd_int4(my_bright);
    cursor_on_at(13, ITEM_CY, 4);

    if (my_bright > 0)
      {
        set_led_brightness(100);
        set_disp_brightness(100-my_bright);
      }
    else
      {
        set_disp_brightness(100);
        set_led_brightness(100+my_bright);
      }
  }

static void _save_balance(void)
  {
    if (my_bright > 0)
      {
        eeprom_write_byte((void *)eadr_global_bright_led, 100);
        eeprom_write_byte((void *)eadr_global_bright_disp, 100-my_bright);
      }
    else
      {
        eeprom_write_byte((void *)eadr_global_bright_led, 100+my_bright);
        eeprom_write_byte((void *)eadr_global_bright_disp, 100);
      }
  }

static void _restore_balance(void)
  {
    unsigned char my_led, my_disp;

    my_led = eeprom_read_byte((void *)eadr_global_bright_led);
    my_disp = eeprom_read_byte((void *)eadr_global_bright_disp);
    
    if (my_led > my_disp)  my_bright = 100-my_disp;
    else                   my_bright = -(100-my_led);
  }



void HandleBrightnessBalance( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            my_bright += get_knob_rotate();
            if (my_bright < -100) my_bright = -100;
            if (my_bright > 100) my_bright = 100;
            _display_balance();
            turn_led_red_immediate(LED_D);
            break;
        case 0:         					// init
            // no break;
        case KEY_A:  // DEF
            _restore_balance();
            turn_led_green_immediate(LED_D);
            _display_balance();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            _save_balance();
            break;
      }
  }

//-----------------------------------------------------------------------------------

static void _save_bright_night(void)
  {
    eeprom_write_byte((void *)eadr_global_bright_night, my_bright);
  }

static void _restore_bright_night(void)
  {
    my_bright = eeprom_read_byte((void *)eadr_global_bright_night);
  }


void HandleBrightnessNight( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            my_bright += get_knob_rotate();
            if (my_bright < 0) my_bright = 0;
            if (my_bright > 100) my_bright = 100;
            display_my_bright();
            turn_led_red_immediate(LED_D);
            break;
        case 0:         					// init
            // no break;
        case KEY_A:  // DEF
            _restore_bright_night();
            turn_led_green_immediate(LED_D);
            display_my_bright();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            _save_bright_night();
            break;
      }
  }

//-----------------------------------------------------------------------------------

static int my_switch_time;

static void display_my_time(void)
  {
    set_cursor(14,ITEM_CY);
    lcd_int4(my_switch_time * (50000L / 1000L));  // TURNOUT_TICK_PERIOD
    lcd_putc('m');
    lcd_putc('s');
    cursor_on_at(14,ITEM_CY, 4);
  }

void HandleSwitchTime( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_switch_time += get_knob_rotate();
            if (my_switch_time < 0) my_switch_time = 0;
            if (my_switch_time > 199) my_switch_time = 199;
            display_my_time();
            break;
        case 0:         					// init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_switch_time = turnout_delay;
            display_my_time();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            save_turnout_delay(my_switch_time);
			break;
      }
  }

//-----------------------------------------------------------------------------------

const char text_make_off[]        PROGMEM = "Off  ";
const char text_make_stop[]       PROGMEM = "Stop ";
const char text_make_stop_locos[] PROGMEM = "myLoc";

static t_STOP_action my_stop_action;

static void display_stop_action(void)
  {
    set_cursor(14,ITEM_CY);
    switch(my_stop_action)
      {
        default:  // no break;
        case MAKES_OFF:                lcd_string_P(text_make_off);         break;
        case MAKES_STOP:               lcd_string_P(text_make_stop);        break;
        case MAKES_STOP_MY_LOCOS:      lcd_string_P(text_make_stop_locos);  break;
      }
    cursor_on_at(14,ITEM_CY,sizeof(text_make_off)-1);   // x, y, size
  }

void HandleStopKey( unsigned char key )
  {
    int temp;
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            temp = get_knob_rotate();
            if (temp > 0) { my_stop_action++; if (my_stop_action == sizeof_t_STOP_action) my_stop_action=0; }
            if (temp < 0) { if (my_stop_action == 0) my_stop_action = sizeof_t_STOP_action; my_stop_action--; }
            display_stop_action();
            break;
		case 0:         					// init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_stop_action = STOP_action;
            display_stop_action();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            STOP_action = my_stop_action;
            eeprom_write_byte((void *)eadr_STOP_action, STOP_action);
            break;
      }
  }

//-----------------------------------------------------------------------------------

const char text_knob_dual[]        PROGMEM = "as DC";
const char text_knob_single[]      PROGMEM = "as AC";


static t_knob_mode my_knob_mode;

static void display_knob_mode(void)
  {
    set_cursor(14,ITEM_CY);
    switch(my_knob_mode)
      {
        default:  // no break;
        case KNOB_MODE_DUALSIDED:    lcd_string_P(text_knob_dual);     break;
        case KNOB_MODE_SINGLESIDED:  lcd_string_P(text_knob_single);   break;
      }
    cursor_on_at(14,ITEM_CY, sizeof(text_knob_dual)-1);  // x, y, size
  }

void HandleKnobMode( unsigned char key )
  {
    int temp;
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            temp = get_knob_rotate();
            if (temp > 0) { my_knob_mode++; if (my_knob_mode == sizeof_t_knob_mode) my_knob_mode=0; }
            if (temp < 0) { if (my_knob_mode == 0) my_knob_mode = sizeof_t_knob_mode; my_knob_mode--; }
            display_knob_mode();
            break;
		case 0:         					// init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_knob_mode = knob_mode;
            display_knob_mode();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            knob_mode = my_knob_mode;
            eeprom_write_byte((void *)eadr_knob_mode, knob_mode);
            break;
      }
  }

//-----------------------------------------------------------------------------------

const char text_knob_fs0[]        PROGMEM = "Speed 0";
const char text_knob_estop[]      PROGMEM = "Em.Stop";


static t_knob_push my_knob_push;

static void display_knob_push(void)
  {
    set_cursor(14,ITEM_CY);
    switch(my_knob_push)
      {
        default:  // no break;
        case KNOB_PUSH_FS0:    lcd_string_P(text_knob_fs0);     break;
        case KNOB_PUSH_ESTOP:  lcd_string_P(text_knob_estop);   break;
      }
    cursor_on_at(14,ITEM_CY, sizeof(text_knob_fs0)-1);  // x, y, size
  }

void HandleKnobPush( unsigned char key )
  {
    int temp;
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            temp = get_knob_rotate();
            if (temp > 0) { my_knob_push++; if (my_knob_push == sizeof_t_knob_push) my_knob_push=0; }
            if (temp < 0) { if (my_knob_push == 0) my_knob_push = sizeof_t_knob_push; my_knob_push--; }
            display_knob_push();
            break;
		case 0:         					// init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_knob_push = knob_push;
            display_knob_push();
            break;                    
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            knob_push = my_knob_push;
            eeprom_write_byte((void *)eadr_knob_push, knob_push);
            break;
      }
  }

//-----------------------------------------------------------------------------------

static t_format my_format;

void display_dcc_default(t_format f)
  {
    set_cursor(15,ITEM_CY);
    switch(f)
      {
        default: break;
        case DCC14: lcd_int3(14);            break;
        case DCC27: break;
        case DCC28: lcd_int3(28);            break;
        case DCC128:lcd_int3(126);           break;
      }
    cursor_on_at(15, ITEM_CY, 3);
  }

void HandleDCCDefault( unsigned char key )
  {
    int temp;
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            temp = get_knob_rotate();
            switch(my_format)
              {
                default:  break;
                case DCC14: if (temp > 0) my_format = DCC28;
                            if (temp < 0) my_format = DCC128;
                            break;
                case DCC27:
                            break;
                case DCC28: if (temp > 0) my_format = DCC128;
                            if (temp < 0) my_format = DCC14;
                            break;
                case DCC128:if (temp > 0) my_format = DCC14;
                            if (temp < 0) my_format = DCC28;
                            break;
              }
            display_dcc_default(my_format);
            break;
		case 0:         					// init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_format = default_dcc_format;
            display_dcc_default(my_format);
            break;
        case KEY_KNOB:
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            eeprom_write_byte((void *)eadr_default_dcc_format, my_format);
            default_dcc_format = my_format;
            break;

      }
  }


//-----------------------------------------------------------------------------------

static const char text_on[]         PROGMEM = "ON     ";
static const char text_psave1[]     PROGMEM = "PSAVE 1";
static const char text_psave2[]     PROGMEM = "PSAVE 2";
static const char text_no_module[]  PROGMEM = "n.avail.";
static char my_xbee_mode;

static void display_XBEEmode(void)
  {
    set_cursor(13,ITEM_CY);
    if(xbee_module_detected)
      {
        switch (my_xbee_mode)
          {
            case 0: lcd_string_P(text_on); break;
            case 1: lcd_string_P(text_psave1); break;
            case 2: lcd_string_P(text_psave2); break;
          }
      }
    else
      {
        lcd_string_P(text_no_module);
      }
    cursor_on_at(13,ITEM_CY,sizeof(text_on)-1);
  }


void HandleXBEEmode( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            if(xbee_module_detected)
              {
                int8_t temp = get_knob_rotate();
                if (temp > 0)
                  {
                    my_xbee_mode++;
                    if (my_xbee_mode==3) my_xbee_mode=0;
                  }
                if (temp < 0)
                  {
                    if (my_xbee_mode==0) my_xbee_mode=3;
                    my_xbee_mode--; 
                  }
              }
            display_XBEEmode();
            break;
        case 0:      // init, no break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_xbee_mode = eeprom_read_byte((void *)eadr_XBEE_mode);
            display_XBEEmode();
            break;                    
        case KEY_KNOB:          
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            if(xbee_module_detected)
              {
                if (my_xbee_mode == eeprom_read_byte((void *)eadr_XBEE_mode) ) return;  // no change
                eeprom_write_byte((void *)eadr_XBEE_mode, my_xbee_mode);
              }
            break;
      }
  }

 

//-----------------------------------------------------------------------------------

static const char text_xp[]         PROGMEM = "XP  ";
static const char text_wrp[]        PROGMEM = "WRP ";
static const char text_none[]       PROGMEM = "none";
static char my_connect;


static void display_Connect(void)
  {
    set_cursor(14,ITEM_CY);
    
    switch (my_connect)
      { 
        default:
        case 0:  lcd_string_P(text_none);   break;
        case 1:  lcd_string_P(text_xp);   break;
        case 2:  lcd_string_P(text_wrp);   break;
      }
    cursor_on_at(14,ITEM_CY,sizeof(text_xp)-1);
  }


void HandleConnect( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            if (get_knob_rotate()) 
              {
                my_connect++;
                if (my_connect == 3) my_connect = 0;
              }
            display_Connect();
            break;
        case 0:      // init, no break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_connect = eeprom_read_byte((void *)eadr_connect);
            display_Connect();
            break;                    
        case KEY_KNOB:          
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            if (my_connect == eeprom_read_byte((void *)eadr_connect) ) return;  // no change
            eeprom_write_byte((void *)eadr_connect, my_connect);
            break;
      }
  }

 

//-----------------------------------------------------------------------------------

static const char text_auto[]        PROGMEM = "AUTO";
static int my_slot_id;
#ifndef SLOT_AUTOMATIC
#define SLOT_AUTOMATIC 1
#define SLOT_MANUAL 0
#endif

static void display_XPSlot(void)
  {
    set_cursor(14,ITEM_CY);
    if(my_slot_id > 0)
      {
        lcd_putc('M');
        lcd_putc(':');
        lcd_int02(my_slot_id);
      }
    else
      {
        lcd_string_P(text_auto);
      }
    cursor_on_at(14,ITEM_CY,sizeof(text_auto)-1);
  }


void HandleXPSlot( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_slot_id += get_knob_rotate();
            if (my_slot_id < 0) my_slot_id = 0;
            if (my_slot_id > 31) my_slot_id = 31;   // max range
            display_XPSlot();
            break;
        case 0:      // init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            if (eeprom_read_byte((void *)eadr_xp_slot_mode) == SLOT_AUTOMATIC) my_slot_id = 0;
            else my_slot_id = eeprom_read_byte((void *)eadr_xp_slot_addr);
            display_XPSlot();
            break;                    
        case KEY_KNOB:          
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            if (my_slot_id == 0)
              {
                eeprom_write_byte((void *)eadr_xp_slot_mode, SLOT_AUTOMATIC);
              }
            else
              {
                eeprom_write_byte((void *)eadr_xp_slot_mode, SLOT_MANUAL);
                eeprom_write_byte((void *)eadr_xp_slot_addr, my_slot_id);
              }
            break;
      }
  }

 
//-----------------------------------------------------------------------------------

static int my_pc_tid;

static void display_pc_tid(void)
  {
    set_cursor(14,ITEM_CY);
    lcd_int02(my_pc_tid);
    cursor_on_at(14,ITEM_CY,2);
  }


void HandlePC_TID( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            turn_led_red_immediate(LED_D);
            my_pc_tid += get_knob_rotate();
            if (my_pc_tid < 1) my_pc_tid = 1;
            if (my_pc_tid > 31) my_pc_tid = 31;   // max range
            display_pc_tid();
            break;
        case 0:      // init
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            my_pc_tid = eeprom_read_byte((void *)eadr_pc_tid);
            display_pc_tid();
            break;                    
        case KEY_KNOB:          
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            eeprom_write_byte((void *)eadr_pc_tid, my_pc_tid);
            throttle_id = my_pc_tid;
            break;
      }
  }


//-----------------------------------------------------------------------------------

static int my_para_addr;
static int my_para_value;
static unsigned char p_edit_focus;

static void display_edit_cursor(void)
  {
    if (p_edit_focus) cursor_on_at(10+6, ITEM_CY, 3);   // data
    else              cursor_on_at(10+1, ITEM_CY, 4);   // addr
  }

static void display_para_addr(void)
  {
    set_cursor(10,ITEM_CY);
    lcd_putc('#');
    lcd_int4(my_para_addr);
    lcd_putc(':');
    display_edit_cursor();
  }

static void display_para_value(void)
  {
    set_cursor(10+6,ITEM_CY);
    lcd_int3(my_para_value);
    display_edit_cursor();
  }


void HandleParamE( unsigned char key )
  {
    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            if (p_edit_focus)
              {
                turn_led_red_immediate(LED_D);
                my_para_value += get_knob_rotate();
                if (my_para_value < 0) my_para_value = 0;
                if (my_para_value > 255) my_para_value = 255;   // byte range
                display_para_value();
              }
             else
              {
                turn_led_green_immediate(LED_D);
                my_para_addr += get_knob_rotate();
                if (my_para_addr < 0) my_para_addr = 0;
                if (my_para_addr > 4095) my_para_addr = 4095;   // 4k range
                display_para_addr();
                my_para_value = eeprom_read_byte((void *)my_para_addr);
                display_para_value();
              }
            break;
        case KEY_KNOB:
            if (p_edit_focus) p_edit_focus = 0;
            else p_edit_focus = 1;
            display_edit_cursor();
            break;
        case 0:      // init
            p_edit_focus = 0; // no break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            display_para_addr();
            my_para_value = eeprom_read_byte((void *)my_para_addr);
            display_para_value();
            break;                    
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            eeprom_write_byte((void *)my_para_addr, my_para_value);
            break;
      }
  }

//-----------------------------------------------------------------------------------

static int my_para_addr;
static int my_para_value;
static unsigned char p_edit_focus;

// double
//static void display_edit_cursor(void)
//  {
//    if (p_edit_focus) cursor_on_at(10+6, ITEM_CY, 3);   // data
//    else              cursor_on_at(10+1, ITEM_CY, 4);   // addr
//  }

// double
// static void display_para_addr(void)
//  {
//    set_cursor(10,ITEM_CY);
//    lcd_putc('#');
//    lcd_int4(my_para_addr);
//    lcd_putc(':');
//    display_edit_cursor();
//  }

// doublestatic void display_para_value(void)
//  {
//    set_cursor(10+6,ITEM_CY);
//    lcd_int3(my_para_value);
//    display_edit_cursor();
//  }



void HandleParamR( unsigned char key )
  {
    uint8_t *ptr;

    switch (key) 
      {
        default: break;
        case KEY_KNOB_TURNED:
            if (p_edit_focus)
              {
                turn_led_red_immediate(LED_D);
                my_para_value += get_knob_rotate();
                if (my_para_value < 0) my_para_value = 0;
                if (my_para_value > 255) my_para_value = 255;   // byte range
                display_para_value();
              }
             else
              {
                turn_led_green_immediate(LED_D);
                my_para_addr += get_knob_rotate();
                if (my_para_addr < 0) my_para_addr = 0;
                if (my_para_addr > 4095) my_para_addr = 4095;   // 4k range
                display_para_addr();
                ptr = (uint8_t *)my_para_addr;
                my_para_value = *ptr;
                display_para_value();
              }
            break;
        case KEY_KNOB:
            if (p_edit_focus) p_edit_focus = 0;
            else p_edit_focus = 1;
            display_edit_cursor();
            break;
        case 0:      // init
            p_edit_focus = 0; // no break;
        case KEY_A:  // DEF
            turn_led_green_immediate(LED_D);
            display_para_addr();
            ptr = (uint8_t *)my_para_addr;
            my_para_value = *ptr;
            display_para_value();
            break;                    
        case KEY_D:  // SAV
            turn_led_green_immediate(LED_D);
            ptr = (uint8_t *)my_para_addr;
            *ptr = my_para_value;
            break;
      }
  }

//-----------------------------------------------------------------------------------
const struct SetupEntry SetupMenu[] PROGMEM =
  {
    // 012345678901, function pointer
    { "Brightness",  HandleBrightness },
    #if (DISPLAY == LCD_EA128x64)
    { "Contrast",    HandleContrast },
    #endif
    { "LED vs. Back",HandleBrightnessBalance },
    { "Bright-Night",HandleBrightnessNight },
    { "XBEE Module", HandleXBEEmode },
    { "Def. Connect",HandleConnect },
    { "XP Slot",     HandleXPSlot },
    { "PC ID",       HandlePC_TID },
    { "DCC Default", HandleDCCDefault },
    { "Stop Key",    HandleStopKey },
    { "Knob Action", HandleKnobMode },
    { "Knob Push",   HandleKnobPush },
    { "Switch Time", HandleSwitchTime },
    { "Splash Scrn", HandleSplashTime },
    { "int.EEPROM",  HandleParamE },
    { "int.RAM",     HandleParamR },
  };



//   01234567890123456789
//  |--------------------|
// 0|Setup          STOP |
// 1| Brightness         |
// 2|                    |
// 3| DEF            SAVE|
//  |--------------------|


void show_setup_item(unsigned char index)
  {
    unsigned char i;
    char data;
 
    const char *str = SetupMenu[setup_index].Text;  // a pointer to progmem (prog_char)
    
    set_cursor(1,ITEM_CY);
    for (i=0; i<sizeof(SetupMenu[0].Text); i++)
      {
        if ((data = pgm_read_byte (str++))) lcd_putc(data);
        else lcd_putc(' ');                     // fill rest with blank
      }
    for ( ; i<20; i++)
      {
        lcd_putc(' ');                          // fill rest of line with blank
      }
    ((HandleFnct) pgm_read_word( &(SetupMenu[setup_index].Function) ))(0);      // call the handler
  }


void action_setup(t_navicode par)
  {
    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            turn_led_off_soft(LED_B);
            turn_led_off_soft(LED_C);

            lcd_clear();
            lcd_string_P(text_setup);
            check_status_display(0);
            
            display_softkey_P(0, text_sk_def);
			turn_led_green_immediate(LED_A);
            display_softkey_P(1, text_sk_blank);
			display_softkey_P(2, text_sk_blank);
			display_softkey_P(3, text_sk_sav);
			turn_led_green_immediate(LED_D);

            show_setup_item(setup_index);
            break;
        case MENU_UPDATE:
            // we want: SetupMenu[setup_index].Function(par.key); 
            // this construct is required - the call is in flash memory            
            ((HandleFnct) pgm_read_word( &(SetupMenu[setup_index].Function) ))(par.key);      // call the handler
            
            switch (par.key)
              {
                case KEY_UP:
                    if (setup_index > 0) setup_index--;
                    show_setup_item(setup_index);
                    break;
                case KEY_DOWN:
                    if (setup_index < (sizeof(SetupMenu)/sizeof(SetupMenu[0]) -1) ) setup_index++;
                    show_setup_item(setup_index);
                    break;
                case KEY_KNOB:
                    // save: is processed by handle
                    break;
              }
            break;
        case MENU_CLOSE:
			turn_led_off(LED_A);
            turn_led_off(LED_D);
            break;
      }
  }
