//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2008-2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_splash.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-07-18 V0.01 kw  started
//            2009-08-23 V0.02 kw  added debug screens for Rx and Tx
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   menu for splash_screen
//
//========================================================================

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>
#include <stdio.h>
#include <util/delay.h>

#include "config.h"
#include "hardware.h"               // a58
#include "cortos.h"                 // real time kernel
#include "keycodes.h"
#include "keyboard.h"
#include "timer.h"
#include "database.h"
#include "messages_objects.h"
#include "display.h"


#include "menu_navigator.h"
#include "menu_lok.h"               // init for lok

#include "serial.h"                 // debug screen (Rx & Tx)

#include "lcd_text.h"
#include "lcd_ks0073_charset.h"



//------------------------------------------------------------------------------
//
//  Menu - the ABOUT screen
//
//------------------------------------------------------------------------------

//-----------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//-----------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   01234567890123456789
//  |--------------------|
// 0| My First Throttle  |
// 1| (c) www.opendcc.de |
// 2| Vers: 00.01  S:0000|
// 3| Build: 2009-03-26  |
//  |--------------------|

char about_MFT_0[] PROGMEM = 
  {//0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','M','y',' ','F','i','r','s','t',' ','T','h','r','o','t','t','l','e',' ',' ',0
  };
char about_MFT_1[] PROGMEM = 
  {//0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','(','c',')',' ','w','w','w','.','o','p','e','n','d','c','c','.','d','e',' ',0
  };
char about_MFT_2[] PROGMEM = 
  {//    1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','V','e','r','s',':',' ','0','0','.','0','1',' ',' ','S',':','0','0','0','0',0
  };
char about_MFT_3[] PROGMEM = 
  {//    1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9  
    ' ','B','u','i','l','d',':',' ','2','0','0','9','-','0','8','-','1','0',0
  };


void show_about_screen(void)
  {
    lcd_clear();
    set_cursor(0,0);
    lcd_string_P(about_MFT_0);
    set_cursor(0,1);
    lcd_string_P(about_MFT_1);
    set_cursor(0,2);
    lcd_string_P(about_MFT_2);
    set_cursor(7,2);
    lcd_int02(MFT_VERSION);
    set_cursor(10,2);
    lcd_int02(MFT_SUBVERSION);
    set_cursor(16,2);
    lcd_int04(serial_number);
    

    set_cursor(0,3);
    lcd_string_P(about_MFT_3);
    set_cursor(8,3);
    lcd_int4(2000+BUILD_YEAR);
    set_cursor(13,3);
    lcd_int02(BUILD_MONTH);
    set_cursor(16,3);
    lcd_int02(BUILD_DAY);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

//   01234567890123456789
//  |--------------------|
// 0| My First Throttle  |
// 1| (c) www.opendcc.de |
// 2|                    |
// 3| Version:    00.01  |
// 4| Serial:      0000  |
// 5| Build: 2009-03-26  |
// 6|                    |
// 7|                    |
//  |--------------------|

const char about_MFT_0[] PROGMEM = 
  {//0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','M','y',' ','F','i','r','s','t',' ','T','h','r','o','t','t','l','e',' ',' ',0
  };
const char about_MFT_1[] PROGMEM = 
  {//0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','(','c',')',' ','w','w','w','.','o','p','e','n','d','c','c','.','d','e',' ',0
  };
const char about_MFT_2[] PROGMEM = 
  {//    1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','V','e','r','s','i','o','n',':',' ',' ',' ',' ','0','1','.','0','0',' ',' ',0
  };
const char about_MFT_3[] PROGMEM = 
  {//    1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9 
    ' ','S','e','r','i','a','l',':',' ',' ',' ',' ',' ',' ','0','0','0','0',' ',' ',0
  };
const char about_MFT_4[] PROGMEM = 
  {//    1   2   3   4   5   6   7   8   9   0   1   2   3   4   5   6   7   8   9  
    ' ','B','u','i','l','d',':',' ','2','0','0','9','-','0','1','-','2','6',0
  };


void show_about_screen(void)
  {
    lcd_clear();
    set_cursor(0,0);
    lcd_string_P(about_MFT_0);
    set_cursor(0,1);
    lcd_string_P(about_MFT_1);
    set_cursor(0,3);
    lcd_string_P(about_MFT_2);
    set_cursor(13,3);
    lcd_int02(MFT_VERSION);
    set_cursor(16,3);
    lcd_int02(MFT_SUBVERSION);
    set_cursor(0,4);
    lcd_string_P(about_MFT_3);
    set_cursor(14,4);
    lcd_int04(serial_number);
    

    set_cursor(0,5);
    lcd_string_P(about_MFT_4);
    set_cursor(8,5);
    lcd_int4(2000+BUILD_YEAR);
    set_cursor(13,5);
    lcd_int02(BUILD_MONTH);
    set_cursor(16,5);
    lcd_int02(BUILD_DAY);
    if (a58_check())
      {
        LCD_putchar('-');
        LCD_putchar('M');
      }
    else
      {
        LCD_putchar('-');
        LCD_putchar('F');    // Falko    
      }
  }


#define DEBUG_ROWS   7
#define DEBUG_COLS   7


const char text_debugrx[]  PROGMEM = "RxBuffer ";

static void debug_screen_rx(void)
  {
    lcd_clear();
    lcd_string_P(text_debugrx);

    LCD_putchar('w');
    LCD_putchar(':');
    lcd_hex2(rx_write_ptr);
    LCD_putchar(' ');
    
    LCD_putchar('r');
    LCD_putchar(':');
    lcd_hex2(rx_read_ptr);

    unsigned char r,c;
    unsigned char rx_temp;

    rx_temp = rx_write_ptr + RxBuffer_Size;
    rx_temp -= (DEBUG_COLS*DEBUG_ROWS);
    if (rx_temp >= RxBuffer_Size) rx_temp -= RxBuffer_Size;

    for (r=0; r<DEBUG_ROWS; r++)
      {
        set_cursor(0,r+1);
        for  (c=0; c<DEBUG_COLS; c++)
          {
            lcd_hex2(RxBuffer[rx_temp]);
            lcd_blank(1);
            rx_temp++;
            if (rx_temp == RxBuffer_Size) rx_temp = 0;
          }
      }
  }


const char text_debugtx[]  PROGMEM = "TxBuffer ";

static void debug_screen_tx(void)
  {
    lcd_clear();
    lcd_string_P(text_debugtx);

    LCD_putchar('w');
    LCD_putchar(':');
    lcd_hex2(tx_write_ptr);
    LCD_putchar(' ');
    
    LCD_putchar('r');
    LCD_putchar(':');
    lcd_hex2(tx_read_ptr);

    unsigned char r,c;
    unsigned char tx_temp;

    tx_temp = tx_write_ptr + TxBuffer_Size;
    tx_temp -= (DEBUG_COLS*DEBUG_ROWS);
    if (tx_temp >= TxBuffer_Size) tx_temp -= TxBuffer_Size;

    for (r=0; r<DEBUG_ROWS; r++)
      {
        set_cursor(0,r+1);
        for  (c=0; c<DEBUG_COLS; c++)
          {
            lcd_hex2(TxBuffer[tx_temp]);
            lcd_blank(1);
            tx_temp++;
            if (tx_temp == TxBuffer_Size) tx_temp = 0;
          }
      }
  }


//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY



void action_about(t_navicode par)
  {
    // here we don't show status!
    switch (par.menucode)
      {
        case MENU_OPEN:
            show_about_screen(); 
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                default:
                    break;
                case KEY_DOWN:
                    // this is a hidden debug screen!
                    #if  (DISPLAY == LCD_EA128x64)
                      debug_screen_rx();
                    #endif
                    break;
                case KEY_UP:
                    // this is a hidden debug screen!
                    #if  (DISPLAY == LCD_EA128x64)
                      debug_screen_tx();
                    #endif
                    break;               
              }
            break;
        case MENU_CLOSE:
            break;
      }
    cursor_off();
  }






//------------------------------------------------------------------------------
//
//  Menu - the SPLASH screen
//
//------------------------------------------------------------------------------


#if (DISPLAY == LCD_EA204)
//   012345678901234567890
//  |--------------------|
// 0|  XX XX  XXX  XXX   |
// 1|  X X X  X     X    |
// 2|  X   X  XX    X    |
// 3|  X   X  X     X    |
//  |--------------------|
//
//


char splash_mft_0[] PROGMEM = 
  {
    ' ',' ',C_BLOCK_FULL,C_BLOCK_FULL,' ',C_BLOCK_FULL,C_BLOCK_FULL,' ',' ',    // M
    C_BLOCK_FULL,C_BLOCK_FULL,C_BLOCK_FULL,' ',' ',                             // F
    C_BLOCK_FULL,C_BLOCK_FULL,C_BLOCK_FULL,' ',' ',                             // T
    0
  };
char splash_mft_1[] PROGMEM = 
  {
    ' ',' ',C_BLOCK_FULL,' ',C_BLOCK_FULL,' ',C_BLOCK_FULL,' ',' ',             // M
    C_BLOCK_FULL,' ',' ',' ',' ',                                               // F
    ' ',C_BLOCK_FULL,' ',' ',' ',                                               // T
    0
  };
char splash_mft_2[] PROGMEM = 
  {
    ' ',' ',C_BLOCK_FULL,' ',' ',' ',C_BLOCK_FULL,' ',' ',                     // M
    C_BLOCK_FULL,C_BLOCK_FULL,' ',' ',' ',                                     // F
    ' ',C_BLOCK_FULL,' ',' ',' ',                                              // T
    0
  };
char splash_mft_3[] PROGMEM = 
  {
    ' ',' ',C_BLOCK_FULL,' ',' ',' ',C_BLOCK_FULL,' ',' ',                      // M
    C_BLOCK_FULL,' ',' ',' ',' ',                                               // F
    ' ',C_BLOCK_FULL,' ',' ',                                                   // T
    0
  };
    

void show_splash_mft(void)
  {
    lcd_clear();
    set_cursor(0,0);
    lcd_string_P(splash_mft_0);
    set_cursor(0,1);
    lcd_string_P(splash_mft_1);
    set_cursor(0,2);
    lcd_string_P(splash_mft_2);
    set_cursor(0,3);
    lcd_string_P(splash_mft_3);
  }

//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)

#include "bmp/splash_glcd.h"

void show_splash_mft(void)
  {
    LCD_fill_with_BMP((unsigned char *)splash_glcd_bmp);
    // enable_display_update();
    update_display_complete(); // there is no task running
  }


//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY



void action_splash(t_navicode par)
  {
    // here we don't show status!
    switch (par.menucode)
      {
        case MENU_OPEN:
            show_splash_mft(); 
            cursor_off();           
            break;
        case MENU_UPDATE:
            break;
        case MENU_CLOSE:
            break;
      }
  }
