//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_sysstat.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-02-09 V0.01 kw  start
//            2009-02-28 V0.02 kw  added xp state and # of Loks
//            2009-03-02 V0.03 kw  added handling of 'analog' KEYS
//            2009-09-24 V0.04 kw  conditional assembly for displays
//            2009-07-16 V0.05 kw  XBEE infos added
//            2010-08-14 V0.06 kw  XBEE rssi added
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   System status and debug page
//            see menu_navigator.c and .h on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "analog.h"                 // adc
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_ks0073_charset.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "xp_client_if.h"           // slot_id
#include "menu_navigator.h"
#include "xbee_if.h"                // status of xbee

//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------

typedef enum 
  {
    VIEW_LOK,
    VIEW_VOLTAGE,
    VIEW_XP,
    VIEW_SLOTS,
    VIEW_XBEE_STATUS,
    VIEW_XBEE_INFO,
    N_of_SYSITEM
  } t_sys_item;

static t_sys_item sys_item;

//------------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//
//   01234567890123456789
//  |--------------------|
// 0|SysMonitor     STOP |
// 1|  Batt: 2560mV      |
// 2|  Power: *          |
// 3|                    |
//  |--------------------|

static char text_system[]  PROGMEM = "SysMonitor";


void display_sys_item(void)
  {
    // uint16_t voltage;

    set_cursor(2,1); lcd_blank(18);
    set_cursor(2,2); lcd_blank(18);
    set_cursor(2,3); lcd_blank(18);

    switch(sys_item)
      {
        case VIEW_VOLTAGE:
            //adc_init(INTERNAL_VREF, NO_LEFT_ADJUST, 2);
            //voltage = 5 * adc_single_conversion(2);         // measure ADC2 - VINT/2
            
            set_cursor(2,1);                // display batt
            static char text_batt[]    PROGMEM = "Batt:    ";
            lcd_string_P(text_batt);
            lcd_int4(batt_voltage);
            lcd_putc('m');
            lcd_putc('V'); 
                                            // display_vccio
            set_cursor(2,2);
            static char text_power[]   PROGMEM = "Power:   ";
            lcd_string_P(text_power);
            lcd_bool(g_vccio_state);
            break;
        case VIEW_XP:
            set_cursor(2,1);                // display xp states
            static char text_xp[]      PROGMEM = "XP       ";
            lcd_string_P(text_xp);
            static char text_xp_on[]      PROGMEM = " on";
            static char text_xp_off[]      PROGMEM = "off";
            if (xp_connect == XP_CONNECTED) lcd_string_P(text_xp_on);
            else lcd_string_P(text_xp_off);
            
            set_cursor(2,2);
            static char text_slotid[]  PROGMEM = "SlotID:  ";
            lcd_string_P(text_slotid);
            lcd_int2(slot_id);
            lcd_putc(',');
            if (slot_mode == SLOT_AUTOMATIC) lcd_putc('A');
            else lcd_putc('M');
            break;
        case VIEW_LOK:
            set_cursor(2,1);
            static char text_locos[]   PROGMEM = "#ofLok:  ";
            lcd_string_P(text_locos);
            lcd_int3(get_loco_in_database());
            set_cursor(2,2);
            static char text_locom[]   PROGMEM = "maxLok:  ";
            lcd_string_P(text_locom);
            lcd_int3(SIZE_LOCO_DATABASE-1);
            break;
        case VIEW_SLOTS:
            set_cursor(2,1);
            static char text_slots[]   PROGMEM = "XP-Slots:  ";
            lcd_string_P(text_slots);
            unsigned char i;
            set_cursor(2,2);
            for (i=0; i < 15; i++)
              {
                if (i==8) lcd_putc(' ');
                if (i == slot_id) lcd_putc('*');
                else lcd_bool(slots_in_use[i]);
              }
            set_cursor(2,3);
            for (i=16; i < 31; i++)
              {
                if (i==24) lcd_putc(' ');
                if (i == slot_id) lcd_putc('*');
                else lcd_bool(slots_in_use[i]);
              }
            break;
        case VIEW_XBEE_STATUS:
            set_cursor(2,1);
            static char text_xb_sts[]   PROGMEM = "XBEE Status";
            lcd_string_P(text_xb_sts);
            set_cursor(2,2);
            if (xbee_module_detected)
              {
                static char text_xb_base_macH[]  PROGMEM = "Base-#h: ";
                lcd_string_P(text_xb_base_macH);
                lcd_hex2(base_station_mac[0]);
                lcd_hex2(base_station_mac[1]);
                lcd_hex2(base_station_mac[2]);
                lcd_hex2(base_station_mac[3]);
                set_cursor(2,3);
                static char text_xb_base_macL[]  PROGMEM = "Base-#l: ";
                lcd_string_P(text_xb_base_macL);
                lcd_hex2(base_station_mac[4]);
                lcd_hex2(base_station_mac[5]);
                lcd_hex2(base_station_mac[6]);
                lcd_hex2(base_station_mac[7]);
              }
            else
              {
                static char text_xb_miss[]   PROGMEM = "XBEE: no module";
                lcd_string_P(text_xb_miss);
              }
            break;
        case VIEW_XBEE_INFO:
            set_cursor(2,1);
            static char text_xb_info[]   PROGMEM = "XBEE: Info";
            lcd_string_P(text_xb_info);
            set_cursor(2,2);
            if (xbee_module_detected)
              {
                static char text_xb_info_pan[]  PROGMEM = "PAN:   ";     // ATID
                lcd_string_P(text_xb_info_pan);
                lcd_hex4(pan_id);
                set_cursor(2,3);
                lcd_hex8(xbee_serial_high);
                lcd_hex8(xbee_serial_low);
              }
            else
              {
                static char text_xb_miss[]   PROGMEM = "XBEE: no module";
                lcd_string_P(text_xb_miss);
              }
            break;
        default:
            break;
        case N_of_SYSITEM:
            break;
      }
  }


static void show_header(void)
  {
    cursor_off();
    lcd_clear();
    lcd_string_P(text_system);
    check_status_display(0);            
  }

//------------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//------------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)



//
//   01234567890123456789
//  |--------------------|
// 0|SysMonitor     STOP |
// 1|                    |
// 2|  Batt: 2560mV      |
// 3|  Power: *          |
// 4|   additional       |
// 5|   additional       |
// 6|                    |
// 7|    softkeys        |
//  |--------------------|

const static char text_system[]    PROGMEM = "SysMonitor";
const static char text_sk_blank[]  PROGMEM = " ";
const static char text_sk_save[]   PROGMEM = "SAVE";
const static char text_sk_read[]   PROGMEM = "READ";
                



void display_sys_item(void)
  {
    // uint16_t voltage;

    set_cursor(2,2); lcd_blank(18);
    set_cursor(2,3); lcd_blank(18);
    set_cursor(2,4); lcd_blank(18);
    set_cursor(2,5); lcd_blank(18);
    set_cursor(2,6); lcd_blank(18);

    clear_softkey(LED_D); 

    switch(sys_item)
      {
        case VIEW_VOLTAGE:
            //adc_init(INTERNAL_VREF, NO_LEFT_ADJUST, 2);
            //voltage = 5 * adc_single_conversion(2);         // measure ADC2 - VINT/2
            
            set_cursor(2,2);                // display batt
            const static char text_batt[]    PROGMEM = "Batt:    ";
            lcd_string_P(text_batt);
            lcd_int4(batt_voltage);
            lcd_putc('m');
            lcd_putc('V'); 
                                            // display_vccio
            set_cursor(2,3);
            const static char text_power[]   PROGMEM = "Power:   ";
            lcd_string_P(text_power);
            lcd_bool(g_vccio_state);
            turn_led_off(LED_D);
            break;
        case VIEW_XP:
            set_cursor(2,2);                // display xp states
            const static char text_xp[]      PROGMEM = "XP       ";
            lcd_string_P(text_xp);
            const static char text_xp_on[]      PROGMEM = " on";
            const static char text_xp_off[]      PROGMEM = "off";
            if (xp_connect == XP_CONNECTED) lcd_string_P(text_xp_on);
            else lcd_string_P(text_xp_off);
            
            set_cursor(2,3);
            const static char text_slotid[]  PROGMEM = "SlotID:  ";
            lcd_string_P(text_slotid);
            lcd_int2(slot_id);
            lcd_putc(',');
            if (slot_mode == SLOT_AUTOMATIC) lcd_putc('A');
            else lcd_putc('M');
            turn_led_off(LED_D);
            break;
        case VIEW_LOK:
            set_cursor(2,2);
            const static char text_locos[]   PROGMEM = "#ofLok:  ";
            lcd_string_P(text_locos);
            lcd_int3(get_loco_in_database());
            set_cursor(2,3);
            const static char text_locom[]   PROGMEM = "maxLok:  ";
            lcd_string_P(text_locom);
            lcd_int3(SIZE_LOCO_DATABASE-1);
            turn_led_off(LED_D);
            break;
        case VIEW_SLOTS:
            set_cursor(2,2);
            const static char text_slots[]   PROGMEM = "XP-Slots:  ";
            lcd_string_P(text_slots);
            unsigned char i;
            set_cursor(2,3);
            for (i=0; i < 15; i++)
              {
                if (i==8) lcd_putc(' ');
                if (i == slot_id) lcd_putc('*');
                else lcd_bool(slots_in_use[i]);
              }
            set_cursor(2,4);
            for (i=16; i < 31; i++)
              {
                if (i==24) lcd_putc(' ');
                if (i == slot_id) lcd_putc('*');
                else lcd_bool(slots_in_use[i]);
              }
            turn_led_off(LED_D);
            break;
        case VIEW_XBEE_STATUS:
            set_cursor(2,2);
            const static char text_xb_sts[]   PROGMEM = "XBEE: Status";
            lcd_string_P(text_xb_sts);
            set_cursor(2,3);
            if (xbee_module_detected)
              {
                const static char text_xb_sts_rssi[]   PROGMEM = "RSSI:    -";
                lcd_string_P(text_xb_sts_rssi);
                lcd_int02(xbee_rssi);
                const static char text_xb_sts_dBm[]   PROGMEM = " dBm";
                lcd_string_P(text_xb_sts_dBm);
                
                set_cursor(2,4);
                const static char text_xb_base_macH[]  PROGMEM = "Base-#h: ";
                lcd_string_P(text_xb_base_macH);
                lcd_hex2(base_station_mac[0]);
                lcd_hex2(base_station_mac[1]);
                lcd_hex2(base_station_mac[2]);
                lcd_hex2(base_station_mac[3]);
                
                set_cursor(2,5);
                const static char text_xb_base_macL[]  PROGMEM = "Base-#l: ";
                lcd_string_P(text_xb_base_macL);
                lcd_hex2(base_station_mac[4]);
                lcd_hex2(base_station_mac[5]);
                lcd_hex2(base_station_mac[6]);
                lcd_hex2(base_station_mac[7]);
                
                
                /*
                set_cursor(2,4);
                static char text_xb_sts_good[]   PROGMEM = "#Good:   ";
                lcd_string_P(text_xb_sts_good);
                set_cursor(2,5);
                static char text_xb_sts_bad[]    PROGMEM = "#Bad:    ";
                lcd_string_P(text_xb_sts_bad);
                */
                display_softkey_P(LED_D, text_sk_save); 
                turn_led_red_soft(LED_D);
              }
            else
              {
                const static char text_xb_miss[]   PROGMEM = "XBEE: no module";
                lcd_string_P(text_xb_miss);
              }
            break;
        case VIEW_XBEE_INFO:
            set_cursor(2,2);
            const static char text_xb_info[]   PROGMEM = "XBEE: Info";
            lcd_string_P(text_xb_info);
            set_cursor(2,3);
            if (xbee_module_detected)
              {
                const static char text_xb_info_pan[]  PROGMEM = "PAN:   ";     // ATID
                lcd_string_P(text_xb_info_pan);
                lcd_hex4(pan_id);
                if (xbee_ai_state == 0)
                  {
                    const static char text_joined[]   PROGMEM = ",joined";
                    lcd_string_P(text_joined);
                  }
                else
                  {
                    const static char text_away[]   PROGMEM = ",no net";
                    lcd_string_P(text_away);
                  }
                
                /*set_cursor(2,4);
                const static char text_xb_info_sh[]   PROGMEM = "SH:    ";     // ATSH
                lcd_string_P(text_xb_info_sh);
                lcd_hex8(xbee_serial_high);
                */
                set_cursor(2,4);
                const static char text_xb_info_sl[]   PROGMEM = "SL:    ";     // ATSL
                lcd_string_P(text_xb_info_sl);
                lcd_hex8(xbee_serial_low);
                set_cursor(2,5);
                const static char text_xb_info_vr[]   PROGMEM = "FW-Ver:";     // ATVR
                lcd_string_P(text_xb_info_vr);
                lcd_hex4(xbee_fw_version);
                set_cursor(2,6);
                const static char text_xb_info_hv[]   PROGMEM = "HW-Ver:";     // ATHV
                lcd_string_P(text_xb_info_hv);
                lcd_hex4(xbee_hw_version);

                display_softkey_P(LED_D, text_sk_read); 
                turn_led_green_soft(LED_D);
              }
            else
              {
                const static char text_xb_miss[]   PROGMEM = "XBEE: no module";
                lcd_string_P(text_xb_miss);
                turn_led_off(LED_D);
              }
            break;
        default:
            break;
        case N_of_SYSITEM:
            break;
      }
  }


static void show_header(void)
  {
    cursor_off();
    lcd_clear();
    lcd_string_P(text_system);
    check_status_display(0);            
  }


//------------------------------------------------------------------------
//
// DISPLAY - END
//
//------------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//------------------------------------------------------------------------
//
// Menu Handler
//
//------------------------------------------------------------------------

void action_sysstat(t_navicode par)
  {

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            show_header();
            
            sys_item = VIEW_VOLTAGE;
            display_sys_item();
            break;
        case MENU_UPDATE:
            switch (par.key)
              {
                case KEY_SHIFT:
                    break;
                case KEY_SHIFT_RELEASE:
                     break;
                
                case KEY_SH_A:
                case KEY_SH_B:
                case KEY_SH_C:
                case KEY_SH_D:
                    break;
                case KEY_A:
                    break;
                case KEY_B:
                case KEY_C:
                    break;
                case KEY_D:
                    if (sys_item == VIEW_XBEE_STATUS)
                      {
                        unsigned char j;
                        for(j=0; j<8; j++)
                          {
                            eeprom_write_byte((void *)(eadr_xbee_base_mac0+j), base_station_mac[j]);
                          } 
                        eeprom_write_byte((void *)(eadr_xbee_base),1);     // save as active
                        turn_led_green_soft(LED_D);
                      }
                    if (sys_item == VIEW_XBEE_INFO)
                      {
                        set_task_ready(TASK_INIT_XBEE);   // read again
                        turn_led_red_immediate(LED_D);
                      }
                    break;

                case KEY_KNOB:
                    break;

                case KEY_CENTER:
                    // used for Navigate
                    break;

                case KEY_PARSER_STATE:
                    // no break;
                case KEY_BATT_EMPTY:     
                case KEY_BATT_20_100:    
                case KEY_BATT_40_100:     
                case KEY_BATT_60_100:    
                case KEY_BATT_FULL:
                case KEY_VCCIO: 
                    display_sys_item();
                    break;   
                
                case KEY_DOWN:                      // call up prev item
                    if (sys_item == 0) sys_item = N_of_SYSITEM;
                    sys_item--;
                    display_sys_item();
                    break;

                case KEY_UP:                        // call up next item
                    sys_item++;
                    if (sys_item == N_of_SYSITEM) sys_item = 0;
                    display_sys_item();
                    break;
                 

                case KEY_KNOB_TURNED:
                    
                    break;
              }
            break;
        case MENU_CLOSE:
            break;
      }
  }

