//========================================================================
//
// OpenDCC - Throttle
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      menu_weiche.c
// author:    Jean-Michel Fischer, Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-09-27 V0.01 kw  start
//            2009-02-28 V0.02 kw  changing the turnout addr voids state
//            2009-03-01 V0.03 kw  added query commands
//            2009-03-24 V0.04 kw  conditional assembly for displays
//            2009-03-27 V0.05 kw  softkey for graphic display implemented
//            2009-09-03 V0.06 kw  bugfix in char display
//            2010-06-22 V0.07 kw  msg_transmit
//            2013-12-10 V0.08 kw  added polling for turnouts
//            2014-08-17 V0.09 kw  added english text
//
//------------------------------------------------------------------------
//
// purpose:   Throttle for DCC
//
// content:   The menu screens for lok
//
//========================================================================
// see menu_navigator on how to handle menus
//
//========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // real time kernel
#include "timer.h"
#include "led.h"

#include "keycodes.h"
#include "keyboard.h"
#include "lcd_text.h"
#include "display.h"
#include "database.h"               // loco data base
#include "messages_objects.h"
#include "menu_navigator.h"

//------------------------------------------------------------------------
//
// Global data
//
//------------------------------------------------------------------------
//
// we use the array actual_weiche from messages_objects.c
//
      
static unsigned char edit_focus;

//------------------------------------------------------------------------
//
// DISPLAY - CHAR
//
//------------------------------------------------------------------------

#if (DISPLAY == LCD_EA204)
//   012345678901234567890
//  |--------------------|
// 0|Weiche         ON   |
// 1|                    |
// 2| 345  346  347  348 |
// 3|  .    X    X    .  |
//  |--------------------|
//
#if (LANGUAGE == GERMAN)
	const static char text_adresse[]     PROGMEM = "Adresse";

	const static char text_weiche[]     PROGMEM = "Weiche";
	const static char text_sk_unknown[] PROGMEM = "?";
	const static char text_sk_plus[]    PROGMEM = "+";
	const static char text_sk_minus[]   PROGMEM = "-";
#elif  (LANGUAGE == ENGLISH)
    const static char text_adresse[]     PROGMEM = "Address";

	const static char text_weiche[]     PROGMEM = "Point";
	const static char text_sk_unknown[] PROGMEM = "?";
	const static char text_sk_plus[]    PROGMEM = "+";
	const static char text_sk_minus[]   PROGMEM = "-";;
#else
    #error LANGUAGE is void
#endif

#define WADDR_ZEILE  2


static void display_weichen_addr(unsigned char focus)
  {
  set_cursor(0+focus*5,2);
  lcd_int4(actual_weiche[focus].addr);
  }

static void display_weichen_state(unsigned char focus)
  {
  set_cursor(2+focus*5,3);
  switch(actual_weiche[focus].state) 
      {
        default:   // no break;
        case WS_VOID: lcd_putc('?'); 
                      turn_led_off(focus);
                      break;
        case WS_0:    lcd_putc('-'); 
                      turn_led_red_soft(focus);
                      break;
        case WS_1:    lcd_putc('+');
                      turn_led_green_soft(focus);
                      break;
      }
  }

static void set_cursor_editfocus()
  {
    cursor_on_at(0+edit_focus*5, 2, 3);  // addr
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_weiche);
    check_status_display(0);
  }


//-----------------------------------------------------------------------
//
// DISPLAY - GRAPHICS
//
//-----------------------------------------------------------------------
#elif  (DISPLAY == LCD_EA128x64)


//   012345678901234567890
//  |--------------------|
// 0|Weiche         ON   |
// 1|                    |
// 2| 345  346  347  348 |
// 3|  .    X    X    .  |
//  |--------------------|
//

#if (LANGUAGE == GERMAN)
	const static char text_adresse[]     PROGMEM = "Adresse";

	const static char text_weiche[]     PROGMEM = "Weiche";
	const static char text_sk_unknown[] PROGMEM = "?";
	const static char text_sk_plus[]    PROGMEM = "+";
	const static char text_sk_minus[]   PROGMEM = "-";
#elif  (LANGUAGE == ENGLISH)
    const static char text_adresse[]     PROGMEM = "Address";

	const static char text_weiche[]     PROGMEM = "Point";
	const static char text_sk_unknown[] PROGMEM = "?";
	const static char text_sk_plus[]    PROGMEM = "+";
	const static char text_sk_minus[]   PROGMEM = "-";;
#else
    #error LANGUAGE is void
#endif

#define WADDR_ZEILE  5

static void display_weichen_addr(unsigned char focus)
  {
    cursor_off();
    set_cursor(0+focus*5,WADDR_ZEILE);
    lcd_int4(actual_weiche[focus].addr);
  }

static void display_weichen_state(unsigned char focus)
  {
  switch(actual_weiche[focus].state) 
      {
        default:   // no break;
        case WS_VOID: display_softkey_P(focus, text_sk_unknown); 
                      turn_led_off(focus);
                      break;
        case WS_0:    display_softkey_P(focus, text_sk_minus); 
                      turn_led_red_soft(focus);
                      break;
        case WS_1:    display_softkey_P(focus, text_sk_plus); 
                      turn_led_green_soft(focus);
                      break;
      }
  }

static void set_cursor_editfocus()
  {
    cursor_on_at(edit_focus*5,WADDR_ZEILE,4);
  }

static void show_header(void)
  {
    lcd_clear();
    lcd_string_P(text_weiche);
    check_status_display(0);

    set_cursor(7,WADDR_ZEILE-2);
    lcd_string_P(text_adresse);
  }
//-----------------------------------------------------------------------
//
// DISPLAY - END
//
//-----------------------------------------------------------------------
#else  // DISPLAY
#warning Display not supported
#endif  // DISPLAY


//-----------------------------------------------------------------------
//
// Menu Handler
//
//-----------------------------------------------------------------------


void action_weiche(t_navicode par)
  {
    unsigned char i;
    signed int stemp;

    check_status_display(par.key);

    switch (par.menucode)
      {
        case MENU_OPEN:
            display_time_allowed = 1;
            show_header();

            read_actual_waddr();

            for (i=0; i<4; i++)
              {
                actual_weiche[i].state = WS_VOID;
              }
                        
            for (i=0; i<4; i++)
              {
                display_weichen_addr(i);
              }
            
            msg_clear_accessory_watch(0xFFFF);                    // un-watch all
            for (i=0; i<4; i++)
              {
                msg_send_accessory_query(actual_weiche[i].addr);  // turnout addr
              }
            msg_transmit();
            for (i=0; i<4; i++)
              {
                display_weichen_state(i);
              }
            
            edit_focus = 0;
            set_cursor_editfocus();
            actual_weiche_poll_interval = 37;   // >0: Polling, given in TURNOUT_TICK_PERIOD (50ms)
            break;

        case MENU_UPDATE:
            switch (par.key)
              {
                case KEY_KNOB:
                    edit_focus++;
                    edit_focus &= 0x03;      // mod 4
                    break;
                case KEY_SH_KNOB:
                    edit_focus--;
                    edit_focus &= 0x03;      // mod 4
                    break;
                case KEY_A:
                case KEY_B:
                case KEY_C:
                case KEY_D:
                    i = par.key & 0x03;
                    switch (actual_weiche[i].state)
                      {
                        case WS_VOID: actual_weiche[i].state = WS_0; break; 
                        case WS_0:   actual_weiche[i].state = WS_1; break; 
                        case WS_1:   actual_weiche[i].state = WS_0; break; 
                      }
                    display_weichen_state(i);
                    switch_turnout(actual_weiche[i].addr, actual_weiche[i].state);
                    break;
                case KEY_SH_A:  // with shift: send again
                case KEY_SH_B:
                case KEY_SH_C:
                case KEY_SH_D:
                    i = par.key & 0x03;
                    switch (actual_weiche[i].state)
                      {
                        case WS_VOID: actual_weiche[i].state = WS_1; break; 
                        case WS_0:   actual_weiche[i].state = WS_0; break; 
                        case WS_1:   actual_weiche[i].state = WS_1; break; 
                      }
                    display_weichen_state(i);
                    switch_turnout(actual_weiche[i].addr, actual_weiche[i].state);
                    break;
                case KEY_KNOB_TURNED:
                    stemp = actual_weiche[edit_focus].addr;
                    msg_clear_accessory_watch(actual_weiche[edit_focus].addr);  // remove watch
                    stemp += get_knob_rotate();
                    if (stemp < 1) stemp = 1;
                    if (stemp > 2040) stemp = 2040;
                    actual_weiche[edit_focus].addr = stemp;

                    display_weichen_addr(edit_focus);
                    actual_weiche[edit_focus].state = WS_VOID;  // should be loaded by command station
                    display_weichen_state(edit_focus);
                    msg_send_accessory_query(actual_weiche[edit_focus].addr);
                    msg_transmit();
                    break;
                case KEY_PARSER_SWITCH:
                    for (i=0; i<4; i++)
                      {
                        display_weichen_state(i);
                      }
                    break;
              }
            set_cursor_editfocus();
                    
            break;
        case MENU_CLOSE:
            write_actual_waddr();
            actual_weiche_poll_interval = 0;   // 0: no more Polling
            break;
      }
  }
