//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      messages.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-02-10 V0.01 kw  started
//            2009-03-02 V0.02 kw  added sending of functions and light
//            2009-03-13 V0.03 kw  added config commands
//            2009-04-08 V0.05 kw  wrapper for command_station_status and led control
//            2010-06-14 V0.06 kw  added extended accessory
//            2013-12-10 V0.07 kw  added polling for turnouts
//
//--------------------------------------------------------------------------
//
// purpose:   throttle
// content:   handling of messages issued by menus or turnout
//            (this is an abstraction layer, here we checkout the different
//            interfaces)
//
// 2do:       PoM for accessory still missing
//
//--------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <string.h>

#include "config.h"
#include "hardware.h"
#include "timer.h"
#include "cortos.h"
#include "led.h"
#include "messages_objects.h"
#include "display.h"
#include "database.h"


#include "xp_messages.h"
#include "xp_client_if.h"

#include "wrp_messages.h"
#include "xbee_if.h"



//--------------------------------------------------------------------------
//
// Objects
//
//--------------------------------------------------------------------------

//=== Command Station

volatile t_status command_station_status;    // loaded by corresponding parser
volatile t_status prog_saved_cs_status;   
volatile bool we_are_programming;
volatile bool we_should_ask_again;           // true, if we should query the prog result again
volatile char turnout_state_query;           // true, if we should ask cs about turnout state


//=== Zustand aktuellen LOK

locomem_t current_loc;                      // this is a structure holding parameters

unsigned int active_locos[5];                // here we save addresses of our controlled
                                             // locos
                                             // we can have up to 5 locos running

t_database_entry new_database_entry;         // parser dumps new database entry here


unsigned char turnout_delay;

void restore_turnout_delay(void)
  {
    turnout_delay = eeprom_read_byte((void *)eadr_turnout_delay);
  }

void save_turnout_delay(unsigned char delay)
  {
    if (turnout_delay != delay) 
      {
        turnout_delay = delay;
        eeprom_write_byte((void *)eadr_turnout_delay, turnout_delay);
      }
  }

//------------------------------------------------------------------------
t_turnout actual_weiche[4];                 // we have 4 turnouts to watch
unsigned char actual_weiche_poll_interval;  // 0: no Polling
                                            // >0: Polling, given in TURNOUT_TICK_PERIOD (50ms)
                                            // recommended: 25

t_signal actual_signal[4];    // extended accessory


// -- feedback stuff

int fb_addr[2];        // we watch 2 feedback addresses

volatile unsigned char fb_array[FB_ARRAY_SIZE] = {0x55, 0xAA, 0x33};


//-- tunnel mode

unsigned char pc_message[17];

//--------------------------------------------------------------------------
//
// Command Station
//
//--------------------------------------------------------------------------

void set_command_station_status(t_status new_stat)
  {
    if (new_stat != command_station_status)
      {
        command_station_status = new_stat;
        switch(command_station_status)
          {
            case STS_ON:
                turn_led_green_soft(LED_STOP);
                break;
            case STS_OFF:
                turn_led_red_immediate(LED_STOP);
                break;
            case STS_STOP:
                blink_led_green_soft(LED_STOP);
                break;
            case STS_PROG:
                blink_led_green_fast(LED_STOP);
                break;
            case STS_BUSY:
                blink_led_red_fast(LED_STOP);
                break;
            case STS_PBUSY:
                // nothing
                break;
            case STS_DISCONNECTED:
                blink_led_red_green(LED_STOP);
                break;
            case STS_CMD_ERROR:
                blink_led_red_fast(LED_STOP);
                break;
          }
      }
  }


//--------------------------------------------------------------------------
//
// Object-Tracker
//
//--------------------------------------------------------------------------



#define TURNOUT_TICK_PERIOD  50000L 
#if ((TURNOUT_TICK_PERIOD / SYSTICK_PERIOD) == 0)
  #error: TURNOUT_TICK_PERIOD too small!
#endif

#define NUM_OF_ACTIV_TURNOUTS  16

typedef struct
  {
    unsigned int addr;    // this the real dcc accessory address
    unsigned char delay;  // given in TURNOUT_TICK_PERIOD
    enum {TURN_ON, TURN_OFF, TURN_QUERY} phase;
  } t_do_turnout;

t_do_turnout turnout[NUM_OF_ACTIV_TURNOUTS];


static unsigned char query_lok_other_time;  // here we save the time, until we ask again for this lok
static unsigned char actual_weiche_poll_counter;

t_cr_task check_objects(void)
  {
    unsigned char i;

    // Weichen Pollen (sofern durch actual_weiche_poll_interval aktiviert)
    if (actual_weiche_poll_interval)
      {
        actual_weiche_poll_counter++;
        if (actual_weiche_poll_counter++ >= actual_weiche_poll_interval)
          {
            actual_weiche_poll_counter = 0;
            for (i=0; i<sizeof(actual_weiche)/sizeof(actual_weiche[0]); i++) // 4 turnouts at watch
              {
                msg_send_accessory_query(actual_weiche[i].addr);
              }
          }
      }

    // Weichen schalten (Spule / on / off)
    for (i=0; i<NUM_OF_ACTIV_TURNOUTS; i++)
      {
        if (turnout[i].delay)
          {
            turnout[i].delay--;
            if (turnout[i].delay == 0)
              {
                switch(turnout[i].phase)
                  {
                    default:
                        break;
                    case TURN_ON:
                        break;
                    case TURN_OFF:
                        msg_send_accessory(turnout[i].addr, COIL_OFF);
                        turnout[i].delay = 200000L / TURNOUT_TICK_PERIOD;
                        turnout[i].phase = TURN_QUERY;
                        break;
                    case TURN_QUERY:
                        msg_send_accessory_query((turnout[i].addr / 2) + 1);  // this need turnout addr
                        turnout[i].phase = TURN_ON;
                  }
              }
          }
      }

    // if a lok is stolen - watch it!
    // we ask every second (1000000L / TURNOUT_TICK_PERIOD)

    if (current_loc.other && current_loc.addr != 0) 
      {
        query_lok_other_time++;
        if (query_lok_other_time == 1000000L / TURNOUT_TICK_PERIOD)
          {
            query_lok_other_time = 0;
            msg_send_askforlocinfo(current_loc.addr);
          }
      }

    if (we_are_programming && we_should_ask_again)
      {
        msg_send_askforresult();
      }

    msg_transmit();       // if any of the above created a msg_send - lets transmit it.
    return(TURNOUT_TICK_PERIOD / SYSTICK_PERIOD);
  }


// id: range 1 .. 2040 (also turnout-addr)
void switch_turnout(unsigned int id, t_turnout_state state)
  {
    unsigned char i, offset = 0;
    unsigned int addr;          // real dcc coil addr

    switch(state)
      {
        default:
        case WS_VOID: return; break; 
        case WS_0:    offset = 0; break; 
        case WS_1:    offset = 1; break; 
      }

    if ((id > 0) && (id < 2040))
      {
        id--;
        addr = id * 2 + offset;

        msg_send_accessory(addr, COIL_ON);   // dcc (coil-addr)
    
        for (i=0; i<NUM_OF_ACTIV_TURNOUTS; i++)
          {
            if (turnout[i].phase == TURN_ON)
              {
                turnout[i].addr = addr;
                if (turnout_delay)
                  {
                    turnout[i].delay = turnout_delay+1;
                    turnout[i].phase = TURN_OFF;
                  }
                else  // switchtime is 0 -> no coil OFF command, only a query
                  {
                    turnout[i].delay = 200000L / TURNOUT_TICK_PERIOD;
                    turnout[i].phase = TURN_QUERY;
                  }
                return;                  
              }
          }
        display_error(ERR_TURNOUT_SPACE);
      }
    return;
  }



void init_objects(void)
  {
    unsigned char i;

    turnout_state_query = eeprom_read_byte((void *)eadr_turnout_state_query);
    read_actual_waddr();
    read_actual_saddr();


    query_lok_other_time = 0;

    current_CV.result = PRG_CV_VOID;
    current_CV.number = 1;

    set_command_station_status(STS_DISCONNECTED);    

    restore_turnout_delay();
    
    for (i=0; i<NUM_OF_ACTIV_TURNOUTS; i++)
      {
        turnout[i].delay = 0;
        turnout[i].phase = TURN_ON;
      }

    for ( i=0; i < sizeof(active_locos)/sizeof(active_locos[0]); i++)
      {
        active_locos[i] = 0;
      }
  }

//--------------------------------------------------------------------------
//
// Interface
//
//--------------------------------------------------------------------------

// used_interface is defined in config.h

void msg_start_interface(t_used_interface new_interface)
  {
    set_command_station_status(STS_DISCONNECTED);
    // cr_fifo_put(&key_fifo, KEY_PARSER_STATE);       // force update
    // cr_fifo_put(&key_fifo, KEY_DISCONNECT);       // force update
    
    used_interface = new_interface;
    switch(new_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            ENABLE_USB;
            used_interface = USE_XPRESSNET;
            restart_xpressnet_client();
            xp_send_cs_status_request();
            break;
        case USE_XBEE:
            ENABLE_XBEE;
            used_interface = USE_XBEE;
            restart_xbee();

                                               // Verbindungsaufbau abwarten!
            _delay_ms(20);                     // busy waiting! not fair and not secure !!!!

            wrp_send_cs_status_request();
            break;
        case USE_CAN:
            break;
      }
  }

// this routine must be called, when a block of messages is complete.
void msg_transmit(void)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            // nothing to do here - each message is transmitted alone, no pakets
            break;
        case USE_XBEE:
            wrp_transmit();                              // clean up any pending messages
            break;
        case USE_CAN:
            break;
      }
  }

t_connect_state get_connection_status(void)
  {
    t_connect_state cs;

    switch(used_interface)
      {
        default:
        case USE_NONE:
            cs = DISCONNECTED;
            break;
        case USE_USB:
            cs = DISCONNECTED;
            break;
        case USE_XPRESSNET:
            if (xp_connect == XP_CONNECTED) cs = CONNECT_XPRESSNET;
            else cs = DISCONNECTED;
            break;
        case USE_XBEE:
            if (xbee_connect == XBEE_RUNNING) cs = CONNECT_XBEE;
            else cs = DISCONNECTED; // hier wird das XP disconnected Symbol angezeigt
            break;
        case USE_CAN:
            cs = DISCONNECTED;
            break;
      }
    return(cs);
  }

//--------------------------------------------------------------------------
//
// Messages
// (Note: don't forget to call msg_transmit(); !)
//
//--------------------------------------------------------------------------




void msg_send_cs_status_request(void)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_cs_status_request();
            break;
        case USE_XBEE:
            wrp_send_cs_status_request();
            wrp_transmit();
            break;
        case USE_CAN:
            break;
      }
  }

void msg_send_state(t_status new_state)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_state(new_state);
            break;
        case USE_XBEE:
            wrp_send_state(new_state);
            wrp_transmit();
            break;
        case USE_CAN:
            break;
      }
  }

// parameter: feedback block addr (=module address, bytewise)
void msg_send_feedback_query(unsigned int addr)
  {
    if (turnout_state_query)  // this is flag loaded from eeprom
      {
        switch(used_interface)
          {
            default:
            case USE_NONE:
                break;
            case USE_USB:
                break;
            case USE_XPRESSNET:
                xp_send_feedback_query(addr);
                break;
            case USE_XBEE:
                wrp_send_feedback_query(addr);
                wrp_send_feedback_subscribe();
                break;
            case USE_CAN:
                break;
          } 
      }
  }


//----------------------------------------------------------------------------
//
// Turnouts
//
//----------------------------------------------------------------------------

// parameter: coil addr [0...
void msg_send_accessory(unsigned int addr, bool activate)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_accessory(addr, activate);
            break;
        case USE_XBEE:
            wrp_send_accessory(addr, activate);
            break;
        case USE_CAN:
            break;
      } 
  }


// parameter: turnout addr (range 1..2040)
void msg_send_accessory_query(unsigned int addr)
  {
    if (turnout_state_query)  // this is flag loaded from eeprom
      {
        switch(used_interface)
          {
            default:
            case USE_NONE:
                break;
            case USE_USB:
                break;
            case USE_XPRESSNET:
                xp_send_accessory_query(addr);
                break;
            case USE_XBEE:
                wrp_send_accessory_query(addr);
                break;
            case USE_CAN:
                break;
          } 
      }
  }

void msg_clear_accessory_watch(unsigned int addr)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            // nothing to do here, always as broadcast
            break;
        case USE_XBEE:
            wrp_clear_accessory_watch(addr);
            break;
        case USE_CAN:
            break;
      } 
  }

void msg_send_extended_accessory(unsigned int addr, unsigned char aspect)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_extended_accessory(addr, aspect);
            break;
        case USE_XBEE:
            wrp_send_extended_accessory(addr, aspect);
            break;
        case USE_CAN:
            break;
      } 
  }


//----------------------------------------------------------------------------
//
// Lok-command
//
//----------------------------------------------------------------------------


void record_inactive_loco(unsigned int addr)
  {
    unsigned char i;
     for ( i=0; i<sizeof(active_locos)/sizeof(active_locos[0]); i++)
      {
        if (active_locos[i] == addr)
          {
            active_locos[i] = 0;                // remove from record
            return;
          }
      }
  }

void record_active_loco(unsigned int addr, unsigned char speed)
  {
    unsigned char i;
    if (speed > 0)
      {
        for ( i=0; i<sizeof(active_locos)/sizeof(active_locos[0]); i++)
          {
            if (active_locos[i] == addr)  return;   // already recorded
          }
        for ( i=0; i<sizeof(active_locos)/sizeof(active_locos[0]); i++)
          {
            if (active_locos[i] == 0)
              {
                active_locos[i] = addr;             // add to record
                return;
              }
          }
        // if we come here - error, record is full
      }
    else
      {
        record_inactive_loco(addr);
      }
  }



// parameter:
// addr: dcc addr
// format: DCC14, DCC28, DCC128
// speed:  0..126
// dir:    0 = forward, 1 = revers
void msg_send_loc_drive(unsigned int addr, t_format format, unsigned char speed, unsigned char dir)
  {
    
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_loc_drive(addr, format, speed, dir);
            break;
        case USE_XBEE:
            wrp_send_loc_drive(addr, format, speed, dir);
            break;
        case USE_CAN:
            break;
      } 
  }

// note: current_loc.f?? must be loaded completely
// we transmit only the according group to save bandwidth
// parameter: fnum, range 0..27 corresponds to f1 ... f28
void msg_send_loc_function(unsigned char fnum)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_loc_function(fnum);
            break;
        case USE_XBEE:
            wrp_send_loc_function();    // with wrp, we always do the complete set of functions
            break;
        case USE_CAN:
            break;
      } 
  }

// note: current_loc.light and others must be up to date
// we transmit only the neccesary command for light to save bandwidth
// parameter: none
void msg_send_loc_light(void)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_loc_light();
            break;
        case USE_XBEE:
            wrp_send_loc_light();
            break;
        case USE_CAN:
            break;
      } 
  }

void msg_send_loc_stop(unsigned int addr)
  {
    record_inactive_loco(addr);         // remove this from our records
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_loc_stop(addr);
            break;
        case USE_XBEE:
            wrp_send_loc_stop(addr);    // nothalt
            break;
        case USE_CAN:
            break;
      } 
  }

void msg_send_all_loc_stop(void)
  {
    unsigned char i;
     for ( i=0; i<sizeof(active_locos)/sizeof(active_locos[0]); i++)
      {
        if (active_locos[i] != 0)
          {
            msg_send_loc_stop(active_locos[i]);
            return;
          }
      }
  }

// we send a query for parameters of this lok
// these will (hopefully) show up in current_loc soon.
void msg_send_askforlocinfo(unsigned int addr)
  {
    current_loc.addr = addr;
    current_loc.valid = 0;
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_askforlocinfo(addr);
            break;
        case USE_XBEE:
            wrp_send_askforlocinfo(addr);
            break;
        case USE_CAN:
            break;
      } 
  }


//----------------------------------------------------------------------------
//
// CV-stuff
//
//----------------------------------------------------------------------------

void check_prog_save_status(void)
  {
    we_are_programming = TRUE;
    if (command_station_status <= STS_STOP)   // off, on or stop
      {
        prog_saved_cs_status = command_station_status;    
      } 
  }

void msg_send_cv_read(unsigned int cv)  // 1..1024
  {
    check_prog_save_status();
    current_CV.prog_request_timestamp = get_systick_int();
    current_CV.result = PRG_CV_WAIT_FOR_ANSWER;
    current_CV.number = cv;
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_cv_read(cv);
            break;
        case USE_XBEE:
            wrp_send_cv_read(cv);
            break;
        case USE_CAN:
            break;
      } 
    msg_send_askforresult();
  }


void msg_send_cv_write(unsigned int cv, unsigned char value)
  {
    check_prog_save_status();
    current_CV.prog_request_timestamp = get_systick_int();
    current_CV.result = PRG_CV_WAIT_FOR_ANSWER;
    current_CV.number = cv;
    current_CV.value = value;
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_cv_write(cv, value);
            break;
        case USE_XBEE:
            wrp_send_cv_write(cv, value);
            break;
        case USE_CAN:
            break;
      } 
    msg_send_askforresult();
  }


void msg_send_cv_end(void)
  {
    if (we_are_programming)
      {
        we_are_programming = 0;
        if (prog_saved_cs_status <= STS_STOP)   // off, on or stop
          {
            msg_send_state(prog_saved_cs_status);   
          } 
        else
            msg_send_state(STS_ON);
      }
  };



void msg_send_askforresult(void)
  {
    we_should_ask_again = FALSE;

    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_askforresult();
            break;
        case USE_XBEE:
            wrp_send_askforresult();
            break;
        case USE_CAN:
            break;
       } 
    }


// PoM
void msg_send_PoM_cv_read(unsigned int addr, unsigned int cv)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_PoM_cv_read(addr, cv);
            break;
        case USE_XBEE:
            wrp_send_PoM_cv_read(addr, cv);
            break;
        case USE_CAN:
            break;
      } 
  }

// PoM
void msg_send_PoM_cv_write(unsigned int addr, unsigned int cv, unsigned char value)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_PoM_cv_write(addr, cv, value);
            break;
        case USE_XBEE:
            wrp_send_PoM_cv_write(addr, cv, value);
            break;
        case USE_CAN:
            break;
      } 
  }

//----------------------------------------------------------------------------
//
// command station configure
//
//----------------------------------------------------------------------------

void msg_send_cs_cfg_read(unsigned int cv)
  {
    current_CV.result = PRG_CV_WAIT_FOR_ANSWER;
    current_CV.number = cv;
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_cs_cfg_read(cv);
            break;
        case USE_XBEE:
            wrp_send_cs_cfg_read(cv);
            break;
        case USE_CAN:
            break;
      } 
  }


void msg_send_cs_cfg_write(unsigned int cv, unsigned char value)
  {
    current_CV.result = PRG_CV_WAIT_FOR_ANSWER;
    current_CV.number = cv;
    current_CV.value = value;
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_cs_cfg_write(cv, value);
            break;
        case USE_XBEE:
            wrp_send_cs_cfg_write(cv, value);
            break;
        case USE_CAN:
            break;
      } 
  }

void msg_send_clock_set(t_fast_clock *my_clock)
  {
    switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_clock_set(my_clock);
            break;
        case USE_XBEE:
            wrp_send_clock_set(my_clock);
            break;
        case USE_CAN:
            break;
      } 
  }


//----------------------------------------------------------------------------
//
// pc interface
//
//----------------------------------------------------------------------------
// 

void msg_send_pc_hello(void)
  {
     switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_pc_hello();
            break;
        case USE_XBEE:
            break;
        case USE_CAN:
            break;
      }
  }

void msg_send_pc_goodbye(void)
  {
     switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_pc_goodbye();
            break;
        case USE_XBEE:
            break;
        case USE_CAN:
            break;
      } 
  }

void msg_send_PID(void)
  {
     switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_PID();
            break;
        case USE_XBEE:
            break;
        case USE_CAN:
            break;
      }
  }



void msg_send_pc_key(unsigned char key, signed char value)
  {
     switch(used_interface)
      {
        default:
        case USE_NONE:
            break;
        case USE_USB:
            break;
        case USE_XPRESSNET:
            xp_send_pc_key(key, value);
            break;
        case USE_XBEE:
            break;
        case USE_CAN:
            break;
      } 
  }

  
 
