//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      messages_objects.h
// author:    Wolfgang Kufer, Jean Michel Fischer
// contact:   kufer@gmx.de
// history:   2009-02-10 V0.01 kw  started
//            2009-02-26 V0.02 kw  added actual_weiche
//            2009-03-01 V0.03 kw  added query for turnouts
//            2009-03-14 V0.04 kw  added turnout_state_query
//            2009-04-08 V0.05 kw  wrapper for command_station_status
//            2010-06-16 V0.02 kw  added actual_signal (extended accessory)
//
//--------------------------------------------------------------------------
//
// purpose:   throttle
// content:   handling of messages issued by menus or turnout
//            depending on the used interface, these messages go to
//            - Xpressnet
//            - ZigBEE
//            - CAN
//            definition of objects - like lok, turnouts, feedback
//--------------------------------------------------------------------------


#ifndef _MESSAGES_H_
#define _MESSAGES_H_

#define COIL_OFF   0
#define COIL_ON    1


#include <stdbool.h>


//-------------------------------------------------------------------------------
//
// Objects
//
//-------------------------------------------------------------------------------
//
//=== Command Station

extern volatile t_status command_station_status;    // loaded by parser
extern volatile bool we_are_programming;            // true, if we own the command station
extern volatile bool we_should_ask_again;           // true, if we should query the prog result again
extern volatile char turnout_state_query;           // true, if we should ask cs about turnout state

//=== Zustand aktuellen LOK

extern locomem_t current_loc;



typedef struct                                      // Lok-Datenbank
  {
    unsigned char idx;
    unsigned char size;
    unsigned int  addr;
    char name[LOK_NAME_LENGTH];
  } t_database_entry;

extern t_database_entry new_database_entry;


//=============== Zustand der aktuellen Weiche

extern unsigned char turnout_delay;       // Switch time for turnouts, given in 50ms

typedef enum {WS_VOID, WS_0, WS_1} t_turnout_state; 

typedef struct
  {
    unsigned int addr;
    t_turnout_state state;
  } t_turnout;

extern t_turnout actual_weiche[4];
extern unsigned char actual_weiche_poll_interval;    // 0: no Polling
                                                     // >0: Polling, given in TURNOUT_TICK_PERIOD (50ms)
                                                     // recommended: 25

//================== Zustand der aktuellen Signals (extended Accessory)

typedef struct
  {
    unsigned int addr;
    unsigned char state;
  } t_signal;

extern t_signal actual_signal[4];


//================== CV-Access

typedef enum {  PRG_CV_VOID,
                PRG_CV_WAIT_FOR_ANSWER,                      
                PRG_CV_SHORT, 
                PRG_CV_NOLOK, 
                PRG_CV_OKAY,
                PRG_CV_CHANGED,            // changed in editor, but not written
                sizeof_t_prog_result,           
            } t_prog_result;

typedef struct                      // this is parameter passed along for menus
  {
    t_prog_result result;
    int prog_request_timestamp;
    unsigned int  addr;
    unsigned int  number;           // range 1..1024
    unsigned char value;
  }
  t_cv_info;

t_cv_info current_CV;


//=== Feedback

#define FB_ARRAY_SIZE 128

volatile unsigned char fb_array[FB_ARRAY_SIZE];

extern int fb_addr[2];        // we watch 2 feedback addresses

// volatile unsigned int cFeedbackIndex;


//=== Tunnel

extern unsigned char pc_message[17];




void set_command_station_status(t_status new_stat);

//-------------------------------------------------------------------------------
//
// Calls (Task and Control)
//
//-------------------------------------------------------------------------------

t_cr_task check_objects(void);

// send both COIL_ON and COIL_OFF to the turnout
void switch_turnout(unsigned int id, t_turnout_state state);

void restore_turnout_delay(void);
void save_turnout_delay(unsigned char delay);


void init_objects(void);



//-------------------------------------------------------------------------------
//
// Calls (Messages)
//
//-------------------------------------------------------------------------------

// first call to establish the link on 'used_interfaces'
void msg_start_interface(t_used_interface new_interface);

void msg_transmit(void);

unsigned char get_connection_status(void);


// tell command station the request for a new state
void msg_send_state(t_status new_state);

// ask command station for actual state
void msg_send_cs_status_request(void);

// addr = module (byte wise)
void msg_send_feedback_query(unsigned int addr);


// parameter:
// addr: dcc addr
// format: DCC14, DCC28, DCC128
// speed:  0..126
// dir:    0 = forward, 1 = revers
void msg_send_loc_drive(unsigned int addr, t_format format, unsigned char speed, unsigned char dir);

void msg_send_askforlocinfo(unsigned int locidid);

void msg_send_loc_stop(unsigned int addr);

void msg_send_all_loc_stop(void);

void msg_send_loc_function(unsigned char fnum);

void msg_send_loc_light(void);

void record_inactive_loco(unsigned int addr);


// parameter:
// addr: dcc addr of coil (not turnout)
// activate:    COIL_ON or COIL_OFF
void msg_send_accessory(unsigned int addr, bool activate);

// addr: turnout, range 1..2040
void msg_send_accessory_query(unsigned int addr);

void msg_clear_accessory_watch(unsigned int addr);

void msg_send_extended_accessory(unsigned int addr, unsigned char aspect);

// read a config var.
void msg_send_cv_read(unsigned int cv);

void msg_send_cv_write(unsigned int cv, unsigned char value);

void msg_send_askforresult(void);

void msg_send_cv_end(void);

void msg_send_PoM_cv_read(unsigned int addr, unsigned int cv);

void msg_send_PoM_cv_write(unsigned int addr, unsigned int cv, unsigned char value);


//cs config
void msg_send_cs_cfg_read(unsigned int cv);

void msg_send_cs_cfg_write(unsigned int cv, unsigned char value);

void msg_send_clock_set(t_fast_clock *my_clock);


void msg_send_pc_hello(void);

void msg_send_pc_goodbye(void);

void msg_send_PID(void);

void msg_send_pc_key(unsigned char key, signed char value);





#endif // _MESSAGES_H_
