//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-07-01 V0.01 kw  started
//            2009-07-04 V0.02 kw  get_arg now accepts 0x**** as hex
//                                 debug commands (all start with X...)
//            2009-07-07 V0.03 kw  added mirror of wrong command 
//
//--------------------------------------------------------------------------
//
// purpose:   Throttle
// content:   command line parser for host interface
//
//--------------------------------------------------------------------------

#include <inttypes.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>

#include "cortos.h"             // task system

#include "config.h"
#include "serial.h"
#include "pc_send.h"
#include "database.h"

#include "flash_driver_m25p80.h"


//----------------------------------------------------------------------------
//
// Variables and Tools for the Parser
//
// Im Gegensatz zu sonstigen C Gepflogenheiten halten wir den Pointer auf pcc
// global; damit kann der String mit einem Pointer durchsucht werden.
// 
unsigned char pcc[48];          // pc_message speicher

unsigned char pcc_size;         // if known: size of current command (for p50x_binary)
unsigned char pcc_index;        // current write or read index

unsigned char *ibcmd;           // globaler Pointer auf commandstring


enum
  {
    IDLE,
    WAIT_FOR_COMMAND,
    WAIT_FOR_PARAM
  } parser_state;


//------------------------------------------------------------------------------
// Hilfsrountinen:

signed char char2hex (unsigned char *arg)
  {
    if ((*arg >= '0') && (*arg <= '9')) return (*arg - '0');
    if ((*arg >= 'A') && (*arg <= 'F')) return (*arg - 'A' + 10);
    if ((*arg >= 'a') && (*arg <= 'f')) return (*arg - 'a' + 10);
    return (-1);  // no hex
  }

signed char char2bin (unsigned char *arg)
  {
    if (*arg == '0') return(0);
    if (*arg == '1') return(1);
    return (-1);  // no bin
  }

//-----------------------------------------------------------------------------------
// Diese Funktionen nochmal, jetzt als Spezialversion mit
// globalen Variablen; Der pcc_index bleibt auf dem ersten falschen
// Zeichen stehen.

// axtoi bricht beim ersten non-hex-Wert ab

unsigned int get_cmdvalx(void)
  {
    unsigned int intValue = 0;      // integer value of hex string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)                 // security limit against wrong string    
      {
        if ((temp=char2hex(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 4) + temp;
        ibcmd++;
      }
    return (intValue);
  }

unsigned int get_cmdvalb(void)
  {
    unsigned int intValue = 0;      // integer value of bin string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)             
      {
        if ((temp=char2bin(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 1) + temp;
        ibcmd++;
      }
    return (intValue);
  }
// the following routine reads one argument from commandline
// parameter:   pointer to int, where to put the result
//              parsing rules:
//              leading %: the number is parsed as binary
//              leading $: the number is parsed as hex
//              char r and g are parsed as 0 and 1
// return:      -1: EOL reached
//              -2: no valid arg, syntax error
//              0:  no parameter found, *result will be 0xffff;
//              1:  there is an arg, value is returned in *result
//
// side effect: ibcmd steht auf dem ersten Zeichen nach Delimiter
//
int get_px50arg(unsigned int *result)
  {
    unsigned int argValue = 0;      // integer value of hex string
    
    while (*ibcmd == 0x20) ibcmd++;  // skip leading blanks
    
    if (*ibcmd == 0)
      {
        return(-1);  // empty arg, end of line!
      }
    if (*ibcmd == ',')
      {
        ibcmd++;
        *result = 0xffff;
        return(0);  // empty arg 
      }
    if (*ibcmd == '$')
      {
        ibcmd++;
        argValue = get_cmdvalx();
        if (*ibcmd == ',') ibcmd++;   // skip trailing ',' 
      }
    else if ((*ibcmd == '0') && ((*(ibcmd+1) | 0x20) == 'x'))
      {
        ibcmd+=2;
        argValue = get_cmdvalx();
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if (*ibcmd == '%')
      {
        ibcmd++;
        argValue = get_cmdvalb();
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if ((*ibcmd | 0x20) == 'r')
      {
        ibcmd++;
        argValue = 0;
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if ((*ibcmd | 0x20) == 'g')
      {
        ibcmd++;
        argValue = 1;
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if ((*ibcmd >= '0') && (*ibcmd <= '9'))
      {
        argValue = atoi((char *)ibcmd);
        while ((*ibcmd >= '0') && 
               (*ibcmd <= '9'))  ibcmd++;  // skip numbers

        if (*ibcmd == ',') ibcmd++;   // skip ',' 
      }
    else
          {
            return(-2);                    // void arg
          }
    *result = argValue;
    return(1);
  }

unsigned char is_delimiter(unsigned char test)
  {
    if (test == 0) return(1);   // EOL
    if (test == ' ') return(1);
    if (test == ',') return(1);
    return(0);
  }

unsigned char is_delimiter_or_num(unsigned char test)
  {
    if (test == 0) return(1);   // EOL
    if (test == ' ') return(1);
    if (test == ',') return(1);
    if (test == '%') return(1);
    if (test == '$') return(1);
    if ((test >= '0') && (test <= '9'))  return(1);  // skip numbers
    return(0);
  }


// advance parameter pointer to next arg.
void next_arg(void)
  {
    while (*ibcmd) 
      {
        if (is_delimiter(*ibcmd))
          {
             ibcmd++;
             while (*ibcmd == 0x20) ibcmd++;  // skip leading blanks
             return;
          }
        ibcmd++;
      }
  }


unsigned int parameter[10];

void parse_parameters(void)
  {
    unsigned char i;
    
    for (i=0; i<(sizeof(parameter)/sizeof(unsigned int)); i++)
      {
        parameter[i] = 0xfffe;          // mark as void
      }

    for (i=0; i<(sizeof(parameter)/sizeof(unsigned int)); i++)
      {
        if (get_px50arg(&parameter[i]) < 0)
          {
            // eol or void parameter reached
            break;
          }
      }
  }

//-------------------------------------------------------------------------------

unsigned char lok_addr_out_of_range(unsigned int *addr)       // Hilfsroutine: range der Lok Addr prfen
  {
    if ((*addr >= 1) && (*addr <= 10239)) return(0);
    else return(1);
  }


//------------------------------------------------------------------------
// 2. Command Parser
//------------------------------------------------------------------------
//
// a) run_parser: collects a complete input command and starts to parse it,
// when a terminating white char is detected.
//
// b) parse: checks for opcodes and parameter
//  and calls the corresponding function
//
// c) PA_xx: this are the functions
//


const static char text_info1[]  PROGMEM = "MFT_";
const static char text_info2[]  PROGMEM = " 1.0, version ";
const static char text_info3[]  PROGMEM = ", api 01";
           
void PA_info(void)
  {
    #if (SIZE_LOCO_DATABASE > 250)
       error (char - int)
    #endif
    pc_send_string_P(text_info1);
    #if (DISPLAY == LCD_EA204)
       pc_send('C');
    #elif (DISPLAY == LCD_EA128x64)
       pc_send('G');
    #else
      #warning DISPLAY??
    #endif
    pc_send_string_P(text_info2);
    pc_send_int02(MFT_VERSION);
    pc_send('.');
    pc_send_int02(MFT_SUBVERSION);
    pc_send_answer_P(text_info3); 
  }

const static char text_help01[]  PROGMEM = "Commands:";
const static char text_help02[]  PROGMEM = " HELP: this helptext; HELP 1 shows debug commands";
const static char text_help03[]  PROGMEM = " BACKUP: dump a complete list of loco and parameters";
const static char text_help04[]  PROGMEM = " INFO: show device type, software version, API version";
const static char text_help05[]  PROGMEM = " LOCADD <addr>, <steps>, <format>, <name>: add a locomotive";
const static char text_help06[]  PROGMEM = " LOCDEL <addr>: delete a locomotive";
const static char text_help07[]  PROGMEM = " LOCDUMP: print all known locomotives: addr, steps, format, name, picture-id";
const static char text_help08[]  PROGMEM = " LOCFREE: print # of free loco memory";
const static char text_help09[]  PROGMEM = " LOCPIC <addr> <picture-id>: assign picture to loco";
const static char text_help10[]  PROGMEM = " LOCQUERY <addr>: show parameters of this loco";

const static char text_debug01[] PROGMEM = "Debug-Commands:";
const static char text_debug02[] PROGMEM = " XER <addr>: read eeprom";
const static char text_debug03[] PROGMEM = " XEW <addr> <data>: write eeprom";
const static char text_debug04[] PROGMEM = " XFR <addr>: read external flash";
const static char text_debug05[] PROGMEM = " XFW <addr> <data>: write external flash";
const static char text_debug06[] PROGMEM = " XFERA: erase external flash";


void PA_help(void)
  {
    if (parameter[0] == 1)
      {
        pc_send_answer_P(text_debug01);
        pc_send_answer_P(text_debug02);
        pc_send_answer_P(text_debug03);
        pc_send_answer_P(text_debug04);
        pc_send_answer_P(text_debug05);
        pc_send_answer_P(text_debug06);
      }
    else
      {
        pc_send_answer_P(text_help01);
        pc_send_answer_P(text_help02);
        pc_send_answer_P(text_help03);
        pc_send_answer_P(text_help04);
        pc_send_answer_P(text_help05);
        pc_send_answer_P(text_help06);
        pc_send_answer_P(text_help07);
        pc_send_answer_P(text_help08);
        pc_send_answer_P(text_help09);
        pc_send_answer_P(text_help10);
      }
  }


const static char text_error[]    PROGMEM = "Error: ";
const static char text_ok[]       PROGMEM = "OK";
const static char text_no[]       PROGMEM = "No ";


const static char text_unknown[]  PROGMEM = "Unknown command - type 'help'";
void PA_unknown(void)
  {
    unsigned char j;
    pc_send_string_P(text_error);
    pc_send_string_P(text_unknown);
    //
    ibcmd = &pcc[0];
    for (j=0; j < 20; j++)
      {
        if ((*ibcmd == 0x0a) || (*ibcmd == 0x0d))       // terminator
          {
            pc_send_crlf(); 
            return;
          }
        pc_send( *ibcmd++);
      } 
    pc_send_crlf(); 
  }

void PA_locfree(void)
  {
    unsigned char free;
    free = SIZE_LOCO_DATABASE - 1 - get_loco_in_database();
    pc_send_int(free); 
    pc_send_crlf(); 
  }



const char text_locaddr_outrange[]      PROGMEM = "Lok addr out of range";
const char text_locformat[]             PROGMEM = "Format not DCC";
const char text_buffer_full[]           PROGMEM = "Lok buffer full";
const char text_speedsteps[]            PROGMEM = "wrong speed steps";

const char text_loco[]                 PROGMEM = "Loco ";
const char text_pid[]                  PROGMEM = ", Pict-ID=0x";

void PA_send_no_lok(void)
  {
    pc_send_string_P(text_error);
    pc_send_string_P(text_no);
    pc_send_string_P(text_loco);
    pc_send_num(parameter[0]);
    pc_send_crlf();
  }


void PA_locpic(void)
  {
    unsigned int my_pic;
    if (lok_addr_out_of_range(&parameter[0]))
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_locaddr_outrange); return;
      }
    if (parameter[1] != 0xFFFE)
      {
        if (assign_picture_to_loco(parameter[0],parameter[1]))
          {
            pc_send_answer_P(text_ok); return;
          }
        else
          {
            PA_send_no_lok(); return;
          }
      }
    else
      { // query        
        if (query_picture_to_loco(parameter[0], &my_pic))
         {
            pc_send_string_P(text_loco);
            pc_send_num(parameter[0]);pc_send_string_P(text_pid);
            pc_send_hex(my_pic / 256);
            pc_send_hex(my_pic);
            pc_send_crlf(); return;
          }
        else
          {
            PA_send_no_lok(); return;
          }
      }
  }


const char text_pict_id[]      PROGMEM = ", Pict-ID=0x";
const char text_rpt_dcc[]      PROGMEM = ", DCC, ";
const char text_rpt_name[]     PROGMEM = ", Name=";

void PA_send_lok_report(unsigned char index)
  {
    unsigned char i;
    t_locoentry my_lok;
    
    pc_send_string_P(text_loco);
    get_loco_data(index, &my_lok);
    
    pc_send_num(my_lok.w.addr);
    pc_send_string_P(text_rpt_dcc);   // pc_send_string(", DCC, ");

    pc_send_num(format_to_uint8(my_lok.w.format));

    pc_send_string_P(text_pict_id);      // pc_send_string(", Pict-ID=0x");
    pc_send_hex(my_lok.picture_id / 256);
    pc_send_hex(my_lok.picture_id);
        
    pc_send_string_P(text_rpt_name);   
    for(i=0; i< LOK_NAME_LENGTH; i++)
      {
        if (my_lok.name[i]) pc_send(my_lok.name[i]);
        else pc_send(' ');                     // fill rest with blank
      }
    pc_send_crlf();
  }

void PA_locquery(void)
  {
    if (lok_addr_out_of_range(&parameter[0]))
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_locaddr_outrange); return;
      }

    unsigned char i;
    if ((i = search_loco(parameter[0])) < SIZE_LOCO_DATABASE)
      {
        // Lok found, now report it
        PA_send_lok_report(i);
      }
    else
      {
        PA_send_no_lok();
      }
  }


const char text_all_deleted[]      PROGMEM = "Lok database cleared";
const char text_lok_deleted[]      PROGMEM = "Lok removed";


void PA_locdel(void)
  {
    if (parameter[0] == 0xfffe)
      {
        clear_loco_database();
        pc_send_answer_P(text_all_deleted);
      }
    else if (lok_addr_out_of_range(&parameter[0]))
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_locaddr_outrange); return;
      }
    else
      {
        unsigned char i;
        if ((i = search_loco(parameter[0])) < SIZE_LOCO_DATABASE)
          {
            // Lok found, now delete it
            delete_loco_data(i);
            pc_send_answer_P(text_lok_deleted);
          }
        else
          {
            PA_send_no_lok();
          }
      }
  }

 



//LOCADD: Eintrag zur Lokdatenbank hinzufgen
//   Syntax: LOCADD <Adr>, <Speed>, <Fmt>, <Name>
//           LOCADD  1234,     126,   DCC, BR03
//   Die Angaben entsprechen den Ausgaben von LOCDUMP. Der Name sollte in einfache
//   oder doppelte Hochkommas eingeschlossen werden, wenn er Leerzeichen oder
//   Kommata enthlt.
void PA_locadd(void)
  {
    t_format my_format;
    if (lok_addr_out_of_range(&parameter[0]))
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_locaddr_outrange); return;
      }
    if (parameter[1] == 14) my_format = DCC14;  
    else if (parameter[1] == 28) my_format = DCC28;  
    else if (parameter[1] == 126) my_format = DCC128;  
    else
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_speedsteps); return;
      }
    ibcmd = &pcc[1];
    next_arg();             // now at addr
    next_arg();             // now at speed
    next_arg();             // now at format
    if ((*ibcmd | 0x20) != 'd')
      {
        pc_send_string_P(text_error);
        pc_send_answer_P(text_locformat); return;
      }
    next_arg();             // now at name

    // now check for this loco

    unsigned char index;
    if ((index = search_loco(parameter[0])) < SIZE_LOCO_DATABASE)
      {
        // Lok found, now replace it.
 
        t_locoentry my_lok;
        my_lok.w.addr = parameter[0];
        my_lok.w.format = my_format;
        my_lok.picture_id = 0;
        unsigned char i;
        for(i=0; i< LOK_NAME_LENGTH; i++)
          {
            if (*ibcmd)
              {
                my_lok.name[i] = *ibcmd++;
              }
            else
              {
                my_lok.name[i] = 0;
              }
          }
        save_loco_data(index, &my_lok);
        PA_send_lok_report(index);
      }
    else
      {
        if ((index = insert_new_loco_data(0)) > 0)       // try to insert after index 0
          {
            t_locoentry my_lok;
            my_lok.w.addr = parameter[0];
            my_lok.w.format = my_format;
            my_lok.picture_id = 0;
            unsigned char i;
            for(i=0; i< LOK_NAME_LENGTH; i++)
              {
                my_lok.name[i] = *ibcmd++;
              }
            save_loco_data(index, &my_lok);
            PA_send_lok_report(index);
          }
        else
          {
            pc_send_string_P(text_error);
            pc_send_answer_P(text_buffer_full);  // failed - to many locos
          }
      }
  }


const char text_end[]        PROGMEM = "*END*";;
void PA_locdump(void)
  {
    unsigned char i, j, max;
    unsigned char index = 0;

    t_locoentry my_lok;
    
    max = get_loco_in_database();
    for (j=0; j<max; j++)
      {
        index = scan_loco_database_forward(index);
        
        get_loco_data(index, &my_lok);
    
        pc_send_num(my_lok.w.addr);
        pc_send(',');
        pc_send_num(format_to_uint8(my_lok.w.format));
        pc_send_string_P(text_rpt_dcc); // pc_send_string(", DCC, ");

        for(i=0; i< sizeof(my_lok.name); i++)
          {
            if (my_lok.name[i]) pc_send(my_lok.name[i]);
            else pc_send(' ');                     // fill rest with blank
          }
        pc_send_crlf(); 
      }
    pc_send_answer_P(text_end);   
  }




const char text_bckup_cmd[]     PROGMEM = "LocAdd ";
const char text_bckup_pic[]     PROGMEM = "LocPic ";


void PA_backup(void)
  {
    unsigned char i, j, max;
    unsigned char index = 0;

    t_locoentry my_lok;
    
    max = get_loco_in_database();
    for (j=0; j<max; j++)
      {
        index = scan_loco_database_forward(index);
        
        get_loco_data(index, &my_lok);
    
        pc_send_string_P(text_bckup_cmd);
        pc_send_num(my_lok.w.addr);
        pc_send(','); pc_send(' ');

        pc_send_num(format_to_uint8(my_lok.w.format));

        pc_send_string_P(text_rpt_dcc); // pc_send_string(", DCC, ");

        for(i=0; i< LOK_NAME_LENGTH; i++)
          {
            if (my_lok.name[i]) pc_send(my_lok.name[i]);
            else pc_send(' ');                     // fill rest with blank
          }
        pc_send_crlf(); 
        
        pc_send_string_P(text_bckup_pic);
        pc_send_num(my_lok.w.addr);
        pc_send(','); pc_send('$');
        pc_send_hex(my_lok.picture_id / 256);
        pc_send_hex(my_lok.picture_id);
        pc_send_crlf();
      }
  }

// these are debug functions - no checking is performed, use at own risk :-)

void PA_read_eeprom(void)
  {
    pc_send('e');
    pc_send(':');
    pc_send_hex(parameter[0]/256);
    pc_send_hex(parameter[0]);
    pc_send(':');
    pc_send_hex(eeprom_read_byte((void *)parameter[0]));
    pc_send_crlf();
  }

void PA_write_eeprom(void)
  {
    uint8_t content;
    content = (uint8_t) parameter[1];   

    pc_send('w');
    eeprom_write_byte((void *)parameter[0], content);
    PA_read_eeprom();
  }

void PA_read_flash(void)
  {
    uint8_t content;
    
    pc_send('f');
    pc_send(':');
    pc_send_hex(parameter[0]/256);
    pc_send_hex(parameter[0]);
    pc_send(':');
    //            addr, * buffer, count
    flash_read(parameter[0], &content, 1);
    pc_send_hex(content);
    pc_send_crlf();
  }

void PA_write_flash(void)
  {
    uint8_t content;

    content = (uint8_t) parameter[1];   

    pc_send('w');
    //            addr, * buffer, count
    flash_write(parameter[0], &content, 1);
    PA_read_flash();
  }

const char text_era_flash[]     PROGMEM = "Erasing Flash ... 20s .. ";    
const char text_done[]          PROGMEM = "done ";    
     
void PA_erase_flash(void)
  {
    pc_send_string_P(text_era_flash);
    flash_erase_complete();
    pc_send_answer_P(text_done);   
  }

//------------------------------------------------------------------------
// Note: the scan stops after the first match, therefore a single char command
//       must be the last command with this character.
//       There must be an empty command at the end of the list - otherwise the
//       scan will not terminate.

#define COMMAND_LENGTH  (7+1)
typedef struct
{
  unsigned char command[COMMAND_LENGTH];
  void (*fnAction)(void);
} T_AACTION;

const T_AACTION AsciiParseTab[] PROGMEM =
  {
    // Command / Call               // implemented / simulated / tested / comment 
    {{"BACKUP"},  PA_backup },      // i / s / t / help string
    {{"HELP"},    PA_help },        // i / s / t / help string
    {{"INFO"},    PA_info },        // i / s / t
    {{"LOCADD"},  PA_locadd },      // i /
    {{"LOCDEL"},  PA_locdel },      // - /
    {{"LOCDUMP"}, PA_locdump },     // i /
    {{"LOCFREE"}, PA_locfree },     // i /
    {{"LOCPIC"},  PA_locpic },      // i /         assign picture id to loc address or query
    {{"LOCQ"},    PA_locquery },    // i /         query complete record
    {{"XER"},     PA_read_eeprom }, // i /         read eeprom (addr)
    {{"XEW"},     PA_write_eeprom },// i /         write eeprom (addr, data)
    {{"XFR"},     PA_read_flash },  // i /         read external flash (addr)
    {{"XFW"},     PA_write_flash }, // i /         write external flash (addr, data)
    {{"XFERA"},   PA_erase_flash }, // i /         erase external flash
    {{""},        PA_unknown},      // i / s / - / empty command is always true
  };
//--------------------------------------------------------------------------------
// This routine checks for opcodes and parameter
//  and calls the corresponding function
//
void parse (void)
  {
    unsigned char *arg;
    unsigned char i, j;
    void (*fp)(void);                   // declare pointer to function


    j=0;
    while(1)
      {
        arg = ibcmd;                    // reset pointer to arg
        for (i=0; i<COMMAND_LENGTH; i++)
          {
            if (pgm_read_byte(&AsciiParseTab[j].command[i]) == 0)   // target found
              {
                ibcmd = ibcmd+i;        // advance parse pointer to arguments
                while (!is_delimiter_or_num(*ibcmd)) ibcmd++;
                
                parse_parameters();
                fp = (void *) pgm_read_word(&AsciiParseTab[j].fnAction);   // assign function addr read from flash
                (*fp)();                  // now call this function
                 
                return;
              }
            if (toupper(*arg++) != pgm_read_byte(&AsciiParseTab[j].command[i]) )
              {
                break;                // this target is wrong
              }
          }
        j++;
      }
  }


//------------------------------------------------------------------------
// 


void init_parser(void)
  {
    parser_state = IDLE;
  }


bool input_ready(void)
  {
    if(rs232_break_detected == 1)
      {
        init_parser();
        init_rs232(DEFAULT_BAUD, TASK_PARSER);  // reinit rs232 and get rid of old connection
      }
    return(rx_fifo_ready());
  }


// new: drop lf after cr
unsigned char last_received;

unsigned char drop_linefeed_after_cr(unsigned char test)
  {
    unsigned char retval = FALSE;
    if (test == 0x0a)
      {
        if (last_received == 0x0d)          // if <cr><lf> -> drop <lf>
          {
            retval = TRUE;
          }
      }
    last_received = test;
    return(retval);
  }


unsigned char input_read(void)
  {
    return(rx_fifo_read());
  }


// return: 0: ready again
//        -1: blocked by fifo

t_cr_task run_parser(void)
  {
    unsigned char received;

    switch (parser_state)
      {
        case IDLE:                                      
            parser_state = WAIT_FOR_COMMAND;
            pcc_index = 0;
            ibcmd = &pcc[0];
            break;
        case WAIT_FOR_COMMAND:
            if (!input_ready()) return(-1);      // this stops TASK_PARSER
            
            received = input_read();
            if (drop_linefeed_after_cr(received))  return(0);
            pcc_index = 1;
            parser_state = WAIT_FOR_PARAM;
            *ibcmd = toupper(received);
            break;
        case WAIT_FOR_PARAM:
            if ((*ibcmd == 0x0a) || (*ibcmd == 0x0d))       // terminator
              {
                *ibcmd = 0;                                 // terminate string
                ibcmd = &pcc[0];
                parse();
                parser_state = IDLE;
                return(0);
              }
            if (!input_ready()) return(-1);
            received = input_read();
            if (drop_linefeed_after_cr(received))  return(0);
                 
            if (pcc_index < (sizeof(pcc)-1))
              {
                pcc_index++;
                ibcmd++;                                    // advance read pointer
              }
            else 
              {
                // - fatal error - string too long (keep only last char)
              }
            *ibcmd = received;                    // read char and dump to command buffer
            break;
     }
    return(0);
  }


