//------------------------------------------------------------------------------
//
// OpenDCC - Throttle - Programmer
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//-----------------------------------------------------------------------------
//
// file:      programmer.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-03-02 V0.02 kw  startet
//            2009-03-15 V0.03 kw  added some code for read
//            2009-03-17 V0.04 kw  added some code for write
//            2010-06-23 V0.05 kw  added msg_transmit();
//
//-----------------------------------------------------------------------------
//
// purpose:   sequencer for reading and writing short or long addresses to locos
//
//------------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>

#include "config.h"
#include "cortos.h"
#include "timer.h"          				// access to systick
#include "keyboard.h"

#include "messages_objects.h"
#include "programmer.h"

#define PROG_TIMEOUT  20000000L             // 20s

signed char is_prog_result_ready(void)
  {
    // check for answer
    switch (current_CV.result)
      {
        case PRG_CV_VOID:
        case PRG_CV_WAIT_FOR_ANSWER:  
            // timeout prfen
            if ((get_systick_int() - current_CV.prog_request_timestamp) > (PROG_TIMEOUT / SYSTICK_PERIOD) )
              {
                prg_state = PRG_FAILED;
                return(-1);
              }
            else
              {
                return(0);  // wait again
              }                   
        case PRG_CV_OKAY:
            return(1);
        case PRG_CV_SHORT:
            prg_state = PRG_FAILED;
            return(-1); 
        case PRG_CV_NOLOK:
            prg_state = PRG_FAILED;
            return(-1); 
        default:
        case sizeof_t_prog_result:
            prg_state = PRG_FAILED;
            return(-1);
            break; 
      }
  }

t_prg_state prg_state;   // we also export this to enable others to display our state
                         // write access only from here!

int progged_lok_addr;

void start_programmer_readaddr(void)
  {
    if (prg_state <= PRG_FAILED)
      {
        prg_state = PRG_READ_ADDR;
        set_task_ready(TASK_PROGRAMMER);
      }
    else
      {
        //!warning! // error(Programming already active)
      }
  }

void start_programmer_writeaddr(unsigned int addr)
  {
    if (prg_state <= PRG_FAILED)
      {
        progged_lok_addr = addr;
        prg_state = PRG_WRITE_ADDR;
        set_task_ready(TASK_PROGRAMMER);
      }
    else
      {
        // error(Programming already active)
      }
  }


t_cr_task programmer (void)
  {
    switch (prg_state)
      {
        default:
            break;
        case PRG_DONE:
            cr_fifo_put(&key_fifo, KEY_PROGRAMMER);   //notify gui if change detected
            return(-1);  // stop me
            break;

        case PRG_FAILED:
            progged_lok_addr = 0;
            cr_fifo_put(&key_fifo, KEY_PROGRAMMER);   //notify gui if change detected
            return(-1);  // stop me
            break;
        
        case PRG_READ_ADDR:
            progged_lok_addr = 0;
            msg_send_cv_read(29);
            msg_transmit();
            prg_state = PRG_RA_CV29;
            break;
        case PRG_RA_CV29:
            if (is_prog_result_ready() == 1)
              {
                if (current_CV.value & 0x20)            // check long addressing bit
                  {                                     // read long addr
                    msg_send_cv_read(17);
                    msg_transmit();
                    prg_state = PRG_RA_CV17;
                  }
                else
                  {                                     // read short addr
                    msg_send_cv_read(1);
                    msg_transmit();
                    prg_state = PRG_RA_CV1;
                  }
                break;
              }  
            break;
        case PRG_RA_CV1:
            if (is_prog_result_ready() == 1)
              {
                progged_lok_addr = current_CV.value;
                prg_state = PRG_DONE;
                break;
              }  
            break;
        case PRG_RA_CV17:
            if (is_prog_result_ready() == 1)
              {
                progged_lok_addr = current_CV.value;
                if ((progged_lok_addr < 192) || (progged_lok_addr >= 232))
                  {
                    progged_lok_addr = 0;       // wrong
                    prg_state = PRG_DONE;
                  }
                else
                  {
                    progged_lok_addr = (progged_lok_addr - 192) * 256;
                    msg_send_cv_read(18);
                    msg_transmit();
                    prg_state = PRG_RA_CV18;
                  }
              }
            break;
        case PRG_RA_CV18:
            if (is_prog_result_ready() == 1)
              {
                progged_lok_addr += current_CV.value;
                prg_state = PRG_DONE;
              }
            break;

        case PRG_WRITE_ADDR:
            msg_send_cv_read(29);
            msg_transmit();
            prg_state = PRG_WA_CV29;
            break;
        case PRG_WA_CV29:
            if (is_prog_result_ready() == 1)
              {
                if (current_CV.value & 0x20)            // check long addressing bit
                  {                                     // is set to long addr
                    if (progged_lok_addr < 100)
                      {
                        // we must clear this bit in CV29
                        msg_send_cv_write(29, current_CV.value & ~0x20);
                        msg_transmit();
                        prg_state = PRG_WA_CV29_OKAY;
                      }
                    else
                      {
                        msg_send_cv_write(17, (progged_lok_addr / 256) + 192);
                        msg_transmit();
                        prg_state = PRG_WA_CV17;
                      }
                  }
                else
                  {                                     // is set to short addr
                    if (progged_lok_addr < 100)
                      {
                        msg_send_cv_write(1, progged_lok_addr);
                        msg_transmit();
                        prg_state = PRG_WA_CV1;
                      }
                    else
                      {
                        // we must set this bit in CV29
                        msg_send_cv_write(29, current_CV.value | 0x20);
                        msg_transmit();
                        prg_state = PRG_WA_CV29_OKAY;
                      }  
                  }
              }  
            break;
        case PRG_WA_CV29_OKAY:
            if (is_prog_result_ready() == 1)
              {
                if (progged_lok_addr < 100)
                  {
                    msg_send_cv_write(1, progged_lok_addr);
                    msg_transmit();
                    prg_state = PRG_WA_CV1;
                  }
                else
                  {
                    msg_send_cv_write(17, (progged_lok_addr / 256) + 192);
                    msg_transmit();
                    prg_state = PRG_WA_CV17;
                  }  
              }  
            break;
        case PRG_WA_CV1:
            if (is_prog_result_ready() == 1)
              {
                prg_state = PRG_DONE;
              }  
            break;
        case PRG_WA_CV17:
            if (is_prog_result_ready() == 1)
              {
                msg_send_cv_write(18, progged_lok_addr % 256);
                msg_transmit();
                prg_state = PRG_WA_CV18;
              }  
            break;
        case PRG_WA_CV18:
            if (is_prog_result_ready() == 1)
              {
                prg_state = PRG_DONE;
              }  
            break;
      }
   return(100000L / SYSTICK_PERIOD);     // check again in 100ms
  }
