//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006-2008, 2009 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      serial.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-13 V0.01 started
//            2006-04-28 V0.02 added set_led*** and timeout
//            2006-05-15 V0.03 removed early sei();
//            2006-10-17 V0.04 baudrate types for intellibox mode
//            2006-11-16 V0.05 added push_to_rx for easier simulation
//            2007-01-27 V0.06 changed to 2 stop bits to clear some 
//                             trouble with USB-to-Serial Converter
//            2007-06-09 V0.07 new function tx_all_sent
//                             reset txc on data transmission 
//            2008-04-04 V0.08 used for sniffer - runs on Atmega162,
//                             UART0; 
//            2008-04-17 V0.09 fixed a bug in init U2X was cleared
//                             unintentionally; added better cast!
//            2009-04-09 V0.10 UART0 / UART1 support (compile switch)
//            2009-08-08 V0.11 task wake-up now configurable
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   routines for RS232
//            see also:
//            http://www.roboternetz.de/wissen/index.php/UART_mit_avr-gcc
//
//-----------------------------------------------------------------


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <string.h>

#include "config.h"               // general structures and definitions
#include "hardware.h"
#include "cortos.h"               // make consumer ready

#include "serial.h"

#define LED_ACTION  FALSE         // this is TRUE if we should handle a
                                  // a status led -> see status.c from OpenDCC

#define UART   1
#define UART0  2                    // define some numbers (!=0)
#define UART1  3

#define MY_UART  UART1

#if (MY_UART == UART)
    #define my_UDR      UDR
    #define my_UCSRA    UCSRA
        #define my_RXC      RXC
        #define my_TXC      TXC
        #define my_U2X      U2X
        #define my_FE       FE
        #define my_DOR      DOR
        #define my_UDRE     UDRE
    #define my_UCSRB    UCSRB
        #define my_RXCIE    RXCIE
        #define my_TXCIE    TXCIE
        #define my_UDRIE    UDRIE
        #define my_RXEN     RXEN
        #define my_TXEN     TXEN
        #define my_UCSZ2    UCSZ2
        #define my_RXB8     RXB8
        #define my_TXB8     TXB8
    #define my_UCSRC    UCSRC
        #ifdef URSEL
        #define my_URSEL    URSEL
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL
        #define my_UPM1     UPM1
        #define my_UPM0     UPM0
        #define my_USBS     USBS
        #define my_UCSZ1    UCSZ1
        #define my_UCSZ0    UCSZ0
        #define my_UCPOL    UCPOL
    #define my_UBRRH    UBRRH
    #define my_UBRRL    UBRRL

#elif (MY_UART == UART0)
    #define my_UDR      UDR0
    #define my_UCSRA    UCSR0A
        #define my_RXC      RXC0
        #define my_TXC      TXC0
        #define my_U2X      U2X0
        #define my_FE       FE0
        #define my_DOR      DOR0
        #define my_UDRE     UDRE0
    #define my_UCSRB    UCSR0B
        #define my_RXCIE    RXCIE0
        #define my_TXCIE    TXCIE0
        #define my_UDRIE    UDRIE0
        #define my_RXEN     RXEN0
        #define my_TXEN     TXEN0
        #define my_UCSZ2    UCSZ02
        #define my_RXB8     RXB80
        #define my_TXB8     TXB80
    #define my_UCSRC    UCSR0C
        #ifdef URSEL0
        #define my_URSEL    URSEL0
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL0
        #define my_UPM1     UPM01
        #define my_UPM0     UPM00
        #define my_USBS     USBS0
        #define my_UCSZ1    UCSZ01
        #define my_UCSZ0    UCSZ00
        #define my_UCPOL    UCPOL0
    #define my_UBRRH    UBRR0H
    #define my_UBRRL    UBRR0L

#elif (MY_UART == UART1)
    #define my_UDR      UDR1
    #define my_UCSRA    UCSR1A
        #define my_RXC      RXC1
        #define my_TXC      TXC1
        #define my_U2X      U2X1
        #define my_FE       FE1
        #define my_DOR      DOR1
        #define my_UDRE     UDRE1
    #define my_UCSRB    UCSR1B
        #define my_RXCIE    RXCIE1
        #define my_TXCIE    TXCIE1
        #define my_UDRIE    UDRIE1
        #define my_RXEN     RXEN1
        #define my_TXEN     TXEN1
        #define my_UCSZ2    UCSZ12
        #define my_RXB8     RXB81
        #define my_TXB8     TXB81
    #define my_UCSRC    UCSR1C
        #ifdef URSEL1
        #define my_URSEL    URSEL1
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL1
        #define my_UPM1     UPM11
        #define my_UPM0     UPM10
        #define my_USBS     USBS1
        #define my_UCSZ1    UCSZ11
        #define my_UCSZ0    UCSZ10
        #define my_UCPOL    UCPOL1
    #define my_UBRRH    UBRR1H
    #define my_UBRRL    UBRR1L
#else
  #warning wrong uart!
#endif // MY_UART


//
//=====================================================================
//
// RS232
//
// purpose:   send and receive messages from pc
//
// how:       uart acts with interrupt on fifos.
//            ohter programs access only the fifos.
//            hardware handshake with RTS and CTS.
//            display of status with one LED (see status.c)
//
// uses:      set_led_rs232
//            no_timeout.rs232 (see status.c)
// interface: see rs232.h
//
// 2do:       zur Zeit wird RTS nur als connected Erkennung benutzt,
//            das Senden wird NICHT angehalten.
//
//-----------------------------------------------------------------


// FIFO-Objekte und Puffer fr die Ein- und Ausgabe
// max. Size: 255

#define RxBuffer_Size  64              // mind. 16
unsigned char RxBuffer[RxBuffer_Size];


#define TxBuffer_Size  128
unsigned char TxBuffer[TxBuffer_Size];

unsigned char rx_read_ptr = 0;        // point to next read
unsigned char rx_write_ptr = 0;       // point to next write
unsigned char rx_fill = 0;
unsigned char tx_read_ptr = 0;
unsigned char tx_write_ptr = 0;
unsigned char tx_fill = 0;

const unsigned long baudrate[] PROGMEM =    // ordered like in Lenz Interface!
    {9600L,   //  = 0,
     19200L,  //  = 1,
     38400L,  //  = 2,
     57600L,  //  = 3,
     115200L, //  = 4,
     2400L,   //  = 5,  // used for Intellibox 
	};

t_baud actual_baudrate;                      // index to field above

volatile unsigned char rs232_break_detected = 0;      // flag, that a break was detected
                                                    // volatile, weil aus ISR bearbeitet wird.

t_task_id serial_consumer_task;                    // this is our consumer

void serial_set_consumer_task(t_task_id newtask)
  {
    if (newtask != serial_consumer_task)
      {
        if (serial_consumer_task) set_task_blocked(serial_consumer_task);      // stop old consumer        
        serial_consumer_task = newtask;
      }
  }

//-------------------------------------------------------------------
// this goes to usart 0

void init_rs232(t_baud new_baud, t_task_id consumer)
  {
    uint16_t ubrr;
    uint8_t sreg = SREG;

    cli();

    // serial_consumer_task = TASK_PARSER;           // default is parser
    serial_set_consumer_task(consumer);

    my_UCSRB = 0;                  // stop everything

    actual_baudrate = new_baud;

    // note calculations are done at mult 100
    // to avoid integer cast errors +50 is added

    // real rates:  @16MHz (div16)(div8)    @10MHz (div16)(div8)
    // BAUD_2400:     2398,1      2401,0         2403,8      2399,2
    // BAUD_4800:     4807,7      4796,2         4807,7      4807,7
    // BAUD_9600:     9615,4      9615,4         9615,4      9615,4
    // BAUD_19200:   19230,8     19230,8        18939,4	    19230,8
    // BAUD_38400:   38461,5     38461,5        39062,5     37878,8
    // BAUD_57600:   58823,5     57142,9        56818,2     56818,2
    // BAUD_115200: 111111,1    117647,1       125000,0	   113636,4

    switch(new_baud)
	  {
	    default:
		case BAUD_2400:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*2400L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*24L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC);
		    break;
		case BAUD_4800:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*4800L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*48L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC);
		    break;
		case BAUD_9600:
		    //   UBRRL = 103; // 19200bps @ 16.00MHz
			// ubrr = (uint16_t) ((uint32_t) F_CPU/(16*9600L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*96L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC);
		    break;
		case BAUD_19200:
		    //   UBRRL = 51; // 19200bps @ 16.00MHz
			ubrr = (uint16_t) ((uint32_t) F_CPU/(16*19200L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*192L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC);
		    break;
		case BAUD_38400:
		    //   UBRRL = 25; // 38400bps @ 16.00MHz
			// ubrr = (uint16_t) ((uint32_t) F_CPU/(16*38400L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*384L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC) ; // High Speed Mode, nur div 8
		    break;
		case BAUD_57600:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(8*57600L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(8*576L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC) | (1 << my_U2X);  // High Speed Mode, nur div 8
		    break;
		case BAUD_115200:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(8*115200L) - 1);
            ubrr = (uint16_t) ((uint32_t) (F_CPU/(8*1152L) - 100L + 50L) / 100);
            my_UBRRH = (uint8_t) (ubrr>>8);
            my_UBRRL = (uint8_t) (ubrr);
			my_UCSRA = (1 << my_RXC) | (1 << my_TXC) | (1 << my_U2X);  // High Speed Mode
		    break;
	  }
   
    // FIFOs fr Ein- und Ausgabe initialisieren
    rx_read_ptr = 0;      
    rx_write_ptr = 0;
    rx_fill = 0;      
    tx_read_ptr = 0;
    tx_write_ptr = 0;
    tx_fill = 0;

	
	// UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 8N1, asynchron
    
    my_UCSRB = (1 << my_RXEN) | (1 << my_TXEN) | (1 << my_RXCIE);
    my_UCSRC = (1 << my_URSEL)        // must be one
           | (0 << my_UMSEL)        // 0 = asyn mode
           | (0 << my_UPM1)         // 0 = parity disabled
           | (0 << my_UPM0)         // 0 = parity disabled
           | (1 << my_USBS)         // 1 = tx with 2 stop bits
           | (1 << my_UCSZ1)        // 11 = 8 or 9 bits
           | (1 << my_UCSZ0)
           | (0 << my_UCPOL);


    #ifdef RS232_EXTRA_STOP
        my_UCSRB |= (1<<my_TXB8);     // set 9th bit to 1
        my_UCSRB |= (1<<my_UCSZ2);    // set frame length to 9bit
    #endif

    // Flush Receive-Buffer
    do
      {
        my_UDR;
      }
    while (my_UCSRA & (1 << RXC0));

    // Rcksetzen von Receive und Transmit Complete-Flags
    my_UCSRA |= (1 << my_RXC);
    my_UCSRA |= (1 << my_TXC);
    
    my_UDR;

    rs232_break_detected = 0;
    
    SREG = sreg;

    #if (LED_ACTION == TRUE)
       set_led_rs232(DISCONNECTED);
    #endif

}


//-------------------------------------------------------------------------------
// Empfangene Zeichen werden in die Eingabgs-FIFO gespeichert und warten dort
// Der aktuell gelinkte consumer wird ready gesetzt und mu sich das Zeichen aus
// dem Fifo abholen.
// Wenn bis auf einen Rest von 10 gefllt ist: CTS senden.
//
#if (MY_UART == UART)
ISR(USART_RXC_vect)
#elif (MY_UART == UART0)
 #ifndef USART0_RXC_vect
 #define USART0_RXC_vect USART0_RX_vect
 #endif
ISR(USART0_RXC_vect)
#elif (MY_UART == UART1)
 #ifndef USART1_RXC_vect
 #define USART1_RXC_vect USART1_RX_vect
 #endif
ISR(USART1_RXC_vect)
#else
   #warning ISR undefined
#endif
{
    if (my_UCSRA & (1<< my_FE))
	  { // Frame Error
        rs232_break_detected = 1;       // always on error
		if (my_UDR == 0)                   // this is a break sent!
	      {
		     rs232_break_detected = 1;  // set flag for parser and discard
		  }
	  }
    else
      {
        if (my_UCSRA & (1<< my_DOR))
          { // DATA Overrun -> Fatal
            // !!! 
          }

       	RxBuffer[rx_write_ptr] = my_UDR;

        rx_write_ptr++;
        if (rx_write_ptr == RxBuffer_Size) rx_write_ptr=0;
        rx_fill++;
        if (rx_fill > (RxBuffer_Size - 10))
          {
            // we are full, stop remote Tx -> set CTS off !!!!
    		#if (LED_ACTION == TRUE)
              set_led_rs232(STOPPED);
            #endif
            // removed here done with polling of CTS in status.c
    		MY_CTS_HIGH;
          }
      }

    // make consumer ready

    #ifdef _CORTOS_H_
      set_task_ready(serial_consumer_task);
    #endif
}

//----------------------------------------------------------------------------
// Ein Zeichen aus der Ausgabe-FIFO lesen und ausgeben
// Ist das Zeichen fertig ausgegeben, wird ein neuer SIG_UART_DATA-IRQ getriggert
// Ist das FIFO leer, deaktiviert die ISR ihren eigenen IRQ.

// Bei 9 Bit knnte noch ein Stopbit erzeugt werden: UCSRB |= (1<<TXB8);


#if (DEBUG == 3)
   unsigned  char tx_char;
#endif

#if (MY_UART == UART)
ISR(USART_UDRE_vect)
#elif (MY_UART == UART0)
ISR(USART0_UDRE_vect)
#elif (MY_UART == UART1)
ISR(USART1_UDRE_vect)
#else
   #warning ISR undefined
#endif
  {

    if (tx_read_ptr != tx_write_ptr)
      {
        my_UCSRA |= (1 << my_TXC);              // writing a one clears any existing tx complete flag
        my_UDR = TxBuffer[tx_read_ptr];
        #if (DEBUG == 3)
            tx_char = TxBuffer[tx_read_ptr];
        #endif
        tx_read_ptr++;
        if (tx_read_ptr == TxBuffer_Size) tx_read_ptr=0;
        tx_fill--;
      }
    else
        my_UCSRB &= ~(1 << my_UDRIE);           // disable further TxINT
  }



//=============================================================================
// Upstream Interface
//-----------------------------------------------------------------------------
// TX:
bool tx_fifo_ready (void)
{
    if (tx_fill < (TxBuffer_Size-16))     // keep space for one complete message (16)
      {
        return(1);                        // true if enough room
      }
    else
      {
        return(0);
      }
}


// ret 1 if full
bool tx_fifo_write (const unsigned char c)
{
    TxBuffer[tx_write_ptr] = c;

    tx_write_ptr++;
    if (tx_write_ptr == TxBuffer_Size) tx_write_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    tx_fill++;
    SREG = mysreg;

    my_UCSRB |= (1 << my_UDRIE);   // enable TxINT

    if (tx_fill < (TxBuffer_Size-16))
      {
        return(1);
      }

    return(0);
}

// ret 1 if all is sent
bool tx_all_sent (void)
{
    if (tx_fill == 0)
      {
        if (!(my_UCSRA & (1 << my_UDRE)))  return(0);    // UDR not empty
        if (!(my_UCSRA & (1 << my_TXC)))  return(0);    // TX Completed not set
        return(1);                        
      }
    else
      {
        return(0);
      }
}

void uart_puts (const char *s)
{
    do
      {
        tx_fifo_write (*s);
      }
    while (*s++);
}


//------------------------------------------------------------------------------
// RX:
bool rx_fifo_ready (void)
{
    if (rx_read_ptr != rx_write_ptr)
      {
        return(1);     // there is something
      }
    else
      {
        #if (LED_ACTION == TRUE)
            if (no_timeout.rs232)
              {
                // kein timeout -> nichts machen
              }
            else
              {
                if (MY_RTS_DETECT) set_led_rs232(CONNECTED);
                else set_led_rs232(DISCONNECTED);
              } 
        #endif
        return(0);  
      }
}

//-------------------------------------------------------------------
// rx_fifo_read gets one char from the input fifo
//
// there is no check whether a char is ready, this must be
// done before calling with a call to rx_fifo_ready();

unsigned char rx_fifo_read (void)
  {
    unsigned char retval;

    retval = RxBuffer[rx_read_ptr];
    rx_read_ptr++;
    if (rx_read_ptr == RxBuffer_Size) rx_read_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    rx_fill--;
    SREG = mysreg;

    if (rx_fill < (RxBuffer_Size - 14))
      {
        if (MY_CTS_STATE)
          {
            // CTS is high, but we are no longer full
            // enable remote Tx -> reset CTS 
            #if (LED_ACTION == TRUE)
               set_led_rs232(RUNNING);
            #endif
            MY_CTS_LOW;
          }
      }
    #if (LED_ACTION == TRUE)
       no_timeout.rs232 = 1000000L / TICK_PERIOD;      // 1 sec
    #endif
    return(retval);
  }


//-----------------------------------------------------------------------------------
// testroutinen zu Parsertest
// Ersatzroutine, um ein Zeichen in den Empfangsbuffer abzulegen
// Sendepuffer kann im Simulator mit einen Databreakpoint auf UDR berwacht werden.
void push_to_rx(unsigned char d)
  {
	RxBuffer[rx_write_ptr] = d;

    rx_write_ptr++;
    if (rx_write_ptr == RxBuffer_Size) rx_write_ptr=0;
    rx_fill++;
    if (rx_fill > (RxBuffer_Size-5))
      {
        // we are full, stop remote Tx -> set CTS off !!!!
		#if (LED_ACTION == TRUE)
           set_led_rs232(STOPPED);
		#endif
        MY_CTS_HIGH;
        // 
      }
  }



