//------------------------------------------------------------------------
//
// OpenDCC - Throttle
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      timer.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-12 V0.1 kw started
//
//------------------------------------------------------------------------
//
// purpose:   start up timers, supply systick
//
//------------------------------------------------------------------------
//
#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "config.h"
#include "timer.h"
// #include "hardware.h"

//---------------------------------------------------------------------------------
// assign timer resources
//
//          prescaler   tick    mode    interrupt   hardware    software
// timer0:   8          0.8us   STD     COMP        -           Xpressnet tx-rx switch time
// timer1:  64          6.4us   PWM     -           Backlight
// timer2:  64          6.4us   PWM     OVF         LED_A       systick
// timer3:  64          6.4us   PWM     -           LED_B,C,D
//
//---------------------------------------------------------------------------------


//---------------------------------------------------------------------------

#if   (T0_PRESCALER==1)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(0<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==8)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==64)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==256)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==1024)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(1<<CS00))
#else
  #error T0_PRESCALER is not defined or is not a supported value
#endif

#if defined (__AVR_AT90CAN128__)
void init_timer0(void)
  {
    TCCR0A = (0<<FOC0A)   // reserved set to 0
          | (0<<WGM00)  // all normal mode (Top = 0xFF)
          | (0<<COM0A1)  // no output compare
          | (0<<COM0A0)
          | (0<<WGM01)
          | T0_PRESCALER_BITS;

    TCNT0 = 0;

    TIMSK0 =  (0<<TOIE0)
            | (0<<OCIE0A);
  }
#elif defined (__AVR_ATmega128__)   // this swappes timer0 and timer1
  void - see change not from atmel

#else
  #warning undefined CPU
#endif



//---------------------------------------------------------------------------------

// we run Timer1 + 3 with 10MHz, Prescaler 64: This is 6,4us timer grid;
// Complete cycle 1.6384ms -> SystickPrescaler: 16 (1.6us Tick)
//
// We use Mode 5, Fast PWM, TOP at 0xFF (WGM = 0101)

#define T1_PRESCALER 64

#if   (T1_PRESCALER==1)
  #define T1_PRESCALER_BITS   ((0<<CS12)|(0<<CS11)|(1<<CS10))
#elif (T1_PRESCALER==8)
  #define T1_PRESCALER_BITS   ((0<<CS12)|(1<<CS11)|(0<<CS10))
#elif (T1_PRESCALER==64)
  #define T1_PRESCALER_BITS   ((0<<CS12)|(1<<CS11)|(1<<CS10))
#elif (T1_PRESCALER==256)
  #define T1_PRESCALER_BITS   ((1<<CS12)|(0<<CS11)|(0<<CS10))
#elif (T1_PRESCALER==1024)
  #define T1_PRESCALER_BITS   ((1<<CS12)|(0<<CS11)|(1<<CS10))
#else
  #error T1_PRESCALER is not defined or is not a supported value
#endif

#define T3_PRESCALER_BITS  T1_PRESCALER_BITS           // same speed

void init_timer1(void)
  {
    TCCR1A = (0 << COM1A1) | (0 << COM1A0)  // 00=normal port  01=normal port   10=clear on compare match, set at TOP   11 set on compare match
           | (0 << COM1B1) | (0 << COM1B0)
           | (0 << COM1C1) | (0 << COM1C0)
           | (0 << WGM11) | (1 << WGM10);

    TCCR1B = (0 << ICNC1)       // input noise canceler
           | (0 << ICES1)       // input cature edge
           | (0 << WGM13)
           | (1 << WGM12)
           | T1_PRESCALER_BITS;
    OCR1A = 0;
    OCR1B = 0;
    OCR1C = 0;
  }

void init_timer3(void)
  {
    TCCR3A = (0 << COM3A1) | (0 << COM3A0)  // 00=normal port  01=normal port   10=clear on compare match, set at TOP   11 set on compare match
           | (0 << COM3B1) | (0 << COM3B0)
           | (0 << COM3C1) | (0 << COM3C0)
           | (0 << WGM31) | (1 << WGM30);

    TCCR3B = (0 << ICNC3)       // input noise canceler
           | (0 << ICES3)       // input cature edge
           | (0 << WGM33)
           | (1 << WGM32)
           | T3_PRESCALER_BITS;
    OCR3A = 0;
    OCR3B = 0;
    OCR3C = 0;
  }

//---------------------------------------------------------------------------
// Define the TIMER2, this is a 8 Bit Timer, Overflow at 256
//
//
// we run Timer2 with 10MHz, Prescaler 64: This is 6,4us timer grid;
// Complete cycle 1.6384ms -> Systick
// Knob and keyboard can hook to this tick
//
// #define T2_PRESCALER   64 (in timer.h)

/*
typedef union
{
    uint16_t as_uint16;
    int16_t  as_int16;
    uint8_t  as_uint8[2];
    int8_t   as_int8[2];
} t_data16;  */

volatile t_data16  tick;

// volatile signed char systick;           // gets incremented every 1.6ms
// volatile signed int  systick_i;         // gets incremented every 1.6ms
volatile signed int  systick_10ms;      // gets incremented every 10ms
                                        // atomic access required, range 0..655sec

//---------------------------------------------------------------------------

#if   (T2_PRESCALER==1)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==8)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==32)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==64)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==128)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==256)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==1024)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(1<<CS20))
#else
  #error T2_PRESCALER is not defined or is not a supported value
#endif


void init_timer2(void)
  {
    TCCR2A = (0<<FOC2A)    // reserved set to 0
           | (1<<WGM20)    // Mode 3: Fast PWM (Top = 0xFF)
           | (0<<COM2A1)   // 1,0 Clear OC2 on Compare Match, set OC2 at TOP.
           | (0<<COM2A0)
           | (1<<WGM21)
           | T2_PRESCALER_BITS;

    TCNT2 = 0;

    TIMSK2 |= (1<<TOIE2);  // enable overflow interrupt

    // systick = 0;
    // systick_i = 0;
    systick_10ms = 0;

    tick.as_int16 = 0;
  }



signed char local_systick;

ISR(TIMER2_OVF_vect)
  {
    // We get an Interrupt every 1638us
    // We do a software extension to systick -> systick_10ms
    // 
    signed char compare;

    tick.as_int16++;

    compare = tick.as_int8[0] - local_systick;
    if (compare >= 6)
      {
        systick_10ms++;
        local_systick += 6;
      }

        
    /*
    systick++;
    systick_i++;

    compare = systick - local_systick;
    if (compare >= 6)
      {
        systick_10ms++;
        local_systick += 6;
      }
    */
    // hook for keyboard
    // hook for rtos kernel
  }  


// this returns a 16 bit int, resolution 6,4us, total range ~400ms
// this is to be used for delay calculations
// here we access hardware (TCNT) and interrupt variable, so we
// must ensure the sync 

signed int get_systick_us(void)
  {
    t_data16 result;
    signed char sw_tick;

    cli();
    sw_tick = tick.as_int8[0];
    result.as_uint8[0] = TCNT2;
    if (result.as_uint8[0] < 250)
      {  // low counter value, may be we missed a tick)
        if (TIFR2 & (1<<TOV2))  // pending overflow interrupt not yet processed
          sw_tick++;
      }
    result.as_int8[1] = sw_tick;
    sei();
    return(result.as_int16);
  }

signed int get_systick_10ms(void)
  {
    signed int mstick;

    cli();
    mstick = systick_10ms;       // atomic access
    sei();
    return(mstick);
  }

signed int get_systick_int(void)
  {
    signed int stick;

    cli();
    stick = tick.as_int16;            // atomic access
    sei();
    return(stick);
  }

signed char get_systick(void)
  {
    signed char stick;

    cli();
    stick = tick.as_int8[0];            // atomic access
    sei();
    return(stick);
  }


void init_timer(void)
  {
    init_timer0(); 
    init_timer1();
    init_timer2();
    init_timer3();
  }
