//------------------------------------------------------------------------------
//
// OpenDCC - WRP Wireless Railroad Protocol
//
// Copyright (c) 2009,2010 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------------
//
// file:      wrp_messages.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2010-02-10 V0.01 kw  started
//            2010-06-23 V0.02 kw  changed to payload
//
//------------------------------------------------------------------------------
//
// purpose:   throttle / gateway
// content:   handling of messages in a client
//            message are entered to data-field "payload", where they get
//            transmitted later. May be some by kind of scheduler or a clean-up by
//            the menus or tasks (turnout, programmer).
//
//------------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <string.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"
#include "timer.h"
#include "led.h"
#include "display.h"                // display_error

#include "wrp.h"                    // typedef for wireless railroad protocol

#include "messages_objects.h"       // current_loc
#include "xbee_if.h"


// check compile switch (see xbee_if.h)

#ifdef XBEE_IF
  #if (XBEE_IF == 0)
    #error XBEE_IF not selected - must be MASTER (=1) or CLIENT (=2)
  #endif
#else
    #error XBEE_IF not defined - must be MASTER (=1) or CLIENT (=2)
#endif


//====================================================================================
//
// DATA 
//
//====================================================================================

// here we build messages

wrp_pkt_t my_message;

// here we build the accumulate payload (this structure is defined in xbee_if.h)
payload_t payload = 
  { 
     0,    // size
     {0,0},// message
  };


//====================================================================================
//
// WRP common routines
//
//====================================================================================


// new version: akkumulate wrp pakets, fire separate
void wrp_send(wrp_pkt_t *load)
  {
    // check size of message buffer
    if ((payload.message_size + (load->header & 0x7F) +1) > XBEE_MAX_DATA_LEN)
      {
        display_error(ERR_XBEE_TX_OVERRIDE);
        return;  // This really dirty - we just drop this message, UGH!!
      }
    // okay, lets append
    memcpy(&payload.message[payload.message_size],
              load,
                 (load->header & 0x7F) +1);
    payload.message_size += (load->header & 0x7F) + 1;
  }

// clean-up routine 
// (to called after all sends: this actually forwards the message to the host)
void wrp_transmit(void)
  {
    if (payload.message_size > 0)
      {
        xb_send_payload_to_host();          
        payload.message_size = 0;           // now this is empty again
      }
  }


// special routine - check for speed paket to same lok (but only at start of payload)

static unsigned char wrp_check_for_speed_replace(wrp_pkt_t *load)
  {
    wrp_pkt_t *existing;
    if (payload.message_size > 0)
      {
        existing = (wrp_pkt_t *) payload.message;
        if (existing->header != load->header)    return(0);   // different
        if (existing->type != load->type)        return(0);   // different
        if (existing->opcode != load->opcode)    return(0);   // different
        if (existing->addr != load->addr)        return(0);   // different
        
        existing->data[0] = load->data[0];     
        existing->data[1] = load->data[1]; 
        return(1);    
      }
    return(0);
  }



#if (1==0)
// old version: pack load and fire immediately
void wrp_send(wrp_pkt_t *load)
  {
    xb_make_message_to_host(load);    // pack it to an XBEE API message
    send_xbee_message();              // and forward this packet to the module
  }

void wrp_transmit(void)                 // this is a dummy in old version 
  {
  }
#endif
//====================================================================================
//
// WRP Messages issued by Client
//
//====================================================================================

void create_wrp_sys_idle(void)
  {
    my_message.header = 1;
    my_message.type = WRP_TYPE_SYSTEM;
    my_message.opcode = WRP_OPCODE_IDLE;
  }




void wrp_send_sys_idle(void)
  {
    #if (1==0)   // debug only
    static uint8_t sys_idle_debug_led;
    if (sys_idle_debug_led)
       {
         turn_led_green_immediate(LED_A);
         sys_idle_debug_led = 0;
       }
    else
       {
         turn_led_red_immediate(LED_A);
         sys_idle_debug_led = 1;
       }
    #endif
    create_wrp_sys_idle();
    wrp_send(&my_message);
  }


//----------------------------------------------------------------------------
// bookkeeping-stuff
//----------------------------------------------------------------------------


static void create_wrp_sys_apply(void)
  {
    my_message.header = 1 + 2 + 1;
    my_message.type = WRP_TYPE_SYSTEM;
    my_message.opcode = WRP_OPCODE_APPLY;
    my_message.addr = 1;                     // Hardware ENUM for MFT
    my_message.data[0] = throttle_id;
  }

void wrp_send_sys_apply(void)
  {
    create_wrp_sys_apply();
    wrp_send(&my_message);
  }



//----------------------------------------------------------------------------
// command station
//----------------------------------------------------------------------------

void wrp_send_cs_status_request(void)
  {
	my_message.header = 1;
    my_message.type = WRP_TYPE_STATE;
    my_message.opcode = WRP_OPCODE_SQUERY;
    wrp_send(&my_message);
  }
  
void wrp_send_state(t_status new_state)
  {
    my_message.header = 1;
    my_message.type = WRP_TYPE_STATE;
    switch (new_state)
      {
        case STS_OFF:
            my_message.opcode = WRP_OPCODE_OFF;
            wrp_send(&my_message);
            break;
        case STS_ON:
            my_message.opcode = WRP_OPCODE_GO;
            wrp_send(&my_message);
            break;
        case STS_STOP:
            my_message.opcode = WRP_OPCODE_STOP;     // WRP_OPCODE_SOFT_STOP 
            wrp_send(&my_message);
            break;
        case STS_PROG:
            // no command required
            break;
        case STS_BUSY:
            // no command required
            break;
        case STS_PBUSY:
            // no command required
            break;
        case STS_DISCONNECTED:
            // no command required
            break;
        case STS_CMD_ERROR:
            // no command required
            break;
      }
  }

// parameter: module addr (module means a 8 bit range)
void wrp_send_feedback_query(unsigned int addr)
  {
    my_message.header = 1 + 2;
    my_message.type = WRP_TYPE_FEEDBACK;
    my_message.opcode = WRP_OPCODE_FQUERY8;
    my_message.addr = addr;
    wrp_send(&my_message);
  }

void wrp_send_feedback_subscribe(void)
  {
    my_message.header = 1 + 2;
    my_message.type = WRP_TYPE_FEEDBACK;
    my_message.opcode = WRP_OPCODE_FSUBSCRIBE;
    my_message.addr = 0;                    // unsubscribe all
    wrp_send(&my_message);

    my_message.header = 1 + 2 + 2;
    my_message.type = WRP_TYPE_FEEDBACK;
    my_message.opcode = WRP_OPCODE_FSUBSCRIBE;
    my_message.addr = 1;                    // subscribe new
    my_message.data[0] = fb_addr[0];
    my_message.data[1] = fb_addr[1];
    wrp_send(&my_message);
  }

// parameter: true dcc coil addr (raw)
void wrp_send_accessory(unsigned int addr, bool activate)
  {
    unsigned int trntaddr = addr / 2;       // convert to turnout-addr
    unsigned char aspect = addr &0x01;      // 0 = red, 1 = green
    
    my_message.header = 1 + 2 + 1;
    my_message.type = WRP_TYPE_ACCESSORY;
    if (activate) my_message.opcode = WRP_OPCODE_A_SET;
    else          my_message.opcode = WRP_OPCODE_A_RESET;
    my_message.addr = trntaddr;
    my_message.data[0] = aspect;
    wrp_send(&my_message);
  }
 
// parameter: turnout addr [1..]
void wrp_send_accessory_query(unsigned int addr)
  {
    addr--;
    my_message.header = 1 + 2;
    my_message.type = WRP_TYPE_ACCESSORY;
    my_message.opcode = WRP_OPCODE_A_QUERY;
    my_message.addr = addr;
    wrp_send(&my_message);
  }


void wrp_clear_accessory_watch(unsigned int addr)
  {
    addr--;
    my_message.header = 1 + 2;
    my_message.type = WRP_TYPE_ACCESSORY;
    my_message.opcode = WRP_OPCODE_A_UNWATCH;
    my_message.addr = addr;
    wrp_send(&my_message);
  }  

void wrp_send_extended_accessory(unsigned int addr, unsigned char aspect)
  {
    addr--;
    my_message.header = 0x00 + 4;   // 5 = type + addr(2) + data
    my_message.type = WRP_TYPE_ACCESSORY;
    my_message.opcode = WRP_OPCODE_XA_SET;
    my_message.addr = addr;
    my_message.data[0] = aspect;
    wrp_send(&my_message);
  }



// parameter:
//      addr: dcc addr
//      format: DCC14, DCC28, DCC128
//      speed:  0..126
//      dir:    0 = forward, 1 = revers
// note: loco must be loaded completely, we transmit all

void wrp_send_loc_drive(unsigned int addr, t_format format, unsigned char speed, unsigned char dir)
  {
    wrp_loco_bits_t temp;
    temp.as_byte = 0;

    if(addr < 1)  return;
    if(addr > 10239)  return;  // // 10239 = 0x27FF

    my_message.header = 0x00 + 5;   // 5 = type + addr(2) + bitfield + speed
    my_message.type = WRP_TYPE_LOCO;
    my_message.opcode = WRP_OPCODE_DRIVE;
    my_message.addr = addr;
    
    temp.format = format;
    if (dir)    temp.dir = 1;
    my_message.data[0] = temp.as_byte;
    
    my_message.data[1] = speed;  
    
    // die bisherige payload durchsuchen und fallweise ein Speedkommando an die gleiche Lok ueberschreiben.   
    if (wrp_check_for_speed_replace(&my_message)) return;
    
    wrp_send(&my_message);
  }
  

// sends an complete function record for loco
void wrp_send_loc_control(locomem_t *loco)
  {
    wrp_loco_bits_t temp;
    temp.as_byte = 0;

    if(loco->addr < 1)  return;
    if(loco->addr > 10239)  return;  // // 10239 = 0x27FF

    my_message.header = 0x00 + 9;    // 9 = type + addr(2) + bitfield + speed + func (4)
    my_message.type = WRP_TYPE_LOCO;
    my_message.opcode = WRP_OPCODE_CONTROL;
    my_message.addr = loco->addr;
    temp.format = loco->format;
    if (loco->dir)    temp.dir = 1;
    my_message.data[0] = temp.as_byte;
    my_message.data[1] = loco->speed;
    
    my_message.data[2] = loco->f4_f1;
    if (loco->light)  my_message.data[2] |= 0x10;       // f0
    my_message.data[3] = loco->f12_f9  << 4 | loco->f8_f5;
    my_message.data[4] = loco->f20_f17 << 4 | loco->f16_f13;
    my_message.data[5] = loco->f28_f25 << 4 | loco->f24_f21;

    wrp_send(&my_message);
  }

  
// sends an emergency stop to this loco
void wrp_send_loc_stop(unsigned int addr)
  {
    my_message.header = 3;
    my_message.type = WRP_TYPE_LOCO;
    my_message.opcode = WRP_OPCODE_ESTOP;
    my_message.addr = addr;
    wrp_send(&my_message);    
  }

// ask command station about actual lok data
// we will receive: format, actual speed, functions f0 ... f12

void wrp_send_askforlocinfo(unsigned int addr)
  {
    my_message.header = 3;
    my_message.type = WRP_TYPE_LOCO;
    my_message.opcode = WRP_OPCODE_LQUERY;
    my_message.addr = addr;
    wrp_send(&my_message);    
  }

// note: loco.f?? must be loaded completely
// we transmit the complete parameters of the loco
// void xp_send_loc_function(unsigned char fnum)
void wrp_send_loc_function(void)
  {
    wrp_send_loc_control(&current_loc);
  }

void wrp_send_loc_light(void)
  {
    wrp_send_loc_control(&current_loc);
  }

//----------------------------------------------------------------------------
// CV-stuff
//----------------------------------------------------------------------------

void wrp_send_askforresult(void)
  {
    my_message.header = 1;
    my_message.type = WRP_TYPE_PROG;
    my_message.opcode = WRP_OPCODE_PQUERY;
    wrp_send(&my_message);
  }
  
void wrp_send_cv_read(unsigned int cv)  // 1..1024
  {
    my_message.header = 1 + 2;
    my_message.type = WRP_TYPE_PROG;
    my_message.opcode = WRP_OPCODE_RDCV;
    my_message.addr = cv - 1;
    
    wrp_send(&my_message);
  }

void wrp_send_cv_write(unsigned int cv, unsigned char value)   // 1..1024
  {
    my_message.header = 1 + 2 + 1;
    my_message.type = WRP_TYPE_PROG;
    my_message.opcode = WRP_OPCODE_WRCV;
    my_message.addr = cv - 1;
    
    my_message.data[0] = value; 
    
    wrp_send(&my_message);
  }


// PoM
//  
void wrp_send_PoM_cv_write(unsigned int addr, unsigned int cv, unsigned char value)
  {                  // type+oc  addr  data(3)
    my_message.header = 1 +      2 +    3;
    my_message.type = WRP_TYPE_PROG;
    my_message.opcode = WRP_OPCODE_POM_WRCV;
    my_message.addr = addr;
    
    unsigned int* temp;
    temp = (unsigned int *) &my_message.data[0];
    *temp = cv - 1;
    
    my_message.data[2] = value; 
    
    wrp_send(&my_message);
  }

void wrp_send_PoM_cv_read(unsigned int addr, unsigned int cv)
  {                  // type+oc  addr  data(2)
    my_message.header = 1 +      2 +    2;
    my_message.type = WRP_TYPE_PROG;
    my_message.opcode = WRP_OPCODE_POM_RDCV;
    my_message.addr = addr;
    
    unsigned int* temp;
    temp = (unsigned int *) &my_message.data[0];
    *temp = cv - 1;
    
    wrp_send(&my_message);
  }



// command station config

void wrp_send_cs_cfg_read(unsigned int cv)
  {
	my_message.header = 1 + 2 ;
    my_message.type = WRP_TYPE_SYSTEM;
    my_message.opcode = WRP_OPCODE_CS_CV_QRY;
    my_message.addr = cv;
    
    wrp_send(&my_message);
  }

void wrp_send_cs_cfg_write(unsigned int cv, unsigned char value)
  {
	my_message.header = 1 + 2 + 1;
    my_message.type = WRP_TYPE_SYSTEM;
    my_message.opcode = WRP_OPCODE_CS_CV_SET;
    my_message.addr = cv;
    
    my_message.data[0] = value; 
    
    wrp_send(&my_message);
  }

void wrp_send_clock_set(t_fast_clock *my_clock)
  {
    my_message.header = 1 + 2 + 4;
    my_message.type = WRP_TYPE_CLOCK;
    my_message.opcode = WRP_OPCODE_CLKSET;
    my_message.addr = 0;
    
    my_message.data[0] = 0x00 | my_clock->minute;
	my_message.data[1] = 0x80 | my_clock->hour;
	my_message.data[2] = 0x40 | my_clock->day_of_week;
    my_message.data[3] = 0xC0 | my_clock->ratio;

    wrp_send(&my_message);
  }
