//========================================================================
//
// OpenDCC - MFT
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      wrp_parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-07-10 V0.01 kw  start
//            2010-07-25 V0.02 kw  added programming messages
//            2013-07-25 V0.03 kw  Einarbeitung der nderung von H.Kurscheidt
//                                 (wenn SILENT / GRANTED State-Auswertung)
//
//------------------------------------------------------------------------
//
// purpose:   central station for dcc - parser for messages
//            transported with xbee
//
// content:   
//
// to do:     
//
//========================================================================

#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // cooperative scheduler and fifo
#include "timer.h"
#include "led.h"
#include "lcd_text.h"
#include "display.h"                // midlevel display routines
#include "keyboard.h"
#include "wrp.h"                    // defines for wireless railroad protocol
#include "wrp_messages.h"            
#include "xbee_if.h"            

#include "debug_graphics.h"         // analyze debug graphics message

#include "analog.h"
#include "messages_objects.h"
#include "serial.h"                 // trivial serial io, no isr



//-------------------------------------------------------------------------------------
//
static void parse_led(uint16_t led_enum, uint8_t r, uint8_t g, uint8_t b)
  {
    if (led_enum < NUM_OF_LED)
      {
        if (r > 0)
          {
            if (g > 0)  led_orange(led_enum);
            else        turn_led_red_immediate(led_enum);
          }
        else
            if (g > 0)  turn_led_green_immediate(led_enum);
            else        turn_led_off(led_enum);
      }
  }


//-------------------------------------------------------------------------------------
// xbee offers a byte wide payload interface
// when a payload is decoded, we get a pointer to a buffer with the content
// This buffer contains (possibly concatenated) messages of type wrp_pkt_t
//
// Here we do the slave side
//

void parse_mr_cmd(wrp_pkt_t *command)
  {
    unsigned char temp;
    wrp_pkt_t *pkt;
    pkt = (wrp_pkt_t *)command;

    while (pkt->header != 0)
      {     
        if (xbee_connect != XBEE_RUNNING)		//hk we allow only SILENT or GRANTED
          {
            if (pkt->type == WRP_TYPE_SYSTEM)
              {
                if (pkt->opcode == WRP_OPCODE_SILENT )
                  {                             // we get this message if we are considered 'SILENT'
                    restart_xbee();
                  }
                else if  (pkt->opcode == WRP_OPCODE_GRANTED)    // the network access is granted
                  {
                    set_xbee_state(XBEE_VOID);
                    set_xbee_state(XBEE_RUNNING);   // fresh!
                  }	
              }
          }
        else
          {
    		switch(pkt->type)
              {
                case WRP_TYPE_STATE:
                    switch(pkt->opcode)
                      {
                        case WRP_OPCODE_OFF:
                            set_command_station_status(STS_OFF);
                            cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                            break;
                        case WRP_OPCODE_STOP:
                            set_command_station_status(STS_STOP);
                            cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                            break;
                        case WRP_OPCODE_SOFT_STOP:
                            break;
                        case WRP_OPCODE_GO:
                            set_command_station_status(STS_ON);
                            cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                            break;
                        case WRP_OPCODE_PROGMODE:
                            set_command_station_status(STS_PROG);
                            cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                            break;
                        default:
                            break;
                      }
                    break;
                case WRP_TYPE_LOCO:
                    switch((wrp_oc_loco_t)pkt->opcode)
                      {
                        case WRP_OPCODE_ESTOP:   // ESTOP
                            break;
                        case WRP_OPCODE_CONTROL:   // CONTROL
                        
                            if(current_loc.addr != pkt->addr)  return;     // not our loc
    
                            {
                            wrp_loco_bits_t temp;
                            temp.as_byte = pkt->data[0];
    
                            if (temp.occupied)
                              {
                                record_inactive_loco(pkt->addr);           // tell messages_objects about the change
                                current_loc.other = 1;
                                cr_fifo_put(&key_fifo, KEY_PARSER_STOLEN);          // notify GUI
                              }
                            
                            current_loc.dir = temp.dir;
                            current_loc.format = temp.format & 0x03;
                            
                            }
                            
                            #warning: ist Speed korrekt
                            
                            current_loc.speed = pkt->data[1];  // noch aufs Format konvertieren??
    
                            if (pkt->data[2] & 0x10) current_loc.light = 1; 
                            else                     current_loc.light = 0;                 
                            
                            current_loc.f4_f1   = pkt->data[2] & 0x0F;
                            current_loc.f8_f5   = pkt->data[3] & 0x0F;
                            current_loc.f12_f9  = (pkt->data[3] & 0xF0) >> 4;
                            current_loc.f16_f13 = pkt->data[4] & 0x0F;
                            current_loc.f20_f17 = (pkt->data[4] & 0xF0) >> 4;
                            current_loc.f24_f21 = pkt->data[5] & 0x0F;
                            current_loc.f28_f25 = (pkt->data[5] & 0xF0) >> 4;
                         
                            cr_fifo_put(&key_fifo, KEY_PARSER_LOK);
                            break;
                        case WRP_OPCODE_DRIVE:   // DRIVE only
                            if(current_loc.addr != pkt->addr)  return;     // not our loc
    
                            {
                            wrp_loco_bits_t temp;
                            temp.as_byte = pkt->data[0];
                            
                            if (temp.occupied)
                              {
                                record_inactive_loco(pkt->addr);           // tell messages_objects about the change
                                current_loc.other = 1;
                                cr_fifo_put(&key_fifo, KEY_PARSER_STOLEN);          // notify GUI
                              }
    
                            current_loc.dir = temp.dir;
                            current_loc.format = temp.format & 0x03;                        
                            }
                            
                            #warning: ist speed korrekt
                            
                            current_loc.speed = pkt->data[1];  // noch aufs Format konvertieren??
    
    
                            cr_fifo_put(&key_fifo, KEY_PARSER_LOK);
                            break;
    
    
                        case WRP_OPCODE_LQUERY:    // QUERY
                            break;
                        case WRP_OPCODE_LSTOLEN:   // STOLEN
                            record_inactive_loco(pkt->addr);           // tell messages_objects about the change
                            if (pkt->addr == current_loc.addr)
                              {
                                current_loc.other = 1;
                                cr_fifo_put(&key_fifo, KEY_PARSER_STOLEN);          // notify GUI
                              }
                            break;
                        case WRP_OPCODE_LOCATE:    // LOCATE
                            break;
                        default:
                            break;
                      }
                    break;
                case WRP_TYPE_ACCESSORY:
                    switch((wrp_oc_accessory_t)pkt->opcode)
                      {
                        unsigned int addr;
                        unsigned char i;
                        case WRP_OPCODE_A_QUERY:
                            break;
                        case WRP_OPCODE_A_SET:
                            break;
                        case WRP_OPCODE_A_RESET:
                            break;
                        case WRP_OPCODE_A_POSIT:
                            addr = pkt->addr;
                            addr++;  // we code from 1 to 2040
                            for (i=0; i<sizeof(actual_weiche)/sizeof(actual_weiche[0]); i++)
                              {
                                if (addr == actual_weiche[i].addr)
                                  {
                                    switch (pkt->data[0])
                                      {
                                        default:       actual_weiche[i].state = WS_VOID;  break;
                                        case 0:        actual_weiche[i].state = WS_0;     break;
                                        case 1:        actual_weiche[i].state = WS_1;     break;
                                      }
                                    cr_fifo_put(&key_fifo, KEY_PARSER_SWITCH);
                                  }
                              }
                            break;
                        case WRP_OPCODE_A_WATCH:
                            break;
                        case WRP_OPCODE_A_UNWATCH:  // if a watch fails
                            break;
                        case WRP_OPCODE_XA_SET:
                            break;
                        case WRP_OPCODE_XA_QUERY:
                            break;
                        case WRP_OPCODE_XA_POSIT:
                            break;
                      }
                    break;
                case WRP_TYPE_FEEDBACK:
                    switch((wrp_oc_feedback_t)pkt->opcode)
                      {
                        case WRP_OPCODE_FEVENT:     // single event
                            break;
                        case WRP_OPCODE_FBITS8:     // 8 bit vektor, addr, vektor in data 0
    
                            if (pkt->addr < FB_ARRAY_SIZE)
                              {
                                if ( (pkt->addr == fb_addr[0]) ||  (pkt->addr == fb_addr[1])  )
                                  {
                                    cr_fifo_put(&key_fifo, KEY_PARSER_FB);   //notify gui if change detected
                                                                             // here: notify anyway
                                  }
                                fb_array[pkt->addr] = pkt->data[0];
                              }
                            break;
                        case WRP_OPCODE_FBIDI:
                            break;
                        case WRP_OPCODE_FCONFIG:
                            break;
                        case WRP_OPCODE_FQUERY1:
                            break;
                        case WRP_OPCODE_FQUERY8:
                            break;
                        case WRP_OPCODE_FSUBSCRIBE:     // subscription failed; not nice, but we ignore it.
                            break;
                        case WRP_OPCODE_FQBIDI:
                            break;
                        case WRP_OPCODE_FQCONFIG:
                            break;
                      }
                    break;
                case WRP_TYPE_PROG:
                    switch((wrp_oc_prog_t)pkt->opcode)
                      {
                        case WRP_OPCODE_PROG_STAT:   // status: addr = 0; data[0]=state,
                            if (pkt->addr == 0 || (pkt->addr & 0x3FFF) == current_CV.addr)
                              {
                                switch(pkt->data[0])
                                  {
                                    default:
                                        current_CV.result = PRG_CV_VOID;
                                        break;
                                    case 0:
                                        current_CV.result = PRG_CV_OKAY;
                                        break;
                                    case 1:
                                        current_CV.result = PRG_CV_WAIT_FOR_ANSWER;
                                        set_command_station_status(STS_PBUSY);
                                        we_should_ask_again = TRUE;                     // tell messages_objects to repeat the query
                                        break;
                                    case 2:
                                        current_CV.result = PRG_CV_NOLOK;
                                        break;
                                    case 3:
                                        current_CV.result = PRG_CV_SHORT;
                                        break;
                                  }
                                current_CV.number = ((pkt->data[2] << 8) | pkt->data[1]) + 1;
                                current_CV.value = pkt->data[3];
                                cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                                cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                              }
                            break;
                        case WRP_OPCODE_RDCV:
                            break;
                        case WRP_OPCODE_WRCV:
                            break;
                        case WRP_OPCODE_POM_RDCV:
                            break;
                        case WRP_OPCODE_POM_WRCV:
                            break;
                        case WRP_OPCODE_PAC_RDCV:
                            break;
                        case WRP_OPCODE_PAC_WRCV:
                            break;
                        case WRP_OPCODE_PEX_RDCV:
                            break;
                        case WRP_OPCODE_PEX_WRCV:
                            break;
                        case WRP_OPCODE_PQUERY:
                            break;
                        case WRP_OPCODE_POM_QUERY:
                            break;
                      }
                    break;
                case WRP_TYPE_CLOCK:
                    switch((wrp_oc_clock_t)pkt->opcode)
                      {
                        uint8_t i,j;
                        case WRP_OPCODE_CLKOFF:
                            break;
                        case WRP_OPCODE_CLKON:
                            break;
                        case WRP_OPCODE_CLKSET:
                            for(i = 0; i < 4; i++ )
                              {
                                j = 99;     // use j as temp
                                switch(pkt->data[i] & 0xC0)
                                  {
                                    case 0x00:  j = pkt->data[i] & 0x3F;
                                                if (j < 60) fast_clock.minute = j;
                                                break;
                                    case 0x80:  j = pkt->data[i] & 0x3F;
                                                if (j < 24) fast_clock.hour = j;
                                                break;
                                    case 0x40:  j = pkt->data[i] & 0x3F;
                                                if (j < 7) fast_clock.day_of_week = j;
                                                break;
                                    case 0xC0:  j = pkt->data[i] & 0x3F;
                                                if (j < 32) fast_clock.ratio = j;
                                                break;
                                  }
                              }                            
                            cr_fifo_put(&key_fifo, KEY_PARSER_TIME);
                            break;
                        case WRP_OPCODE_CLKQRY:
                            break;
                        default:
                            break;
                      }
                    break;
                case WRP_TYPE_SYSTEM:
                    switch((wrp_oc_system_t)pkt->opcode)
                      {
                        case WRP_OPCODE_IDLE:       // kind of beacon / Heartbeat
                            #warning we do a ACK anyhow, so this isnt required
                            wrp_send_sys_idle();    // send idle back
                            wrp_transmit();
                            break;
                        case WRP_OPCODE_SILENT:     // we get this message if we are considered 'SILENT'
                            // wrp_send_sys_apply();   // (re-)apply
                            // wrp_transmit();
                            break;
                        case WRP_OPCODE_APPLY:      // a slave applies for network access
                            break;
                        case WRP_OPCODE_GRANTED:    // the network access is granted
                            set_xbee_state(XBEE_VOID);
                            set_xbee_state(XBEE_RUNNING);   // fresh!
                            break;
                        case WRP_OPCODE_CONNECT:
                            break;
                        case WRP_OPCODE_SYSINFO:    // network quality report   
                            break;
                        case WRP_OPCODE_SYSQUERY:   
                            break;
                        case WRP_OPCODE_CS_CV_QRY:
                            break;
                        case WRP_OPCODE_CS_CV_SET:  // master tells us the answer to a cs_cv_qry
                            current_CV.number = pkt->addr;
                            current_CV.value = pkt->data[0];
                            current_CV.result = PRG_CV_OKAY;
                            cr_fifo_put(&key_fifo, KEY_PARSER_CFG);
                            break;
                        case WRP_OPCODE_PICTURE:   
                            break;
                        case WRP_OPCODE_LOCO_DB:   
                            break;
                        default:
                            break;
                      }
                    break;
                case WRP_TYPE_DEBUG:
                    switch((wrp_oc_debug_t)pkt->opcode)
                      {
                        case WRP_OPCODE_PING:
                            break;
                        case WRP_OPCODE_ANNOUNCE:
                            cr_fifo_put(&key_fifo, KEY_XBEE_ANNOUNCE);
                            break;
                        case WRP_OPCODE_FIRMWARE:
                            break;
                        case WRP_OPCODE_DISPLAY:
                            parse_display(&pkt->data[0], pkt->addr);  // addr: size, data: display commands
                            break;
                        case WRP_OPCODE_LED:
                            parse_led(pkt->addr, pkt->data[0], pkt->data[1], pkt->data[2]);
                            break;
                        case WRP_OPCODE_KEYBOARD:
                            break;
                      } 
                default:
                    break;
              }
          }         
        temp = (pkt->header & 0x7F) + 1;
        pkt = (wrp_pkt_t *) ((uint8_t*) pkt + temp);       // evaluate next part of message        
    }
  }






