////========================================================================
//
// OpenDCC - MFT - Gateway
//
// Copyright (c) 2008-10 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      xbee_if.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-07-10 V0.01 kw start
//            2010-05-30 V0.02 kw first transmissions
//            2010-06-14 V0.03 kw changed init
//                             kw timeout checking changed
//            2010-08-05       mp additional code to allow Serie 1
//            2010-08-06       kw added xbee_hw_version
//            2010-08-08 V0.04 mp added support of XBEE Series1
//            2010-08-13 V0.05 mp added readout of RSSI
//            2013-07-26 V0.06 kw abfrage auf XBEE_CONNECTED bei Heartbeat
//
// Bitte   'Frage an hk' klren
//------------------------------------------------------------------------
//
// purpose:   throttle - client for xbee modules
//
// content:   collector for incoming byte
//            packetizing
//            pecket decoding
//
// note:      xbee uses big endian for length, we are little endian
//            ntohs is done in send and receive routines
//
// to do:
//
//========================================================================
//
// THIS IS THE CODE for  CLIENT
//
//========================================================================

#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#ifdef AVR
 #include <avr/io.h>
 #include <avr/interrupt.h>
 #include <avr/eeprom.h>
 #include <util/delay.h>
#endif

#include "config.h"                 // general
#include "hardware.h"               // used hardware
#include "cortos.h"                 // cooperative scheduler and fifo
#include "timer.h"
#include "led.h"
#include "lcd_text.h"
#include "keyboard.h"
#include "wrp_parser.h"             // this decodes the payload


#include "messages_objects.h"
#include "serial.h"                 // serial, isr based

#include "xbee_protocol.h"          // defines of packets
#include "xbee_if.h"

#include "wrp_messages.h"           // create_wrp_sys_idle

#include "display.h"                // display_error


// check compile switch (see xbee_if.h)

#ifdef XBEE_IF
  #if (XBEE_IF == 0)
    #error XBEE_IF not selected - must be MASTER (=1) or CLIENT (=2)
  #endif
#else
    #error XBEE_IF not defined - must be MASTER (=1) or CLIENT (=2)
#endif

//------------------------------------------------------------------------
//
//  0. GLOBALS
//
//------------------------------------------------------------------------

#define XBEE_PERIOD           200000L       // abfrage-Intervall auf Beacon erneut senden
#define BEACON_PERIOD        2000000L
#define BASESTATION_TIMEOUT  4000000L

uint8_t  xbee_at_cmd_in_work = 0;           // flag, is set on every at_cmd response
                                            // 1: at_command in progress, don't send a new one

uint8_t  xbee_module_detected;              // we hold here the global state (filled during init)
uint8_t  xbee_error;                        // error number
uint8_t  xbee_ai_state = XBEE_AI_NET_SCAN;  // Association indication

uint16_t pan_id = 0xffff;                   // the PAN of the network (ZigBEE has 64 Bit, but we take only lower 16 Bit)
uint8_t  xbee_channel;                      // the Channel we are operation on (read only)
uint8_t  xbee_rssi;                         // Signal strength on last receive (given in dBm *-1)
uint16_t xbee_short;                        // Local Adress, Coordinator is 0x0000

uint16_t xbee_hw_version = 0xffff;          // hw version
uint16_t xbee_fw_version = 0xffff;          // firmware version of module
uint32_t xbee_serial_high = 0xfedcba98;     // our serial number, high bits
uint32_t xbee_serial_low  = 0x76543210;     // low bits.

// when ever we send something - this is our host

uint8_t base_station_valid;

uint8_t base_station_mac[8] = {0x12,0x34,0x00,0x00,0x00,0x00,0x00,0x90};

uint8_t broadcast_mac[8] = {0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0xFF};

uint8_t coord_mac[8] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};



static int xbee_active_timestamp;   // this is time of the last active (RF !!!) receive


//------------------------------------------------------------------------
//
//  1. Connect Status & Tools
//
//------------------------------------------------------------------------

// ret true if a module is detected
// (side effect: we access the module, thus disabling USB)
// Note: after power up, XBEE Module need approx. 200ms to boot up
uint8_t detect_xbee_module(void)
  {
    ENABLE_XBEE;
    _delay_us(1);
    if (MY_RTS_DETECT) xbee_module_detected = 1;
    else xbee_module_detected = 0;          // no avtice /CTS detected (line is high)
    // DISABLE_USB_XBEE;                       // deactivate both
    return(xbee_module_detected);
  }


// Generate Checksum for an XBee packet
// this includes: all data, including command_id, but not length
// parameter: pointer to packet (to the delimiter)
uint8_t xbee_checksum(const xbee_pkt_t *pkt)
  {
    uint8_t *pkt_data = ((uint8_t *)pkt) + sizeof(pkt->header.delimiter) + sizeof(pkt->header.len);
    uint16_t i;
    uint8_t checksum = 0;

    for (i = 0; i < pkt->header.len.as_uint16; i++)
        checksum += *(pkt_data++);

    return ~checksum;
  }

// if the lib detects a change in connection status, this routine is called.
void set_xbee_state(xbee_connect_t neu)
  {
    if (neu != xbee_connect)
      {
        xbee_connect = neu;
        switch (xbee_connect)
          {
            default:
            case XBEE_DISCONNECTED:
                if (used_interface == USE_XBEE)
                  {
                    cr_fifo_put(&key_fifo, KEY_DISCONNECT);   // this enforces display update (heart beat)
                    set_command_station_status(STS_DISCONNECTED);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    //!warning! hat das noch weitergehende Folgen?
                  }
                break;
            case XBEE_ENDPOINT:
            case XBEE_COORDINATOR:
                break;
            case XBEE_RUNNING:
                if (used_interface == USE_XBEE)
                  {
                    cr_fifo_put(&key_fifo, KEY_CONNECT);
                    xbee_active_timestamp = get_systick_int();
                  }
                break;
            case XBEE_VOID:     // just a dummy to be able to force an update
                break;
          }
      }
    else if (xbee_connect == XBEE_RUNNING)
      {
        cr_fifo_put(&key_fifo, KEY_HEARTBEAT);   // this enforces display update (heart beat)
        xbee_active_timestamp = get_systick_int();
      }
  }

void update_xbee_connected(void)
  {
    if (xbee_connect == XBEE_RUNNING) cr_fifo_put(&key_fifo, KEY_HEARTBEAT);   // this enforces display update (heart beat)
    // hk else nothing, because we were disconnected
    xbee_active_timestamp = get_systick_int();
  }

//-------------------------------------------------------------------------------------
//
//  2. RECEIVE
//
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
// we use a structure of type xbee_t to handle packetizing
//
// we keep this structure global and only once - so processing
// must be finished within this task!

xbee_pkt_t xbee_rx;

//----------------------------------------------------------------------------------
//
//  2.a) Receive Part / XBEE Parsers
//
//----------------------------------------------------------------------------------

void parse_at_cmd_response(uint8_t *command)
  {
    t_data16 *code;

    code = (t_data16 *)command;
    switch (code->as_uint16)
      {
        case (XBEE_CMD_RSS):                        // 'DB'
            command += 3;                           // skip at-command and qualifier
            xbee_rssi = *command;                 // Just one Byte
            break;
        case (XBEE_CMD_ASSOC_STATUS):               // 'AI'
            command += 3;                           // skip at-command and qualifier
            xbee_ai_state = *command;               // one Byte
            cr_fifo_put(&key_fifo, KEY_XBEE_AI);
            break;
        case (XBEE_CMD_CHANNEL):                    // 'CH'
            command += 3;                           // skip at-command and qualifier
            xbee_channel = *command;                // one Byte
            break;
        case (XBEE_CMD_SRC_ADDR16):                 // 'MY'
            command += 3;                           // skip at-command and qualifier
            xbee_short = *command++ * 256;          // assume big endian
            xbee_short += *command;
            break;
        case (XBEE_CMD_PAN_ID):                     // 'ID' (this response is different, so we check message size)
            if (xbee_rx.header.len.low8==0x07)
              {
                command += 3;                       // skip at-command and qualifier
                pan_id = *command++ * 256;          // assume big endian
                pan_id += *command;
              }
            else
              {
                command += 9;                       // skip at-command, qualifier and 6 byte of pan
                pan_id = *command++ * 256;          // assume big endian
                pan_id += *command;
              }
            break;
        case (XBEE_CMD_VERS_FIRMWARE):              // 'VR' = Version
            command += 3;                           // skip at-command and qualifier
            xbee_fw_version = *command++ * 256;     // assume big endian
            xbee_fw_version += *command;
            break;
        case (XBEE_CMD_VERS_HARDWARE):              // 'HV' = Version
            command += 3;                           // skip at-command and qualifier
            xbee_hw_version = *command++ * 256;     // assume big endian
            xbee_hw_version += *command;
            break;
        case (XBEE_CMD_SER_HI):                     // 'SH' S=0x53 H=0x48
            command += 3;
            xbee_serial_high = (uint32_t) *command++ << 24;
            xbee_serial_high += (uint32_t) *command++ << 16;
            xbee_serial_high += (uint32_t) *command++ << 8;
            xbee_serial_high += (uint32_t) *command++;
            break;
        case (XBEE_CMD_SER_LO):                     // 'SL' S=0x53 L=0x4C
            command += 3;
            xbee_serial_low = (uint32_t) *command++ << 24;
            xbee_serial_low += (uint32_t) *command++ << 16;
            xbee_serial_low += (uint32_t) *command++ << 8;
            xbee_serial_low += (uint32_t) *command++;
            break;
        default:
            display_error2(ERR_XBEE_UNKNOWN_COMMAND,command[2]);
            break;
      }
    xbee_at_cmd_in_work = 0;               // reset flag
  }

//-------------------------------------------------------------------------------------
// whenever we received a complete packet - we check what is in it.

void parse_xbee_packet(void)
  {
    switch(xbee_rx.header.cmd_id)
      {
        case XBEE_PKT_TYPE_MODEM_STATUS:            // 0x8A: 00 kommt am Anfang - XBEE reset
            switch(xbee_rx.modem_status.status)
              {
                case XBEE_STATUS_COORD_STARTED:
                    // set_xbee_state(XBEE_COORDINATOR);   // Frage an hk: warum sollte man das kommentieren?
                    break;
                default:
                case XBEE_STATUS_HW_RESET:
                case XBEE_STATUS_WD_RESET:
                    break;
                case XBEE_STATUS_JOINED:
                    set_xbee_state(XBEE_ENDPOINT);
                    xbee_ai_state = XBEE_AI_OKAY;       // hk
                    break;
                case XBEE_STATUS_DISASSOC:
                    xbee_ai_state = XBEE_AI_NET_SCAN;   // hk
                    break;
                case XBEE_STATUS_NETWORK_KEY:
                case XBEE_STATUS_VCC_LIMIT:
                case XBEE_STATUS_KEY_ESTABLISHED:
                    // ignored
                    break;
                case XBEE_STATUS_STACK_ERROR:
                    display_error(ERR_XBEE_FAILURE);
                    break;
              }
            break;
        case XBEE_PKT_TYPE_ATCMD_RESP:              // 0x88: response to at-command
            {
              switch(xbee_rx.cmd_resp.status)
                {
                    case XBEE_CMDSTATUS_OK:
                        // valid command response receive
                        parse_at_cmd_response(xbee_rx.cmd_resp.command);  // pointer to command
                        break;
                    default:
                        break;
                    case XBEE_CMDSTATUS_ERR:
                        // ignored
                        display_error(ERR_ATCMD_FAILED);
                        break;
                }

            }
            break;
        case XBEE_PKT_TYPE_TX_STATUS:               // 0x89: tx status for series 1 (normal mode)
            switch(xbee_rx.tx_status.status)
              {
                case 0x00:  // = Success
                    break;
                case 0x01:  // = MAC ACK Failure
                case 0x02:  // = CCA Failure
                case 0x03:  // = PURGE
                    // set_xbee_state(XBEE_DISCONNECTED);     // Frage an hk: warum kommentieren?
                    break;
                default:
                    blink_led_red_fast(LED_A);
                    blink_led_red_fast(LED_B);
                    blink_led_red_fast(LED_C);
                    blink_led_red_fast(LED_D);
                    display_error(xbee_rx.zb_tx_status.status);
                    // set_xbee_state(XBEE_DISCONNECTED);     // Frage an hk: warum kommentieren?
                    break;
              }
              break;
        case XBEE_PKT_TYPE_ZB_TX_STATUS:            // 0x8B: tx status
            switch(xbee_rx.zb_tx_status.status)
              {
                case 0x00:  // = Success
                    break;
                case 0x01:  // = MAC ACK Failure    
                case 0x15:  // = Invalid destination endpoint    
/*
Der Koordinator hat eine sog. Child table, wo er alle Endpoints (Childs) eintrgt. 
Da diese tabelle nur 9 Eintrge zult, trgt er die Children, die nicht mehr erreichbar sind,
nach einem time out aus der Tabelle aus. Der Time out wird wieder durch den
SP (sleep) parameter und den SL parameter gebildet. SP(3*n*30ms) * SL. n 0x20...0xaf0.
Der Endpoint versucht beim Wiedereintritt in den Funkbereich, 
seine Eltern wiederzufinden und macht einen sog. Orphan scan (Waise). 
Der EP behlt alle Daten den Koord. 
Der Koordinator nimmt den EP wieder auf, vergibt allerdings eine neue 16 bit
Netzwerk Adresse. Diese Adresse wird als random neu erzeugt!
Da das Gateway nicht weiss, wann der EP verloren gegangen ist und ausgetragen
wurde, sendet er Silent msgs an die alte 16 bit Adresse. Der MFT sendet nichts,
weil er auf die Silent msg wartet, um aus seinem disconnect heraus zu kommen.
Da der MFT nichts sendet, bekommt der Gateway auch die neue 16 bit adr. nicht mit.
-->
Wir senden hier ein idle paket, damit der GWY anhand der mac addr. die neue
16 bit adr. herausfischen kann.   */

					break;
                case 0x22:  // = Not Joined to Network
                    // set_xbee_state(XBEE_DISCONNECTED);  // Frage an hk: warum kommentieren?
                    break;
                case 0x24:  // = Address Not Found   !Hier wre fallweise ein Retry interessant
                case 0x25:  // = Route Not Found
                    // set_xbee_state(XBEE_DISCONNECTED);  // Frage an hk: warum kommentieren?
                    break;
                case 0x21:  // = Network ACK Failure
                    // set_xbee_state(XBEE_DISCONNECTED);  // Frage an hk: warum kommentieren?
                    break;
                case 0x91:  // = Unknown
                    blink_led_red_fast(LED_A);
                    blink_led_green_fast(LED_B);
                    blink_led_green_fast(LED_C);
                    blink_led_red_fast(LED_D);
                    display_error(xbee_rx.zb_tx_status.status);
                    // set_xbee_state(XBEE_DISCONNECTED);  // Frage an hk: warum kommentieren?
                    break;
                default:
                case 0x02:  // = CCA Failure
                case 0x23:  // = Self-addressed
                case 0x26:  // = Broadcast source failed to hear a neighbor relay the message
                case 0x2B:  // = Invalid binding table index
                case 0x2C:  // = Resource error lack of free buffers, timers, etc.
                case 0x32:  // = Resource error lack of free buffers, timers, etc.
                case 0x74:  // = Data payload too large
                case 0x75:  // = Indirect message unrequested
                    blink_led_red_fast(LED_A);
                    blink_led_red_fast(LED_B);
                    blink_led_red_fast(LED_C);
                    blink_led_red_fast(LED_D);
                    display_error(xbee_rx.zb_tx_status.status);
                    // set_xbee_state(XBEE_DISCONNECTED);  // Frage an hk: warum kommentieren?
                    break;
              }
            break;
        case XBEE_PKT_TYPE_ZB_RX:                   // 0x90: rx packet
            if (xbee_rx.zb_rx.opt == 0x02)
              {
                // we received a broadcast
                // check msb of payload header
                // if set: this is a message from our master -> we learn the mac

                #warning das muss hier noch ein compare werden - nur granted laedt die base_mac (broadcast?)
                if (xbee_rx.zb_rx.data[0] & 0x80)
                  {
                    memcpy(base_station_mac, xbee_rx.zb_rx.src64, sizeof(base_station_mac));
                    base_station_valid = 1;         // einstweilen immer true

                  }
              }

            update_xbee_connected();                        // we got something - connection still alive

            parse_mr_cmd((wrp_pkt_t *) xbee_rx.zb_rx.data);  // now decode the payload
            if (xbee_at_cmd_in_work == 0)
              {
                xb_make_at_command(XBEE_CMD_RSS);            // "DB", read signal strength
                send_xbee_message();
                xbee_at_cmd_in_work = 1;
              }
            break;
        case XBEE_PKT_TYPE_RX64:                    // 0x80: standard64 rx packet
            if (xbee_rx.a64_rx.opt == 0x02)
              {
                // 0x02 = we received a address broadcast
                // 0x04 = we received a PAN broadcast
                // check msb of payload header
                // if set: this is a message from our master -> we learn the mac

                #warning das mu hier noch ein compare werden - nur granted laedt die base_mac
                if (xbee_rx.a64_rx.data[0] & 0x80)
                  {
                    memcpy(base_station_mac, xbee_rx.a64_rx.src64, sizeof(base_station_mac));
                    base_station_valid = 1;         // einstweilen immer true

                  }
              }
            xbee_rssi = xbee_rx.a64_rx.rssi;
            update_xbee_connected();

            parse_mr_cmd((wrp_pkt_t *) xbee_rx.a64_rx.data);  // now decode the payload            break;
            break;
        case XBEE_PKT_TYPE_RX16:
            if (xbee_rx.a16_rx.opt == 0x02)
              {
                // 0x02 = we received a address broadcast
                // 0x04 = we received a PAN broadcast
                // check msb of payload header
                // if set: this is a message from our master -> we learn the mac

                #warning das mu hier noch ein compare werden - nur granted laedt die base_mac
                if (xbee_rx.a16_rx.data[0] & 0x80)
                  {
//                    memcpy(base_station_mac, xbee_rx.a16_rx.src64, sizeof(base_station_mac));
                    base_station_valid = 1;         // einstweilen immer true

                  }
              }
            xbee_rssi = xbee_rx.a16_rx.rssi;
            update_xbee_connected();

            parse_mr_cmd((wrp_pkt_t *) xbee_rx.a16_rx.data);  // now decode the payload            break;
            break;
        case XBEE_PKT_TYPE_ZB_X_RX:                 // 0x91: explizit rx packet (AO=1)
        case XBEE_PKT_TYPE_ATCMD:                   // send commands
        case XBEE_PKT_TYPE_QATCMD:                  // 0x09: queued commands
        case XBEE_PKT_TYPE_REMOTE_ATCMD:            // remote command
        case XBEE_PKT_TYPE_REMOTE_ATCMD_RESP:       // response to remote command
        case XBEE_PKT_TYPE_ZB_TX_REQ:               // 0x10: transmit request
        case XBEE_PKT_TYPE_ZB_CMD_FRAME:            // 0x11: expizit addressing
        default:
            display_error2(ERR_XBEE_UNEXPECTED_FRAME,xbee_rx.header.cmd_id);
            // display_error(ERR_XBEE_UNEXPECTED_FRAME);
            break;
      }
  }

//----------------------------------------------------------------------------------
//
//  2.b) Receive Part / Collect routine for incoming bytes (API Mode)
//
//----------------------------------------------------------------------------------
// This is done with:
//   - an ISR, feeding received bytes to a fifo (see serial.c)
//   - this ISR notifies the real time kernel (cortos) to call
//     the following consumer routine, so whenever we receive something
//     this task gets active
//   - it reads data from the serial fifo and forms a xbee_pkt (xbee_rx)
//
// used globals:
static uint8_t bytes_left;  // bytes still to receive for this packet
static uint8_t bytes_rcvd;  // bytes already in this packet
static uint8_t x_index;     // pointer to data

static uint8_t escape_hot;

t_cr_task run_xbee_client(void)
  {
    unsigned char data;

    if ((get_systick_int() - xbee_active_timestamp) > (BASESTATION_TIMEOUT / SYSTICK_PERIOD) )
      {
        set_xbee_state(XBEE_DISCONNECTED);
      }

    while (rx_fifo_ready())
      {
        data = rx_fifo_read();                 // reads one byte
        if (data == XBEE_PKT_ESC)              // XBEE_PKT_ESC = 0x7D
          {
            escape_hot = 1;
          }
        else
          {
            if (escape_hot)
              {
                data ^= 0x20;
                escape_hot = 0;
              }

            switch(bytes_rcvd++)
              {
                case 0:
                    if (data == XBEE_PKT_START)
                      {
                        xbee_rx.header.delimiter = data;
                        escape_hot = 0;            // play it secure
                      }
                    else
                      {
                        // ignore data and wait for delimiter
                        bytes_rcvd = 0;
                      }
                    break;

                case 1:
                    xbee_rx.header.len.high8 = data; // resolve endiness with structure
                    break;

                case 2:
                    xbee_rx.header.len.low8 = data;
                    if ( xbee_rx.header.len.as_uint16 > XBEE_MAX_DATA_LEN )
                      {
                        bytes_rcvd = 0;
                        // xbee_rx_err(xbee);
                        // #warning handle error missing
                      }
                    bytes_left = xbee_rx.header.len.low8;
                    break;
                case 3:
                    xbee_rx.header.cmd_id = data;
                    x_index = 0;
                    bytes_left--;
                    break;

                default:
                    if (bytes_left == 0)
                      {
                        if ( xbee_checksum(&xbee_rx) != data)
                          {
                            // failure in checksum - drop this packet
                            bytes_rcvd = 0;
                            // xbee_rx_checksum_err(xbee);
                            // !warning! handle of checksum error missing
                            // we have very short lines, so this should never happen
                          }
                        else
                          {
                            // now process this packet
                            xbee_rx.raw.data[x_index] = 0;    // set end delimiter (instead of checksum)
                            parse_xbee_packet();
                          }
                        bytes_rcvd = 0;
                      }
                    else
                      {
                        xbee_rx.raw.data[x_index++] = data;
                        bytes_left--;
                      }
                    break;
              }
          }
      }
                                            // everything processed - no longer ready
                                            // we return with a sleeptime (for timeout checking!)
    return(200000L / SYSTICK_PERIOD);       // wait 200ms

                                            // The incoming fifo will issue a wakeup (if required).
  }



//-------------------------------------------------------------------------------------
//
//  3. TRANSMIT
//
//-------------------------------------------------------------------------------------

// to save memory, we use one global packet only
// (this must be tramsmitted before we can send a new one)

signed int xbee_last_tx;        // time of last transmission

uint8_t tx_frame_id = 1;        // we use a frame-id (offensichtlich koppeln die XBEE intern eine
                                // frame-id = 0 bei normalen frames auf AT-Commands ueber

xbee_pkt_t xbee_tx;

void xbee_send(unsigned char c)
  {
    //        0x7E                    0x7D                  0x11                  0x13
    if ((c == XBEE_PKT_START) | (c == XBEE_PKT_ESC) | (c == XBEE_PKT_XON) | (c == XBEE_PKT_XOFF))
      {
        while (!tx_fifo_ready()) {};        // busy waiting
        tx_fifo_write(XBEE_PKT_ESC);        // escape char
        c = c ^ 0x20;
      }
    while (!tx_fifo_ready()) {};  // busy waiting
    tx_fifo_write(c);
  }

void xbee_send_delimiter(void)
  {
    while (!tx_fifo_ready()) {};  // busy waiting
    tx_fifo_write(XBEE_PKT_START);
  }

// send_xbee_message puts a prepared message in xbee_tx to output fifo
// (busy waiting if fifo is getting full)
// 26.04.2010 - API mode 2 with escaping
void send_xbee_message(void)
  {
    unsigned char i=0, len;

    len = xbee_tx.header.len.low8;                  // len = [cmd_id+data]
    while (!tx_fifo_ready()) ;                      // busy waiting!!!
                                                    // vorlufig kein Timeout
    xbee_send_delimiter();
    xbee_send(xbee_tx.header.len.high8);            // resolve endianess
    xbee_send(xbee_tx.header.len.low8);
    xbee_send(xbee_tx.header.cmd_id);
    while (i < len)
      {
        xbee_send(xbee_tx.raw.data[i++]);               // we send one more data than payload size
                                                    //  (this is the checksum byte)
      }
  }

// this sends the accumulated payload to the host (as TX64-Paket, standard 64 addressing)
void xb_send_payload_to_host_S1(void)
  {
    unsigned char payload_size;
    unsigned char checksum = 0;
    uint8_t i;
    t_data16            len;            // this allows to resolve endianess

    xbee_send_delimiter();

    xbee_last_tx = get_systick_10ms();

    payload_size = payload.message_size;  // inkl. header
    //              cmd_id txheader
    len.as_uint16 = 1    + 10 +     payload_size;
    xbee_send(len.high8);            // resolve endianess
    xbee_send(len.low8);
    xbee_send(XBEE_PKT_TYPE_TX64);
    checksum += XBEE_PKT_TYPE_TX64;

    xbee_send(tx_frame_id);
    checksum += tx_frame_id;
    tx_frame_id++;
    if (tx_frame_id == 0) tx_frame_id++;

    for (i=0; i<sizeof(coord_mac); i++)
      {
        xbee_send(coord_mac[i]);
        checksum += coord_mac[i];
      }

    xbee_send(0);    // opt = 0;
    checksum += 0;

    for (i=0; i<payload_size; i++)
      {
        xbee_send(payload.message[i]);
        checksum += payload.message[i];
      }

    xbee_send(0xFF - checksum);
  }



//create the api frame (general zb_tx form)
void xb_make_xbee_tx(uint8_t size, uint8_t *payload)
  {
    xbee_tx.header.delimiter = XBEE_PKT_START;
    xbee_tx.header.len.as_uint16 = (size + 1);
    xbee_tx.header.cmd_id = XBEE_PKT_TYPE_ZB_TX_REQ;
    memcpy(xbee_tx.raw.data, payload, size);
    xbee_tx.raw.data[size] = xbee_checksum(&xbee_tx);
  }



// this sends the accumulated payload to the host (as ZB_TX-Paket, for Series 2)
void xb_send_payload_to_host_S2(void)
  {
    unsigned char payload_size;
    unsigned char checksum = 0;
    uint8_t i;
    t_data16            len;            // this allows to resolve endianess

    xbee_send_delimiter();

    xbee_last_tx = get_systick_10ms();

    payload_size = payload.message_size;  // inkl. header
    //              cmd_id txheader
    len.as_uint16 = 1    + 13 +     payload_size;
    xbee_send(len.high8);            // resolve endianess
    xbee_send(len.low8);
    xbee_send(XBEE_PKT_TYPE_ZB_TX_REQ);
    checksum += XBEE_PKT_TYPE_ZB_TX_REQ;

    xbee_send(tx_frame_id);
    checksum += tx_frame_id;
    tx_frame_id++;
    if (tx_frame_id == 0) tx_frame_id++;

    for (i=0; i<sizeof(coord_mac); i++)
      {
        xbee_send(coord_mac[i]);
        checksum += coord_mac[i];
      }

    xbee_send(0xFF);                // netaddr.high8
    checksum += 0xFF;
    xbee_send(0xFE);                // netaddr.low8
    checksum += 0xFE;

    xbee_send(0);    // radius = 0;
    checksum += 0;
    xbee_send(0);    // opt = 0;
    checksum += 0;

    for (i=0; i<payload_size; i++)
      {
        xbee_send(payload.message[i]);
        checksum += payload.message[i];
      }

    xbee_send(0xFF - checksum);
  }

// xb_send_payload_to_host sends the paket, but handling is different
// on module type
// we distinguish according to hardware version
//     Byte 1   | Part Number
//     0x17**   | XB24 (series 1)
//     0x18**   | XBP24 (series 1) (pro)
//     0x19**   | XB24-B (sereies 2)
//     0x1A**   | XBP24-B (series 2) (pro)
void xb_send_payload_to_host(void)
  {
    uint8_t main_ver;
    main_ver = (xbee_hw_version & 0xFF00) >> 8;

    switch(main_ver)
      {
        case 0x17:
        case 0x18:
            xb_send_payload_to_host_S1();
            break;
        case 0x19:
        case 0x1A:
        default:
            xb_send_payload_to_host_S2();
            break;
      }
  }

// create a api message to transport an at-command (query only)
// the AT-Command (two bytes) are packed to an 16 bit integer, e.g. "AI" ist coded as 'A'+'I'*256
void xb_make_at_command(uint16_t at_command)
  {
    xbee_tx.at_cmd.command[0] = (at_command & 0xFF);
    xbee_tx.at_cmd.command[1] = (at_command & 0xFF00) >> 8;

    xbee_tx.at_cmd.frame_id = tx_frame_id;
    tx_frame_id++;                       // advance frame_id
    if (tx_frame_id == 0) tx_frame_id++;

    xbee_tx.header.delimiter = XBEE_PKT_START;
    //                             cmd_id frame command
    xbee_tx.header.len.as_uint16 = 1    + 1   + 2;
    xbee_tx.header.cmd_id = XBEE_PKT_TYPE_ATCMD;
    xbee_tx.raw.data[xbee_tx.header.len.low8 - 1] = xbee_checksum(&xbee_tx);
  }


#if (1==0)
#warning test
wrp_pkt_t test_message;
void wrp_send(wrp_pkt_t *load);

void xbee_test(void)
  {
    test_message.header = 4;
    test_message.type = WRP_TYPE_SYSTEM;
    test_message.opcode = WRP_OPCODE_APPLY;
    test_message.addr = 1;                     // Hardware ENUM
    test_message.data[0] = throttle_id;

    wrp_send(&test_message);
    wrp_transmit();

    while(1);

  }
#endif
//-------------------------------------------------------------------------------------
//
//  5. INIT
//
//-------------------------------------------------------------------------------------

static void restart_xbee_comm(void)
  {
    xbee_active_timestamp = get_systick_int();    // reset timing control
  }

void restart_xbee(void)
  {
    set_xbee_state(XBEE_DISCONNECTED);
    xbee_ai_state = XBEE_AI_NET_SCAN;             // vorsichtshalber alles zurck
    restart_xbee_comm();
                                                  // Association Indication absetzen
    xb_make_at_command(XBEE_CMD_ASSOC_STATUS);    // "AI"
    send_xbee_message();
    xbee_at_cmd_in_work = 1;

    _delay_ms(10);                                // not fair and not secure
                                                  // we wan't to know joined state

    wrp_send_sys_apply();                         // Versuch der Anmeldung
    wrp_transmit();
  }


//-------------------------------------------------------------------------------------
//
//  5.a) INIT-Task and Connection
//
//-------------------------------------------------------------------------------------

static int my_systick;
enum {
       IDLE,
       WAIT_FOR_MODULE,
       WAIT_FOR_ID,
       WAIT_FOR_SH,
       WAIT_FOR_SL,
       WAIT_FOR_HV,
       WAIT_FOR_VR,              // up to here cold boot
       WAIT_FOR_AI,
       WAIT_FOR_CH,
       WAIT_FOR_MY,
       INIT_DONE,
     } init_xbee_state;

//      my_systick = get_systick_int();  // atomic!

t_cr_task init_xbee(void)
  {
    switch (init_xbee_state)
      {
        default:
        case IDLE:
            my_systick = get_systick_int();
            xbee_last_tx = get_systick_10ms();
            restart_xbee_comm();

            base_station_valid = eeprom_read_byte((void *)(eadr_xbee_base));

            if (base_station_valid)
              {
                unsigned char j;
                for(j=0; j<8; j++)
                  {
                    base_station_mac[j] = eeprom_read_byte((void *)(eadr_xbee_base_mac0+j));
                  }
              }

            init_xbee_state = WAIT_FOR_MODULE;
            cr_fifo_put(&key_fifo, KEY_XBEE_AI);
            return(100000L / SYSTICK_PERIOD);           // wait 100ms
            break;
        case WAIT_FOR_MODULE:
//          if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
//              {
//                // timeout, module does not power up as endpoint
//                xbee_error = 1;                     // no endpoint
//                set_xbee_state(XBEE_DISCONNECTED);
//              }
//            if (xbee_connect == XBEE_ENDPOINT)
//              {
                xb_make_at_command(XBEE_CMD_PAN_ID);            // "ID", this reads pan_id
                send_xbee_message();
                init_xbee_state = WAIT_FOR_ID;
                xbee_at_cmd_in_work = 1;
//              }
            break;
        case WAIT_FOR_ID:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                // timeout, module does not answer
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_SER_HI);            // "SH", read serial number high bits
                send_xbee_message();
                init_xbee_state = WAIT_FOR_SH;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_SH:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                // timeout, module does not answer
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_SER_LO);            // "SL", read serial number low bits
                send_xbee_message();
                init_xbee_state = WAIT_FOR_SL;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_SL:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                // timeout, module does not answer
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_ASSOC_STATUS);          // "AI". read associated indication
                send_xbee_message();
                init_xbee_state = WAIT_FOR_AI;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_AI:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                // AI was answer - how is the state?

                if (xbee_ai_state == XBEE_AI_OKAY)
                  {
                    // joined
                    xb_make_at_command(XBEE_CMD_CHANNEL);           // "CH", read channel
                    send_xbee_message();
                    init_xbee_state = WAIT_FOR_CH;
                    xbee_at_cmd_in_work = 1;
                  }
                else
                  {
                    // not associated - query again
                    xb_make_at_command(XBEE_CMD_ASSOC_STATUS);      // "AI". read associated indication
                    send_xbee_message();
                    init_xbee_state = WAIT_FOR_AI;
                    xbee_at_cmd_in_work = 1;
                    return(100000L / SYSTICK_PERIOD);               // wait 100ms
                  }
              }
            break;
        case WAIT_FOR_CH:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_SRC_ADDR16);            // "MY", read short net addr
                send_xbee_message();
                init_xbee_state = WAIT_FOR_MY;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_MY:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_VERS_HARDWARE);     // "HV", read version
                send_xbee_message();
                init_xbee_state = WAIT_FOR_HV;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_HV:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                // timeout, module does not answer
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                xb_make_at_command(XBEE_CMD_VERS_FIRMWARE);     // "VR", read version
                send_xbee_message();
                init_xbee_state = WAIT_FOR_VR;
                xbee_at_cmd_in_work = 1;
              }
            break;
        case WAIT_FOR_VR:
            if ((get_systick_int() - my_systick) > 5000000L / SYSTICK_PERIOD)
              {
                xbee_error = 2;                     // no answer
                set_xbee_state(XBEE_DISCONNECTED);
              }
            if (!xbee_at_cmd_in_work)
              {
                init_xbee_state = INIT_DONE;
              }
            break;
        case INIT_DONE:
            // set_xbee_state(XBEE_ENDPOINT);
            set_xbee_state(XBEE_VOID);
            set_xbee_state(XBEE_DISCONNECTED);		// wait for granted

            restart_xbee_comm();
            wrp_send_sys_apply();                         // Versuch der Anmeldung
            wrp_transmit();	        //hk
			wrp_send_cs_status_request();
            return(-1);                             // stops this task
            break;

      }
    return(10);  // ready again
  }


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// TEST CODE

// #define DEBUG_TEST  -1
#if (DEBUG_TEST == -1)

void test_make_xbee_rx_payload(uint8_t size, uint8_t *payload)
  {
    xbee_tx.header.delimiter = 0x7e;
    xbee_tx.header.len.as_uint16 = (size + 1);
    xbee_tx.header.cmd_id = XBEE_PKT_TYPE_ZB_RX;
    memcpy(xbee_tx.raw.data, payload, size);
    xbee_tx.raw.data[size] = xbee_checksum(&xbee_tx);
  }


//                          frame        okay
unsigned char atcmd_resp_sh[] = {0x01,'S','H',0x00,0x00,0x13,0xA2,0x00 };
unsigned char atcmd_resp_sl[] = {0x01,'S','L',0x00,0x40,0x52,0x66,0x2B};
unsigned char atcmd_resp_id[] = {0x01,'I','D',0x00,0x02,0x34 };

//        90=normal rx packet  source 64 bit                            source 16 Bit  opt header type 3, opcode 2
unsigned char test1[] = {0x90, 0x12,0x34,0x56,0x78,0x01,0xFF,0x01,0xFF, 0xAF,0xFE,     0x55,0x01, 0x23};

void test_make_xbee_atcmd_response(uint8_t size, uint8_t *data)
  {
    xbee_tx.header.delimiter = 0x7e;
    xbee_tx.header.len.as_uint16 = (size + 1);
    xbee_tx.header.cmd_id = XBEE_PKT_TYPE_ATCMD_RESP;
    memcpy(xbee_tx.raw.data, data, size);
    xbee_tx.raw.data[size] = xbee_checksum(&xbee_tx);
  }



void xb_put_tx_to_rx(void)
  {
    unsigned char i, total;
    uint8_t* txptr;
    xbee_pkt_t* xptr;

    xptr = &xbee_tx;
    txptr = (uint8_t*)xptr;
    //                           total len
    total = xptr->header.len.low8 + 1;       // add 1 for checksum

    push_to_rx(xptr->header.delimiter);
    push_to_rx(xptr->header.len.high8);      // transmit big endian (htons())
    push_to_rx(xptr->header.len.low8);
    txptr+=3;

    for (i=0; i<total; i++)
      {
        push_to_rx(*txptr++);
      }
  }

//                                del   length                             esc               check
unsigned char atcmd_resp_id_full[] = {0x7e, 0, 7, 0x88, 0x01,'I','D', 0x00, 0x7D, 0x33, 0x10, 0xc6 };

void test_xbee_id_rec(void)
  {
    unsigned char i;
    uint8_t* txptr;

    txptr = atcmd_resp_id_full;


    //                           total len
    for (i=0; i<sizeof(atcmd_resp_id_full); i++)
      {
        push_to_rx(*txptr++);
      }
  }







// CV_SET                   // addr 64                              , PAN     ,mode,hdr, oc+type, addr     dat
// unsigned char testload[] = {0x12, 0x34,0x56,0x78,0x01,0xFF,0x01,0xFF,0xAF,0xFE,0x55,0x04,0x3F, 0x34, 0x12, 0x55};

unsigned char txload[] = {0x04, 0x01,0x02,0x03,0x04,0x05};


void test_xbee(void)
  {
    unsigned int i;

    // make_message_to_host((wrp_pkt_t *)txload);

    // test_make_xbee_rx_payload(sizeof(testload),testload);

    // test_make_xbee_atcmd_response(sizeof(atcmd_resp_id),atcmd_resp_id);
    // xb_put_tx_to_rx();

    // test_make_xbee_atcmd_response(sizeof(atcmd_resp_sh),atcmd_resp_sh);
    // xb_put_tx_to_rx();

    // test_make_xbee_atcmd_response(sizeof(atcmd_resp_sl),atcmd_resp_sl);
    // xb_put_tx_to_rx();

    // read this stuff


    test_xbee_id_rec();

    for (i=0; i<255; i++)
      {
          run_xbee_client();
      }
  }


#endif

/////////////////////////////////////////////////


#if (OLD_CODE == -1)   // never


//========================================================================
//
// downstream Interface
//
//========================================================================


void XbeeTxByte(char byte)
  {
    if (tx_fifo_write(byte))
      {
        // throw exeption
        #warning code missing (temp)
      }
  }


unsigned char XbeeRxByte(void)
  {
    while (!rx_fifo_ready()) {}

    return(rx_fifo_read());

    // throw exeption if timeout
    //#warning code missing

  }






//========================================================================

void XbeeTxString(char *string)
  {
    while(*string)
       XbeeTxByte(*string++);
  }


uint8_t XbeeWaitOK(void)
  {
    if( XbeeRxByte() == 'O' )
        if( XbeeRxByte() == 'K' )
            if( XbeeRxByte() == '\r' )
              {
                _delay_ms(10);    // need mysterious delay to function properly -> h??
                //!warning
                return 0;
              }
            _delay_ms(10);
            // USART_Flush();
            return 1;
  }


void XbeeEnterCmdMode(void)
  {
    // assume: we ar in text mode, so we have to send a good old AT escape

    // No characters sent for one second [GT (Guard Times) parameter = 0x3E8]

    _delay_ms(1100);
    // Input three plus characters ("+++") within one second [CC (Command Sequence Character) parameter = 0x2B.]

    XbeeTxString("+++");
    XbeeWaitOK();

  }



//========================================================================
//
// upstream interface
//
//========================================================================

// get RSSI out of received packet

int getRSSI(xbee_pkt_t *packet)
  {
    xbee_pkt_t *pkt = (xbee_pkt_t *)packet;


    if (pkt->header.cmd_id == XBEE_PKT_TYPE_RX16) {
        return ((xbee_a16_rx_pkt_t *)pkt)->rssi;
    } else if (pkt->header.cmd_id == XBEE_PKT_TYPE_RX64) {
        return ((xbee_a64_rx_pkt_t *)pkt)->rssi;
    } else if (pkt->header.cmd_id == XBEE_PKT_TYPE_RX16_IO) {
        return ((xbee_io_a16_rx_pkt_t *)pkt)->rssi;
    } else if (pkt->header.cmd_id == XBEE_PKT_TYPE_RX64_IO) {
        return ((xbee_io_a64_rx_pkt_t *)pkt)->rssi;
    }

    return -1;  // not contained
  }


// old!!!!!!
void XbeeInit_old(void)
  {

    // set hardware to address XBEE

    // init_rs232(t_baud new_baud);

    xbee_active_timestamp = get_systick_int();    // reset timing control

    XbeeEnterCmdMode();
    XbeeTxString("ATRE\r");   // restore to factory defaults
    XbeeWaitOK();
    XbeeTxString("ATPL 4\r"); // power level to 0
    XbeeWaitOK();
    XbeeTxString("ATD7 1\r"); // set up hardware CTS flow control
    XbeeWaitOK();
    XbeeTxString("ATAC\r");   // apply changes
    XbeeWaitOK();
    XbeeTxString("ATWR\r");   // save changes
    XbeeWaitOK();
    XbeeTxString("ATCN\r");   // exit command mode
    XbeeWaitOK();
  }

#endif   // old code
