//========================================================================
//
// OpenDCC - MFT
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      xbee_protocol.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-07-10 V0.01 kw start
//                                This file was initially taken from the
//                                arduino project
//            2010-04-23 V0.02 kw added common structure
//            2010-05-12 V0.03 kw changed src to src64
//            2010-06-07 V0.04 kw added zb_tx_status
//            2010-08-04 V0.05 mp changed at command definitions to
//                                16 bit integer
//            2010-08-05       kw fixed XOFF
//
//------------------------------------------------------------------------
//
// purpose:   wireless throttle for dcc - client for xbee modules
//
// content:   API protocol definitions for XBEE
//
// to do:     
//
//========================================================================

#ifndef XBEE_PROTOCOL_H
#define XBEE_PROTOCOL_H

#ifdef AVR
# define __LITTLE_ENDIAN 1234
# define BYTE_ORDER LITTLE_ENDIAN
#elif defined(ARM)
# include <endian.h>
#else
# include <arpa/inet.h>
#endif

#include <stdint.h>

// #include "xbee.h"

#if !defined(ntohs) && (BYTE_ORDER == LITTLE_ENDIAN)
# define ntohs(n) ((((short)(n)) & 0xff00) >> 8 | (((short)(n)) & 0xff) << 8)
# define htons(n) ntohs(n)
#elif !defined(ntohs)
# define ntohs(n) ((short)(n))
# define htons(n) ntohs(n)
#endif

#if !defined(ntohl) && (BYTE_ORDER == LITTLE_ENDIAN)
# define ntohl(x) ((((x)&0xff000000)>>24) \
                  |(((x)&0x00ff0000)>>8)  \
                  |(((x)&0x0000ff00)<<8)  \
                  |(((x)&0x000000ff)<<24))
# define htonl(n) ntohl(n)
#elif !defined(ntohl)
# define ntohl(n) ((long)(n))
# define htonl(n) ntohs(n)
#endif

/* p2p  CE=0 (end devices) A1=0 (no end dev assoc) same ID/CH
 * coordinator: CE=1, A2=n (coordinator assoc)
 *   SP= sleep perd ST= time betw sleep (should be same on
 *   coord/noncoord)
 * assoc - coord'd only; comm between modules thru coord'r
 *  PAN's - need coordinator.  A1 allows totally dynamic assoc
 */
 
// --- General XBee Definitions --- 
 
// Frame Delimiter "Start of packet" byte; always sent as the first
//  byte of each packet
 
#define XBEE_PKT_START 	    (0x7e)      // delimiter
#define XBEE_PKT_ESC        (0x7d)      // excape sequence
#define XBEE_PKT_XON        (0x11)
#define XBEE_PKT_XOFF       (0x13)

// Maximum packet size; datasheet basically says 100 payload bytes max
#define XBEE_MAX_DATA_LEN        100

//----------------------------------------------------------------------------
//             Definitions for commands the XBee recognizes
//----------------------------------------------------------------------------
#define MK_ATCMD(a,b)					(a+(b<<8))
/* Basic communication parameters/values */
#define XBEE_CMD_CHANNEL                MK_ATCMD('C','H')
#define XBEE_CMD_PAN_ID                 MK_ATCMD('I','D')
#define XBEE_CMD_DEST_ADDR64_HI         MK_ATCMD('D','H')
#define XBEE_CMD_DEST_ADDR64_LO         MK_ATCMD('D','L')
#define XBEE_CMD_SRC_ADDR16             MK_ATCMD('M','Y')
#define XBEE_CMD_SER_HI                 MK_ATCMD('S','H')
#define XBEE_CMD_SER_LO                 MK_ATCMD('S','L')
#define XBEE_CMD_RAND_DLY_SLOTS         MK_ATCMD('R','N')
#define XBEE_CMD_MAC_MODE               MK_ATCMD('M','M')
#define XBEE_CMD_COORD_ENA              MK_ATCMD('C','E')
#define XBEE_CMD_SCAN                   MK_ATCMD('S','C')
#define XBEE_CMD_SCAN_DURATION          MK_ATCMD('S','D')
#define XBEE_CMD_ASSOC_END              MK_ATCMD('A','1')
#define XBEE_CMD_ASSOC_COORD            MK_ATCMD('A','2')
#define XBEE_CMD_ASSOC_STATUS           MK_ATCMD('A','I')
#define XBEE_CMD_RSSI                   MK_ATCMD('D','B')

/* Transceiver Control */
#define XBEE_CMD_PWR_LEVEL              MK_ATCMD('P','L')
#define XBEE_CMD_CCA_THRESH             MK_ATCMD('C','A')

/* Sleep Parameters */
#define XBEE_CMD_SLEEP_MODE             MK_ATCMD('S','M')
#define XBEE_CMD_SLEEP_TIMEOUT          MK_ATCMD('S','T')
#define XBEE_CMD_SLEEP_PERIOD           MK_ATCMD('S','P')
#define XBEE_CMD_SLEEP_PERIOD_DISASSOC  MK_ATCMD('D','P')

/* Interface parameters */
#define XBEE_CMD_DATA_RATE              MK_ATCMD('B','D')
#define XBEE_CMD_PACKETIZATION_TIMEOUT  MK_ATCMD('R','O')
#define XBEE_CMD_DIO7_CONFIG            MK_ATCMD('D','7')
#define XBEE_CMD_DIO6_CONFIG            MK_ATCMD('D','6')
#define XBEE_CMD_DIO5_CONFIG            MK_ATCMD('D','5')
#define XBEE_CMD_DIO4_CONFIG            MK_ATCMD('D','4')
#define XBEE_CMD_DIO3_CONFIG            MK_ATCMD('D','3')
#define XBEE_CMD_DIO2_CONFIG            MK_ATCMD('D','2')
#define XBEE_CMD_DIO1_CONFIG            MK_ATCMD('D','1')
#define XBEE_CMD_DIO0_CONFIG            MK_ATCMD('D','0')
#define XBEE_CMD_PWM0_CONFIG            MK_ATCMD('P','O')
#define XBEE_CMD_API_ENA                MK_ATCMD('A','P')
#define XBEE_CMD_PULLUP_ENA             MK_ATCMD('P','R')

/* Version Info */
#define XBEE_CMD_VERS_FIRMWARE          MK_ATCMD('V','R')
#define XBEE_CMD_VERS_HARDWARE          MK_ATCMD('H','V')
#define XBEE_CMD_VERS_FIRM_VERBOSE      MK_ATCMD('V','L')

/* Received Signal Strength */
#define XBEE_CMD_RSSI_PWM_TIMER         MK_ATCMD('R','P')
#define XBEE_CMD_RSS                    MK_ATCMD('D','B')

/* Error counters */
#define XBEE_CMD_CCA_FAILS              MK_ATCMD('E','C')
#define XBEE_CMD_ACK_FAILS              MK_ATCMD('E','A')

/* AT Command Params */
#define XBEE_CMD_AT_MODE_TIMEOUT        MK_ATCMD('C','T')
#define XBEE_CMD_AT_GUARD_TIME          MK_ATCMD('G','T')
#define XBEE_CMD_AT_CMD_CHAR            MK_ATCMD('C','C')
#define XBEE_CMD_AT_EXIT                MK_ATCMD('C','N')

/* XBEE specific routing */
#define XBEE_CMD_NODE_FIND_DEST         MK_ATCMD('D','N')
#define XBEE_CMD_NODE_DISCOVER          MK_ATCMD('N','D')
#define XBEE_CMD_NODE_ID                MK_ATCMD('N','I')
#define XBEE_CMD_ACTIVE_SCAN            MK_ATCMD('A','S')
#define XBEE_CMD_FORCE_DISASSOC         MK_ATCMD('D','A')
#define XBEE_CMD_ENERGY_SCAN            MK_ATCMD('E','D')
#define XBEE_CMD_FORCE_POLL             MK_ATCMD('F','P')

/* IO Line Passing / Sensor Interfacing */
#define XBEE_CMD_SAMPLE_RATE            MK_ATCMD('I','R')
#define XBEE_CMD_SAMPLES_BEFORE_TX      MK_ATCMD('I','T')

/* Misc */
#define XBEE_CMD_WRITE_PARAMS           MK_ATCMD('W','R')
#define XBEE_CMD_RESET_SOFT             MK_ATCMD('F','R')
#define XBEE_CMD_APPLY_CHANGES          MK_ATCMD('A','C')
#define XBEE_CMD_RESTORE_DEFAULTS       MK_ATCMD('R','E')


#define XBEE_AI_OKAY                   0x00 // okay
#define XBEE_AI_NO_PANs                0x21 // Scan found no PANs
#define XBEE_AI_NO_VALID_PANs          0x22 //  Scan found no valid PANs based on current SC and ID settings
#define XBEE_AI_NO_JOIN                0x23 //  Valid Coordinator or Routers found, but they are not allowing joining (NJ expired)n
#define XBEE_AI_ROUTER_FULL_R          0x24 //  Router Full - Allow Join set, but cannot allow any more Routers to join 
#define XBEE_AI_ROUTER_FULL_E          0x25 //  Router Full - Allow Join set, but cannot allow any more End Devices to join 
#define XBEE_AI_JOIN_FAILED            0x27 //  Node Joining attempt failed (typically due to incompatible security settings)
#define XBEE_AI_COORD_FAILED           0x2A //  Coordinator Start attempt failed
#define XBEE_AI_COORD_SCAN             0x2B //  Checking for an existing coordinator
#define XBEE_AI_NO_RESPONSE            0xAB //  Attempted to join a device that did not respond.
#define XBEE_AI_UNSECURE               0xAC //  Secure join error - network security key received unsecured
#define XBEE_AI_NO_KEY                 0xAD //  Secure join error - network security key not received
#define XBEE_AI_NO_LINK_KEY            0xAF //  Secure join error - joining device does not have the right preconfigured link key
#define XBEE_AI_NET_SCAN               0xFF //  Scanning for a ZigBee network (routers and end devices)


//----------------------------------------------------------------------------
//     Structures useful for communicating with the XBee in API mode
//----------------------------------------------------------------------------

typedef enum
  {
    XBEE_PKT_TYPE_TX64              = 0x00,
    XBEE_PKT_TYPE_TX16              = 0x01,
    XBEE_PKT_TYPE_ATCMD             = 0x08,
    XBEE_PKT_TYPE_QATCMD            = 0x09, /* wait til an immed param or apply cmd */
    XBEE_PKT_TYPE_ZB_TX_REQ         = 0x10,
    XBEE_PKT_TYPE_ZB_CMD_FRAME      = 0x11, /* Not yet impl */
    XBEE_PKT_TYPE_REMOTE_ATCMD      = 0x17,
    XBEE_PKT_TYPE_RX64              = 0x80,
    XBEE_PKT_TYPE_RX16              = 0x81,
    XBEE_PKT_TYPE_RX64_IO           = 0x82,
    XBEE_PKT_TYPE_RX16_IO           = 0x83,
    XBEE_PKT_TYPE_ATCMD_RESP        = 0x88,
    XBEE_PKT_TYPE_TX_STATUS         = 0x89,
    XBEE_PKT_TYPE_MODEM_STATUS      = 0x8a,
    XBEE_PKT_TYPE_ZB_TX_STATUS      = 0x8b,
    XBEE_PKT_TYPE_ADV_MODEM_STATUS  = 0x8c, /* Not yet impl */
    XBEE_PKT_TYPE_ZB_RX             = 0x90,
    XBEE_PKT_TYPE_ZB_X_RX           = 0x91,
    XBEE_PKT_TYPE_ZB_RX_IO          = 0x92, /* Not yet impl */
    XBEE_PKT_TYPE_RX_SENSOR         = 0x94, /* Not yet impl */
    XBEE_PKT_TYPE_NODE_IDENT        = 0x95, /* Not yet impl */
    XBEE_PKT_TYPE_REMOTE_ATCMD_RESP = 0x97,
  } xbee_pkt_cmdid_t;


// in general, packets are wrapped with a start & length 
typedef struct
  {
    uint8_t             delimiter;      // 0x7e
    t_data16            len;            // this allows to resolve endianess
    xbee_pkt_cmdid_t    cmd_id;
  } __attribute__ ((__packed__)) xbee_pkt_hdr_t;



// data is a structure consisting of cmdID and cmdData
// cmdData is dependant on cmdID


























/* --- Bits in packets --- */

/* Communication status bits (0x8A) */

#define XBEE_STATUS_HW_RESET         0x00
#define XBEE_STATUS_WD_RESET         0x01
#define XBEE_STATUS_JOINED           0x02
#define XBEE_STATUS_DISASSOC         0x03
#define XBEE_STATUS_COORD_STARTED    0x06
#define XBEE_STATUS_NETWORK_KEY      0x07
#define XBEE_STATUS_VCC_LIMIT        0x0D
#define XBEE_STATUS_KEY_ESTABLISHED  0x10
#define XBEE_STATUS_STACK_ERROR      0x80

/* Command status bits */

#define XBEE_CMDSTATUS_OK  0
#define XBEE_CMDSTATUS_ERR 1

/* Transmit options */

#define XBEE_TX_FLAG_NO_ACK 0x01
#define XBEE_TX_FLAG_SEND_BCAST_PAN_ID 0x04

/* Transmit status bits */

#define XBEE_TXSTATUS_SUCCESS  0x00
#define XBEE_TXSTATUS_NO_ACK   0x01
#define XBEE_TXSTATUS_CCA_FAIL 0x02
#define XBEE_TXSTATUS_PURGES   0x03

/* Received options */

#define XBEE_RX_FLAG_ADDR_BCAST 0x02
#define XBEE_RX_FLAG_PAN_BCAST  0x04


/* --- Definitions & macros for library use --- */

/* For tracking memory allocations  */
#define XBEE_RECV                0x00
#define XBEE_XMIT                0x01


//-------------------------------------------------------------------------------
// --- API Packet layouts (all packets are wrapped with a header)

typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t             frame_id;
    uint8_t             command[2];
    uint8_t             param[0];
} __attribute__ ((__packed__)) xbee_at_cmd_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t dest64[8];
    uint8_t dest16[2];
    uint8_t apply;
    uint8_t command[2];
    uint8_t param[0];
} __attribute__ ((__packed__)) xbee_remote_at_cmd_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t dest64[8];
    uint8_t opt;
    uint8_t data[0];
} __attribute__ ((__packed__)) xbee_a64_tx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t dest16[2];
    uint8_t opt;
    uint8_t data[0];
} __attribute__ ((__packed__)) xbee_a16_tx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t status;
} __attribute__ ((__packed__)) xbee_tx_status_pkt_t;


typedef struct {                                        // XBEE_PKT_TYPE_ZB_TX_STATUS      = 0x8b,
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t dest16[2];
    uint8_t retries;
    uint8_t status;
    uint8_t discovery_status;
} __attribute__ ((__packed__)) xbee_zb_tx_status_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t status;
} __attribute__ ((__packed__)) xbee_modem_status_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t command[2];
    uint8_t status;
    uint8_t param[0];
} __attribute__ ((__packed__)) xbee_cmd_resp_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t src64[8];
    uint8_t src16[2];
    uint8_t command[2];
    uint8_t status;
    uint8_t param[0];
} __attribute__ ((__packed__)) xbee_remote_cmd_resp_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t src64[8];
    uint8_t rssi; /* signal strength */
    uint8_t opt;
    uint8_t data[0];
} __attribute__ ((__packed__)) xbee_a64_rx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t src16[2];
    uint8_t rssi;
    uint8_t opt;
    uint8_t data[0];
} __attribute__ ((__packed__)) xbee_a16_rx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t src64[8];
    uint8_t rssi; /* signal strength */
    uint8_t opt;
    uint8_t num_samples;
    uint16_t ch_ind; /* bits 14-9: a5-a0 bits 8-0: d8-d0 active */
    uint16_t data[0]; /* First sample digital if any digital chan active
                          rest are 16-bit analog rdgs                  */
} __attribute__ ((__packed__)) xbee_io_a64_rx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t src16[2];
    uint8_t rssi;
    uint8_t opt;
    uint8_t num_samples;
    uint16_t ch_ind; /* bits 14-9: a5-a0 bits 8-0: d8-d0 active */
    uint16_t data[0]; /* First sample digital if any digital chan active
                          rest are 16-bit analog rdgs                  */
} __attribute__ ((__packed__)) xbee_io_a16_rx_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t frame_id;
    uint8_t dest64[8];
    uint8_t dest16[2];
    uint8_t radius;
    uint8_t opt;      /* Multicast = bit 3 */
    uint8_t data[0];  /* Up to 72 bytes/pkt */
} __attribute__ ((__packed__)) xbee_zb_tx_req_pkt_t;


typedef struct {
    // xbee_pkt_hdr_t      header;
    uint8_t src64[8];
    uint8_t src16[2];
    uint8_t opt;
    uint8_t data[0];  /* Up to 72 bytes/pkt */
} __attribute__ ((__packed__)) xbee_zb_rx_pkt_t;

typedef struct
  {
    // xbee_pkt_hdr_t      header;
    uint8_t             data[XBEE_MAX_DATA_LEN];
  } xbee_raw_pkt_t;

// common structure for all packets

typedef struct {
    xbee_pkt_hdr_t  header;
    union {
        xbee_raw_pkt_t              raw;
        xbee_at_cmd_pkt_t           at_cmd;
        xbee_remote_at_cmd_pkt_t    remote_at_cmd;
        xbee_a64_tx_pkt_t           a64_tx;
        xbee_a16_tx_pkt_t           a16_tx;
        xbee_tx_status_pkt_t        tx_status;
        xbee_zb_tx_status_pkt_t     zb_tx_status;
        xbee_modem_status_pkt_t     modem_status;
        xbee_cmd_resp_pkt_t         cmd_resp;
        xbee_remote_cmd_resp_pkt_t  remote_cmd_resp;
        xbee_a64_rx_pkt_t           a64_rx;
        xbee_a16_rx_pkt_t           a16_rx;
        xbee_io_a64_rx_pkt_t        io_a64_rx;
        xbee_io_a16_rx_pkt_t        io_a16_rx;
        xbee_zb_tx_req_pkt_t        zb_tx_req;
        xbee_zb_rx_pkt_t            zb_rx;
    };
} xbee_pkt_t;


#endif /* #ifndef XBEE_PROTOCOL_H ... */


