//------------------------------------------------------------------------
//
// OpenDCC - Xpressnet Interface
//           (Railcom Bridge, Throttle)
//
// Copyright (c) 2006, 2007, 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//------------------------------------------------------------------------
//
// file:      xp_client_parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-09-15 V0.01 kw  start
//            2009-02-12 V0.02 kw  ported to throttle
//            2009-02-28 V0.03 kw  surveillance of xpressnet (timeout 500ms)
//            2009-03-01 V0.04 kw  extended loco functions
//            2009-06-30 V0.05 kw  added fast time command
//            2013-05-13 V0.06 kw  no infinite sleep - come here from time to time
//                                 to check broken connection
//
//------------------------------------------------------------------------
//
// purpose:   brigde message to xpressnet
//            
//            here: xpressnet client
//            we do:
//               automatically select our slot_id -> we slip away ...
//
// ====== >>>> config.h  is the central definition file for the project
//             xp_client_hardware.h is our hardware
//
//------------------------------------------------------------------------
//


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"                 // general definitions
#include "hardware.h"
#include "timer.h"                  // access to systick for surveillance
#include "cortos.h"                 // fifo access
#include "keycodes.h"               
#include "keyboard.h"
#include "display.h"               
#include "xp_client_hardware.h"     // port definitions for target
#include "xp_client_if.h"           // this is the low level backend

#include "messages_objects.h"

#if (DEBUG != 0)                    // generell debug allowed
    #define DEBUG_RX   0            // 0: no debug messages
                                    // 1: dump down messages to our slot
                                    // 2: dump first received byte
#endif


//------------------------------------------------------------------------
//
//  Connect Status
//
//------------------------------------------------------------------------
//

void set_xp_state(xp_connect_t neu)
  {
    if (neu != xp_connect)
      {
        xp_connect = neu;
        switch (xp_connect)
          {
            default:
            case XP_DISCONNECTED:                    
                if (used_interface == USE_XPRESSNET)
                  {
                    cr_fifo_put(&key_fifo, KEY_DISCONNECT);
                    set_command_station_status(STS_DISCONNECTED);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    //!warning! hat das noch weitergehende Folgen? Kann/Soll ein anderer bernehmen
                  }
                break;
            case XP_CONNECTED:
                if (used_interface == USE_XPRESSNET)
                  {
                    cr_fifo_put(&key_fifo, KEY_CONNECT);
                  }
                break;
          }
      }
  }

//------------------------------------------------------------------------
//
//  Slothandling
//
//------------------------------------------------------------------------
//

t_slot_mode slot_mode;

unsigned char slots_in_use[32];

void mark_slot_as_used(unsigned char slot)
  {
    unsigned char i;

    slots_in_use[slot] = 255;
    if (slot == slot_id)
      {
        // oops, there is somebody on our slot
        if (slot_mode == SLOT_AUTOMATIC)
          {
            for (i=1; i < 32; i++)
              {
                if (slots_in_use[i] == 0)
                  {
                    //!warning! slot change missing
                    // slot_id = i;  // take this slot
                    return;
                  }
              }
            // no slot free - error!
            display_error(ERR_NO_SLOT);

          }
        else
          {
            // slot manual selected and occupied - error!
            display_error(ERR_SLOT_OCCUPIED);
          }
      }
  }


unsigned char rx_message[17];
unsigned char rx_index;
unsigned char rx_size;

//------------------------------------------------------------------------
//
//  Parser for Feedback messages
//
//------------------------------------------------------------------------

void xp_parser_feedback(void)
  {
    unsigned char temp0 = fb_array[fb_addr[0]];
    unsigned char temp1 = fb_array[fb_addr[1]];
    
    unsigned int myfb_addr;
    unsigned char msg_size = rx_message[0] & 0xf;

    if ( (rx_message[0] & 0xf0) == 0x40 && msg_size >=2)
      {
        unsigned char j;
        for (j=1; j<msg_size; j+=2)
          {
            myfb_addr = rx_message[j];
            if (myfb_addr < FB_ARRAY_SIZE)
              {
                if (rx_message[j+1] & 0x10)
                  {
                    fb_array[myfb_addr] = (fb_array[myfb_addr] & 0x0F) | ((rx_message[j+1] & 0x0F) << 4);
                  }
                else
                  {
                    fb_array[myfb_addr] = (fb_array[myfb_addr] & 0xF0) | (rx_message[j+1] & 0x0F);
                  }
              }
          }
      }
    if ( (temp0 != fb_array[fb_addr[0]]) || (temp1 != fb_array[fb_addr[1]]) )
      {
        cr_fifo_put(&key_fifo, KEY_PARSER_FB);   //notify gui if change detected
      }
  }


//------------------------------------------------------------------------
//
//  Parser (all messages)
//
//  Tracks down xpressnet messages and does:
//  1. update objects
//  2. send a notify to the GUI (navigator); this is done with
//     virtual keycodes (see keycodes.h), we use:
//        KEY_PARSER          0xE0   // general Event (obsolete)
//        KEY_PARSER_STATE    0xE1   // State change
//        KEY_PARSER_FB       0xE2   // Feedback event
//        KEY_PARSER_SWITCH   0xE3   // Accessory event
//        KEY_PARSER_LOK      0xE4   // Lok event
//        KEY_PARSER_STOLEN   0xE5   // Lok owner event
//        KEY_PARSER_PROG     0xE6   // Prog event
//
//------------------------------------------------------------------------

uint8_t xp_parser(void)
  {
    uint8_t Actioned = 0;
    uint8_t i,j;
    uint16_t addr;

    #if (DEBUG_RX == 1)
        #warning debug printout of RX messages enabled
        set_cursor(0,0);
        lcd_hex2(rx_message[0]);
        lcd_string(".");
        lcd_hex2(rx_message[1]);
        lcd_string(".");
        lcd_hex2(rx_message[2]);
        lcd_string(".");
        lcd_hex2(rx_message[3]);
        lcd_string(":");
    #endif

    switch (rx_message[0] >> 4)
      {
        case 0x0:
            switch(rx_message[1])
              {
                case 0xF1:
                    // set clock
                    for(i = 2; i <= (rx_message[0] & 0x0F); i++ )
                      {
                        j = 99;     // use speed as temp
                        switch(rx_message[i] & 0xC0)
                          {
                            case 0x00:  j = rx_message[i] & 0x3F;
                                        if (j < 60) fast_clock.minute = j;
                                        break;
                            case 0x80:  j = rx_message[i] & 0x3F;
                                        if (j < 24) fast_clock.hour = j;
                                        break;
                            case 0x40:  j = rx_message[i] & 0x3F;
                                        if (j < 7) fast_clock.day_of_week = j;
                                        break;
                            case 0xC0:  j = rx_message[i] & 0x3F;
                                        if (j < 32) fast_clock.ratio = j;
                                        break;
                          }
                      }                            
                    cr_fifo_put(&key_fifo, KEY_PARSER_TIME);
                    break;
              }
            break;
        case 0x2:
            // 0    1    2     3     4   
            // 0x24 0x28 AddrH AddrL DATA [XOR]  SO from Command Station
            if (rx_message[1] == 0x28)
              {
                current_CV.number = rx_message[2] * 256 + rx_message[3];
                current_CV.value = rx_message[4];
                current_CV.result = PRG_CV_OKAY;
                cr_fifo_put(&key_fifo, KEY_PARSER_CFG);
              }
            break;
        case 0x3:
            // 0x3*: Parser extensions, not an offcial xpressnet protocol
            // 0x3* are tunnel messages - talking directly to a client
            if ((rx_message[2] == 0) && (rx_message[3] == 1))
              {
                // 0x33 0x?? 0x00 0x00 leave tunnel mode
                // 0x33 0x?? 0x00 0x01 enter tunnel mode
                cr_fifo_put(&key_fifo, KEY_PARSER_REMOTE);
              }        
            memcpy (pc_message, rx_message, sizeof(pc_message));
            cr_fifo_put(&key_fifo, KEY_PARSER_MESSAGE);
            break;
        case 0x4:
            // 0x42 Addr DATA [XOR] "Accessory Decoder information"
            // Data=ITTNZZZZ: 
            // TT=00: Schaltempf. ohne Feedback, TT=01: mit feedback, TT=10 feedback only, TT11void
            // determine Address
            addr = (rx_message[1] << 2) + (unsigned char)((rx_message[2] & 0x10) >> 3);
            addr++;  // we code from 1 to 2040
            // now we have state info of two turnouts
            // xpressnet has an errornous coding of the I-bit: this exits only
            // once, but there a two turnouts in this command :-(
            // so we ingore the whole message if I bit is set
            // up to now, TT bits are also ignored
            if (!(rx_message[2] & 0x80))
              {
                for (i=0; i<sizeof(actual_weiche)/sizeof(actual_weiche[0]); i++)
                  {
                    if (addr == actual_weiche[i].addr)
                      {
                        switch (rx_message[2] & 0b0011)
                          {
                            default:   // no break;
                            case 0b0011:   // broken
                            case 0b0000:   actual_weiche[i].state = WS_VOID;  break;
                            case 0b0001:   actual_weiche[i].state = WS_0;     break;
                            case 0b0010:   actual_weiche[i].state = WS_1;     break;
                          }
                        cr_fifo_put(&key_fifo, KEY_PARSER_SWITCH);
                      }
                    if ((addr+1) == actual_weiche[i].addr)
                      {
                        switch (rx_message[2] & 0b1100)
                          {
                            default:  // no break;
                            case 0b1100: 
                            case 0b0000:   actual_weiche[i].state = WS_VOID;  break;
                            case 0b0100:   actual_weiche[i].state = WS_0;     break;
                            case 0b1000:   actual_weiche[i].state = WS_1;     break;
                          }
                        cr_fifo_put(&key_fifo, KEY_PARSER_SWITCH);
                      }
                  }
              }
            // also look at feed
            xp_parser_feedback();
            break;
        case 0x6:
            switch (rx_message[1])
              {
                case 0x00:
                    // 0x61 0x00 0x61 "Everything off (emergency off)"
                    set_command_station_status(STS_OFF);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x01:
                    // 0x61 0x01 0x60 "Normal operation resumed"
                    set_command_station_status(STS_ON);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x02:
                    // 0x61 0x02 0x63 "Service Mode entry"
                    set_command_station_status(STS_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                    // 0x63 0x10 EE DATA [XOR] "Service Mode response for Register and Paged Mode"
                case 0x11:
                    // 0x61 0x11 [XOR] "Programming info: Command station ready"
                    current_CV.result = PRG_CV_OKAY;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    Actioned = 1;
                    break;
                case 0x12:
                    // 0x61 0x12 [XOR] "Programming info: short-circuit"
                    current_CV.result = PRG_CV_SHORT;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    Actioned = 1;
                    break;
                case 0x13:
                    // 0x61 0x13 [XOR] "Programming info: Data byte not found"
                    current_CV.result = PRG_CV_NOLOK;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    Actioned = 1;
                    break;
                case 0x14:
                    // 0x63 0x14 CV DATA [XOR] "Service Mode response for Direct CV mode"
                    if (rx_message[2] == 0) current_CV.number = 1024;
                    else current_CV.number = rx_message[2];
                    current_CV.value = rx_message[3];
                    current_CV.result = PRG_CV_OKAY;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x15:
                    // 0x63 0x15 CV DATA [XOR] "Service Mode response for Direct CV mode (256-511)"
                    current_CV.number = rx_message[2] + 256;
                    current_CV.value = rx_message[3];
                    current_CV.result = PRG_CV_OKAY;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x16:
                  // 0x63 0x16 CV DATA [XOR] "Service Mode response for Direct CV mode (512-767)"
                    current_CV.number = rx_message[2] + 512;
                    current_CV.value = rx_message[3];
                    current_CV.result = PRG_CV_OKAY;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x17:
                    // 0x63 0x17 CV DATA [XOR] "Service Mode response for Direct CV mode (768-1023)"
                    current_CV.number = rx_message[2] + 768;
                    current_CV.value = rx_message[3];
                    current_CV.result = PRG_CV_OKAY;
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x1F:
                    // 0x61 0x1f [XOR] "Programming info: Command station busy"
                    set_command_station_status(STS_PBUSY);
                    we_should_ask_again = TRUE;                     // tell messages_objects to repeat the query
                    cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x21:
                    // 0x62 0x21 DATA [XOR] "Software-version"
                    // 0x63 0x21 DATA_1 DATA_2 [XOR] "Software-version"
                    break;
                case 0x22:
                    // 0x62 0x22 DATA [XOR] "Command station status indication"
                    // Bits:
                    //   7      6       5       4       3       2       1       0
                    //   RAM    Restart rsv     rsv     PROG    Start   Notaus  Stop
                    //   Error                                  mode

                    if (rx_message[2] & 0x02) set_command_station_status(STS_OFF);
                    else if (rx_message[2] & 0x01) set_command_station_status(STS_STOP);
                    else if (rx_message[2] & 0x08) set_command_station_status(STS_PROG);
                    else set_command_station_status(STS_ON);   
                    
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);

                    break;
                case 0x81:
                    set_command_station_status(STS_BUSY);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                case 0x82:
                    // 0x61 0x82 [XOR] "Instruction not supported by command station"
                    set_command_station_status(STS_CMD_ERROR);
                    cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                    Actioned = 1;
                    break;
                    // 0x61 0x80+F [XOR] "Double Header error"
              }
            break;
        case 0x7:   // BIDI
            // 0x76 0xE1 AddrH AddrL CV_H CV_L DAT [XOR]
           switch(rx_message[1])
              {
                default:
                    break;
                case 0xE1:  
                    addr = (unsigned int) (rx_message[2] << 8) + rx_message[3];
                    if (addr & (1<<14) )
                      {
                        // accessory
                      }
                    else
                      { // lok
                        if ((addr & 0x3FFF) == current_CV.addr)
                          {
                            addr = (unsigned int) (rx_message[4] << 8) + rx_message[5] + 1;
                            if (current_CV.number == addr)
                              {
                                current_CV.value = rx_message[6];
                                current_CV.result = PRG_CV_OKAY;
                                cr_fifo_put(&key_fifo, KEY_PARSER_PROG);
                                Actioned = 1;
                              }
                          }
                      }
                    break;
              }
            break;
        case 0x8:
          switch (rx_message[1])
            {
              case 0x00:  // 0x81 0x00 0x81 "Everything stopped (emergency stop)"
                set_command_station_status(STS_STOP);
                cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
                break;

                // 0x83 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is available for operation"
                // 0x84 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is available for operation"
            }

            break;
        case 0xA:
            // this is a old command, only V1 or V2
            // 0xA3 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is being operated by another device"
            // 0xA4 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is being operated by another device"
            break;
        case 0xC:
            // 0xC5 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header available"
            // 0xC5 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header occupied"
            // 0xC6 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header available"
            // 0xC6 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header occupied"
            break;
        case 0xE:
            switch(rx_message[1] & 0xF0)
              {
                case 0x00:
                    // 0xE4 Identification Speed FKTA FKTB [XOR] "Locomotive information normal locomotive"
                    switch(rx_message[1] & 0x07)  // FFF=Fahrstufen: 000=14, 001=27, 010=28, 100=128
                      {
                        case 0: current_loc.format = DCC14; break;
                        case 1: current_loc.format = DCC27; break;
                        case 2: current_loc.format = DCC28; break;
                        case 4: current_loc.format = DCC128; break;
                      }
                    if (rx_message[1] & 0x08) current_loc.other = 1; 
                    else                      current_loc.other = 0;  
                    
                    if (rx_message[2] & 0x80) current_loc.dir = 0;                   // 0 = forward, 1=revers
                    else                      current_loc.dir = 1;  
                    
                    switch(current_loc.format)
                      {
                        case DCC14:
                            current_loc.speed = rx_message[2] & 0x0F;
                            if (current_loc.speed == 0) current_loc.speed = 1;
                            current_loc.speed--;
                            Actioned = 1;   
                            break;
                        case DCC27:
                        case DCC28:
                            if ((rx_message[2] & 0x0F) <= 1)             // map 0x?0 to 0 and 0x?1 to 1
                                current_loc.speed = 0;                              // stop or nothalt
                            else 
                              {
                                current_loc.speed = ((rx_message[2] & 0x0F) << 1) | ((rx_message[2] & 0x10) >> 4);
                                current_loc.speed = current_loc.speed - 3;                  // map 4..31 to 1..28
                              }
                            Actioned = 1;
                            break;
                        case DCC128:
                            current_loc.speed = rx_message[2] & 0x7F;
                            Actioned = 1;
                            break;
                      }
                    
                    if (rx_message[3] & 0x10) current_loc.light = 1; 
                    else                      current_loc.light = 0;  
                    
                    current_loc.f4_f1 = rx_message[3] & 0x0F;
                    current_loc.f8_f5 = rx_message[4] & 0x0F;
                    current_loc.f12_f9 = (rx_message[4] & 0xF0) >> 4;
                    current_loc.valid = 1;
                    
                    cr_fifo_put(&key_fifo, KEY_PARSER_LOK);
                    break;
                    // 0xE5 Identification Speed FKTA FKTB MTR [XOR] "Locomotive information for a locomotive in a multi-unit"
                    // 0xE2 Identification Speed [XOR] "Locomotive information for the Multi-unit address"
                    // 0xE6 Identification Speed FKTA FKTB AddrH AddrL [XOR] "Locomotive information for a locomotive in a Double Header"
                    // 0xE3 0x30+K AddrH AddrL [XOR] "Locomotive information response for address retrieval requests"
                    // Roco:
                    // 0xE7 0x02 0x80 0x20+FKT1-4 FKT5-12 FKT13-20 0x00 0x00 -> mehr Funktionen
                case 0x40:
                    // 0xE3 0x40 AddrH AddrL [XOR] "Locomotive is being operated by another device"
                    addr = (rx_message[2] & 0x3F) * 256 + rx_message[3];
                    record_inactive_loco(addr);                             // tell messages_objects about the change
                    if (addr == current_loc.addr)
                      {
                        current_loc.other = 1;
                        cr_fifo_put(&key_fifo, KEY_PARSER_STOLEN);          // notify GUI
                      }
                    break;
                case 0x50:
                    // 0xE3 0x50 STAT0 STAT1 [XOR] "Function status F0..F12 (tastend)"
                    // 0xE3 0x51 STAT0 STAT1 [XOR] "Function status F13--F28 (tastend)"
                    // 0xE3 0x52 DAT0 DAT1 [XOR] "Function level F13--F28"
                    switch(rx_message[1] & 0x07)
                      {
                        default: break;
                        case 2:
                            i = current_loc.f20_f17 << 4 | current_loc.f16_f13;
                            j = current_loc.f28_f25 << 4 | current_loc.f24_f21;
                            PORTA = 0xAA;
                            if ( (i != rx_message[2]) || (j != rx_message[3]) ) 
                              {
                                current_loc.f16_f13 = rx_message[2] & 0x0F;
                                current_loc.f20_f17 = (rx_message[2] & 0xF0) >> 4;
                                current_loc.f24_f21 = rx_message[3] & 0x0F;
                                current_loc.f28_f25 = (rx_message[3] & 0xF0) >> 4;
                                cr_fifo_put(&key_fifo, KEY_PARSER_LOK);
                              }
                            break;
                      }
                    break;
                case 0x80:
                    // 0xE1 0x80+F [XOR] "XpressNet error message" (multi-unit)
                    break;
                case 0xF0:
                    // 0    1    2     3     4   5    6...
                    // 0xE9 0xF1 AddrH AddrL IDX SIZE [NAME] [XOR] "Locomotive Data Base Entry"

                    // Note: Up to now, this wont work - Lok download is Call message, not Master!

                    new_database_entry.addr = (rx_message[2] & 0x3F) * 256 + rx_message[3];
                    new_database_entry.idx  = rx_message[4];
                    new_database_entry.size = rx_message[5];
                      {
                        unsigned char i;
                        i = rx_message[0] & 0x0F;
                        i -= 5;
                        if (i > LOK_NAME_LENGTH) i=LOK_NAME_LENGTH;
                        memcpy(new_database_entry.name,&rx_message[6],i);
                        if (i < LOK_NAME_LENGTH) new_database_entry.name[i] = 0;  // string terminator
                      }
                    cr_fifo_put(&key_fifo, KEY_PARSER_DATABASE);
                  break;
              }
            break;
      }
    return(Actioned);
  }


//------------------------------------------------------------------------
//
//  Client Task
//
//------------------------------------------------------------------------
//  This is a cooperative task, see cortos.c
//  It does:
//    1. get as many as available samples from xpressnet
//    2. trace down the link layer (see xp_state)
//    3. if a valid packet for our ID is detected:
//       call the parser
//
//  It should be called at least every 10ms (see X_RxBuffer_Size)
//------------------------------------------------------------------------


// internal and static
enum xp_states
  {                                     // actual state for the Xpressnet Task
    XP_WAIT_FOR_CALLBYTE,
    XP_INQUIRE_SLOT,
    XP_COLLECT_MESSAGE,
    XP_WAIT_FOR_MESSAGE_COMPLETE,
    XP_COLLECT_FEEDBACK,
    XP_WAIT_FOR_FEEDBACK_COMPLETE,
  } xp_state;

static int xp_active_timestamp;
static unsigned char actual_slot;          // this is the slot currently receiving

t_cr_task run_xpressnet_client(void)
  {
    unsigned int xdata;

    if ((get_systick_int() - xp_active_timestamp) > (500000L / SYSTICK_PERIOD) )
      {
        set_xp_state(XP_DISCONNECTED);
        xp_active_timestamp = get_systick_int();
        //!warning! hat noch weitergehende Folgen: wir sollten auch TX Fifo-flush
      }

    while (XP_rx_ready())
      {
        xdata = XP_rx_read();                 // reads one byte
        
        if (xdata >= 256)
          { // new call byte
            set_xp_state(XP_CONNECTED);

            xp_active_timestamp = get_systick_int();
            actual_slot = xdata & 0x1F;

            // debug only: if(actual_slot == slot_id){
            // debug only: lcd_int(actual_slot); lcd_string("#"); }

            switch(xdata & 0x60)
              {
                default:  // no break;
                case 0x00:      // ACK
                    xp_state = XP_WAIT_FOR_CALLBYTE;
                    break;
                case 0x20:      // Future or Feedback
                    if (actual_slot == 0)
                      {
                        xp_state = XP_COLLECT_FEEDBACK;
                      }
                    else
                      {
                        xp_state = XP_WAIT_FOR_CALLBYTE;
                      }
                    break;
                case 0x40:      // Inquiry
                    xp_state = XP_INQUIRE_SLOT;
                    break;
                case 0x60:      // A message from the master
                    if (actual_slot == 0 || actual_slot == slot_id)
                      {
                         xp_state = XP_COLLECT_MESSAGE;
                      }
                    else
                      {
                         xp_state = XP_WAIT_FOR_CALLBYTE;  // not our slot - ignore it.
                      }
                    break;
              }
          }
        else
          {
            #if (DEBUG_RX == 2)
                #warning debug printout of first RX byte
                static char dumped;
                if (!dumped)
                  {
                    set_cursor(0,0);
                    lcd_hex(xdata / 256);
                    lcd_hex(xdata % 256);
                    dumped = 1;
                  }
            #endif

            switch (xp_state)
              {
                case XP_WAIT_FOR_CALLBYTE:                  // our state is not known - just drop any message
                    break;

                case XP_INQUIRE_SLOT:
                    mark_slot_as_used(actual_slot);         // there is something after an inquiry -> must be used
                    // Note: when we receive an inquiry to our slot-id,
                    //       the ISR itself handles the tx_message (if any).
                    break;

                case XP_COLLECT_MESSAGE:
                    xp_state = XP_WAIT_FOR_MESSAGE_COMPLETE;
                    rx_message[0] = xdata;                  // save header
                    rx_size = rx_message[0] & 0x0F;         // length is without xor
                    rx_size++;                              // now including xor
                    rx_index = 1;
                    break;

                case XP_COLLECT_FEEDBACK:
                    xp_state = XP_WAIT_FOR_FEEDBACK_COMPLETE;
                    rx_message[0] = xdata;                  // save header
                    rx_size = rx_message[0] & 0x0F;         // length is without xor
                    rx_size++;                              // now including xor
                    rx_index = 1;
                    break;

                case XP_WAIT_FOR_MESSAGE_COMPLETE:
                    rx_message[rx_index] = xdata;
                    if (rx_index == rx_size)
                      {
                        unsigned char i, my_check = 0;
                        // all data and xor read, now check xor
                        for (i=0; i<=rx_size; i++) my_check ^= rx_message[i];


                        if (my_check == 0)
                          {
                            // packet is received and okay, now parse it
                            xp_parser();
                            xp_state = XP_WAIT_FOR_CALLBYTE;
                          }
                        else
                          {
                            // XOR is wrong!
                            xp_state = XP_WAIT_FOR_CALLBYTE;
                          }
                      }
                    else
                      {
                        rx_index++;
                        if (rx_index == 17-1) rx_index = 17-1;   // overrun!
                      }
                    break;

                case XP_WAIT_FOR_FEEDBACK_COMPLETE:
                    rx_message[rx_index] = xdata;
                    if (rx_index == rx_size)
                      {
                        unsigned char i, my_check = 0;
                        // all data and xor read, now check xor
                        for (i=0; i<=rx_size; i++) my_check ^= rx_message[i];


                        if (my_check == 0)
                          {
                            // packet is received and okay, now parse it
                            xp_parser_feedback();
                            xp_state = XP_WAIT_FOR_CALLBYTE;
                          }
                        else
                          {
                            // XOR is wrong!
                            xp_state = XP_WAIT_FOR_CALLBYTE;
                          }
                      }
                    else
                      {
                        rx_index++;
                        if (rx_index == 17-1) rx_index = 17-1;   // overrun!
                      }
                    break;
              }
          }
      }
    return(500000L / SYSTICK_PERIOD);   // wake up, to check timeout
 // return(-1); // everything processed - no more ready
                // The incoming fifo will issue a wakeup.
  }


void restart_xpressnet_client(void)
  {
    init_xp_client_if();                        // init low level
    xp_state = XP_WAIT_FOR_CALLBYTE;

    set_xp_state(XP_DISCONNECTED);
    
    xp_active_timestamp = get_systick_int();    // reset timing control

    slot_mode = eeprom_read_byte((void*)eadr_xp_slot_mode);
    slot_id = eeprom_read_byte((void*)eadr_xp_slot_addr);
    if ((slot_id < 1) || (slot_id > 31))
      {
        slot_id = 17;  // default
        eeprom_write_byte((void*)eadr_xp_slot_addr, slot_id);
      }
  }

void stop_xpressnet_client(void)
  {
    // disable interrupts
    stop_xp_client_if();
    // the consumer task will terminate itself
  }

void init_xpressnet_client(void)
  {
    restart_xpressnet_client();
  }


void test_xp_parser(void)
  {
    rx_message[0] = 0x62;
    rx_message[1] = 0x22;
    rx_message[2] = 0x01;
    xp_parser();
  }
