//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      xp_messages.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2009-02-10 V0.01 kw  started
//            2009-02-26 V0.02 kw  added pc interface routinen
//            2009-03-01 V0.03 kw  added function + light messages
//            2010-03-31 V0.04 kw  added PoM read
//            2010-06-14 V0.05 kw  added extended accessory
//            2010-06-17 V0.06 kw  added PoM commands for accessory and ext.acc.
//            2011-07-07 V0.07 kw  bugfix bei xp_send_cv_read
//
//--------------------------------------------------------------------------
//
// purpose:   throttle
// content:   handling of messages issued by menus or turnout
//            (this is message layer for xpressnet)
//
//--------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <string.h>

#include "config.h"
#include "cortos.h"

#include "messages_objects.h"
#include "xp_client_if.h"

t_xp_message xp_tx_msg_new;                 // next message to be sent


static void __setLocAddr( int addr, unsigned char * addrbytes ) {
  if( addr < 100 ) {
    addrbytes[0] = 0;
    addrbytes[1] = addr & 0x00FF;
  }
  else {
    addrbytes[0] = ((addr & 0xFF00) >> 8) + 0xC0;
    addrbytes[1] = addr & 0x00FF;
  }
}


//-----------------------------------------------------------------------------

void xp_send_cs_status_request(void)
  {
	xp_tx_msg_new.header = 0x21;
	xp_tx_msg_new.data[0] = 0x24;
	xp_tx_msg_new.data[1] = 0x05;
    fifo_put_xp_out(&xp_tx_msg_new);
  }

void xp_send_state(t_status new_state)
  {
    switch (new_state)
      {
        case STS_ON:
            xp_tx_msg_new.header = 0x21;
            xp_tx_msg_new.data[0] = 0x81;   // go
            fifo_put_xp_out(&xp_tx_msg_new);
            break;
        case STS_OFF:
            xp_tx_msg_new.header = 0x21;
            xp_tx_msg_new.data[0] = 0x80;   // off
            fifo_put_xp_out(&xp_tx_msg_new);
            break;
        case STS_STOP:
            xp_tx_msg_new.header = 0x80;
            xp_tx_msg_new.data[0] = 0x80;   // stop
            fifo_put_xp_out(&xp_tx_msg_new);
            break;
        case STS_PROG:
            // no command required
            break;
        case STS_BUSY:
            // no command required
            break;
        case STS_PBUSY:
            // no command required
            break;
        case STS_DISCONNECTED:
            // no command required
            break;
        case STS_CMD_ERROR:
            // no command required
            break;
      }
  }

// parameter: module addr (in Lenz this is the same address area as accessory
void xp_send_feedback_query(unsigned int addr)
  {
    xp_tx_msg_new.header = 0x42;
    xp_tx_msg_new.data[0] = addr;
    xp_tx_msg_new.data[1] = 0x80;     // lower nibble
    fifo_put_xp_out(&xp_tx_msg_new);

    xp_tx_msg_new.header = 0x42;
    xp_tx_msg_new.data[0] = addr;
    xp_tx_msg_new.data[1] = 0x81;     // upper nibble
    fifo_put_xp_out(&xp_tx_msg_new);
  }


// parameter: true dcc coil addr (raw)
void xp_send_accessory(unsigned int addr, bool activate)
  {
    unsigned int decaddr = addr / 8;
    unsigned int port = addr % 8;

    xp_tx_msg_new.header = 0x52;
    xp_tx_msg_new.data[0] = decaddr;
    if (activate) xp_tx_msg_new.data[1] = 0x08;
    else          xp_tx_msg_new.data[1] = 0x00;
    xp_tx_msg_new.data[1] |= 0x80 | port;

    fifo_put_xp_out(&xp_tx_msg_new);
  }

// parameter: turnout addr
void xp_send_accessory_query(unsigned int addr)
  {
    addr--;
    unsigned int decaddr = addr / 4;
    unsigned char nibble = (addr & 0x2) >> 1;

    xp_tx_msg_new.header = 0x42;
    xp_tx_msg_new.data[0] = decaddr;
    xp_tx_msg_new.data[1] = 0x80 + nibble;

    fifo_put_xp_out(&xp_tx_msg_new);
  }

// 0x13 0x01 B+AddrH AddrL   DCC extended accessory command request. (compact form)
//                           AddrH (bit A10..A8) and AddrL (bit A7..A0) form a 11-bit decoder address, 
//                           B (=B4..B0) form the desired aspect. 
//                           All bits B4..B0 equal 0 define the 'stop'-aspect. 
void xp_send_extended_accessory(unsigned int addr, unsigned char aspect)
  {
    t_data16 temp;
    temp.as_uint16 = addr;

    xp_tx_msg_new.header = 0x13;
    xp_tx_msg_new.data[0] = 0x01;
    xp_tx_msg_new.data[1] = aspect << 3;                // only 5 bits
    xp_tx_msg_new.data[1] |= temp.as_uint8[1] & 0x07;   // upper 3 bits
    xp_tx_msg_new.data[2] = temp.as_uint8[0];           // lower 8 bits

    fifo_put_xp_out(&xp_tx_msg_new);
  }

// parameter:
// addr: dcc addr
// format: DCC14, DCC28, DCC128
// speed:  0..126
// dir:    0 = forward, 1 = revers
void xp_send_loc_drive(unsigned int addr, t_format format, unsigned char speed, unsigned char dir)
  {
    if(addr < 1)  return;
    if(addr > 10239)  return;  // // 10239 = 0x27FF

    unsigned char lenzspeed;
    unsigned char reqid = 0x12; /* default 28 speed steps */
    switch(format)
      {
        default:
        case DCC28:
            reqid = 0x12;
            if (speed < 1)
              {
                lenzspeed = speed; 
              }
            else
              {          
                lenzspeed = (speed & 0x1F) + 3;    // map internal speed 1..28 to external 4..31
                lenzspeed = (lenzspeed>>1) | ((lenzspeed & 0x01) <<4);
              }
            break;
        case DCC27:
            reqid = 0x11;
            if (speed < 1)
              {
                lenzspeed = speed; 
              }
            else
              {          
                lenzspeed = (speed & 0x1F) + 3;    // map internal speed 1..28 to external 4..31
                lenzspeed = (lenzspeed>>1) | ((lenzspeed & 0x01) <<4);
              }
            break;
        case DCC14:
            reqid = 0x10;
            if (speed < 1) lenzspeed = speed; 
            else           lenzspeed = speed + 1;
            break;
        case DCC128:
            reqid = 0x13;
            if (speed < 1) lenzspeed = speed; 
            else           lenzspeed = speed + 1;
            break;
      }
    xp_tx_msg_new.header = 0xE4;
    xp_tx_msg_new.data[0] = reqid;
    __setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = dir ? 0x00:0x80;
    xp_tx_msg_new.data[3] |= lenzspeed;

    fifo_put_xp_out(&xp_tx_msg_new);
  }

// sends an emergency stop to this loco
void xp_send_loc_stop(unsigned int addr)
  {
    xp_tx_msg_new.header = 0x92;
    __setLocAddr( addr, xp_tx_msg_new.data );

    fifo_put_xp_out(&xp_tx_msg_new);
  }

// ask command station about actual lok data
// we will receive: format, actual speed, functions f0 ... f12

void xp_send_askforlocinfo(unsigned int addr)
  {
    xp_tx_msg_new.header = 0xE3;
    xp_tx_msg_new.data[0] = 0x00;
    __setLocAddr( addr, xp_tx_msg_new.data+1 );

    fifo_put_xp_out(&xp_tx_msg_new);

    #if (DCC_F13_F28)
        // also query extended functions
        xp_tx_msg_new.header = 0xE3;
        xp_tx_msg_new.data[0] = 0x09;
        __setLocAddr( addr, xp_tx_msg_new.data+1 );

        fifo_put_xp_out(&xp_tx_msg_new);
    #endif
  }


// note: current_loc.f?? must be loaded completely
// we transmit only the according group to save bandwidth
// parameter: fnum, range 0..27 corresponds to f1 ... f28
void xp_send_loc_function(unsigned char fnum)
  {
    // 1. pick corresponding function group
    if (fnum < 4)
      {
        xp_tx_msg_new.header  = 0xE4;
        xp_tx_msg_new.data[0] = 0x20;
        __setLocAddr( current_loc.addr, xp_tx_msg_new.data+1 );
        xp_tx_msg_new.data[3] = current_loc.f4_f1;
        if (current_loc.light) xp_tx_msg_new.data[3] |=0x10;
        fifo_put_xp_out(&xp_tx_msg_new);
      }
    else if (fnum < 8)
      {
        xp_tx_msg_new.header  = 0xE4;
        xp_tx_msg_new.data[0] = 0x21;
        __setLocAddr( current_loc.addr, xp_tx_msg_new.data+1 );
        xp_tx_msg_new.data[3] = current_loc.f8_f5;
        fifo_put_xp_out(&xp_tx_msg_new);
      }
    else if (fnum < 12)
      {
        xp_tx_msg_new.header  = 0xE4;
        xp_tx_msg_new.data[0] = 0x22;
        __setLocAddr( current_loc.addr, xp_tx_msg_new.data+1 );
        xp_tx_msg_new.data[3] = current_loc.f12_f9;
        fifo_put_xp_out(&xp_tx_msg_new);        
      }
    else if (fnum < 20)
      {
        xp_tx_msg_new.header  = 0xE4;
        xp_tx_msg_new.data[0] = 0x23;
        __setLocAddr( current_loc.addr, xp_tx_msg_new.data+1 );
        xp_tx_msg_new.data[3] = (current_loc.f20_f17 << 4) | current_loc.f16_f13;
        fifo_put_xp_out(&xp_tx_msg_new);
      }
    else // if (fnum < 28)
      {
        xp_tx_msg_new.header  = 0xE4;
        xp_tx_msg_new.data[0] = 0x28;
        __setLocAddr( current_loc.addr, xp_tx_msg_new.data+1 );
        xp_tx_msg_new.data[3] = (current_loc.f28_f25 << 4) | current_loc.f24_f21;
        fifo_put_xp_out(&xp_tx_msg_new);
      }
  }

void xp_send_loc_light(void)
  {
    xp_send_loc_function(0);            // this is coded combined with fuction group 1)
  }


//----------------------------------------------------------------------------
//
// CV-stuff
//
//----------------------------------------------------------------------------

void xp_send_askforresult(void)
  {
	xp_tx_msg_new.header = 0x21;
	xp_tx_msg_new.data[0] = 0x10;
	xp_tx_msg_new.data[1] = 0x31;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

void xp_send_cv_read(unsigned int cv)
  {
    if (cv < 255)      xp_tx_msg_new.data[0] = 0x15; // this would be syntax for 3.0 protocol
                                                     // only 1..255
	else if (cv==1024) xp_tx_msg_new.data[0] = 0x18;
    else if (cv < 256) xp_tx_msg_new.data[0] = 0x18;
    else if (cv < 512) xp_tx_msg_new.data[0] = 0x19;
    else if (cv < 768) xp_tx_msg_new.data[0] = 0x1A;
    else               xp_tx_msg_new.data[0] = 0x1B;

	xp_tx_msg_new.header = 0x22;
	xp_tx_msg_new.data[1] = cv & 0xFF;
	fifo_put_xp_out(&xp_tx_msg_new);

	// OpenDCC sendet das Ergebnis nur auf Nachfrage nochmal.
	// Der Lenz-Handregler fragt aber nicht nochmal nach.
	// Voerschlag: OpenDCC sendet das Ergebnis wenn es soweit ist
	// auf jeden Fall noch mal. (Bisher ist das anscheinend nicht der Fall).
    // jmf? macht das die Lenz Zentrale?
  }


void xp_send_cv_write(unsigned int cv, unsigned char value)
  {
    if (cv < 255)      xp_tx_msg_new.data[0] = 0x16; // this would be syntax for 3.0 protocol
                                                     // only 1..255
	else if (cv==1024) xp_tx_msg_new.data[0] = 0x1C;
    else if (cv < 256) xp_tx_msg_new.data[0] = 0x1C;
    else if (cv < 512) xp_tx_msg_new.data[0] = 0x1D;
    else if (cv < 768) xp_tx_msg_new.data[0] = 0x1E;
    else               xp_tx_msg_new.data[0] = 0x1F;

	xp_tx_msg_new.header = 0x23;
	xp_tx_msg_new.data[1] = cv & 0xFF;
	xp_tx_msg_new.data[2] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }


// PoM READ (loco)
// 0xE6 0x30 AddrH AddrL 0xE4+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_cv_read(unsigned int addr, unsigned int cv)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE5;
    xp_tx_msg_new.data[0] = 0x30;
	__setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xE4 + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// PoM WRITE (loco)
// 0xE6 0x30 AddrH AddrL 0xEC+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_cv_write(unsigned int addr, unsigned int cv, unsigned char value)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE6;
    xp_tx_msg_new.data[0] = 0x30;
	__setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xEC + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	xp_tx_msg_new.data[5] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// PoM READ (accessory)
// 0xE6 0x30 AddrH AddrL 0xF4+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_acc_cv_read(unsigned int addr, unsigned int cv)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE5;
    xp_tx_msg_new.data[0] = 0x30;
	__setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xF4 + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// PoM WRITE (accessory)
// 0xE6 0x30 AddrH AddrL 0xF0+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_acc_cv_write(unsigned int addr, unsigned int cv, unsigned char value)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE6;
    xp_tx_msg_new.data[0] = 0x30;
	__setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xF0 + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	xp_tx_msg_new.data[5] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// PoM READ (extended accessory)
// 0xE6 0x30 AddrH AddrL 0xF4+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_extacc_cv_read(unsigned int addr, unsigned int cv)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE5;
    xp_tx_msg_new.data[0] = 0x30;
    __setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xFC + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// PoM WRITE (extended accessory)
// 0xE6 0x30 AddrH AddrL 0xF0+C CV DAT [XOR] 
// Note: CV is here addressed in the range 0 .. 1023
void xp_send_PoM_extacc_cv_write(unsigned int addr, unsigned int cv, unsigned char value)
  {
    cv = cv - 1;
    xp_tx_msg_new.header = 0xE6;
    xp_tx_msg_new.data[0] = 0x30;
	__setLocAddr( addr, xp_tx_msg_new.data+1 );
    xp_tx_msg_new.data[3] = 0xF8 + (unsigned char)((cv >> 8) & 0x03);
	xp_tx_msg_new.data[4] = cv & 0xFF;
	xp_tx_msg_new.data[5] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }


//----------------------------------------------------------------------------
//
// CS-config
//
//----------------------------------------------------------------------------

void xp_send_cs_cfg_read(unsigned int cv)
  {
	xp_tx_msg_new.header = 0x23;
	xp_tx_msg_new.data[0] = 0x28;
	xp_tx_msg_new.data[1] = (unsigned char)(cv >> 8);
	xp_tx_msg_new.data[2] = (unsigned char) cv;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

void xp_send_cs_cfg_write(unsigned int cv, unsigned char value)
  {
	xp_tx_msg_new.header = 0x24;
	xp_tx_msg_new.data[0] = 0x29;
	xp_tx_msg_new.data[1] = (unsigned char)(cv >> 8);
	xp_tx_msg_new.data[2] = (unsigned char) cv;
	xp_tx_msg_new.data[3] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

//0x05 0xF1 TCODE1 TCODE2 TCODE3 TCODE4 [XOR] (0=m, 80=h, 40=week, C0=ratio)
// we do only hour and minute, and ratio
void xp_send_clock_set(t_fast_clock *my_clock)
  {
  	xp_tx_msg_new.header = 0x05;
	xp_tx_msg_new.data[0] = 0xF1;
	xp_tx_msg_new.data[1] = 0x00 | my_clock->minute;
	xp_tx_msg_new.data[2] = 0x80 | my_clock->hour;
	xp_tx_msg_new.data[3] = 0x40 | my_clock->day_of_week;
	xp_tx_msg_new.data[4] = 0xC0 | my_clock->ratio;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

// special functions for direct pc interaction

void xp_send_pc_hello(void)
  {
    xp_tx_msg_new.header = 0x33;
    xp_tx_msg_new.data[0] = slot_id;
    xp_tx_msg_new.data[1] = 0x00;       // 00 = hello/goodbye
    xp_tx_msg_new.data[2] = 0x01;
	fifo_put_xp_out(&xp_tx_msg_new);
  }


void xp_send_pc_goodbye(void)
  {
    xp_tx_msg_new.header = 0x33;
    xp_tx_msg_new.data[0] = slot_id;
    xp_tx_msg_new.data[1] = 0x00;       // 00 = hello/goodbye
    xp_tx_msg_new.data[2] = 0x00;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

void xp_send_PID(void)
  {
    // send 0x38 0x01 VID PID SerialH SerialL Version Subversion;
            
    xp_tx_msg_new.header = 0x38;
    xp_tx_msg_new.data[0] = slot_id;
    xp_tx_msg_new.data[1] = 0x01;       // 01 = PID/version
    xp_tx_msg_new.data[2] = 0x13;       // 13 = OpenDCC
    #if (DISPLAY == LCD_EA204)
        xp_tx_msg_new.data[3] = 0x01;       // 01 = MFT with char display 4 x 20
    #else
        xp_tx_msg_new.data[3] = 0x02;       // 02 = MFT with graphic display 128 x 64
    #endif
    xp_tx_msg_new.data[4] = serial_number >> 8;
    xp_tx_msg_new.data[5] = serial_number & 0xFF;
    
    xp_tx_msg_new.data[6] = MFT_VERSION;
    xp_tx_msg_new.data[7] = MFT_SUBVERSION;
    fifo_put_xp_out(&xp_tx_msg_new);
  }


void xp_send_pc_key(unsigned char key, signed char value)
  {
    xp_tx_msg_new.header = 0x34;
    xp_tx_msg_new.data[0] = slot_id;
    xp_tx_msg_new.data[1] = 0x20;
    xp_tx_msg_new.data[2] = key;
    xp_tx_msg_new.data[3] = value;
	fifo_put_xp_out(&xp_tx_msg_new);
  }

  
 
